/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* This code was taken from 'BaseLibs' project
* Class CGe3dPoint implementation
********************************************************************/
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <new.h>
#include "Geom.h"

static double _Angle  = 0.0;
static double _cosAng = 1.0;
static double _sinAng = 0.0;
/*
double CGe3dPoint::m00 = 0.0;
double CGe3dPoint::m10 = 0.0;
double CGe3dPoint::m20 = 0.0;
double CGe3dPoint::m01 = 0.0;
double CGe3dPoint::m11 = 0.0;
double CGe3dPoint::m21 = 0.0;
double CGe3dPoint::m02 = 0.0;
double CGe3dPoint::m12 = 0.0;
double CGe3dPoint::m22 = 0.0;
double CGe3dPoint::m03 = 0.0;
double CGe3dPoint::m13 = 0.0;
double CGe3dPoint::m23 = 0.0;
*/

double CGe3dPoint::m_X0;
double CGe3dPoint::m_Y0;
double CGe3dPoint::m_Z0;
BYTE   CGe3dPoint::m_Axis = 0;
double CGe3dPoint::m_cosAng = 0.0;
double CGe3dPoint::m_sinAng = 1.0;


//-----------------------------------------------
CGe3dPoint::CGe3dPoint (const CGePoint& pnt)
{
  x = pnt.x;
  y = pnt.y;
  z = 0.;
}


//-----------------------------------------------
void CGe3dPoint::SetVector (const CGe3dPoint& pnt)
{
  x = pnt.x;
  y = pnt.y;
  z = pnt.z;
  VectorNormalize();
}

//-----------------------------------------------
void CGe3dPoint::SetVector (const CGe3dPoint& pt0, const CGe3dPoint& pt1)
{
  x = pt1.x - pt0.x;
  y = pt1.y - pt0.y;
  z = pt1.z - pt0.z;
  VectorNormalize();
}

//-----------------------------------------------
double CGe3dPoint::Distance (const CGe3dPoint& p) const
{
  double dx = p.x - x;
  double dy = p.y - y;
  double dz = p.z - z;
  dx = p.x - x;
  dy = p.y - y;
  dz = p.z - z;
  return sqrt( dx*dx + dy*dy + dz*dz );
}

//-----------------------------------------------
// Angle from this point to Pnt (0..2PI)
//-----------------------------------------------
double CGe3dPoint::Angle (const CGe3dPoint& Pnt, EnPlane Plane) const
{
  double a, b;
  switch( Plane ){
    case EnPlane::GE_PLANE_TOP:
      a = Pnt.x - x;
      b = Pnt.y - y;
      break;
    case EnPlane::GE_PLANE_BOTTOM:
      a = x - Pnt.x;
      b = Pnt.y - y;
      break;
    case EnPlane::GE_PLANE_FRONT:
      a = Pnt.x - x;
      b = Pnt.z - z;
      break;
    case EnPlane::GE_PLANE_BACK:
      a = x - Pnt.x;
      b = Pnt.z - z;
      break;
    case EnPlane::GE_PLANE_RIGHT:
      a = Pnt.y - y;
      b = Pnt.z - z;
      break;
    case EnPlane::GE_PLANE_LEFT:
      a = y - Pnt.y;
      b = Pnt.z - z;
      break;
    default:
      return 0.0;
  }
  if (a == 0.) {
    if (b == 0.){
      return 0.;
    }
    return (b>0.)? GE_PI2 : GE_DEG270;
  }else{
    if (b == 0.){
      return (a>0.)? 0. : GE_PI;
    }
    if (a > 0.){
      if (b > 0.){
        return atan( b / a );
      }else{
        return atan( b / a ) + GE_2PI;
      }
    }else{
      return atan( b / a ) + GE_PI;
    }
  }
}

//-----------------------------------------------
// Angle from 0,0,0 point to this point (0..2PI)
//-----------------------------------------------
double CGe3dPoint::Angle (EnPlane Plane) const
{
  CGe3dPoint pt0;
  CGe3dPoint pt1(x,y,z);
  return pt0.Angle( pt1, Plane );
}

//-----------------------------------------------
void CGe3dPoint::GetAngles (const CGe3dPoint& Pnt, double* pAngHorz, double* pAngVert) const
{
  CGePoint Pt0, Pt1;
  double Dist, H;
  Pt0.Set( x, y );
  Pt1.Set( Pnt.x, Pnt.y );
  *pAngHorz = Pt0.Angle( Pt1 );
  Dist = Pt0.Distance( Pt1 );
  H = Pnt.z - z;
  *pAngVert = atan( H / Dist );
}


// Scale a vector to 1-unit in length:
//-----------------------------------------------
void CGe3dPoint::VectorNormalize ()
{
  double Len;
  Len = VectorLength();
  if (Len > 0.0){
    VectorScale( 1.0 / Len );
  }
}



//-----------------------------------------------
CGe3dPoint CGe3dPoint::VectorCross (const CGe3dPoint& B) const
{
  CGe3dPoint V;
  V.x = y * B.z - B.y * z;
  V.y = z * B.x - B.z * x;
  V.z = x * B.y - B.x * y;
  return V;
}


// The one and only Arbitrary Axis Algorithm
// Fills in the XAxis and YAxis arrays with
// the X and Y axis unit vectors.
// from: http://www.caddzone.com/delphi/ucs.htm
//-----------------------------------------------
void CGe3dPoint::GetArbitraryAxes (CGe3dPoint* XAxis, CGe3dPoint* YAxis) const
{
  double phi;
  CGe3dPoint WorldY;
  CGe3dPoint WorldZ;

  phi = 1.0 / 64.0;
  WorldY.Set( 0.0, 1.0, 0.0 );
  WorldZ.Set( 0.0, 0.0, 1.0 );
  // "this" vector is a normal vector of OCS
  if (fabs(x) < phi && fabs(y) < phi){
    *XAxis = WorldY.VectorCross( *this );
  }else{
    *XAxis = WorldZ.VectorCross( *this );
  }
  XAxis->VectorNormalize();
  *YAxis = VectorCross( *XAxis);
  YAxis->VectorNormalize();
}

//-----------------------------------------------
void CGe3dPoint::Transform (const CGe3dPoint& Xaxis, const CGe3dPoint& Yaxis, const CGe3dPoint& Zaxis)
{
  double x1;
  double y1;
  double z1;
  x1 = x;
  y1 = y;
  z1 = z;
  x = (x1 * Xaxis.x) + (y1 * Yaxis.x) + (z1 * Zaxis.x);
  y = (x1 * Xaxis.y) + (y1 * Yaxis.y) + (z1 * Zaxis.y);
  z = (x1 * Xaxis.z) + (y1 * Yaxis.z) + (z1 * Zaxis.z);
}

//-----------------------------------------------
void CGe3dPoint::Transform (const CGe3dPoint& Xaxis, const CGe3dPoint& Yaxis, const CGe3dPoint& Zaxis, CGe3dPoint* pOutPt) const
{
  pOutPt->x = (x * Xaxis.x) + (y * Yaxis.x) + (z * Zaxis.x);
  pOutPt->y = (x * Xaxis.y) + (y * Yaxis.y) + (z * Zaxis.y);
  pOutPt->z = (x * Xaxis.z) + (y * Yaxis.z) + (z * Zaxis.z);
}

//-----------------------------------------------
CGe3dPoint CGe3dPoint::TransformBy (const CGe3dPoint& Xaxis, const CGe3dPoint& Yaxis, const CGe3dPoint& Zaxis) const
{
  CGe3dPoint Pt;
  Pt.x = (x * Xaxis.x) + (y * Yaxis.x) + (z * Zaxis.x);
  Pt.y = (x * Xaxis.y) + (y * Yaxis.y) + (z * Zaxis.y);
  Pt.z = (x * Xaxis.z) + (y * Yaxis.z) + (z * Zaxis.z);
  return Pt;
}

//-----------------------------------------------
void CGe3dPoint::Polar (double Angle, double Dist, EnPlane Plane)
{
  if (_Angle != Angle){
    _Angle  = Angle;
    _cosAng = cos(_Angle);
    _sinAng = sin(_Angle);
  }
  switch( Plane ){
    case EnPlane::GE_PLANE_TOP:
    default:
      x += Dist * _cosAng;
      y += Dist * _sinAng;
      break;
    case EnPlane::GE_PLANE_BOTTOM:
      x -= Dist * _cosAng;
      y += Dist * _sinAng;
      break;
    case EnPlane::GE_PLANE_FRONT:
      x += Dist * _cosAng;
      z += Dist * _sinAng;
      break;
    case EnPlane::GE_PLANE_BACK:
      x -= Dist * _cosAng;
      z += Dist * _sinAng;
      break;
    case EnPlane::GE_PLANE_RIGHT:
      y += Dist * _cosAng;
      z += Dist * _sinAng;
      break;
    case EnPlane::GE_PLANE_LEFT:
      y -= Dist * _cosAng;
      z += Dist * _sinAng;
      break;
  }
}

//-----------------------------------------------
void CGe3dPoint::Polar (double AngHorz, double AngVert, double Dist)
{
  double d  = Dist * cos( AngVert );
  double dy = Dist * sin( AngVert );
  y = y + dy;
  x = x + (d * cos(AngHorz));
  z = z - (d * sin(AngHorz));
}



//-----------------------------------------------
CGe3dPoint CGe3dPoint::PolarBy (double Angle, double Dist, EnPlane Plane) const
{
  CGe3dPoint Pt;
  if (_Angle != Angle){
    _Angle  = Angle;
    _cosAng = cos(_Angle);
    _sinAng = sin(_Angle);
  }
  switch( Plane ){
    case EnPlane::GE_PLANE_TOP:
    default:
      Pt.Set( x+Dist*_cosAng, y+Dist*_sinAng, z );
      break;
    case EnPlane::GE_PLANE_BOTTOM:
      Pt.Set( x-Dist*_cosAng, y+Dist*_sinAng, z );
      break;
    case EnPlane::GE_PLANE_FRONT:
      Pt.Set( x+Dist*_cosAng, y, z+Dist*_sinAng );
      break;
    case EnPlane::GE_PLANE_BACK:
      Pt.Set( x-Dist*_cosAng, y, z+Dist*_sinAng );
      break;
    case EnPlane::GE_PLANE_RIGHT:
      Pt.Set( x, y+Dist*_cosAng, z+Dist*_sinAng );
      break;
    case EnPlane::GE_PLANE_LEFT:
      Pt.Set( x, y-Dist*_cosAng, z+Dist*_sinAng );
      break;
  }
  return Pt;
}

//-----------------------------------------------
CGe3dPoint CGe3dPoint::PolarBy (double AngHorz, double AngVert, double Dist) const
{
  double d  = Dist * cos( AngVert );
  double dz = Dist * sin( AngVert );
  double z2 = z + dz;
  double x2 = x + (d * cos(AngHorz));
  double y2 = y + (d * sin(AngHorz));
  return CGe3dPoint( x2,y2,z2 );
}


//-----------------------------------------------
void CGe3dPoint::Rotate (const CGe3dPoint& BasePt, double Angle, EnPlane Plane)
{
  double dx, dy, dz;
  if (_Angle != Angle){
    _Angle  = Angle;
    _cosAng = cos(_Angle);
    _sinAng = sin(_Angle);
  }
  switch( Plane ){
    case EnPlane::GE_PLANE_TOP:
    default:
      dx = x - BasePt.x;
      dy = y - BasePt.y;
      x = (dx * _cosAng) - (dy * _sinAng) + BasePt.x;
      y = (dx * _sinAng) + (dy * _cosAng) + BasePt.y;
      break;
    case EnPlane::GE_PLANE_BOTTOM:
      dx = BasePt.x - x;
      dy = y - BasePt.y;
      x = BasePt.x - ((dx * _cosAng) - (dy * _sinAng));
      y = (dx * _sinAng) + (dy * _cosAng) + BasePt.y;
      break;
    case EnPlane::GE_PLANE_FRONT:
      dx = x - BasePt.x;
      dz = z - BasePt.z;
      x = dx * _cosAng - dz * _sinAng + BasePt.x;
      z = dx * _sinAng + dz * _cosAng + BasePt.z;
      break;
    case EnPlane::GE_PLANE_BACK:
      dx = BasePt.x - x;
      dz = z - BasePt.z;
      x = BasePt.x - ((dx * _cosAng) - (dz * _sinAng));
      z = (dx * _sinAng) + (dz * _cosAng) + BasePt.z;
      break;
    case EnPlane::GE_PLANE_RIGHT:
      dy = y - BasePt.y;
      dz = z - BasePt.z;
      y = (dy * _cosAng) - (dz * _sinAng) + BasePt.y;
      z = (dy * _sinAng) + (dz * _cosAng) + BasePt.z;
      break;
    case EnPlane::GE_PLANE_LEFT:
      dy = BasePt.y - y;
      dz = z - BasePt.z;
      y = BasePt.y - ((dy * _cosAng) - (dz * _sinAng));
      z = (dy * _sinAng) + (dz * _cosAng) + BasePt.z;
      break;
  }
}

/*
double FunAngle = 1000.0;
double FunCosAng = 0;
double FunSinAng = 0;
CGe3dPoint Base;

// rotate about a line centred on A
//-----------------------------------------------
void CGe3dPoint::SetRotate (const CGe3dPoint& A, double theta, double phi, double alpha)
{ 
  double cosAlpha, sinAlpha, cosPhi, sinPhi, cosTheta, sinTheta;
  double cosPhi2, sinPhi2, cosTheta2, sinTheta2;
  double c, a1, a2, a3;

  if (A.IsNull()){
    a1 = a2 = a3 = 0.0;
  }else{
    a1 = A.x; 
    a2 = A.y; 
    a3 = A.z;
  }
  cosPhi = cos(phi); 
  sinPhi = sin(phi);
  cosPhi2 = cosPhi * cosPhi; 
  sinPhi2 = sinPhi * sinPhi;
  cosTheta = cos(theta);
  sinTheta = sin(theta);
  cosTheta2 = cosTheta * cosTheta;
  sinTheta2 = sinTheta * sinTheta;
  cosAlpha = cos(alpha);
  sinAlpha = sin(alpha);
  c = 1.0 - cosAlpha;
  m00 = cosTheta2 * (cosAlpha * cosPhi2 + sinPhi2) + cosAlpha * sinTheta2;
  m10 = sinAlpha * cosPhi + c * sinPhi2 * cosTheta * sinTheta;
  m20 = sinPhi * (cosPhi * cosTheta * c - sinAlpha * sinTheta);
  m01 = sinPhi2 * cosTheta * sinTheta * c - (sinAlpha * cosPhi);
  m11 = sinTheta2 * ((cosAlpha * cosPhi2) + sinPhi2) + (cosAlpha * cosTheta2);
  m21 = sinPhi * ((cosPhi * sinTheta * c) + (sinAlpha * cosTheta));
  m02 = sinPhi * ((cosPhi * cosTheta * c) + (sinAlpha * sinTheta));
  m12 = sinPhi * ((cosPhi * sinTheta * c) - (sinAlpha * cosTheta));
  m22 = cosAlpha * sinPhi2 + cosPhi2;
  m03 = a1 - (a1 * m00) - (a2 * m01) - (a3 * m02);
  m13 = a2 - (a1 * m10) - (a2 * m11) - (a3 * m12);
  m23 = a3 - (a1 * m20) - (a2 * m21) - (a3 * m22);
  Base = A;
  if (FunAngle!=phi){ //angY){
    FunAngle  = phi;
    FunCosAng = cos(phi);
    FunSinAng = sin(phi);
  }
}

//-----------------------------------------------
CGe3dPoint CGe3dPoint::RotateBy () const
{
  CGe3dPoint Pt2;
  Pt2.x = -m00*x - m01*y - m02*z;
  Pt2.y = -m10*x - m11*y - m12*z;
  Pt2.z = -m20*x - m21*y - m22*z;
  return Pt2;
}
*/

// Axis: 0-x, 1-y, 2-z
//-----------------------------------------------
void CGe3dPoint::SetRotate (const CGe3dPoint& Pt0, UINT_PTR Axis, double Angle)
{
  m_X0 = Pt0.x;
  m_Y0 = Pt0.y;
  m_Z0 = Pt0.z;
  m_Axis = (BYTE)Axis;
  m_cosAng = cos( Angle );
  m_sinAng = sin( Angle );
}


//-----------------------------------------------
void CGe3dPoint::Rotate ()
{
/*
  double x2,y2,z2;
  x2 = -m00*x - m01*y - m02*z;
  y2 = -m10*x - m11*y - m12*z;
  z2 = -m20*x - m21*y - m22*z;
  x = x2;
  y = y2;
  z = z2;
*/
  double dx, dy, dz;
  switch( m_Axis ){
    case 0: // Around X-axis:
      dy = y - m_Y0;
      dz = z - m_Z0;
      y = (dy * m_cosAng) - (dz * m_sinAng) + m_Y0;
      z = (dy * m_sinAng) + (dz * m_cosAng) + m_Z0;
      break;
    case 1: // Around Y-axis:
      dx = x - m_X0;
      dz = z - m_Z0;
      x = (dx * m_cosAng) + (dz * m_sinAng) + m_X0;
      z = (dz * m_cosAng) - (dx * m_sinAng) + m_Z0;
      break;
    case 2: // Around Z-axis:
      dx = x - m_X0;
      dy = y - m_Y0;
      x = (dx * m_cosAng) - (dy * m_sinAng) + m_X0;
      y = (dx * m_sinAng) + (dy * m_cosAng) + m_Y0;
      break;
  }
}

//-----------------------------------------------
CGe3dPoint CGe3dPoint::RotateBy () const
{
  CGe3dPoint Pt2;
  double dx, dy;
  switch( m_Axis ){
    case 0: // Around X-axis:
      Pt2.x = x;
      Pt2.y = y;
      Pt2.z = z;
      break;
    case 1: // Around Y-axis:
      Pt2.x = x;
      Pt2.y = y;
      Pt2.z = z;
      break;
    case 2: // Around Z-axis:
      dx = x - m_X0;
      dy = y - m_Y0;
      Pt2.x = (dx * m_cosAng) - (dy * m_sinAng) + m_X0;
      Pt2.y = (dx * m_sinAng) + (dy * m_cosAng) + m_Y0;
      Pt2.z = z;
      break;
  }
  return Pt2;
}


#ifdef HJHJJHBBVV
//-----------------------------------------------
void CGe3dPoint::Rotate (const CGe3dPoint& Base, double /*AngX, double AngY, double AngZ)
{
  double x1,y1,z1, x2,y2,z2;
  double CosAng, SinAng;
/*
  if (FunAngle!=AngY){
    FunAngle  = AngY;
    FunCosAng = cos(AngY);
    FunSinAng = sin(AngY);
  }
*/
  // rotate around X
  if (AngY){
    CosAng = cos(AngX);
    SinAng = sin(AngX);
    y2 = y - Base.y;
    z2 = z - Base.z;
    y1 = (y2 * CosAng) - (z2 * SinAng) + Base.y;
    z1 = (y2 * SinAng) + (z2 * CosAng) + Base.z;
  }
  // rotate around Y
  if (AngY){
    CosAng = cos(AngY);
    SinAng = sin(AngY);
    x2 = x - Base.x;
    z2 = z - Base.z;
    x1 = (x2 * CosAng) - (z2 * SinAng) + Base.x;
    z1 = (x2 * SinAng) + (z2 * CosAng) + Base.z;
  }
  x = x1; 
  z = z1;
}

//-----------------------------------------------
//CGe3dPoint CGe3dPoint::RotateBy (const CGe3dPoint& Base, double AngX, double AngY, double AngZ) const
CGe3dPoint CGe3dPoint::RotateBy (const CGe3dPoint& , double , double , double ) const
{
/*
  if (FunAngle!=angle){
    FunAngle  = angle;
    FunCosAng = cos(FunAngle);
    FunSinAng = sin(FunAngle);
  }
  double x2=x-base.x;
  double y2=y-base.y;
  double x1= x2*FunCosAng - y2*FunSinAng + base.x;
  double y1= x2*FunSinAng + y2*FunCosAng + base.y;
*/
  return CGe3dPoint(x,y,z);  // ???
}

#endif //HJHJJHBBVV

//-----------------------------------------------
CGe3dPoint CGe3dPoint::RotateBy (const CGe3dPoint& BasePt, double Angle, EnPlane Plane) const
{
  double dx, dy, dz;
  CGe3dPoint Pt;

  if (_Angle != Angle){
    _Angle  = Angle;
    _cosAng = cos(_Angle);
    _sinAng = sin(_Angle);
  }
  switch( Plane ){
    case EnPlane::GE_PLANE_TOP:
    default:
      dx = x - BasePt.x;
      dy = y - BasePt.y;
      Pt.x = dx * _cosAng - dy * _sinAng + BasePt.x;
      Pt.y = dx * _sinAng + dy * _cosAng + BasePt.y;
      Pt.z = z;
      break;
    case EnPlane::GE_PLANE_BOTTOM:
      dx = BasePt.x - x;
      dy = y - BasePt.y;
      Pt.x = BasePt.x - (dx * _cosAng - dy * _sinAng);
      Pt.y = dx * _sinAng + dy * _cosAng + BasePt.y;
      Pt.z = z;
      break;
    case EnPlane::GE_PLANE_FRONT:
      dx = x - BasePt.x;
      dz = z - BasePt.z;
      Pt.x = dx * _cosAng - dz * _sinAng + BasePt.x;
      Pt.y = y;
      Pt.z = dx * _sinAng + dz * _cosAng + BasePt.z;
      break;
    case EnPlane::GE_PLANE_BACK:
      dx = BasePt.x - x;
      dz = z - BasePt.z;
      Pt.x = BasePt.x - (dx * _cosAng - dz * _sinAng);
      Pt.y = y;
      Pt.z = dx * _sinAng + dz * _cosAng + BasePt.z;
      break;
    case EnPlane::GE_PLANE_RIGHT:
      dy = y - BasePt.y;
      dz = z - BasePt.z;
      Pt.x = x;
      Pt.y = dy * _cosAng - dz * _sinAng + BasePt.y;
      Pt.z = dy * _sinAng + dz * _cosAng + BasePt.z;
      break;
    case EnPlane::GE_PLANE_LEFT:
      dy = BasePt.y - y;
      dz = z - BasePt.z;
      Pt.x = x;
      Pt.y = BasePt.y - (dy * _cosAng - dz * _sinAng);
      Pt.z = dy * _sinAng + dz * _cosAng + BasePt.z;
      break;
  }
  return Pt;
}


//-----------------------------------------------
void CGe3dPoint::Mirror (double a1, double b1, EnPlane Plane)
{
  double a2, b2, xp, yp, zp;

  a2 = -1.0 / a1;
  switch( Plane ){
    case EnPlane::GE_PLANE_TOP:
    case EnPlane::GE_PLANE_BOTTOM:
    default:
      b2 = y - x*a2;
      yp = (a1*b2 - a2*b1) / (a1-a2);
      xp = (yp - b1) / a1;
      x = xp + xp - x;
      y = yp + yp - y;
      break;
    case EnPlane::GE_PLANE_FRONT:
    case EnPlane::GE_PLANE_BACK:
      b2 = z - x*a2;
      zp = (a1*b2 - a2*b1) / (a1-a2);
      xp = (zp - b1) / a1;
      x = xp + xp - x;
      z = zp + zp - z;
      break;
    case EnPlane::GE_PLANE_RIGHT:
    case EnPlane::GE_PLANE_LEFT:
      b2 = z - y*a2;
      zp = (a1*b2 - a2*b1) / (a1-a2);
      yp = (zp - b1) / a1;
      y = yp + yp - y;
      z = zp + zp - z;
      break;
  }
}


//-----------------------------------------------
CGe3dPoint CGe3dPoint::MirrorBy (double a1, double b1, EnPlane Plane) const
{
  CGe3dPoint OutPt;
  double a2, b2, xp, yp;

  OutPt.Set(x,y,z);
  a2 = -1.0 / a1;
  switch( Plane ){
    case EnPlane::GE_PLANE_TOP:
    default:
      b2 = y - x*a2;
      yp = (a1*b2 - a2*b1) / (a1-a2);
      xp = (yp - b1) / a1;
      OutPt.x = xp + xp - x;
      OutPt.y = yp + yp - y;
      break;
    case EnPlane::GE_PLANE_BOTTOM:
      break;
    case EnPlane::GE_PLANE_FRONT:
      break;
    case EnPlane::GE_PLANE_BACK:
      break;
    case EnPlane::GE_PLANE_RIGHT:
      break;
    case EnPlane::GE_PLANE_LEFT:
      break;
  }
  return OutPt;
}


//-----------------------------------------------
bool CGe3dPoint::IsPlaneTop (double Delta) const 
{
  if (Delta > 0.0){
    if ((-Delta < x) && (x < Delta)){
      if ((-Delta < y) && (y < Delta)){
        if (((1.0 - Delta) < z) && (z < (1.0 + Delta))){
          return true;
        }
      }
    }
  }else{
    if (x==0.0 && y==0.0 && z==1.0){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
bool CGe3dPoint::IsPlaneBottom (double Delta) const 
{
  if (Delta > 0.0){
    if ((-Delta < x) && (x < Delta)){
      if ((-Delta < y) && (y < Delta)){
        if (((-1.0-Delta) < z) && (z < (-1.0+Delta))){
          return true;
        }
      }
    }
  }else{
    if (x==0.0 && y==0.0 && z==-1.0){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
bool CGe3dPoint::IsPlaneRight (double Delta) const 
{
  if (Delta > 0.0){
    if (((1.0-Delta) < x) && (x < (1.0+Delta))){
      if ((-Delta < y) && (y < Delta)){
        if ((-Delta < z) && (z < Delta)){
          return true;
        }
      }
    }
  }else{
    if (x==1.0 && y==0.0 && z==0.0){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
bool CGe3dPoint::IsPlaneLeft (double Delta) const 
{
  if (Delta > 0.0){
    if (((-1.0-Delta) < x) && (x < (-1.0+Delta))){
      if ((-Delta < y) && (y < Delta)){
        if ((-Delta < z) && (z < Delta)){
          return true;
        }
      }
    }
  }else{
    if (x==-1.0 && y==0.0 && z==0.0){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
bool CGe3dPoint::IsPlaneFront (double Delta) const 
{
  if (Delta > 0.0){
    if ((-Delta < y) && (y < Delta)){
      if (((1.0-Delta) < x) && (x < (1.0+Delta))){
        if ((-Delta < z) && (z < Delta)){
          return true;
        }
      }
    }
  }else{
    if (x==0.0 && y==1.0 && z==0.0){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
bool CGe3dPoint::IsPlaneBack (double Delta) const 
{
  if (Delta > 0.0){
    if ((-Delta < y) && (y < Delta)){
      if (((-1.0-Delta) < x) && (x < (-1.0+Delta))){
        if ((-Delta < z) && (z < Delta)){
          return true;
        }
      }
    }
  }else{
    if (x==0.0 && y==-1.0 && z==0.0){
      return true;
    }
  }
  return false;
}


//-----------------------------------------------
bool geGet3dPoint (const CGe3dPoint& A, const CGe3dPoint& B, double Z, CGe3dPoint* pOutPt)
{
  static double dX, dY, dZ, Coef;
 
  if ((A.z < Z) && (Z < B.z)){
    goto m1;
  }
  if ((B.z < Z) && (Z < A.z)){
    goto m1;
  }
  return false;
m1:
  dZ = A.z - B.z;
  Coef = (Z - B.z) / dZ;
  dX = A.x - B.x;
  pOutPt->x = B.x + (Coef * dX);
  dY = A.y - B.y;
  pOutPt->y = B.y + (Coef * dY);
  pOutPt->z = Z;
  return true;
}


