/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* This code was taken from 'BaseLibs' project
* Class CGeArc implementation
********************************************************************/
#include <math.h>
#include <stdlib.h>
#include <memory.h>
#include "Geom.h"

#define MIN_ARCANG 0.0001   // min arc angle

//-----------------------------------------------
CGeRect CGeArc::GeExtents () const
{
  CGeRect Rect;
  GetGeExtents( &Rect );
  return Rect; 
}

//-----------------------------------------------
void CGeArc::GetGeExtents (CGeRect* pRect) const
{
  CGePoint Pt1, Pt2, pnt;
  double  Ang, da;
  int     i;

  Pt1 = StartPoint();
  pRect->Set( Pt1, EndPoint() ); 
  da = m_AngArc / 5.0;
  Ang = m_AngStart;
  for (i=0; i<5; ++i){
    Ang += da;
    Pt2 = m_Center.PolarBy( Ang, m_Radius );
    if ((Pt1.y > m_Center.y && Pt2.y < m_Center.y) || (Pt2.y > m_Center.y && Pt1.y < m_Center.y)){
      if (Pt1.x < m_Center.x){
        pnt.Set( m_Center.x - m_Radius, m_Center.y );
        pRect->UpdateExtentsBy( pnt );
      }else{
        pnt.Set( m_Center.x + m_Radius, m_Center.y );
        pRect->UpdateExtentsBy( pnt );
      }
    }else{
      if ((Pt1.x > m_Center.x && Pt2.x < m_Center.x) || (Pt2.x > m_Center.x && Pt1.x < m_Center.x)){
        if (Pt1.y < m_Center.y){
          pnt.Set( m_Center.x, m_Center.y - m_Radius );
          pRect->UpdateExtentsBy( pnt );
        }else{
          pnt.Set( m_Center.x, m_Center.y + m_Radius );
          pRect->UpdateExtentsBy( pnt );
        }
      }
    }
    Pt1 = Pt2;
  }
  if (IsSector()){
    Ang = fabs( m_AngArc );
    if (Ang < GE_DEG180){
      pRect->UpdateExtentsBy( m_Center );
    }
  }
}

//-----------------------------------------------
bool CGeArc::Set (const CGePoint& Pnt, double R, double Ang0, double AngArc, bool bSec) 
{
  double Ang;
  if (R > 0.0){
    Ang = fabs(AngArc);
    if (MIN_ARCANG < Ang){
      m_Center = Pnt; 
      m_Radius = R; 
      m_AngStart = Ang0; 
      SetArcAngle( AngArc );
      m_bSector = bSec;
      return true;
    }
  }
  return false;
}

// define arc by 3 points (Start, Mid, End)
//-----------------------------------------------
bool CGeArc::Set3P (const CGePoint& Pt1, const CGePoint& Pt2, const CGePoint& Pt3)
{
  CGeLine    Line1, Line2;
  CGePoint   CenPt;
  EnIntersec Ret;
  double     Dir;

  Line1.Set( Pt1, Pt2 );
  Line1.Rot90ccw();
  Line2.Set( Pt2, Pt3 );
  Line2.Rot90ccw();
  Ret = Line1.GetIntersection( Line2, &CenPt );
  if ((Ret == EnIntersec::GE_SKEW_CROSS) || (Ret == EnIntersec::GE_SKEW_NO_CROSS)){
    m_Center = CenPt;
    m_Radius = CenPt.Distance( Pt1 );
    m_AngStart = m_Center.Angle( Pt1 );
    if (m_AngStart < 0.){
      m_AngStart += GE_2PI;
    }
    // define the arc direction
    Dir = (Pt1.x * (Pt2.y - Pt3.y)) + (Pt2.x * (Pt3.y - Pt1.y)) + (Pt3.x * (Pt1.y - Pt2.y));
    if (Dir < 0){
      // clockwise direction
      m_AngArc = -Pt1.LeftAngle( m_Center, Pt3 );
    }else{
      // counter-clockwise direction
      m_AngArc = Pt1.RightAngle( m_Center, Pt3 );
    }
    return true;
  }
  return false;
}


// set by Startpoint, Direction, Endpoint
//-----------------------------------------------
bool CGeArc::SetSDE (const CGePoint& Pt1, double DirAngle, const CGePoint& Pt2)
{
  double hord, Ang, Ang2;
  bool   bLeft;

  hord = Pt1.Distance( Pt2 );
  if (hord == 0.0){
    return false;
  }
  Ang = Pt1.Angle( Pt2 );
  // adjust angle to +/- PI range
  if (Ang > GE_PI){
    Ang -= GE_2PI;
  }
  // get turn angle
  Ang = Ang - DirAngle;
  // adjust angle to +/- PI range
  if (Ang < -GE_PI){
    Ang += GE_2PI;
  }else
  if (Ang > GE_PI){
    Ang -= GE_2PI;
  }
  if ((0.001 < Ang && Ang < (GE_PI-0.001)) ||
      (-0.001 > Ang && Ang > (-GE_PI+0.001)))
  {
    m_Radius = hord / 2.0 / sin( fabs(Ang) );
    bLeft = (Ang > 0.)? true : false;
    if (bLeft){
      Ang2 = DirAngle + GE_PI2;
    }else{
      Ang2 = DirAngle - GE_PI2;
    }
    m_Center.x = Pt1.x + (m_Radius * cos( Ang2 ));
    m_Center.y = Pt1.y + (m_Radius * sin( Ang2 ));
    m_AngStart = m_Center.Angle( Pt1 );
    m_AngArc = 2.0 * Ang;
    return true;
  }
  return false;
}


// set by Startpoint, Direction, Included angle, Radius
//-----------------------------------------------
bool CGeArc::SetSDAR (const CGePoint& Pt1, double DirAngle, double AngArc, double Radius)
{
  if ((AngArc != 0.0) && (Radius > 0.0)){
    if (AngArc < 0.0){
      m_AngStart = geNormalizeAngle( DirAngle + GE_DEG90 );
      m_Center = Pt1.PolarBy( m_AngStart, -Radius );
    }else{
      m_AngStart = geNormalizeAngle( DirAngle - GE_DEG90 );
      m_Center = Pt1.PolarBy( m_AngStart, -Radius );
    }
    m_Radius = Radius;
    m_AngArc = AngArc;
    return true;
  }
  return false;
}


// set by Startpoint, Bulge, Endpoint
//-----------------------------------------------
bool CGeArc::SetSBE (const CGePoint& Pt1, double Bulge, const CGePoint& Pt2)
{
  CGePoint ptm;
  double Ang;
  double len = Pt1.Distance( Pt2 );
  if ((len > 0.0) && (fabs(Bulge) > 0.001)){
    Ang = Pt1.Angle( Pt2 );
    len /= 2.0;
    ptm = Pt1.PolarBy( Ang, len );
    ptm.Polar( Ang - GE_PI2, len * Bulge );
    return Set3P( Pt1, ptm, Pt2 );
  }
  return false;
}

// set by Startpoint, Endpoint, Radius and Arc direction (cw/ccw)
//-----------------------------------------------
bool CGeArc::SetSER (const CGePoint& StartPt, const CGePoint& EndPt, double Radius, bool bClockwise, bool bLong)
{
  CGePoint mpt = StartPt.Center( EndPt );
  double  Ang;
  double  dd  = StartPt.Distance( EndPt ) / 2.0;
  double  delta = 0.00001;
  double  off;

  if (dd > 0.0){
    Ang = StartPt.Angle( EndPt );
    if (Radius < dd){
      Radius = dd;
    }
    m_Radius = Radius;
    if (((dd-delta) < m_Radius) && (m_Radius < (dd+delta))){
      m_Center = mpt;
      m_AngArc = (bClockwise)? -GE_DEG180 : GE_DEG180;
    }else{
      off = sqrt( (Radius * Radius) - (dd * dd) );
      if (bClockwise){
        if (bLong){
          m_Center = mpt.PolarBy( Ang + GE_DEG90, off );
        }else{
          m_Center = mpt.PolarBy( Ang - GE_DEG90, off );
        }
        m_AngArc = -StartPt.LeftAngle( m_Center, EndPt );
      }else{
        if (bLong){
          m_Center = mpt.PolarBy( Ang - GE_DEG90, off );
        }else{
          m_Center = mpt.PolarBy( Ang + GE_DEG90, off );
        }
        m_AngArc = StartPt.RightAngle( m_Center, EndPt );
      }
    }
    m_AngStart = m_Center.Angle( StartPt );
    return true;
  }
  return false;
}

// set by Start, End, arc Length 
//-----------------------------------------------
bool CGeArc::SetSEL (const CGePoint& StartPt, const CGePoint& EndPt, double ArcLen, bool bClockwise)
{
  UINT_PTR i=0;
  double Lch = StartPt.Distance( EndPt );
  double B, Delta, L0, L, DirAng0, DirAng, da=0.1;
  CGeArc Arc;
  CGePoint Pt0, Pt1;

  if ((Lch > 0.0) && (ArcLen > Lch)){
    Pt1.x = Lch;
    DirAng0 = 0.0;
    L0 = Lch;
    DirAng = DirAng0 + da;
    for(;;){
      ++i;  // counter of iterations
      Arc.SetSDE( Pt0, DirAng, Pt1 );
      L = Arc.Length();
      Delta = L - ArcLen;
      if (Delta > 0.0){
        if (Delta < 0.000001){
          B = fabs(Arc.Bulge());
          if (bClockwise){
            B = -B;
          }
          SetSBE( StartPt, B, EndPt );
          return true;
        }
        da /= 5.0;
        DirAng = DirAng0 + da;
      }else{
        DirAng0 = DirAng;
        DirAng += da;
      }
    }
  }
  return false;
}

// set by Start, End, Center 
//-----------------------------------------------
bool CGeArc::SetSEC (const CGePoint& Start, const CGePoint& End, const CGePoint& Center, bool bClockwise)
{
  double   Dist, Ang;
//  bool     bClockwise;
  CGePoint Pt0;
  CGeLine  Line( Start, End );

  Dist = Line.DistToPoint( Center );
  // Dist < 0. means Pnt is placed at left side of the line
  // Dist > 0. means Pnt is placed at right side of the line
  if (Dist != 0.0){
    Ang = Line.Angle();
    Pt0 = Line.Center();
//    bClockwise = (Dist > 0.0)? true : false;
    Pt0.Polar( Ang-GE_DEG90, Dist );
    if (SetCSE( Pt0, Start, End, bClockwise )){
      return true;
    }
  }
  return false;
}


// set by Start, End, included Angle 
//-----------------------------------------------
bool CGeArc::SetSEA (const CGePoint& Start, const CGePoint& End, double AngArc)
{
  CGePoint pnt;
  double  Ang, Ang2, dir, len, SEdist;
  if (AngArc == 0.){
    return false;
  }
  Ang = fabs( AngArc );
  if (Ang == GE_DEG180){
    pnt = Start.Center( End );
    return SetCSA( pnt, Start, AngArc );
  }
  if (Ang < GE_DEG360){
    pnt = Start.Center( End );
    SEdist = Start.Distance( End );
    dir = Start.Angle( End ) - GE_DEG90;
    if (Ang > GE_DEG180){
      Ang2 = GE_DEG90 - (GE_DEG360 - Ang) / 2.;
      len = SEdist / 2. * tan(Ang2);
      pnt.Polar( dir, (AngArc>0.)?len:-len );
      return SetCSA( pnt, Start, AngArc );
    }else{
      Ang2 = GE_DEG90 - Ang / 2.;
      len = SEdist / 2. * tan(Ang2);
      pnt.Polar( dir, (AngArc>0.)?-len:len );
      return SetCSA( pnt, Start, AngArc );
    }
  }
  return false;
}

// set by Start, End, Height from line Start-End 
//-----------------------------------------------
bool CGeArc::SetSEH (const CGePoint& Start, const CGePoint& End, double dH)
{
  double Bulg;
  double Len = Start.Distance( End );
  if (Len > 0.0){
    Bulg = dH / (Len / 2.0);
    return SetSBE( Start, Bulg, End );
  }
  return false;
}


// set by Center, Start, End 
//-----------------------------------------------
bool CGeArc::SetCSE (const CGePoint& Center, const CGePoint& Start, const CGePoint& End, bool bClockwise)
{
  double R, AngArc;
  R = Center.Distance( Start );
  if (R > 0.0){
    if (bClockwise){
      AngArc = Start.LeftAngle( Center, End );
    }else{
      AngArc = Start.RightAngle( Center, End );
    }
    if (AngArc > 0.0001){
      m_Center = Center;
      m_Radius = R;
      m_AngStart = Center.Angle( Start );
      m_AngArc = (bClockwise)? -AngArc : AngArc;
      return true;
    }
  }
  return false;
}


// set by Center, Start, included Angle 
//-----------------------------------------------
bool CGeArc::SetCSA (const CGePoint& Center, const CGePoint& Start, double AngArc)
{
  double R = Center.Distance( Start );
  if ((R > 0.0) && (AngArc != 0.0)){
    m_Center = Center;
    m_Radius = R;
    m_AngStart = Center.Angle( Start );
    m_AngArc = AngArc;
    return true;
  }
  return false;
}


// set by Center, Start, chord Length
//-----------------------------------------------
bool CGeArc::SetCSL (const CGePoint& Center, const CGePoint& Start, double ChLen, bool bClockwise)
{
  double coef, cosang;
  double R = Center.Distance( Start );
  double L = fabs( ChLen );

  if (R > 0.0){
    if ((L > (R+R)) || (L == 0.0)){
      L = R + R;
    }
    m_Center = Center;
    m_Radius = R;
    m_AngStart = Center.Angle( Start );
    if (L == (R+R)){
      if (bClockwise){
        m_AngArc = -GE_DEG180;
      }else{
        m_AngArc = GE_DEG180;
      }
    }else{
      coef = L / R / 2.0;
      cosang = sqrt( 1.0 - (coef * coef));
      m_AngArc = acos(cosang) * 2.0;
      if (bClockwise){
        m_AngArc = GE_DEG360 - m_AngArc;
      }
    }
    return true;
  }
  return false;
}

//-----------------------------------------------
bool CGeArc::SetCRAA (const CGePoint& Pnt, double R, double AngS, double AngE, BOOL bClockwise, bool bSec)
{
  double Ang0, Ang1;
  if (R > 0.0){
    Ang0 = geNormalizeAngle( AngS ); 
    Ang1 = geNormalizeAngle( AngE ); 
    if (bClockwise){
      m_AngArc = geNormalizeAngle( Ang0 - Ang1 );
    }else{
      m_AngArc = geNormalizeAngle( Ang1 - Ang0 );
    }
    if (MIN_ARCANG < m_AngArc){
      m_Center = Pnt; 
      m_Radius = R; 
      m_AngStart = Ang0; 
      if (bClockwise){
        m_AngArc = -m_AngArc;
      }
      m_bSector = bSec;
      return true;
    }
  }
  return false;
}


//-----------------------------------------------
void CGeArc::SetStartAngle (double Angle)
{
  m_AngStart = geNormalizeAngle( Angle );
}


//-----------------------------------------------
bool CGeArc::SetArcAngle (double Angle)
{
  double Ang = fabs(Angle);
  if (MIN_ARCANG < Ang){
    if (Ang <= GE_2PI){
      m_AngArc = Angle;
      return true;
    }else{
      while( Ang > GE_2PI){
        Ang -= GE_2PI;
      }
      m_AngArc = (Angle > 0.0)? Ang : -Ang;
      return true;
    }
  }
  return false;
}


//-----------------------------------------------
UINT_PTR CGeArc::GetPoints (CGePoint* Pnt, double AngStep, UINT_PTR MaxNumPoints) const
{
  double    Ang = 0;
  double    Ang2;
  double    MaxAng = fabs(m_AngArc);
  double    x0 = m_Center.x;
  double    y0 = m_Center.y;
  bool      bCW = (m_AngArc < 0.0)?true:false;
  CGeLine   line1, line2, line1c, line2c;
  UINT_PTR  i = 0;
  CGePoint* pPnt = Pnt;

  MaxAng -= MIN_ARCANG / 2.0;
  pPnt->Set( StartPoint() );
  ++pPnt;
  for ( ; ; ){ //while( 1 ){
    ++i;
    Ang += AngStep;
    if (Ang > MaxAng){
      pPnt->Set( EndPoint() );
      ++pPnt;
      ++i;
      break;
    }
    if (i >= MaxNumPoints){
      return i;
    }
    if (bCW){
      Ang2 = m_AngStart - Ang;
    }else{
      Ang2 = m_AngStart + Ang;
    }
    pPnt->x = x0 + (m_Radius * cos( Ang2 ));
    pPnt->y = y0 + (m_Radius * sin( Ang2 ));
    ++pPnt;
  }
  if (m_bSector){
    pPnt->x = x0;
    pPnt->y = y0;
    ++i;
  }
  return i;
}

//-----------------------------------------------
void CGeArc::GetPointsOld (CGePoint* Pnt, UINT_PTR NumPoints) const
{
  double    AngStep;
  double    Ang = m_AngStart;
  double    x0 = m_Center.x;
  double    y0 = m_Center.y;
  UINT_PTR  i = 1;
  CGeLine   line1, line2, line1c, line2c;
  CGePoint* pPnt = Pnt;

  if (NumPoints < 3){
    NumPoints = 3;
  }
  if (m_bSector){
    --NumPoints;
  }
  pPnt->Set( StartPoint() );
  ++pPnt;
  AngStep = m_AngArc / (NumPoints - 1);  // angle step
  for (i=1; i<NumPoints; ++i){
    Ang += AngStep;
    pPnt->x = x0 + (m_Radius * cos( Ang ));
    pPnt->y = y0 + (m_Radius * sin( Ang ));
    ++pPnt;
  }
  if (m_bSector){
    pPnt->x = x0;
    pPnt->y = y0;
  }
}

// Retrieves on-arc points
// MaxNumPoints - number of points for 360 degrees
// can be only MaxNumPoints = 16, 32, 64, 128, 256, 512
//-----------------------------------------------
UINT_PTR CGeArc::GetPointsA (CGePoint* Pnt, UINT_PTR MaxNumPoints, bool bMin5p) const
{
  CArcPrm   Prm;
  UINT_PTR  nPnts, i;
  double    dval, Ang, AngStep, x0, y0;
  CGePoint* pPnt;

  switch( MaxNumPoints ){
    case 16:  Prm.m_Resolution = GE_ARCRES_16; break;
    case 32:  Prm.m_Resolution = GE_ARCRES_32; break;
    case 64:  Prm.m_Resolution = GE_ARCRES_64; break;
    case 128: Prm.m_Resolution = GE_ARCRES_128; break;
    case 256: Prm.m_Resolution = GE_ARCRES_256; break;
    case 512: Prm.m_Resolution = GE_ARCRES_512; break;
    default: 
      Prm.m_Resolution = GE_ARCRES_64; 
      MaxNumPoints = 64;
      break;
  }
  Prm.m_Center = m_Center;
  Prm.m_Radius = m_Radius;
  Prm.m_Ang0 = m_AngStart;
  if (m_AngArc > 0.0){
    Prm.m_bDirCW = false;
    Prm.m_AngArc = m_AngArc;
  }else{
    Prm.m_bDirCW = true;
    Prm.m_AngArc = -m_AngArc;
  }
  if (bMin5p){
    // add extra points if the bulge is too low (less then 5 points)
    dval = (double)MaxNumPoints / (double)GE_DEG360 * Prm.m_AngArc;
    if (dval < 5){
      nPnts = 5;
      pPnt = Pnt;
      pPnt->Set( StartPoint() ); 
      ++pPnt;
      x0 = m_Center.x;
      y0 = m_Center.y;
      AngStep = m_AngArc / ((double)nPnts - 1.0);  // angle step
      Ang = m_AngStart;
      for (i=1; i<nPnts; ++i){
        Ang += AngStep;
        pPnt->x = x0 + (m_Radius * cos( Ang ));
        pPnt->y = y0 + (m_Radius * sin( Ang ));
        ++pPnt;
      }
      goto m1;
    }
  }
  nPnts = geFastArc( Prm, Pnt );
m1:
  if (m_bSector){
    Pnt[nPnts] = m_Center;
    ++nPnts;
  }
  return nPnts;
}

//-----------------------------------------------
UINT_PTR CGeArc::GetPolylineA (CGePolyline* pPline, UINT_PTR MaxNumPoints, bool bMin5p) const
{
  static CGePoint Pnt[514];
  UINT_PTR nPnts;

  if (MaxNumPoints < 16){
    MaxNumPoints = 16;
  }else{
    if (MaxNumPoints > 512){
      MaxNumPoints = 512;
    }
  }
  nPnts = GetPointsA( Pnt, MaxNumPoints, bMin5p );
  pPline->SetVers( Pnt, nPnts, IsSector() );
  return nPnts;
}



// Retrieves on-arc point, which is placed on specified distance
// from arc startpoint.
//-----------------------------------------------
void CGeArc::GetPointByDist (double Dist, CGePoint* pPnt, double* pAngle, double* pDirAngle) const
{
  double Ang = Dist / m_Radius;
  if (m_AngArc > 0.0){
    Ang = m_AngStart + Ang;
  }else{
    Ang = m_AngStart - Ang;
  }
//  Ang = geNormalizeAngle( Ang );
  if (pAngle){
    *pAngle = geNormalizeAngle( Ang );
  }
  if (pDirAngle){
    if (IsCW()){
      *pDirAngle = geNormalizeAngle( Ang-GE_DEG90 );
    }else{
      *pDirAngle = geNormalizeAngle( Ang+GE_DEG90 );
    }
  }
  if (pPnt){
    pPnt->x = m_Center.x + (m_Radius * cos( Ang ));
    pPnt->y = m_Center.y + (m_Radius * sin( Ang ));
  }
}

// Retrieves on-arc point, which is placed on specified angle
// from arc centerpoint.
// Returns false if found point is outside arc's included angle
//-----------------------------------------------
bool CGeArc::GetPointByAngle (double Angle, CGePoint* pPnt, double* pDist) const
{
  CGePoint Pt0 = StartPoint();
  double  Ang2;

  pPnt->x = m_Center.x + m_Radius * cos( Angle );
  pPnt->y = m_Center.y + m_Radius * sin( Angle );
  if (m_AngArc > 0){
    Ang2 = Pt0.RightAngle( m_Center, *pPnt );
    if (pDist){
      *pDist = m_Radius * Ang2;
    }
    if (Ang2 < m_AngArc){
      return true;
    }
  }else{
    Ang2 = Pt0.LeftAngle( m_Center, *pPnt );
    if (pDist){
      *pDist = m_Radius * Ang2;
    }
    if (Ang2 < -m_AngArc){
      return true;
    }
  }
  return false;
}


// Returns a distanse along arc from startpoint
//-----------------------------------------------
double CGeArc::GetDistByPoint (const CGePoint& Pnt, CGePoint* pArcPt) const
{
  double Dist, Ang;
  CGePoint StartPt = StartPoint();
  if (m_AngArc > 0.0){
    Ang = StartPt.RightAngle( m_Center, Pnt );
  }else{
    Ang = StartPt.LeftAngle( m_Center, Pnt );
  }
  Dist = m_Radius * Ang;
  if (pArcPt){
    Ang = m_Center.Angle( Pnt );
    *pArcPt = m_Center.PolarBy( Ang, m_Radius );
  }
  return Dist;
}

//-----------------------------------------------
double CGeArc::GetNearPoint (const CGePoint& Pnt, CGePoint* pPntNear, double* pDistFromStart) const
{
  double Dist, D2p, Ang, d0, d1;
  CGePoint Pt2, EndPt;
  CGePoint StartPt = StartPoint();
  Dist = -1.0;
  if (m_AngArc > 0.0){
    Ang = StartPt.RightAngle( m_Center, Pnt );
    if (Ang < m_AngArc){
      Dist = m_Radius * Ang;
    }
  }else{
    Ang = StartPt.LeftAngle( m_Center, Pnt );
    if (Ang < -m_AngArc){
      Dist = m_Radius * Ang;
    }
  }
  if (Dist > 0.0){
    Ang = m_Center.Angle( Pnt );
    Pt2 = m_Center.PolarBy( Ang, m_Radius );
  }else{
    EndPt = EndPoint();
    d0 = Pnt.Distance( StartPt );
    d1 = Pnt.Distance( EndPt );
    if (d0 < d1){
      Dist = 0.0;
      Pt2 = StartPt;
    }else{
      Dist = Length();
      Pt2 = EndPt;
    }
  }
  if (pPntNear){
    *pPntNear = Pt2;
  }
  if (*pDistFromStart){
    *pDistFromStart = Dist;
  }
  // distance between Pnt and found point (Pt2)
  D2p = Pnt.Distance( Pt2 ); 
  return D2p;
}


//-----------------------------------------------
int CGeArc::GetTangentPoints (const CGePoint& BasePt, CGePoint* Pnt) const
{
  CGePoint pt;
  int     i=0;
  double  dist, d2, Ang, Ang0, rad;

  dist = m_Center.Distance( BasePt );
  if (dist > m_Radius){
    rad = m_Radius * 0.9999;
    d2 = sqrt( dist*dist - rad*rad );
    Ang = atan( rad / d2 );
    Ang0 = BasePt.Angle( m_Center );
    pt = BasePt.PolarBy( Ang0 + Ang, d2 );
    if (IsArcPoint( pt )){
      Pnt[i] = pt;
      ++i;
    }
    pt = BasePt.PolarBy( Ang0 - Ang, d2 );
    if (IsArcPoint( pt )){
      Pnt[i] = pt;
      ++i;
    }
    return i;
  }
  return 0;
}

// Retrieves an array of points along the line
// nParts - number of parts the line to be divided, >= 2
// bStartEnd - if TRUE, then start & end points are included in pOutPnt array
// Return: number of points in the pOutPnt array
//-----------------------------------------------
UINT_PTR CGeArc::Divide (UINT_PTR nParts, CGePoint* pOutPnt, bool bStartEnd) const
{
  UINT_PTR nPnts = 0;
  UINT_PTR i;
  double   Len, Dpart, Dist;

  if ((nParts < 2) || (1000 < nParts)){
    return 0;
  }
  if (bStartEnd){
    *pOutPnt = StartPoint();
    ++pOutPnt;
    ++nPnts;
  }
  Len = Length();
  Dpart = Len / (double)nParts;
  Dist = Dpart;
  for (i=1; i<nParts; ++i){
    GetPointByDist( Dist, pOutPnt, NULL );
    Dist += Dpart;
    ++pOutPnt;
    ++nPnts;
  }
  if (bStartEnd){
    *pOutPnt = EndPoint();
    ++nPnts;
  }
  return nPnts;
}


// test is the point are placed inside of the arc included angle
//-----------------------------------------------
bool CGeArc::IsArcPoint (const CGePoint& Pnt) const
{
  CGePoint Pt0 = StartPoint();
  double  Ang2;

  if (m_AngArc > 0.0){
    Ang2 = Pt0.RightAngle( m_Center, Pnt );
    if (Ang2 < m_AngArc){
      return true;
    }
  }else{
    Ang2 = Pt0.LeftAngle( m_Center, Pnt );
    if (Ang2 < -m_AngArc){
      return true;
    }
  }
  return false;
}

// test is the point are placed inside of the arc included angle
// and within dR to radius
//-----------------------------------------------
bool CGeArc::IsArcPoint2 (const CGePoint& Pnt, double dR) const
{
  double Dist;
  if (IsArcPoint( Pnt )){
    Dist = m_Center.Distance( Pnt );
    if ((Dist-dR)<=m_Radius && m_Radius<=(Dist+dR)){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
// Returns a distance from specified point to this arc
// Dist < 0. means Pnt is placed at left side of the arc
// Dist > 0. means Pnt is placed at right side of the arc
//-----------------------------------------------
double CGeArc::DistToPoint (const CGePoint& Pnt, bool* pInsideArcAngle) const
{
  bool bInside, bNeg;
  double Dist, d0, d1;

  Dist = m_Center.Distance( Pnt );
  if (Dist > m_Radius){
    if (IsCW()){
      Dist = m_Radius - Dist;
    }else{
      Dist = Dist - m_Radius;
    }
  }else{
    if (IsCCW()){
      Dist = Dist - m_Radius;
    }else{
      Dist = m_Radius - Dist;
    }
  }
  bInside = IsArcPoint( Pnt );
  if (pInsideArcAngle){
    *pInsideArcAngle = bInside;
  }
  if (!bInside){
    bNeg = (Dist < 0.0)? true : false;
    d0 = StartPoint().Distance( Pnt );
    d1 = EndPoint().Distance( Pnt );
    Dist = __min( d0, d1 );
    if (bNeg){
      Dist = -Dist;
    }
  }
  return Dist;
}



//-----------------------------------------------
double CGeArc::Bulge () const
{
  double H = m_Radius * (1.0 - cos(m_AngArc/2.0));
  double Len = m_Radius * sin(m_AngArc/2.0);
  double B = fabs( H / Len );
  if (m_AngArc < 0.0){
    B = -B;
  }
  return B;
}


//-----------------------------------------------
double CGeArc::EndAngle () const 
{
  double Ang = m_AngStart + m_AngArc;
  // normalize to the range 0..2PI
  if (Ang < 0.0){
    Ang += GE_2PI;
  }else{
    if (Ang > GE_2PI){
      Ang -= GE_2PI;
    }
  }
  return Ang;
}


//-----------------------------------------------
double CGeArc::StartTangent () const
{
  double Ang;
  if (m_AngArc < 0.0){
    Ang = m_AngStart + GE_PI2;
  }else{
    Ang = m_AngStart - GE_PI2;
  }
  // normalize to the range 0..2PI
  if (Ang < 0.0){
    Ang += GE_2PI;
  }else{
    if (Ang > GE_2PI){
      Ang -= GE_2PI;
    }
  }
  return Ang;
}


//-----------------------------------------------
double CGeArc::EndTangent () const
{
  double Ang;
  if (m_AngArc < 0.0){
    Ang = EndAngle() - GE_PI2;
  }else{
    Ang = EndAngle() + GE_PI2;
  }
  // normalize to the range 0..2PI
  if (Ang < 0.){
    Ang += GE_2PI;
  }else{
    if (Ang > GE_2PI){
      Ang -= GE_2PI;
    }
  }
  return Ang;
}


//-----------------------------------------------
double CGeArc::ChordLen () const
{
  double len, ca;
  double Ang = fabs(m_AngArc);
  if (Ang > GE_DEG360){
    Ang = geNormalizeAngle( Ang );
  }
  if (Ang == 0.){
    return 0.;
  }
  if (Ang == GE_DEG180){
    return m_Radius + m_Radius;
  }
  if (Ang > GE_DEG180){
    Ang = GE_DEG360 - Ang;
  }
  ca = cos(Ang/2.);
  len = m_Radius * sqrt( 1 - ca*ca ) * 2.;
  return len;
}

//-----------------------------------------------
double CGeArc::Length () const 
{
  return fabs(m_AngArc) * m_Radius;
}

//-----------------------------------------------
double CGeArc::Height () const 
{
  double B = Bulge();
  if (B == 0.0){
    return 0.0;
  }
  return ChordLen() / 2.0 * fabs(B);
}

//-----------------------------------------------
double CGeArc::Area () const
{
  double Ang2 = fabs(m_AngArc) / 2.;
  double r2 = m_Radius * m_Radius;
  double ar, ar2;
  double ar1 = Ang2 * r2;
  if (Ang2 < GE_PI2){
    ar2 = r2 * sin(Ang2) * cos(Ang2);
    ar = ar1 - ar2;
  }else{
    Ang2 = (GE_2PI - fabs(m_AngArc)) / 2.;
    ar2 = r2 * sin(Ang2) * cos(Ang2);
    ar = ar1 + ar2;
  }
  return ar; 
}

/*
//-----------------------------------------------
int CGeArc::ClipBy (const CGeRect& rect, VcGeLine2d* pOutLine) const
{
  return rect.Clip( *this, pOutLine );
}
*/


//-----------------------------------------------
bool CGeArc::Contains (const CGePoint& pnt) const
{
  CGePoint StartPt;
  double  Ang;
  if (m_bSector){
    if (m_Center.Distance( pnt ) < m_Radius){
      StartPt = StartPoint();
      if (m_AngArc > 0.){
        Ang = StartPt.RightAngle( m_Center, pnt );
        if (Ang < m_AngArc){
          return true;
        }
      }else{
        if (m_AngArc < 0.){
          Ang = StartPt.LeftAngle( m_Center, pnt );
          if (Ang < -m_AngArc){
            return true;
          }
        }
      }
    }
  }
  return false;
}


//-----------------------------------------------
bool CGeArc::Contains (const CGeRect& rect) const
{
  if (Contains( rect.LefBot() )){
    if (Contains( rect.LefTop() )){
      if (Contains( rect.RigBot() )){
        if (Contains( rect.RigTop() )){
          return true;
        }
      }
    }
  }
  return false;
}




#define NPTCROSS 256
//-----------------------------------------------
bool CGeArc::Crosses (const CGeRect& Rect, bool bPoint) const
{
  UINT_PTR i, n;
  CGePoint Pnt[NPTCROSS+2];
  CGePoint PickPnt;
  CGeLine  Line;
  double   Dist, Delta;

  if (bPoint){
    PickPnt = Rect.Center();
    Delta = Rect.Height() * 0.6;
    Dist = m_Center.Distance( PickPnt );
    if (((Dist-Delta) < m_Radius) && (m_Radius < (Dist+Delta))){
      if (IsArcPoint( PickPnt )){
        return true;
      }
    }
    if (m_bSector){
      Line.Set( m_Center, StartPoint() );
      if (Rect.Crosses( Line )){
        return true;
      }
      Line.Set( m_Center, EndPoint() );
      if (Rect.Crosses( Line )){
        return true;
      }
    }
    return false;
  }

  n = GetPointsA( Pnt, NPTCROSS );
  for (i=1; i<n; ++i){
    Line.Set( Pnt[i-1], Pnt[i] );
    if (Rect.Crosses( Line )){
      return true;
    }
  }
  if (m_bSector){
    Line.Set( Pnt[n-1], Pnt[0] );
    if (Rect.Crosses( Line )){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
void CGeArc::Move (const CGePoint& Off)
{
  m_Center.Move( Off.x, Off.y );
}
//-----------------------------------------------
void CGeArc::Move (double dx, double dy)
{
  m_Center.Move( dx, dy );
}

//-----------------------------------------------
void CGeArc::Rotate (const CGePoint& Pnt0, double Angle)
{
  m_Center.Rotate( Pnt0, Angle );
  m_AngStart = geNormalizeAngle( m_AngStart + Angle );
}

//-----------------------------------------------
void CGeArc::Scale (const CGePoint& Pnt0, double ScaleX, double ScaleY)
{
  CGePoint Pt[3];
  double Dir, R1;
  if ((ScaleX == 0.0) || (ScaleY == 0.0)){
    return;
  }
  if ((ScaleX > 0.0) && (ScaleY > 0.0)){
    m_Center.Scale( Pnt0, ScaleX, ScaleY );
    if (ScaleX < ScaleY){
      m_Radius *= ScaleX;
    }else{
      m_Radius *= ScaleY;
    }
  }else{
    Pt[0] = StartPoint();
    Pt[1] = MidPoint();
    Pt[2] = EndPoint();
    m_Center.Scale( Pnt0, ScaleX, ScaleY );
    Pt[0].Scale( Pnt0, ScaleX, ScaleY );
    Pt[1].Scale( Pnt0, ScaleX, ScaleY );
    Pt[2].Scale( Pnt0, ScaleX, ScaleY );
    R1 = m_Center.Distance( Pt[0] );
    m_Radius = R1;
    m_AngStart = m_Center.Angle( Pt[0] );
    // get arc direction
    Dir = geRoundValue( Pt[0], Pt[1], Pt[2] );
    // set arc included angle
    if (Dir > 0.0){
      m_AngArc = Pt[0].RightAngle( m_Center, Pt[2] );
    }else{
      m_AngArc = -Pt[0].LeftAngle( m_Center, Pt[2] );
    }
  }
}

//-----------------------------------------------
void CGeArc::Mirror (const CGeLine& Line)
{
  CGePoint pt = m_Center.PolarBy( m_AngStart, m_Radius );
  m_Center.Mirror( Line );
  pt.Mirror( Line );
  m_AngStart = m_Center.Angle( pt );
  m_AngArc = -m_AngArc;
}

//-----------------------------------------------
void CGeArc::Reverse ()
{
  m_AngStart = EndAngle();
  m_AngArc = -m_AngArc;
}

//-----------------------------------------------
void CGeArc::Inverse ()
{
  m_AngStart = EndAngle();
  if (m_AngArc > 0.0){
    m_AngArc = GE_2PI - m_AngArc;
  }else{
    m_AngArc = -m_AngArc - GE_2PI;
  }
}

//-----------------------------------------------
bool CGeArc::Transform (const GE_TRANSPRM* pPrm)
{
  CGePoint Pt2;
  if (geTransHasUniscale( pPrm )){
    Pt2 = StartPoint();
    m_Center.Transform( pPrm );
    Pt2.Transform( pPrm );
    m_Radius = m_Center.Distance( Pt2 );
    m_AngStart = m_Center.Angle( Pt2 );
    return true;
  }
  return false;
}


//-----------------------------------------------
void CGeArc::IncreaseArcLength (double Delta)
{
  double AngDelta = Delta / m_Radius;
  double AngArc = fabs(m_AngArc) + AngDelta + AngDelta;
  if (AngArc < GE_2PI){
    if (m_AngArc < 0.0){
      SetStartAngle( m_AngStart + AngDelta );
      SetArcAngle( -AngArc );
    }else{
      SetStartAngle( m_AngStart - AngDelta );
      SetArcAngle( AngArc );
    }
  }
}

//-----------------------------------------------
void CGeArc::Save (BYTE* pBuf) const
{
  int sz;
  BYTE* ptr = pBuf;
  sz = sizeof(CGePoint);
  memcpy( ptr, &m_Center, sz );  ptr += sz;
  sz = sizeof(double);
  memcpy( ptr, &m_Radius, sz );  ptr += sz;
  memcpy( ptr, &m_AngStart, sz );  ptr += sz;
  memcpy( ptr, &m_AngArc, sz );  ptr += sz;
  sz = sizeof(bool);
  memcpy( ptr, &m_bSector, sz );  
}

//-----------------------------------------------
void CGeArc::Restore (const BYTE* pBuf)
{
  int sz;
  const BYTE* ptr = pBuf;
  sz = sizeof(CGePoint);
  memcpy( &m_Center, ptr, sz );  ptr += sz;
  sz = sizeof(double);
  memcpy( &m_Radius, ptr, sz );  ptr += sz;
  memcpy( &m_AngStart, ptr, sz );  ptr += sz;
  memcpy( &m_AngArc, ptr, sz );  ptr += sz;
  sz = sizeof(bool);
  memcpy( &m_bSector, ptr, sz );  
}


