/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* This code was taken from 'BaseLibs' project
* Class CGeCircle implementation
********************************************************************/
#include <math.h>
#include <stdlib.h>
#include <memory.h>
#include "Geom.h"


//-----------------------------------------------
bool CGeCircle::Set3P (const CGePoint& Pt1, const CGePoint& Pt2, const CGePoint& Pt3)
{
  CGeLine  Line1, Line2;
  CGePoint CenPt;

  if (Pt1.IsEqual( Pt2 ) || Pt1.IsEqual( Pt3 ) || Pt2.IsEqual( Pt3 )){
    return false;
  }
  Line1.Set( Pt1, Pt2 );
  Line1.Rot90ccw();
  Line2.Set( Pt2, Pt3 );
  Line2.Rot90ccw();
  if (Line1.GetInterPoint( Line2, &CenPt, 3 )){
    m_Center = CenPt;
    m_Radius = m_Center.Distance( Pt1 );
    return true;
  }
  return false;
}

//-----------------------------------------------
bool CGeCircle::Set2P (const CGePoint& Pt1, const CGePoint& Pt2)
{
  if (Pt1.IsEqual( Pt2 )){
    return false;
  }
  m_Center = Pt1.Center( Pt2 );
  m_Radius = m_Center.Distance( Pt1 );
  return true;
}

//-----------------------------------------------
bool CGeCircle::SetRadius (double R) 
{
  if (R > 0.0){
    m_Radius = R;
    return true;
  }
  return false;
}

//-----------------------------------------------
bool CGeCircle::SetDiameter (double D) 
{
  if (D > 0.0){
    m_Radius = D/2.0;
    return true;
  }
  return false;
}

//-----------------------------------------------
bool CGeCircle::SetLength (double Len)
{
  if (Len > 0.0){
    m_Radius = Len / GE_2PI;
    return true;
  }
  return false;
}


//-----------------------------------------------
bool CGeCircle::SetArea (double Area)
{
  if (Area > 0.0){
    m_Radius = sqrt( Area / GE_PI );
    return true;
  }
  return false;
}


//-----------------------------------------------
bool CGeCircle::Contains (const CGePoint& pnt) const
{
  if (m_Center.Distance( pnt ) < m_Radius){
    return true;
  }
  return false;
}


//-----------------------------------------------
bool CGeCircle::Contains (const CGeRect& rect) const
{
  if (m_Center.Distance( rect.LefBot() ) < m_Radius){
    if (m_Center.Distance( rect.LefTop() ) < m_Radius){
      if (m_Center.Distance( rect.RigBot() ) < m_Radius){
        if (m_Center.Distance( rect.RigTop() ) < m_Radius){
           return true;
        }
      }
    }
  }
  return false;
}

#define NCROSSPTS 256
//-----------------------------------------------
bool CGeCircle::Crosses (const CGeRect& Rect, bool bPoint) const
{
  INT_PTR   val = 0;
  UINT_PTR  i;
  CGePoint  Pnt[NCROSSPTS+1];
  CGePoint* pPnt0;
  CGePoint* pPnt1;
  CGePoint  PickPnt;
  CGeLine   Line;
  double    Dist, Delta;

  if (bPoint){
    PickPnt = Rect.Center();
    Delta = Rect.Height() * 0.6;
    Dist = m_Center.Distance( PickPnt );
    if (((Dist-Delta) < m_Radius) && (m_Radius < (Dist+Delta))){
      return true;
    }
    return false;
  }
  if (m_Center.Distance( Rect.LefBot() ) >= m_Radius){
    ++val;
  }else{
    --val;
  }
  if (m_Center.Distance( Rect.LefTop() ) >= m_Radius){
    ++val;
  }else{
    --val;
  }
  if (m_Center.Distance( Rect.RigBot() ) >= m_Radius){
    ++val;
  }else{
    --val;
  }
  if (m_Center.Distance( Rect.RigTop() ) >= m_Radius){
    ++val;
  }else{
    --val;
  }
  if ((-4 < val) && (val < 4)){
    return true;
  }
  GetPoints( Pnt, NCROSSPTS );
  pPnt0 = Pnt;
  pPnt1 = Pnt+1;
  Pnt[NCROSSPTS] = *pPnt0;
  for (i=0; i<NCROSSPTS; ++i){
    Line.Set( *pPnt0, *pPnt1 );
    if (Rect.Crosses( Line )){
      return true;
    }
    ++pPnt0;
    ++pPnt1;
  }
  return false;
}

//-----------------------------------------------
UINT_PTR CGeCircle::Divide (double Ang0, bool bClockwise, UINT_PTR nParts, CGePoint* pOutPnt) const
{
  UINT_PTR nPnts = 0;
  UINT_PTR i;
  double   Dang, Ang;

  if ((nParts < 2) || (1000 < nParts)){
    return 0;
  }
  Dang = GE_2PI / (double)nParts;
  if (bClockwise){
    Dang = -Dang;
  }
  Ang = Ang0;
  for (i=0; i<nParts; ++i){
    pOutPnt->x = m_Center.x + (m_Radius * cos( Ang ));
    pOutPnt->y = m_Center.y + (m_Radius * sin( Ang ));
    Ang += Dang;
    ++pOutPnt;
    ++nPnts;
  }
  return nPnts;
}

//-----------------------------------------------
void CGeCircle::Move (const CGePoint& Off)
{
  m_Center.Move( Off.x, Off.y );
}
//-----------------------------------------------
void CGeCircle::Move (double dx, double dy)
{
  m_Center.Move( dx, dy );
}

//-----------------------------------------------
void CGeCircle::Rotate (const CGePoint& Pnt0, double Angle)
{
  m_Center.Rotate( Pnt0, Angle );
}

//-----------------------------------------------
void CGeCircle::Scale (const CGePoint& Pnt0, double ScaleX, double ScaleY)
{
  double asx, asy;

  if ((ScaleX != 0.0) && (ScaleY != 0.0)){
    m_Center.Scale( Pnt0, ScaleX, ScaleY );
    if (ScaleX == ScaleY){
      if (ScaleX < 0.0){
        ScaleX = -ScaleX;
      }
      m_Radius *= ScaleX;
    }else{
      asx = fabs(ScaleX);
      asy = fabs(ScaleY);
      if (asx < asy){
        m_Radius *= asx;
      }else{
        m_Radius *= asy;
      }
    }
  }
}

//-----------------------------------------------
void CGeCircle::Mirror (const CGeLine& Line)
{
  m_Center.Mirror( Line );
}

//-----------------------------------------------
bool CGeCircle::Transform (const GE_TRANSPRM* pPrm)
{
  CGePoint Pt2;
  if (geTransHasUniscale( pPrm )){
    Pt2.x = m_Center.x + m_Radius;
    Pt2.y = m_Center.y;
    m_Center.Transform( pPrm );
    Pt2.Transform( pPrm );
    m_Radius = m_Center.Distance( Pt2 );
    return true;
  }
  return false;
}


// 2 tangent points on circle from BasePt
//-----------------------------------------------
bool CGeCircle::GetTangentPoints (const CGePoint& BasePt, CGePoint* pPnt1, CGePoint* pPnt2) const
{ 
  double  dist, d2, x, y, ang, ang0, sinAng, cosAng;

  dist = m_Center.Distance( BasePt );
  if (dist > m_Radius){
/*
    rad = m_Radius * 0.9999;
    d2 = sqrt( dist*dist - rad*rad );
    ang = atan( rad / d2 );
    ang0 = BasePt.Angle( m_Center );
    *pPnt1 = BasePt.PolarBy( ang0 + ang, d2 );
    *pPnt2 = BasePt.PolarBy( ang0 - ang, d2 );
*/
    d2 = sqrt( dist*dist - m_Radius*m_Radius );  // dist from BasePt to tangent point
    ang = atan( m_Radius / d2 );      // angle between lines <BasePt,m_Center> and <BasePt,tangent point>
    cosAng = cos( ang );
    x = d2 * cosAng;
    y = m_Radius * cosAng; 
    ang0 = BasePt.Angle( m_Center );
    sinAng = sin( ang0 );
    cosAng = cos( ang0 );
    pPnt1->x = (x * cosAng) - (y * sinAng) + BasePt.x;
    pPnt1->y = (x * sinAng) + (y * cosAng) + BasePt.y;
    pPnt2->x = (x * cosAng) + (y * sinAng) + BasePt.x;
    pPnt2->y = (x * sinAng) - (y * cosAng) + BasePt.y;
    return true;
  }
  return false;
}

//-----------------------------------------------
void CGeCircle::GetPoint (double Ang, CGePoint* pPnt) const
{
  pPnt->x = m_Center.x + (m_Radius * cos( Ang ));
  pPnt->y = m_Center.y + (m_Radius * sin( Ang ));
}

//-----------------------------------------------
void CGeCircle::GetPoint (double Ang, double Ang0, bool bClockwise, CGePoint* pPnt, double* pDirAng) const
{
  double Ang2;
  if (bClockwise){
    Ang2 = Ang0 - Ang;
    if (pDirAng){
      *pDirAng = Ang2 - GE_DEG90;
    }
  }else{
    Ang2 = Ang0 + Ang;
    if (pDirAng){
      *pDirAng = Ang2 + GE_DEG90;
    }
  }
  pPnt->x = m_Center.x + (m_Radius * cos( Ang2 ));
  pPnt->y = m_Center.y + (m_Radius * sin( Ang2 ));
}

//-----------------------------------------------
double CGeCircle::GetNearPoint (const CGePoint& Pnt, CGePoint* pPntNear) const
{
  double Ang, D2p;
  CGePoint Pt2;
  Ang = m_Center.Angle( Pnt );
  Pt2 = m_Center.PolarBy( Ang, m_Radius );
  D2p = Pnt.Distance( Pt2 );
  if (pPntNear){
    *pPntNear = Pt2;
  }
  // distance between Pnt and found point (Pt2)
  return D2p;
}


//-----------------------------------------------
void CGeCircle::GetPoints (CGePoint* Pnt, UINT_PTR NumPoints, double Ang0, bool bClockwise) const
{
  CArcPrm   Prm;
  UINT_PTR  i;
  double    AngStep, Ang;
  double    x0 = m_Center.x;
  double    y0 = m_Center.y;
  CGePoint* pPnt = Pnt;

  switch( NumPoints ){
    case 8:   Prm.m_Resolution = GE_ARCRES_8; break;
    case 16:  Prm.m_Resolution = GE_ARCRES_16; break;
    case 32:  Prm.m_Resolution = GE_ARCRES_32; break;
    case 64:  Prm.m_Resolution = GE_ARCRES_64; break;
    case 128: Prm.m_Resolution = GE_ARCRES_128; break;
    case 256: Prm.m_Resolution = GE_ARCRES_256; break;
    case 512: Prm.m_Resolution = GE_ARCRES_512; break;
    default: 
      if (NumPoints < 8){
        NumPoints = 8;
      }
      goto m1;
  }
  Prm.m_Center = m_Center;
  Prm.m_Radius = m_Radius;
  Prm.m_bDirCW = bClockwise;
  Prm.m_AngRot = Ang0;
  geFastArc( Prm, Pnt );
  return;

m1:
  AngStep = GE_2PI / NumPoints;  // angle step
  Ang = Ang0;
  if (bClockwise){
    AngStep = -AngStep;
  }
  for (i=0; i<NumPoints; ++i){
    pPnt->x = x0 + (m_Radius * cos( Ang ));
    pPnt->y = y0 - (m_Radius * sin( Ang ));
    ++pPnt;
    Ang += AngStep;
  }
}

//-----------------------------------------------
void CGeCircle::GetPolyline (CGePolyline* pPline, UINT_PTR NumPoints, double Ang0, bool bClockwise) const
{
  CGePoint Pnt[514];

  if (NumPoints < 8){
    NumPoints = 8;
  }else{
    if (NumPoints > 512){
      NumPoints = 512;
    }
  }
  GetPoints( Pnt, NumPoints, Ang0, bClockwise );
  pPline->SetVers( Pnt, NumPoints, true );
}


//-----------------------------------------------
CGeRect CGeCircle::GeExtents () const
{
  return CGeRect( MinExtents(), MaxExtents() ); 
}

//-----------------------------------------------
void CGeCircle::GetGeExtents (CGeRect* pRect) const
{
  pRect->Set( MinExtents(), MaxExtents() ); 
}



// nPnts - number of output points, 4 or 8
// Pnt - result points along vector from this circle to  Cir2
// 0,1 - left->left
// 2,3 - right->right
// 4,5 - left->right
// 6,7 - right->left
//-----------------------------------------------
bool CGeCircle::GetTangentPoints (const CGeCircle& Cir2, CGePoint* Pnt, int* pnPnts) const
{
  bool   bGrow;  // true if first circle is less then second
  int    npt;
  double r, d, ccd, a, b;
  double rr = m_Radius + Cir2.m_Radius;
  int    i;
  double cca, ang, cosAng, sinAng, x, y;

  a = Cir2.m_Center.x - m_Center.x;
  b = Cir2.m_Center.y - m_Center.y;
  ccd = _hypot( a, b );
  if (a == 0.) {
    cca = (b>=0.)? GE_PI2 : -GE_PI2 ;
  }else{
    if (a < 0){
      cca = atan( b / a ) + GE_PI;
    }else{
      cca = atan( b / a );
    }
  }
  if (m_Radius < Cir2.m_Radius){
    bGrow = true;
    r = Cir2.m_Radius;
    d = ccd + m_Radius;
  }else{
    bGrow = false;
    r = m_Radius;
    d = ccd + Cir2.m_Radius;
  }
  *pnPnts = 0;
  if (r < d){
    npt = (ccd > rr)? 8 : 4;
    if (bGrow){
      r = Cir2.m_Radius - m_Radius;
      ang = GE_PI - acos( r / ccd );
    }else{
      if (m_Radius > Cir2.m_Radius){
        r = m_Radius - Cir2.m_Radius;
        ang = acos( r / ccd );
      }else{
        ang = GE_PI2;
      }
    }
    cosAng = cos( ang );
    sinAng = sin( ang );
    Pnt[0].x = Pnt[2].x = m_Radius * cosAng;
    Pnt[1].x = Pnt[3].x = ccd + Cir2.m_Radius * cosAng;
    Pnt[0].y = m_Radius * sinAng;
    Pnt[1].y = Cir2.m_Radius * sinAng;
    Pnt[2].y = -Pnt[0].y;
    Pnt[3].y = -Pnt[1].y;
    if (npt == 8){
      d = ccd / (Cir2.m_Radius / m_Radius + 1);
      ang = acos( m_Radius / d );
      cosAng = cos( ang );
      sinAng = sin( ang );
      Pnt[4].x = Pnt[6].x = m_Radius * cosAng;
      Pnt[5].x = Pnt[7].x = ccd - Cir2.m_Radius * cosAng;
      Pnt[4].y = m_Radius * sinAng;
      Pnt[5].y = -Cir2.m_Radius * sinAng;
      Pnt[6].y = -Pnt[4].y;
      Pnt[7].y = -Pnt[5].y;
    }
    sinAng = sin( cca );
    cosAng = cos( cca );
    for (i=0; i<npt; ++i){
      x = Pnt[i].x;
      y = Pnt[i].y;
      Pnt[i].x = x * cosAng - y * sinAng + m_Center.x;
      Pnt[i].y = x * sinAng + y * cosAng + m_Center.y;
    }
    *pnPnts = npt;
    return true;
  }
  return false;
}


//-----------------------------------------------
void CGeCircle::Save (BYTE* pBuf) const
{
  int sz;
  BYTE* ptr;

  ptr = pBuf;
  sz = sizeof(CGePoint);
  memcpy( ptr, &m_Center, sz );  ptr += sz;
  sz = sizeof(double);
  memcpy( ptr, &m_Radius, sz );
}


//-----------------------------------------------
void CGeCircle::Restore (const BYTE* pBuf)
{
  int sz;
  const BYTE* ptr;

  ptr = pBuf;
  sz = sizeof(CGePoint);
  memcpy( &m_Center, ptr, sz );  ptr += sz;
  sz = sizeof(double);
  memcpy( &m_Radius, ptr, sz );
}
