/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* This code was taken from 'BaseLibs' project
* Functions to find intersection points
********************************************************************/
#include <math.h>
#include <stdlib.h>
#include "Geom.h"

#define MAX_INTPNT  1000
#define ELL_PREC    256    // number of generated points for ellipse

static CGePoint _IntPnt[MAX_INTPNT];
static UINT_PTR _nIntPnts = 0;
static CGePoint _Pnt[ELL_PREC+4];
static CGePoint _Pnt2[ELL_PREC+4];
static UINT_PTR _nPnts = 0;
static double   _Toler = g_GeTolerance;
static bool     _bPolyPline = false;

const double LEN_XLINE = 1000000000.0;

//-----------------------------------------------
void geInter_SetTolerance (double Tolerance)
{
  _Toler = Tolerance;
}

//-----------------------------------------------
bool geInter_GetPoint (UINT_PTR i, CGePoint* pPnt)
{
  if (i < _nIntPnts){
    *pPnt = _IntPnt[i];
    return true;
  }
  return false;
}


//-----------------------------------------------
UINT_PTR geInter_Line_Line (const CGeLine& Line, const CGeLine& Line2, UINT_PTR Apparent, CGePoint* IntPt)
{
  _nIntPnts = 0;
  if (Line.GetInterPoint( Line2, _IntPnt, Apparent )){
    if (IntPt){
      IntPt[0] = _IntPnt[0];
    }
    ++_nIntPnts;
  }
  return _nIntPnts;
}


//-----------------------------------------------
UINT_PTR geInter_Line_Pline (const CGeLine& Line, const CGePolyline& Pline, UINT_PTR Apparent, CGePoint* IntPt, UINT_PTR MaxPnts)
{
  CGePoint   InterPt;
  CGeLine    Line2;
  UINT_PTR   i;
  UINT_PTR   npt = Pline.NumVers();
  UINT_PTR   nPnts = 0;
  UINT_PTR   Appa = 0;
  bool       bInter;
  bool       bExit = false;
  CGeVertex* pVerPrev;
  CGeVertex* pVer;


  if (IntPt == NULL){
    _nIntPnts = 0;
  }
  if (npt >= 2){
    i = 1;
    pVerPrev = Pline.GetFirstVer();
    pVer = Pline.GetNextVer( pVerPrev );
    while( pVer ){
      Line2.Set( pVerPrev->m_Pos, pVer->m_Pos );
      if (Apparent == 0){
        bInter = Line.GetInterPoint( Line2, &InterPt );
      }else{
        Appa = 0;
        if (Apparent & 1){
          Appa |= 1;
        }
        if (Apparent & 2){
          if (Pline.IsNotClosed()){
            if (i == 1){
              Line2.Inflate( 10000000.0, 0.0 );
              Appa |= 2;
            }else
            if (i == (npt-1)){
              Line2.Inflate( 0.0, 10000000.0 );
              Appa |= 2;
            }
          }
        }
        bInter = Line.GetInterPoint( Line2, &InterPt, Appa );
      }
      if (bInter){
        if (IntPt){
          if (nPnts > 0){
            if (InterPt.IsEqual( IntPt[nPnts-1] )){
              goto m1;
            }
          }
          IntPt[nPnts] = InterPt;
          ++nPnts;
          if (nPnts >= MaxPnts){
            break;
          }
        }else{
          if (_nIntPnts > 0){
            if (InterPt.IsEqual( _IntPnt[_nIntPnts-1] )){
              goto m1;
            }
          }
          _IntPnt[_nIntPnts] = InterPt;
          ++_nIntPnts;
          if (_nIntPnts >= MAX_INTPNT){
            break;
          }
        }
      }
      if (bExit){
        break;
      }
m1:
      ++i;
      pVerPrev = pVer;
      pVer = Pline.GetNextVer( pVer );
      if (pVer == NULL){
        if (Pline.IsClosed()){
          pVer = Pline.GetFirstVer();
          bExit = true;
        }
      }
    }
  }
  if (IntPt){
    return nPnts;
  }else{
    return _nIntPnts;
  }
}

//-----------------------------------------------
UINT_PTR geInter_Line_Circle (const CGeLine& Line, const CGeCircle& Circ, UINT_PTR Apparent, CGePoint* IntPt)
{
  CGePoint Pt0 = Line.Start();
  double   Ang = Line.Angle();
  double   Len = Line.Length();
  double   R = Circ.Radius();
  CGePoint Pnt, Pt1;
  double   dx, dy;
  UINT_PTR nPnts = 0;

  if (IntPt == NULL){
    _nIntPnts = 0;
  }
  Pt1 = Circ.Center().RotateBy( Pt0, -Ang );
  dy = fabs( Pt1.y - Pt0.y );
  if (dy < (R+_Toler)){
    if (((R-_Toler) < dy) && (dy < (R+_Toler))){
      Pnt.Set( Pt1.x, Pt0.y );
      if (Apparent){
        if (IntPt){
          IntPt[nPnts] = Pnt.RotateBy( Pt0, Ang );
          ++nPnts;
        }else{
          _IntPnt[_nIntPnts] = Pnt.RotateBy( Pt0, Ang );
          ++_nIntPnts;
        }
      }else{
        if (Pt0.x <= Pnt.x && Pnt.x <= (Pt0.x+Len) ){
          if (IntPt){
            IntPt[nPnts] = Pnt.RotateBy( Pt0, Ang );
            ++nPnts;
          }else{
            _IntPnt[_nIntPnts] = Pnt.RotateBy( Pt0, Ang );
            ++_nIntPnts;
          }
        }
      }
    }else{
      dx = sqrt( R*R - dy*dy );
      if (Apparent){
        if (IntPt){
          Pnt.Set( Pt1.x-dx, Pt0.y );
          IntPt[nPnts] = Pnt.RotateBy( Pt0, Ang );
          ++nPnts;
          Pnt.Set( Pt1.x+dx, Pt0.y );
          IntPt[nPnts] = Pnt.RotateBy( Pt0, Ang );
          ++nPnts;
        }else{
          Pnt.Set( Pt1.x-dx, Pt0.y );
          _IntPnt[_nIntPnts] = Pnt.RotateBy( Pt0, Ang );
          ++_nIntPnts;
          Pnt.Set( Pt1.x+dx, Pt0.y );
          _IntPnt[_nIntPnts] = Pnt.RotateBy( Pt0, Ang );
          ++_nIntPnts;
        }
      }else{
        if (IntPt){
          Pnt.Set( Pt1.x-dx, Pt0.y );
          if (Pt0.x <= Pnt.x && Pnt.x <= (Pt0.x+Len) ){
            IntPt[nPnts] = Pnt.RotateBy( Pt0, Ang );
            ++nPnts;
          }
          Pnt.Set( Pt1.x+dx, Pt0.y );
          if (Pt0.x <= Pnt.x && Pnt.x <= (Pt0.x+Len) ){
            IntPt[nPnts] = Pnt.RotateBy( Pt0, Ang );
            ++nPnts;
          }
        }else{
          Pnt.Set( Pt1.x-dx, Pt0.y );
          if (Pt0.x <= Pnt.x && Pnt.x <= (Pt0.x+Len) ){
            _IntPnt[_nIntPnts] = Pnt.RotateBy( Pt0, Ang );
            ++_nIntPnts;
          }
          Pnt.Set( Pt1.x+dx, Pt0.y );
          if (Pt0.x <= Pnt.x && Pnt.x <= (Pt0.x+Len) ){
            _IntPnt[_nIntPnts] = Pnt.RotateBy( Pt0, Ang );
            ++_nIntPnts;
          }
        }
      }
    }
  }
  if (IntPt){
    return nPnts;
  }else{
    return _nIntPnts;
  }
}

//-----------------------------------------------
UINT_PTR geInter_Line_Arc (const CGeLine& Line, const CGeArc& Arc, UINT_PTR Apparent, CGePoint* IntPt)
{
  UINT_PTR  i, n;
  UINT_PTR  nPnts = 0;
  CGePoint  Pnt[5];
  CGeCircle Circ( Arc.Center(), Arc.Radius() );

  if (IntPt == NULL){
    _nIntPnts = 0;
  }
  n = geInter_Line_Circle( Line, Circ, Apparent, Pnt );
  if (n > 0){
    if (Apparent){
      if (IntPt){
        for (i=0; i<n; ++i){
          IntPt[nPnts] = Pnt[i];
          ++nPnts;
        }
      }else{
        for (i=0; i<n; ++i){
          _IntPnt[_nIntPnts] = Pnt[i];
          ++_nIntPnts;
        }
      }
    }else{
      if (IntPt){
        for (i=0; i<n; ++i){
          if (Arc.IsArcPoint( Pnt[i] )){
            IntPt[nPnts] = Pnt[i];
            ++nPnts;
          }
        }
      }else{
        for (i=0; i<n; ++i){
          if (Arc.IsArcPoint( Pnt[i] )){
            _IntPnt[_nIntPnts] = Pnt[i];
            ++_nIntPnts;
          }
        }
      }
    }
  }
  if (IntPt){
    return nPnts;
  }else{
    return _nIntPnts;
  }
}

//-----------------------------------------------
UINT_PTR geInter_Line_Ellipse (const CGeLine& Line, const CGeEllipse& Ell, UINT_PTR Apparent, CGePoint* IntPt)
{
  UINT_PTR    nPnts;
  CGePoint    Pnt[5];
  CGePolyline Pline;
/*
  CGeEllipse  Ell2;
  bool        bFull = Ell.IsFull();
  double      Ang;
  UINT_PTR    nPnts = ELL_PREC;

  if (bFull){
    Ell.GetPoints( _Pnt, nPnts );
  }else{
    if (Apparent & 2){
      Ell2.Set( Ell );
      Ell2.SetFull();
//      Ell2.GetPoints( _Pnt, nPnts );
      Ell2.GetPolyline( &Pline, nPnts );
      bFull = true;
      Apparent &= ~2;
    }else{
      Ang = fabs(Ell.ArcAngle());
      nPnts = (UINT_PTR)(Ang * GE_RAD_TO_DEG);
      if (nPnts < 5){
        nPnts = 5;
      }
//      Ell.GetPoints( _Pnt, nPnts );
      Ell.GetPolyline( &Pline, nPnts );
    }
  }
//  Pline.SetPoints( _Pnt, nPnts, bFull );
*/

  nPnts = Ell.GetPolylineE( &Pline, ELL_PREC );
  nPnts = geInter_Line_Pline( Line, Pline, Apparent, Pnt, 2 );
  if (IntPt){
    if (nPnts > 0){
      IntPt[0] = Pnt[0];
      if (nPnts == 2){
        IntPt[1] = Pnt[1];
      }
    }
    return nPnts;
  }else{
    _nIntPnts = nPnts;
    if (nPnts > 0){
      _IntPnt[0] = Pnt[0];
      if (nPnts == 2){
        _IntPnt[1] = Pnt[1];
      }
    }
    return _nIntPnts;
  }
}


//-----------------------------------------------
void geInter_BeginPolyPline ()
{
  _bPolyPline = true;
  _nIntPnts = 0;
}

//-----------------------------------------------
UINT_PTR geInter_EndPolyPline ()
{
  _bPolyPline = false;
  return _nIntPnts;
}

//-----------------------------------------------
UINT_PTR geInter_Pline_Line (const CGePolyline& Pline, const CGeLine& Line, UINT_PTR Apparent)
{
  UINT_PTR Appa = 0;
  if (Apparent & 1){
    Appa |= 2;
  }
  if (Apparent & 2){
    Appa |= 1;
  }
  return geInter_Line_Pline( Line, Pline, Appa );
}


//-----------------------------------------------
UINT_PTR geInter_Pline_Pline (const CGePolyline& Pline, const CGePolyline& Pline2, UINT_PTR Apparent)
{
  CGePoint   InterPt;
  CGeLine    Line, Line2;
  UINT_PTR   i;
  UINT_PTR   j;
  UINT_PTR   npt = Pline.NumVers();
  UINT_PTR   npt2 = Pline2.NumVers();
  UINT_PTR   Appa;
  bool       bInter;
  bool       bExit = false;
  bool       bExit2 = false;
  CGeVertex* pVerPrev;
  CGeVertex* pVer;
  CGeVertex* pVerPrev2;
  CGeVertex* pVer2;

  if (_bPolyPline == false){
    _nIntPnts = 0;
  }
  if ((npt >= 2) && (npt2 >= 2)){
    i = 1;
    pVerPrev = Pline.GetFirstVer();
    pVer = Pline.GetNextVer( pVerPrev );
    while( pVer ){
      Line.Set( pVerPrev->m_Pos, pVer->m_Pos );
      j = 1;
      bExit2 = false;
      pVerPrev2 = Pline2.GetFirstVer();
      pVer2 = Pline2.GetNextVer( pVerPrev2 );
      while( pVer2 ){
        Line2.Set( pVerPrev2->m_Pos, pVer2->m_Pos );
        if (Apparent == 0){
          bInter = Line.GetInterPoint( Line2, &InterPt );
        }else{
          Appa = 0;
          if (Apparent & 1){
            if (Pline.IsNotClosed()){
              if (i == 1){
                Line.Inflate( 10000000.0, 0.0 );
                Appa |= 1;
              }else
              if (i == (npt-1)){
                Line.Inflate( 0.0, 10000000.0 );
                Appa |= 1;
              }
            }
          }
          if (Apparent & 2){
            if (Pline2.IsNotClosed()){
              if (j == 1){
                Line2.Inflate( 10000000.0, 0.0 );
                Appa |= 2;
              }else
              if (j == (npt2-1)){
                Line2.Inflate( 0.0, 10000000.0 );
                Appa |= 2;
              }
            }
          }
          bInter = Line.GetInterPoint( Line2, &InterPt, Appa );
        }
        if (bInter){
          _IntPnt[_nIntPnts] = InterPt;
          ++_nIntPnts;
          if (_nIntPnts >= MAX_INTPNT){
            goto m1;
          }
        }
        if (bExit2){
          break;
        }
        ++j;
        pVerPrev2 = pVer2;
        pVer2 = Pline2.GetNextVer( pVer2 );
        if (pVer2 == NULL){
          if (Pline2.IsClosed()){
            pVer2 = Pline2.GetFirstVer();
            bExit2 = true;
          }
        }
      }
      if (bExit){
        break;
      }
      ++i;
      pVerPrev = pVer;
      pVer = Pline.GetNextVer( pVer );
      if (pVer == NULL){
        if (Pline.IsClosed()){
          pVer = Pline.GetFirstVer();
          bExit = true;
        }
      }
    }
  }
m1:
  return _nIntPnts;
}

//-----------------------------------------------
UINT_PTR geInter_Pline_Circle (const CGePolyline& Pline, const CGeCircle& Circ, UINT_PTR Apparent)
{
  CGePoint   InterPt[2];
  CGeLine    Line;
  CGeVertex* pVerPrev;
  CGeVertex* pVer;
  UINT_PTR   i, j, nip;
  UINT_PTR   n = Pline.NumVers();
  UINT_PTR   Appa;
  bool       bExit = false;

  if (_bPolyPline == false){
    _nIntPnts = 0;
  }
  if (n >= 2){
    i = 1;
    pVerPrev = Pline.GetFirstVer();
    pVer = Pline.GetNextVer( pVerPrev );
    while( pVer ){
      Line.Set( pVerPrev->m_Pos, pVer->m_Pos );
      Appa = 0;
      if (Apparent & 1){
        if (Pline.IsNotClosed()){
          if (i == 1){
            Line.Inflate( 10000000.0, 0.0 );
            Appa |= 1;
          }else
          if (i == (n-1)){
            Line.Inflate( 0.0, 10000000.0 );
            Appa |= 1;
          }
        }
      }
      nip = geInter_Line_Circle( Line, Circ, Appa, InterPt );
      if (nip > 0){
        for (j=0; j<nip; ++j){
          if (_nIntPnts < MAX_INTPNT){
            _IntPnt[_nIntPnts] = InterPt[j];
            ++_nIntPnts;
          }else{
            break;
          }
        }
      }
      if (bExit){
        break;
      }
      ++i;
      pVerPrev = pVer;
      pVer = Pline.GetNextVer( pVer );
      if (pVer == NULL){
        if (Pline.IsClosed()){
          pVer = Pline.GetFirstVer();
          bExit = true;
        }
      }
    }
  }
  return _nIntPnts;
}

//-----------------------------------------------
UINT_PTR geInter_Pline_Arc (const CGePolyline& Pline, const CGeArc& Arc, UINT_PTR Apparent)
{
  CGePoint   InterPt[2];
  CGeLine    Line;
  CGeVertex* pVerPrev;
  CGeVertex* pVer;
  UINT_PTR   i, j, nip;
  UINT_PTR   n = Pline.NumVers();
  UINT_PTR   Appa;
  bool       bExit = false;

  if (_bPolyPline == false){
    _nIntPnts = 0;
  }
  if (n >= 2){
    i = 1;
    pVerPrev = Pline.GetFirstVer();
    pVer = Pline.GetNextVer( pVerPrev );
    while( pVer ){
      Line.Set( pVerPrev->m_Pos, pVer->m_Pos );
      Appa = 0;
      if (Apparent & 1){
        if (Pline.IsNotClosed()){
          if (i == 1){
            Line.Inflate( 10000000.0, 0.0 );
            Appa |= 1;
          }else
          if (i == (n-1)){
            Line.Inflate( 0.0, 10000000.0 );
            Appa |= 1;
          }
        }
      }
      if (Apparent & 2){
        Appa |= 2;
      }
      nip = geInter_Line_Arc( Line, Arc, Appa, InterPt );
      if (nip > 0){
        for (j=0; j<nip; ++j){
          if (_nIntPnts < MAX_INTPNT){
            _IntPnt[_nIntPnts] = InterPt[j];
            ++_nIntPnts;
          }else{
            break;
          }
        }
      }
      if (bExit){
        break;
      }
      ++i;
      pVerPrev = pVer;
      pVer = Pline.GetNextVer( pVer );
      if (pVer == NULL){
        if (Pline.IsClosed()){
          pVer = Pline.GetFirstVer();
          bExit = true;
        }
      }
    }
  }
  return _nIntPnts;
}

//-----------------------------------------------
UINT_PTR geInter_Pline_Ellipse (const CGePolyline& Pline, const CGeEllipse& Ell, UINT_PTR Apparent)
{
  UINT_PTR    nPnts;
  CGePolyline Pline2;
/*
  UINT_PTR    nPnts = ELL_PREC;
  CGePolyline Pline2;
  CGeEllipse  Ell2;
  double      Ang;
  bool        bFull = Ell.IsFull();

  if (bFull){
//    Ell.GetPoints( _Pnt, nPnts );
    Ell.GetPolyline( &Pline2, nPnts );
  }else{
    if (Apparent & 2){
      Ell2.Set( Ell );
      Ell2.SetFull();
//      Ell2.GetPoints( _Pnt, nPnts );
      Ell2.GetPolyline( &Pline2, nPnts );
      bFull = true;
    }else{
      Ang = fabs(Ell.ArcAngle());
      nPnts = (UINT_PTR)(Ang * GE_RAD_TO_DEG);
      if (nPnts < 5){
        nPnts = 5;
      }
//      Ell.GetPoints( _Pnt, nPnts );
      Ell.GetPolyline( &Pline2, nPnts );
    }
  }
//  Pline2.SetPoints( _Pnt, nPnts, bFull );
*/
  nPnts = Ell.GetPolylineE( &Pline2, ELL_PREC );
  Apparent &= ~2;
  return geInter_Pline_Pline( Pline, Pline2, Apparent );
}





//-----------------------------------------------
UINT_PTR geInter_Circle_Line (const CGeCircle& Circ, const CGeLine& Line, UINT_PTR Apparent)
{
  if (Apparent & 2){
    Apparent = 1;
  }else{
    Apparent = 0;
  }
  return geInter_Line_Circle( Line, Circ, Apparent );
}


//-----------------------------------------------
UINT_PTR geInter_Circle_Pline (const CGeCircle& Circ, const CGePolyline& Pline, UINT_PTR Apparent)
{
  if (Apparent & 2){
    Apparent = 1;
  }else{
    Apparent = 0;
  }
  return geInter_Pline_Circle( Pline, Circ, Apparent );
}

//-----------------------------------------------
UINT_PTR geInter_Circle_Circle (const CGeCircle& Circ, const CGeCircle& Circ2)
{
  CGePoint Pnt;
  CGePoint CenPt = Circ.Center();
  CGePoint CenPt2 = Circ2.Center();
  double   Ang, d1, d2, dy, R1R2, Lmin, Lmax;
  double   R1 = Circ.Radius();
  double   R2 = Circ2.Radius();
  double   Dist = CenPt.Distance( CenPt2 );

  _nIntPnts = 0;
  Lmin = Dist - _Toler;
  Lmax = Dist + _Toler;
  R1R2 = R1 + R2;
  if (Lmin >= R1R2){
    return 0;
  }
  if ((Lmin < R1R2) && (R1R2 < Lmax)){
    Ang = CenPt.Angle( CenPt2 );
    _IntPnt[_nIntPnts] = CenPt.PolarBy( Ang, R1 );
    ++_nIntPnts;
    return _nIntPnts;
  }
  Lmin = Dist + R1 - _Toler;
  Lmax = Dist + R1 + _Toler;
  if ((Lmin < R2) && (R2 < Lmax)){
    Ang = CenPt2.Angle( CenPt );
    _IntPnt[_nIntPnts] = CenPt2.PolarBy( Ang, R2 );
    ++_nIntPnts;
    return _nIntPnts;
  }
  Lmin = Dist + R2 - _Toler;
  Lmax = Dist + R2 + _Toler;
  if ((Lmin < R1) && (R1 < Lmax)){
    Ang = CenPt.Angle( CenPt2 );
    _IntPnt[_nIntPnts] = CenPt.PolarBy( Ang, R1 );
    ++_nIntPnts;
    return _nIntPnts;
  }
  d1 = (R2*R2 - R1*R1 - Dist*Dist) / (-2.0 * Dist);
  d2 = R1*R1 - d1*d1;
  if (d2 > 0.000001){
    Ang = CenPt.Angle( CenPt2 );
    Pnt = CenPt.PolarBy( Ang, d1 );
    dy = sqrt( d2 );
    _IntPnt[_nIntPnts] = Pnt.PolarBy( Ang+GE_DEG90, dy );
    ++_nIntPnts;
    _IntPnt[_nIntPnts] = Pnt.PolarBy( Ang+GE_DEG90, -dy );
    ++_nIntPnts;
    return _nIntPnts;
  }
  return 0;
}

//-----------------------------------------------
UINT_PTR geInter_Circle_Arc (const CGeCircle& Circ, const CGeArc& Arc, UINT_PTR Apparent)
{
  UINT_PTR  i, np;
  CGePoint  Pnt[5];
  CGeCircle Circ2( Arc.Center(), Arc.Radius() );

  _nIntPnts = 0;
  geInter_Circle_Circle( Circ, Circ2 );
  if (_nIntPnts > 0 && Apparent&2){
    np = 0;
    for (i=0; i<_nIntPnts; ++i){
      if (Arc.IsArcPoint( _IntPnt[i] )){
        Pnt[np] = _IntPnt[i];
        ++np;
      }
    }
    _nIntPnts = 0;
    if (np > 0){
      for (i=0; i<np; ++i){
        _IntPnt[_nIntPnts] = Pnt[i];
        ++_nIntPnts;
      }
    }
  }
  return _nIntPnts;
}

//-----------------------------------------------
UINT_PTR geInter_Circle_Ellipse (const CGeCircle& Circ, const CGeEllipse& Ell, UINT_PTR /*Apparent*/)
{
  UINT_PTR    nPnts;
  CGePolyline Pline;
/*
  UINT_PTR    nPnts = ELL_PREC;
  CGePolyline Pline;
  CGeEllipse  Ell2;
  bool        bFull = Ell.IsFull();
  double      Ang;

  if (bFull){
//    Ell.GetPoints( _Pnt, nPnts );
    Ell.GetPolyline( &Pline, nPnts );
  }else{
    if (Apparent & 2){
      Ell2.Set( Ell );
      Ell2.SetFull();
//      Ell2.GetPoints( _Pnt, nPnts );
      Ell2.GetPolyline( &Pline, nPnts );
      bFull = true;
    }else{
      Ang = fabs(Ell.ArcAngle());
      nPnts = (UINT_PTR)(Ang * GE_RAD_TO_DEG);
      if (nPnts < 5){
        nPnts = 5;
      }
//      Ell.GetPoints( _Pnt, nPnts );
      Ell.GetPolyline( &Pline, nPnts );
    }
  }
//  Pline.SetPoints( _Pnt, nPnts, bFull );
*/
  nPnts = Ell.GetPolylineE( &Pline, ELL_PREC );
  return geInter_Pline_Circle( Pline, Circ, 0 );
}




//-----------------------------------------------
UINT_PTR geInter_Arc_Line (const CGeArc& Arc, const CGeLine& Line, UINT_PTR Apparent)
{
  UINT_PTR Appa = 0;
  if (Apparent & 1){
    Appa |= 2;
  }
  if (Apparent & 2){
    Appa |= 1;
  }
  return geInter_Line_Arc( Line, Arc, Appa );
}


//-----------------------------------------------
UINT_PTR geInter_Arc_Pline (const CGeArc& Arc, const CGePolyline& Pline, UINT_PTR Apparent)
{
  UINT_PTR Appa = 0;
  if (Apparent & 1){
    Appa |= 2;
  }
  if (Apparent & 2){
    Appa |= 1;
  }
  return geInter_Pline_Arc( Pline, Arc, Appa );
}

//-----------------------------------------------
UINT_PTR geInter_Arc_Circle (const CGeArc& Arc, const CGeCircle& Circ, UINT_PTR Apparent)
{
  UINT_PTR Appa = 0;
  if (Apparent & 1){
    Appa |= 2;
  }
  if (Apparent & 2){
    Appa |= 1;
  }
  return geInter_Circle_Arc( Circ, Arc, Appa );
}

//-----------------------------------------------
UINT_PTR geInter_Arc_Arc (const CGeArc& Arc, const CGeArc& Arc2, UINT_PTR Apparent)
{
  UINT_PTR  i, np;
  CGePoint  Pnt[5];
  CGeCircle Circ( Arc.Center(), Arc.Radius() );
  CGeCircle Circ2( Arc2.Center(), Arc2.Radius() );

  _nIntPnts = 0;
  geInter_Circle_Circle( Circ, Circ2 );
  if (Apparent & 3){
    return _nIntPnts;
  }
  if (_nIntPnts > 0){
    np = 0;
    for (i=0; i<_nIntPnts; ++i){
      if (Apparent == 0){
        if (Arc.IsArcPoint( _IntPnt[i] ) && Arc2.IsArcPoint( _IntPnt[i] )){
          Pnt[np] = _IntPnt[i];
          ++np;
        }
      }else
      if (Apparent & 1){
        if (Arc2.IsArcPoint( _IntPnt[i] )){
          Pnt[np] = _IntPnt[i];
          ++np;
        }
      }else
      if (Apparent & 2){
        if (Arc.IsArcPoint( _IntPnt[i] )){
          Pnt[np] = _IntPnt[i];
          ++np;
        }
      }
    }
    _nIntPnts = 0;
    if (np > 0){
      for (i=0; i<np; ++i){
        _IntPnt[_nIntPnts] = Pnt[i];
        ++_nIntPnts;
      }
    }
  }
  return _nIntPnts;
}

//-----------------------------------------------
UINT_PTR geInter_Arc_Ellipse (const CGeArc& Arc, const CGeEllipse& Ell, UINT_PTR Apparent)
{
  UINT_PTR    nPnts;
  CGePolyline Pline;
/*
  UINT_PTR    nPnts = ELL_PREC;
  CGePolyline Pline;
  CGeEllipse  Ell2;
  bool        bFull = Ell.IsFull();
  double      Ang;

  if (bFull){
//    Ell.GetPoints( _Pnt, nPnts );
    Ell.GetPolyline( &Pline, nPnts );
  }else{
    if (Apparent & 2){
      Ell2.Set( Ell );
      Ell2.SetFull();
//      Ell2.GetPoints( _Pnt, nPnts );
      Ell2.GetPolyline( &Pline, nPnts );
      bFull = true;
    }else{
      Ang = fabs(Ell.ArcAngle());
      nPnts = (UINT_PTR)(Ang * GE_RAD_TO_DEG);
      if (nPnts < 5){
        nPnts = 5;
      }
//      Ell.GetPoints( _Pnt, nPnts );
      Ell.GetPolyline( &Pline, nPnts );
    }
  }
//  Pline.SetPoints( _Pnt, nPnts, bFull );
*/
  Apparent&=~2;
  nPnts = Ell.GetPolylineE( &Pline, ELL_PREC );
  return geInter_Pline_Arc( Pline, Arc, Apparent );
}






//-----------------------------------------------
UINT_PTR geInter_Ellipse_Line (const CGeEllipse& Ell, const CGeLine& Line, UINT_PTR Apparent)
{
  UINT_PTR Appa = 0;
  if (Apparent & 1){
    Appa |= 2;
  }
  if (Apparent & 2){
    Appa |= 1;
  }
  return geInter_Line_Ellipse( Line, Ell, Appa );
}


//-----------------------------------------------
UINT_PTR geInter_Ellipse_Pline (const CGeEllipse& Ell, const CGePolyline& Pline, UINT_PTR Apparent)
{
  UINT_PTR Appa = 0;
  if (Apparent & 1){
    Appa |= 2;
  }
  if (Apparent & 2){
    Appa |= 1;
  }
  return geInter_Pline_Ellipse( Pline, Ell, Appa );
}

//-----------------------------------------------
UINT_PTR geInter_Ellipse_Circle (const CGeEllipse& Ell, const CGeCircle& Circ, UINT_PTR Apparent)
{
  UINT_PTR Appa = 0;
  if (Apparent & 1){
    Appa |= 2;
  }
  if (Apparent & 2){
    Appa |= 1;
  }
  return geInter_Circle_Ellipse( Circ, Ell, Appa );
}

//-----------------------------------------------
UINT_PTR geInter_Ellipse_Arc (const CGeEllipse& Ell, const CGeArc& Arc, UINT_PTR Apparent)
{
  UINT_PTR Appa = 0;
  if (Apparent & 1){
    Appa |= 2;
  }
  if (Apparent & 2){
    Appa |= 1;
  }
  return geInter_Arc_Ellipse( Arc, Ell, Appa );
}

//-----------------------------------------------
UINT_PTR geInter_Ellipse_Ellipse (const CGeEllipse& Ell, const CGeEllipse& Ell2, UINT_PTR /*Apparent*/)
{
  UINT_PTR    nPnts;
  CGePolyline Pline, Pline2;
/*
  UINT_PTR    nPnts = ELL_PREC;
//  UINT_PTR    nPnts2 = ELL_PREC;
  CGePolyline Pline, Pline2;
  CGeEllipse  Ell3;
  double      Ang;
  bool        bFull = Ell.IsFull();

  if (bFull){
//    Ell.GetPoints( _Pnt, nPnts );
    Ell.GetPolyline( &Pline, nPnts );
  }else{
    if (Apparent & 1){
      Ell3.Set( Ell );
      Ell3.SetFull();
//      Ell3.GetPoints( _Pnt, nPnts );
      Ell3.GetPolyline( &Pline, nPnts );
      bFull = true;
    }else{
      Ang = fabs(Ell.ArcAngle());
      nPnts = (UINT_PTR)(Ang * GE_RAD_TO_DEG);
      if (nPnts < 5){
        nPnts = 5;
      }
//      Ell.GetPoints( _Pnt, nPnts );
      Ell.GetPolyline( &Pline, nPnts );
    }
  }
//  Pline.SetPoints( _Pnt, nPnts, bFull );

  bFull = Ell2.IsFull();
  if (bFull){
//    Ell2.GetPoints( _Pnt, nPnts );
    Ell2.GetPolyline( &Pline2, nPnts );
  }else{
    if (Apparent & 2){
      Ell3.Set( Ell2 );
      Ell3.SetFull();
//      Ell3.GetPoints( _Pnt, nPnts );
      Ell3.GetPolyline( &Pline2, nPnts );
      bFull = true;
    }else{
      Ang = fabs(Ell2.ArcAngle());
      nPnts = (UINT_PTR)(Ang * GE_RAD_TO_DEG);
      if (nPnts < 5){
        nPnts = 5;
      }
//      Ell2.GetPoints( _Pnt, nPnts );
      Ell2.GetPolyline( &Pline2, nPnts );
    }
  }
//  Pline2.SetPoints( _Pnt, nPnts, bFull );
*/
  nPnts = Ell.GetPolylineE( &Pline, ELL_PREC );
  nPnts = Ell2.GetPolylineE( &Pline2, ELL_PREC );
  return geInter_Pline_Pline( Pline, Pline2, 0 );
}


