/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* This code was taken from 'BaseLibs' project
* Class CGeLine implementation
********************************************************************/
#include <math.h>
#include <stdlib.h>
#include <memory.h>
#include "Geom.h"

static double   _Ax, _Ay, _Bx, _By, _sa;
static double   _Nx, _Ny, _denom, _num, _dval;
static EnPntPos _aclass;

//-----------------------------------------------
void CGeLine::Set (const CGePoint& Pt0, double Ang, double Len1, double Len2)
{
  m_Start = Pt0.PolarBy( Ang, -Len1 );
  m_End = Pt0.PolarBy( Ang, Len2 );
}

//-----------------------------------------------
CGeRect CGeLine::GeExtents () const
{
  return CGeRect( MinExtents(), MaxExtents() ); 
}


//-----------------------------------------------
void CGeLine::GetGeExtents (CGeRect* pRect) const
{
  pRect->Set( MinExtents(), MaxExtents() ); 
}

//-----------------------------------------------
bool CGeLine::IsDot (double Delta) const
{
  if ((m_Start.x - Delta) < m_End.x && m_End.x < (m_Start.x + Delta) && 
      (m_Start.y - Delta) < m_End.y && m_End.y < (m_Start.y + Delta))
  {
    return true;
  }
  return false;
}

//-----------------------------------------------
double CGeLine::Angle (const CGeLine& Line2) const
{
  double Ang1, Ang2, DA;

  Ang1 = m_Start.Angle( m_End );  // Angle from Start to End (0..2PI)
  Ang2 = Line2.Angle(); 
  DA = Ang2 - Ang1;
  if (DA < 0){
    DA += GE_DEG360;
  }
  return DA;
}


/********************************************************************
* Classify position of point relative to line
*  Pnt - point
* Returns:
*  GE_LEFT  - point is at left side of m_Start-m_End
*  GE_RIGHT - point is at right side of m_Start-m_End
*    Next codes shows that point is on m_Start-m_End direction
*  GE_BEFORE - before m_Start
*  GE_AFTER - after m_End
*  GE_ORIGIN - exactly at m_Start
*  GE_DESTINATION - exactly at m_End
*  GE_BETWEEN - between m_Start and m_End
*********************************************************************/
EnPntPos CGeLine::ClassifyPoint (const CGePoint& Pnt) const
{
  if (Pnt.IsEqual( m_Start, g_GeTolerance )){
    return EnPntPos::GE_ORIGIN;
  }
  if (Pnt.IsEqual( m_End, g_GeTolerance )){
    return EnPntPos::GE_DESTINATION;
  }
  _Ax = m_End.x - m_Start.x;
  _Ay = m_End.y - m_Start.y;
  _Bx = Pnt.x - m_Start.x;
  _By = Pnt.y - m_Start.y;
  _sa = (_Ax*_By) - (_Bx*_Ay);
  if (_sa > 0.0){
    return EnPntPos::GE_LEFT;
  }
  if (_sa < 0.0){
    return EnPntPos::GE_RIGHT;
  }
  if (((_Ax*_Bx) < 0.0) || ((_Ay*_By) < 0.0)){
    return EnPntPos::GE_BEFORE;
  }
  if (sqrt((_Ax*_Ax) + (_Ay*_Ay)) < sqrt((_Bx*_Bx) + (_By*_By))){
    return EnPntPos::GE_AFTER;
  }
  return EnPntPos::GE_BETWEEN;
}

//-----------------------------------------------
EnPntPos CGeLine::ClassifyPoint2 (const CGePoint& Pnt, double Delta) const
{
  double   Len, Ang, Y;
  CGePoint Pt2;

  if (Pnt.IsEqual( m_Start, Delta )){
    return EnPntPos::GE_ORIGIN;
  }
  if (Pnt.IsEqual( m_End, Delta )){
    return EnPntPos::GE_DESTINATION;
  }
  Ang = m_Start.Angle( m_End );
  Len = m_Start.Distance( m_End );
  Pt2 = Pnt.RotateBy( m_Start, -Ang );
  Y = Pt2.y - m_Start.y;
  if (Delta >= Y){
    if (Y >= -Delta){
      if (Pt2.x < m_Start.x){
        return EnPntPos::GE_BEFORE;
      }
      if (Pt2.x > (m_Start.x + Len)){
        return EnPntPos::GE_AFTER;
      }
      return EnPntPos::GE_BETWEEN;
    }else{
      return EnPntPos::GE_RIGHT;
    }
  }else{
    return EnPntPos::GE_LEFT;
  }
}


/********************************************************************
* Classify position between 2 Infinite Lines
*  Line2 - second line
*  T - (output) parametric coefficient of intersect. point's
*      position along this line (A-0 B-1)
* Returns:
*  GE_PARALLEL - the lines are parallel (T is undefined)
*  GE_COLLINEAR - the lines are at one line direction (T is undefined)
*  GE_SKEW - the lines cross with angle (have intersect. point)
*********************************************************************/
EnIntersec CGeLine::ClassifyLine (const CGeLine& Line2, double* pT) const
{
  _Nx = Line2.m_End.y - Line2.m_Start.y;
  _Ny = Line2.m_Start.x - Line2.m_End.x;
  _denom = _Nx * (m_End.x - m_Start.x) + _Ny * (m_End.y - m_Start.y);
  if (-g_GeTolClass <= _denom && _denom <= g_GeTolClass){
    // classify position of point A relative to Line2
    _aclass = Line2.ClassifyPoint( m_Start );
    if (_aclass == EnPntPos::GE_LEFT || _aclass == EnPntPos::GE_RIGHT){
      return EnIntersec::GE_PARALLEL;
    }else{
      return EnIntersec::GE_COLLINEAR;
    }
  }
  _num = _Nx * (m_Start.x - Line2.m_Start.x) + _Ny * (m_Start.y - Line2.m_Start.y);
  if (pT){
    *pT = -_num / _denom;
  }
  return EnIntersec::GE_SKEW;
}


//    
//-----------------------------------------------
double dotProduct (const CGePoint& p, const CGePoint& q)
{
  return (p.x*q.x + p.y*q.y);
}



// used in  "bool Mate (TSPOINT* ..."  (Toposur_Delaunay.cpp)
// same as ClassifyLine ???
//-----------------------------------------------
EnIntersec CGeLine::Intersect (const CGeLine& Line2, double* pT) const
{
  EnPntPos aclass;
  CGePoint a, b, c, d, n;
  double   denom;

  a = m_Start;
  b = m_End;
  c = Line2.m_Start;
  d = Line2.m_End;
  n = CGePoint( (d-c).y, (c-d).x );
  denom = dotProduct( n, b-a );
  if (denom == 0.0){
    aclass = Line2.ClassifyPoint( m_Start );
    if (aclass == EnPntPos::GE_LEFT || aclass == EnPntPos::GE_RIGHT){
      return EnIntersec::GE_PARALLEL;
    }else{
      return EnIntersec::GE_COLLINEAR;
    }
  }
  double num = dotProduct( n, a-c );
  if (pT){
    *pT = -num / denom;
  }
  return EnIntersec::GE_SKEW;
}


/********************************************************************
* Get relative position between 2 lines
*  Line2 - other line
*  pInterPt - (out) intersection point, 
*         if NULL - the point will not be calculated
* Returns:
*  GE_PARALLEL      - the lines are parallel (pInterPt is undefined)
*  GE_COLLINEAR     - the lines are at one line direction (pInterPt is undefined)
*  GE_SKEW_NO_CROSS - intersect. point is outside of lines 
*  GE_SKEW_CROSS    - lines has intersect. point inside of both lines
*********************************************************************/
EnIntersec CGeLine::GetIntersection (const CGeLine& Line2, CGePoint* pInterPt, CGePoint* pNextPt, double* pRightAngle, double Zero1, double One1, double Zero2, double One2) const
{
  double t, ang, ang2;
  EnIntersec crossType;

  crossType = ClassifyLine( Line2, &t );
  if (crossType == EnIntersec::GE_PARALLEL){
    if (pInterPt){
      pInterPt->x = pInterPt->y = 0.;
    }
    return EnIntersec::GE_PARALLEL;  // there are no intersect. point
  }

  if (m_End.IsEqual( Line2.m_Start, g_GeTolerance )){
    ang = m_Start.RightAngle( m_End, Line2.m_End );
    if (pRightAngle){
      *pRightAngle = ang;
    }
    if (pInterPt){
      *pInterPt = m_End;
    }
    if (pNextPt){
      *pNextPt = Line2.m_End;
    }
    if (crossType == EnIntersec::GE_COLLINEAR){
      if (GE_DEG180-0.01 < ang && ang < GE_DEG180+0.01){
        return EnIntersec::GE_SKEW_CROSS;
      }
      return EnIntersec::GE_COLLINEAR;  // there are no intersect. point
    }
    return EnIntersec::GE_SKEW_CROSS;
  }

  if (m_End.IsEqual( Line2.m_End, g_GeTolerance )){
    ang = m_Start.RightAngle( m_End, Line2.m_Start );
    if (pRightAngle){
      *pRightAngle = ang;
    }
    if (pInterPt){
      *pInterPt = m_End;
    }
    if (pNextPt){
      *pNextPt = Line2.m_Start;
    }
    if (crossType == EnIntersec::GE_COLLINEAR){
      if (GE_DEG180-0.01 < ang && ang < GE_DEG180+0.01){
        return EnIntersec::GE_SKEW_CROSS;
      }
      return EnIntersec::GE_COLLINEAR;  // there are no intersect. point
    }
    return EnIntersec::GE_SKEW_CROSS;
  }
  if (crossType == EnIntersec::GE_COLLINEAR){
    return EnIntersec::GE_COLLINEAR;  // there are no intersect. point
  }

  if (pInterPt){
    // intersection point
    pInterPt->x = m_Start.x + t * (m_End.x - m_Start.x);
    pInterPt->y = m_Start.y + t * (m_End.y - m_Start.y);
  }
  if (t < Zero1 || t > One1){
    return EnIntersec::GE_SKEW_NO_CROSS;  // intersect. point is outside of lines 
  }
  Line2.ClassifyLine( *this, &t );
  if (Zero2 <= t && t <= One2){
    if (pNextPt){
      if (t < 0.0001){
        ang = GE_DEG360;
      }else{
        if (pInterPt){
          ang = m_Start.RightAngle( *pInterPt, Line2.m_Start );
        }else{
          ang = 0.0;
        }
      }
      if (t > 0.9999){
        ang2 = GE_DEG360;
      }else{
        if (pInterPt){
          ang2 = m_Start.RightAngle( *pInterPt, Line2.m_End );
        }else{
          ang2 = 0.0;
        }
      }
      if (ang < ang2){
        *pNextPt = Line2.m_Start;
        if (pRightAngle){
          *pRightAngle = ang;
        }
      }else{
        *pNextPt = Line2.m_End;
        if (pRightAngle){
          *pRightAngle = ang2;
        }
      }
    }
    return EnIntersec::GE_SKEW_CROSS;     // lines has intersect. point
  }else{
    return EnIntersec::GE_SKEW_NO_CROSS;  // intersect. point is outside of lines
  }
}


//-----------------------------------------------
int CGeLine::GetIntersection (const CGeCircle& Circ, CGePoint* pInterPt) const
{
  double   ang, len, dx, dy, R;
  CGePoint pnt, pt1;
  int      npt = 0;

  ang = Angle();
  len = Length();
  npt = 0;
  pt1 = Circ.Center();
  pt1.Rotate( m_Start, -ang );
  dy = fabs( pt1.y - m_Start.y );
  R = Circ.Radius();
  if (dy <= R){
    if (dy == R){
      pnt.Set( pt1.x, m_Start.y );
      if (m_Start.x <= pnt.x && pnt.x <= (m_Start.x+len) ){
        pInterPt[npt] = pnt.RotateBy( m_Start, ang );
        ++npt;
      }
    }else{
      dx = sqrt( R*R - dy*dy );
      pnt.Set( pt1.x-dx, m_Start.y );
      if (m_Start.x <= pnt.x && pnt.x <= (m_Start.x+len) ){
        pInterPt[npt] = pnt.RotateBy( m_Start, ang );
        ++npt;
      }
      pnt.Set( pt1.x+dx, m_Start.y );
      if (m_Start.x <= pnt.x && pnt.x <= (m_Start.x+len) ){
        pInterPt[npt] = pnt.RotateBy( m_Start, ang );
        ++npt;
      }
    }
  }
  return npt;
}


//-----------------------------------------------
int CGeLine::GetIntersection (const CGeArc& Arc, CGePoint* pInterPt) const
{
  int  i, n, nIntPt;
  CGePoint Pnt[5];
  CGeCircle Circ;

  nIntPt = 0;
  Circ.Set( Arc.Center(), Arc.Radius() );
  n = GetIntersection( Circ, Pnt );
  if (n > 0){
    for (i=0; i<n; ++i){
      if (Arc.IsArcPoint( Pnt[i] )){
        pInterPt[nIntPt] = Pnt[i];
        ++nIntPt;
      }
    }
  }
  return nIntPt;
}

// Apparent:
// 0 - none
// 1 - first line is infinite
// 2 - second line (Line2) is infinite
// 3 - both lines are infinite
//-----------------------------------------------
bool CGeLine::GetInterPoint (const CGeLine& Line2, CGePoint* pInterPt, UINT_PTR Apparent) const
{
  static double Zero = -0.00001;
  static double One  =  1.00001;
  double Num, T, T2, Nx, Ny, Denom;

  Nx = Line2.m_End.y - Line2.m_Start.y;
  Ny = Line2.m_Start.x - Line2.m_End.x;
  Denom = Nx * (m_End.x - m_Start.x) + Ny * (m_End.y - m_Start.y);
  if (-g_GeTolClass <= Denom && Denom <= g_GeTolClass){
    // lines are parallel or collinear
    if (IsEqual( Line2 ) || IsEqualBack( Line2 )){
      return false;
    }
    if (m_Start == Line2.m_Start || m_Start == Line2.m_End){
      *pInterPt = m_Start;
      return true;
    }else{
      if (m_End == Line2.m_Start || m_End == Line2.m_End){
        *pInterPt = m_End;
        return true;
      }
    }
  }else{
    if (Apparent == 3){
      // both lines are infinite
      Num = Nx * (m_Start.x - Line2.m_Start.x) + Ny * (m_Start.y - Line2.m_Start.y);
      T = -Num / Denom;
      pInterPt->x = m_Start.x + (T * (m_End.x - m_Start.x));
      pInterPt->y = m_Start.y + (T * (m_End.y - m_Start.y));
      return true;
    }else{
      if (Apparent == 0){
        Num = Nx * (m_Start.x - Line2.m_Start.x) + Ny * (m_Start.y - Line2.m_Start.y);
        T = -Num / Denom;
        if (Zero <= T && T <= One){
          Nx = m_End.y - m_Start.y;
          Ny = m_Start.x - m_End.x;
          Denom = Nx * (Line2.m_End.x - Line2.m_Start.x) + Ny * (Line2.m_End.y - Line2.m_Start.y);
          Num = Nx * (Line2.m_Start.x - m_Start.x) + Ny * (Line2.m_Start.y - m_Start.y);
          T2 = -Num / Denom;
          if (Zero <= T2 && T2 <= One){
            pInterPt->x = Line2.m_Start.x + T2 * (Line2.m_End.x - Line2.m_Start.x);
            pInterPt->y = Line2.m_Start.y + T2 * (Line2.m_End.y - Line2.m_Start.y);
            return true;
          }
        }
      }else{
        if (Apparent & 1){
          // only first line is infinite
          Nx = m_End.y - m_Start.y;
          Ny = m_Start.x - m_End.x;
          Denom = Nx * (Line2.m_End.x - Line2.m_Start.x) + Ny * (Line2.m_End.y - Line2.m_Start.y);
          Num = Nx * (Line2.m_Start.x - m_Start.x) + Ny * (Line2.m_Start.y - m_Start.y);
          T2 = -Num / Denom;
          if (Zero <= T2 && T2 <= One){
            pInterPt->x = Line2.m_Start.x + T2 * (Line2.m_End.x - Line2.m_Start.x);
            pInterPt->y = Line2.m_Start.y + T2 * (Line2.m_End.y - Line2.m_Start.y);
            return true;
          }
        }else
        if (Apparent & 2){
          // only second line (Line2) is infinite
          Num = Nx * (m_Start.x - Line2.m_Start.x) + Ny * (m_Start.y - Line2.m_Start.y);
          T = -Num / Denom;
          if (Zero <= T && T <= One){
            pInterPt->x = m_Start.x + T * (m_End.x - m_Start.x);
            pInterPt->y = m_Start.y + T * (m_End.y - m_Start.y);
            return true;
          }
        }
      }
    }
  }
  return false;
}


//-----------------------------------------------
bool CGeLine::IsIntersect (const CGeLine& Line2) const
{
  double T, T2;
  EnIntersec Res;

  Res = ClassifyLine( Line2, &T );
  if ((Res == EnIntersec::GE_SKEW) && (0.0 < T) && (T < 1.0)){
    Res = Line2.ClassifyLine( *this, &T2 );
    if ((0.0 < T2) && (T2 < 1.0)){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
bool CGeLine::IsIntersect (const CGeLine& Line1, const CGeLine& Line2, const CGeLine& Line3, const CGeLine& Line4) const
{
  if (IsIntersect( Line1 )){
    return true;
  }
  if (IsIntersect( Line2 )){
    return true;
  }
  if (IsIntersect( Line3 )){
    return true;
  }
  if (IsIntersect( Line4 )){
    return true;
  }
  return false;
}

//-----------------------------------------------
bool CGeLine::IsEqual (const CGeLine& Line2, double Delta) const
{
  if (Delta > 0.0){
    if (m_Start.IsEqual( Line2.m_Start, Delta )){
      if (m_End.IsEqual( Line2.m_End, Delta )){
        return true;
      }
    }
  }else{
    if (m_Start == Line2.m_Start){
      if (m_End == Line2.m_End){
        return true;
      }
    }
  }
  return false;
}

//-----------------------------------------------
bool CGeLine::IsEqualBack (const CGeLine& Line2, double Delta) const
{
  if (Delta > 0.0){
    if (m_Start.IsEqual( Line2.m_End, Delta )){
      if (m_End.IsEqual( Line2.m_Start, Delta )){
        return true;
      }
    }
  }else{
    if (m_Start == Line2.m_End){
      if (m_End == Line2.m_Start){
        return true;
      }
    }
  }
  return false;
}

// return value:
//   0: no connection
//   1: 1st start - 2nd start
//   2: 1st start - 2nd end
//   3: 1st end - 2nd start
//   4: 1st end - 2nd end
//-----------------------------------------------
UINT_PTR CGeLine::GetConnection (const CGeLine& Line2, double Delta) const
{
  UINT_PTR Ret = 0;
  if (Delta > 0.0){
    if (m_Start.IsEqual( Line2.m_Start, g_GeTolerance )){
      Ret = 1;
    }else
    if (m_Start.IsEqual( Line2.m_End, g_GeTolerance )){
      Ret = 2;
    }else
    if (m_End.IsEqual( Line2.m_Start, g_GeTolerance )){
      Ret = 3;
    }else
    if (m_End.IsEqual( Line2.m_End, g_GeTolerance )){
      Ret = 4;
    }
  }else{
    if (m_Start == Line2.m_Start){
      Ret = 1;
    }else
    if (m_Start == Line2.m_End){
      Ret = 2;
    }else
    if (m_End == Line2.m_Start){
      Ret = 3;
    }else
    if (m_End == Line2.m_End){
      Ret = 4;
    }
  }
  return Ret;
}


//-----------------------------------------------
bool CGeLine::Crosses (const CGeRect& Rect) const
{
  return Rect.Crosses( *this );
}


//-----------------------------------------------
// Line's parameters A,B from 2 points
//-----------------------------------------------
void CGeLine::GetParamAB (double* pA1, double* pB1) const
{
  static double eps = 0.00000000001;

  if (((m_End.x - eps) < m_Start.x) && (m_Start.x < (m_End.x + eps))){
    *pA1 = 1.0 / eps;
  }else{
    if (((m_End.y - eps) < m_Start.y) && (m_Start.y < (m_End.y + eps))){
      *pA1 = eps;
    }else{
      *pA1 = (m_End.y - m_Start.y) / (m_End.x - m_Start.x);
    }
  }
  *pB1 = m_Start.y - m_Start.x * (*pA1);
}


//-----------------------------------------------
// Returns a distance from specified point to this line
// Dist < 0. means Pnt is placed at left side of the line
// Dist > 0. means Pnt is placed at right side of the line
//-----------------------------------------------
double CGeLine::DistToPoint (const CGePoint& Pnt) const
{
  double L, S, Dist;

  L = _hypot( m_Start.x - m_End.x, m_Start.y - m_End.y );
  if (L == 0.0){
    return m_Start.Distance( Pnt );
  }
  S = (m_Start.y - Pnt.y) * (m_End.x - m_Start.x) - (m_Start.x - Pnt.x) * (m_End.y - m_Start.y);
  Dist = S / L;
  return Dist;
}

//-----------------------------------------------
// Get point on line, nearest to given point
//   Pnt - given point
// Output:
//   pPntNear - nearest point on line, can be NULL
// Return:
//   distance to nearest point
//-----------------------------------------------
double CGeLine::GetNearPoint (const CGePoint& Pnt, CGePoint* pPntNear, double* pDistFromStart) const
{
  double Ang, cosAng, sinAng, dx, dist, Len;
  CGePoint Pt2;

  Ang = m_Start.Angle( m_End );
  cosAng = cos(Ang);
  sinAng = sin(Ang);
  dx = ((Pnt.x - m_Start.x) * cosAng) + ((Pnt.y - m_Start.y) * sinAng);
  if (dx < 0){
    // line's start point is nearest
    if (pPntNear){
      *pPntNear = m_Start;
    }
    if (pDistFromStart){
      *pDistFromStart = 0.0;
    }
    dist = _hypot( Pnt.x - m_Start.x, Pnt.y - m_Start.y );
    return dist; 
  }
  Len = m_Start.Distance( m_End );
//  len = ((m_End.x - m_Start.x) * cosAng) + ((m_End.y - m_Start.y) * sinAng);
  if (dx > Len){
    // line's end point is nearest
    if (pPntNear){
      *pPntNear = m_End;
    }
    if (pDistFromStart){
      *pDistFromStart = Length();
    }
    dist = m_End.Distance( Pnt );
    return dist;
  }
  // get perpendicular to line
  dist = ((Pnt.y - m_Start.y) * cosAng) - ((Pnt.x - m_Start.x) * sinAng);
  Pt2.x = m_Start.x + (dx * cosAng);
  Pt2.y = m_Start.y + (dx * sinAng);
  if (pPntNear){
    *pPntNear = Pt2;
  }
  if (pDistFromStart){
    *pDistFromStart = m_Start.Distance( Pt2 );
  }
  return fabs(dist);
}

//-----------------------------------------------
// Get point on line (pOutPnt), which make perpendicular from 'Pnt'
//   Pnt - given point
// Output:
//   pOutPnt - result point on line
//   pAlongDist - distance from line start to 'pOutPnt'
//   pPerpDist - distance from 'Pnt' to 'pOutPnt'
// Return:
//   distance to perpendicular point from 'Pnt'
//-----------------------------------------------
void CGeLine::GetPerpendPoint (const CGePoint& Pnt, CGePoint* pOutPnt, double* pAlongDist, double* pPerpDist) const
{
  double Dist, Ang, cosAng, sinAng, dx;

  Ang = m_Start.Angle( m_End );
  cosAng = cos(Ang);
  sinAng = sin(Ang);
  dx = ((Pnt.x - m_Start.x) * cosAng) + ((Pnt.y - m_Start.y) * sinAng);
  if (pOutPnt){
    pOutPnt->x = m_Start.x + (dx * cosAng);
    pOutPnt->y = m_Start.y + (dx * sinAng);
  }
  if (pAlongDist){
    *pAlongDist = dx;
  }
  if (pPerpDist){
    Dist = ((Pnt.y - m_Start.y) * cosAng) - ((Pnt.x - m_Start.x) * sinAng);
    *pPerpDist = fabs(Dist);
  }
}

//-----------------------------------------------
// Retrieves a point along the line and Dist from start
//-----------------------------------------------
double CGeLine::GetAlongPointA (double Dist, CGePoint* pPnt) const
{
  double DistCoef;

  DistCoef = Dist / Length();
  pPnt->x = m_Start.x + (DistCoef * (m_End.x - m_Start.x));
  pPnt->y = m_Start.y + (DistCoef * (m_End.y - m_Start.y));
  return DistCoef;
}

//-----------------------------------------------
// DistCoef - relative to the line length
//-----------------------------------------------
void CGeLine::GetAlongPointR (double DistCoef, CGePoint* pPnt) const
{
  pPnt->x = m_Start.x + (DistCoef * (m_End.x - m_Start.x));
  pPnt->y = m_Start.y + (DistCoef * (m_End.y - m_Start.y));
}

//-----------------------------------------------
void CGeLine::GetPline (double Delta, CGePolyline* pOutPline, bool bClosed)
{
  CGePoint Pt, Pt0;
  double Ang = Angle() + GE_DEG90;
  pOutPline->Clear();
  Pt = Pt0 = m_Start.PolarBy( Ang, -Delta );
  pOutPline->AddVer( Pt );
  Pt = m_Start.PolarBy( Ang, Delta );
  pOutPline->AddVer( Pt );
  Pt = m_End.PolarBy( Ang, Delta );
  pOutPline->AddVer( Pt );
  Pt = m_End.PolarBy( Ang, -Delta );
  pOutPline->AddVer( Pt );
  pOutPline->SetClosed( bClosed );
  if (!bClosed){
    pOutPline->AddVer( Pt0 );
  }
}


// Retrieves an array of points along the line
// nParts - number of parts the line to be divided, >= 2
// bStartEnd - if TRUE, then start & end points are included in pOutPnt array
// Return: number of points in the pOutPnt array
//-----------------------------------------------
UINT_PTR CGeLine::Divide (UINT_PTR nParts, CGePoint* pOutPnt, bool bStartEnd) const
{
  UINT_PTR nPnts = 0;
  UINT_PTR i;
  double Len, Dpart, Dist, DistCoef, dx, dy;

  if ((nParts < 2) || (1000 < nParts)){
    return 0;
  }
  if (bStartEnd){
    *pOutPnt = m_Start;
    ++pOutPnt;
    ++nPnts;
  }
  dx = m_End.x - m_Start.x;
  dy = m_End.y - m_Start.y;
  Len = _hypot( dx, dy );
  Dpart = Len / (double)nParts;
  Dist = Dpart;
  for (i=1; i<nParts; ++i){
    DistCoef = Dist / Len;
    pOutPnt->x = m_Start.x + (DistCoef * dx);
    pOutPnt->y = m_Start.y + (DistCoef * dy);
    Dist += Dpart;
    ++pOutPnt;
    ++nPnts;
  }
  if (bStartEnd){
    *pOutPnt = m_End;
    ++nPnts;
  }
  return nPnts;
}

//-----------------------------------------------
bool CGeLine::SplitByPoint (const CGePoint& Pnt, double MinLen, CGeLine* pLine1, CGeLine* pLine2) const
{
  EnPntPos Pos;
  double Len1, Len2, Delta;

  Delta = 0.001;
  Pos = ClassifyPoint2( Pnt, Delta );
  if (Pos == EnPntPos::GE_BETWEEN){
    Len1 = m_Start.Distance( Pnt );
    if (Len1 > MinLen){
      Len2 = Pnt.Distance( m_End );
      if (Len2 > MinLen){
        pLine1->Set( m_Start, Pnt );
        pLine2->Set( Pnt, m_End );
        return true;
      }
    }
  }
  return false;
}



//-----------------------------------------------
void CGeLine::Save (BYTE* pBuf) const
{
  static int sz = sizeof(CGePoint);
  BYTE* ptr;

  ptr = pBuf;
  memcpy( ptr, &m_Start, sz );  ptr+=sz;
  memcpy( ptr, &m_End, sz );
}


//-----------------------------------------------
void CGeLine::Restore (const BYTE* pBuf)
{
  static int sz = sizeof(CGePoint);
  const BYTE* ptr;

  ptr = pBuf;
  memcpy( &m_Start, ptr, sz );  ptr+=sz;
  memcpy( &m_End, ptr, sz );
}




// set end point by angle & length
//-----------------------------------------------
void CGeLine::SetEndPolar (double Angle, double Len)
{
  m_End = m_Start.PolarBy( Angle, Len );
}

//-----------------------------------------------
void CGeLine::SetLength (double Len)
{
  if (Len > 0.0){
    m_End = m_Start.PolarBy( Angle(), Len );
  }else{
    if (Len < 0.0){
      m_Start = m_End.PolarBy( Angle(), Len );
    }
  }
}

//-----------------------------------------------
void CGeLine::SetAngle (double Ang)
{
  m_End = m_Start.PolarBy( Ang, Length() );
}

//-----------------------------------------------
void CGeLine::Move (double dx, double dy)
{
  m_Start.Move( dx, dy );
  m_End.Move( dx, dy );
}

// parallel movement, Dist < 0 - to left side, 
//-----------------------------------------------
void CGeLine::Offset (double Dist)
{
  double Ang;

  Ang = Angle() - GE_DEG90;
  m_Start.Polar( Ang, Dist );
  m_End.Polar( Ang, Dist );
}

//-----------------------------------------------
void CGeLine::Rotate (const CGePoint& Pnt0, double Angle)
{
  m_Start.Rotate( Pnt0, Angle );
  m_End.Rotate( Pnt0, Angle );
}

//-----------------------------------------------
void CGeLine::Rotate (UINT_PTR iVer, double Angle)
{
  CGePoint Pt0;
  switch( iVer ){
    case 0: Pt0=m_Start; break;
    case 1: Pt0=Center(); break;
    case 2: Pt0=m_End; break;
    default:
      return;
  }
  Rotate( Pt0, Angle );
}

//-----------------------------------------------
void CGeLine::Scale (const CGePoint& Pnt0, double ScaleX, double ScaleY)
{
  m_Start.Scale( Pnt0, ScaleX, ScaleY );
  m_End.Scale( Pnt0, ScaleX, ScaleY );
}

//-----------------------------------------------
void CGeLine::Scale (const CGePoint& Pnt0, double Scal)
{
  m_Start.Scale( Pnt0, Scal );
  m_End.Scale( Pnt0, Scal );
}

//-----------------------------------------------
void CGeLine::Mirror (const CGeLine& Line)
{
  m_Start.Mirror( Line );
  m_End.Mirror( Line );
}

// 0-start, 1-end
//-----------------------------------------------
void CGeLine::CutByX (double X, int Mode)
{
  double coef, y, dx, dy;

  dx = m_End.x - m_Start.x;
  dy = m_End.y - m_Start.y;
  if (dx == 0.0){
    return;
  }
  coef = dy / dx;
  y = m_Start.y + (X - m_Start.x) * coef;
  if (Mode == 0){
    m_Start.Set( X, y ); 
  }else{
    m_End.Set( X, y );
  }
}

//-----------------------------------------------
void CGeLine::Transform (const GE_TRANSPRM* pPrm)
{
  m_Start.Transform( pPrm );
  m_End.Transform( pPrm );
}


//-----------------------------------------------
void CGeLine::Inflate (double Len0, double Len1)
{
  double Ang;

  Ang = Angle();
  if (Len0){
    m_Start.Polar( Ang, -Len0 );
  }
  if (Len1){
    m_End.Polar( Ang, Len1 );
  }
}

// used for horizontal line (see CHatchGen::Generate())
//-----------------------------------------------
bool CGeLine::Compensate (double LenL, double LenR, double MinLen)
{
  double LineLen;

  if (m_Start.x < m_End.x){
    // line goes from left to right
    LineLen = m_End.x - m_Start.x;
    LineLen -= (LenL + LenR);
    if (LineLen <= MinLen){
      return false;
    }
    m_Start.x += LenL;
    m_End.x -= LenR;
  }else{
    // line goes from right to left
    LineLen = m_Start.x - m_End.x;
    LineLen -= (LenL + LenR);
    if (LineLen <= MinLen){
      return false;
    }
    m_End.x += LenL;
    m_Start.x -= LenR;
  }
  return true;
}


//-----------------------------------------------
void CGeLine::InflateR (double Coef0, double Coef1)
{
  double Ang, Len, L;

  Ang = Angle();
  Len = Length();
  if (Coef0){
    L = Len * Coef0;
    m_Start.Polar( Ang, -L );
  }
  if (Coef1){
    L = Len * Coef1;
    m_End.Polar( Ang, L );
  }
}

//-----------------------------------------------
void CGeLine::Flip ()
{
  _dval = m_Start.x;
  m_Start.x = m_End.x;
  m_End.x = _dval;
  _dval = m_Start.y;
  m_Start.y = m_End.y;
  m_End.y = _dval;
}

// Rotates a line 90 degrees CCW around the line's center
//-----------------------------------------------
void CGeLine::Rot90ccw ()
{
  CGePoint cp;
  double dx, dy;

  cp = Center();
  dx = m_End.x - cp.x;
  dy = m_End.y - cp.y;
  m_Start.x = cp.x + dy;
  m_Start.y = cp.y - dx;
  m_End.x   = cp.x - dy;
  m_End.y   = cp.y + dx;
}

// Rotates a line 90 degrees CW around the line's center
//-----------------------------------------------
void CGeLine::Rot90cw ()
{
  CGePoint cp;
  double dx, dy;

  cp = Center();
  dx = m_End.x - cp.x;
  dy = m_End.y - cp.y;
  m_Start.x = cp.x - dy;
  m_Start.y = cp.y + dx;
  m_End.x   = cp.x + dy;
  m_End.y   = cp.y - dx;
}

//-----------------------------------------------
void CGeLine::AdjustFillet (const CGePoint& Pt0, const CGePoint& Pt1)
{
  double d0 = fabs(DistToPoint( Pt0 ));
  double d1 = fabs(DistToPoint( Pt1 ));

  if (d0 < d1){
    d0 = m_Start.Distance( Pt0 );
    d1 = m_End.Distance( Pt0 );
    if (d0 < d1){
      SetStart( Pt0 );
    }else{
      SetEnd( Pt0 );
    }
  }else{
    d0 = m_Start.Distance( Pt1 );
    d1 = m_End.Distance( Pt1 );
    if (d0 < d1){
      SetStart( Pt1 );
    }else{
      SetEnd( Pt1 );
    }
  }
}

//-----------------------------------------------
void CGeLine::AdjustPoint (const CGePoint& Pt)
{
  double d0 = m_Start.Distance( Pt );
  double d1 = m_End.Distance( Pt );
  if (d0 < d1){
    SetStart( Pt );
  }else{
    SetEnd( Pt );
  }
}

//-----------------------------------------------
void CGeLine::SetEnd (const CGePoint* Pnt, UINT_PTR nPnts, UINT_PTR* piSel)
{
  UINT_PTR i, j;
  double d, dmin;

  j = 0;
  dmin = 1000000000000.0;
  for (i=0; i<nPnts; ++i){
    d = m_Start.Distance( Pnt[i] );
    if (dmin > d){
      dmin = d;
      j = i;
    }
  }
  m_End = Pnt[j];
  if (piSel){
    *piSel = j;
  }
}




/*
//-----------------------------------------------
int CGeLine::GetInterPoints (const CGePolyline* pPline, CGePoint* OutPoint, int MaxPoints) const
{
  CGePoint pt0, pt1, ipt;
  CGeLine Line;
  EnIntersec res;
  int i;
  int nInterPt = 0;
  int npt = pPline->NumPoints();
  if (npt >= 2){
    pPline->FastGetPointAt( 0, &pt0 );
    for (i=1; i<npt; i++){
      pPline->FastGetPointAt( i, &pt1 );
      res = GetIntersection( CGeLine(pt0,pt1), &ipt );
      if (res == GE_SKEW_CROSS){
        if (nInterPt < MaxPoints){
          OutPoint[nInterPt++] = ipt;
        }else{
          break;
        }
      }
      pt0 = pt1;
    }
  }
  return nInterPt;
}
*/

/*
//-----------------------------------------------
bool CGeLine::GetInterPoint (const CGeXline& Xline, CGePoint* pInterPt, bool bApparent) const
{
  static double Zero = -0.00001;
  static double One  =  1.00001;
  static double Num, T, Nx, Ny, Denom;

  Nx = Line2.m_End.y - Line2.m_Start.y;
  Ny = Line2.m_Start.x - Line2.m_End.x;
  Denom = Nx * (m_End.x - m_Start.x) + Ny * (m_End.y - m_Start.y);
  if (-g_GeTolClass <= Denom && Denom <= g_GeTolClass){
    // lines are parallel or collinear
    if (IsEqual( Line2 ) || IsEqualBack( Line2 )){
      return false;
    }
    if (m_Start == Line2.m_Start || m_Start == Line2.m_End){
      *pInterPt = m_Start;
      return true;
    }else{
      if (m_End == Line2.m_Start || m_End == Line2.m_End){
        *pInterPt = m_End;
        return true;
      }
    }
  }else{
    Num = Nx * (m_Start.x - Line2.m_Start.x) + Ny * (m_Start.y - Line2.m_Start.y);
    T = -Num / Denom;
    if (bApparent){
      pInterPt->x = m_Start.x + T * (m_End.x - m_Start.x);
      pInterPt->y = m_Start.y + T * (m_End.y - m_Start.y);
      return true;
    }else{
      if (Zero <= T && T <= One){
        Nx = m_End.y - m_Start.y;
        Ny = m_Start.x - m_End.x;
        Denom = Nx * (Line2.m_End.x - Line2.m_Start.x) + Ny * (Line2.m_End.y - Line2.m_Start.y);
        Num = Nx * (Line2.m_Start.x - m_Start.x) + Ny * (Line2.m_Start.y - m_Start.y);
        T = -Num / Denom;
        if (Zero <= T && T <= One){
          pInterPt->x = Line2.m_Start.x + T * (Line2.m_End.x - Line2.m_Start.x);
          pInterPt->y = Line2.m_Start.y + T * (Line2.m_End.y - Line2.m_Start.y);
          return true;
        }
      }
    }
  }
  return false;
}
*/

