/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* This code was taken from 'BaseLibs' project
* Class CGePoint implementation
********************************************************************/
#include <math.h>
#include <stdlib.h>
#define _GE_GLOBALS_
#include "Geom.h"

double g_GeTolerance  = 0.0000001;
double g_GeTolClass   = 0.0000001;   // tolerance for CGeLine::ClassifyLine() and CGeLine::ClassifyPoint()
double g_GeMiterLimit = 10.0;

static double _Angle  = 0.0;
static double _cosAng = 1.0;
static double _sinAng = 0.0;

//-----------------------------------------------
void CGePoint::SetVector (double Angle)
{
  x = cos( Angle );
  y = sin( Angle );
}

//-----------------------------------------------
void CGePoint::SetVector (const CGePoint& pnt)
{
  x = pnt.x;
  y = pnt.y;
  VectorNormalize();
}

//-----------------------------------------------
void CGePoint::SetVector (const CGePoint& pt0, const CGePoint& pt1)
{
  x = pt1.x - pt0.x;
  y = pt1.y - pt0.y;
  VectorNormalize();
}

// Return the length of a vector
//-----------------------------------------------
double CGePoint::VectorLength () const
{
  return sqrt( x*x + y*y );
}

// Uniformly Scale a vector by a specified factor
//-----------------------------------------------
void CGePoint::VectorScale (double Scale)
{
  x *= Scale;
  y *= Scale;
}

// Scale a vector to 1-unit in length:
//-----------------------------------------------
void CGePoint::VectorNormalize ()
{
  double len = VectorLength();
  if (len > 0.){
    VectorScale( 1.0 / len );
  }
}
/*
//-----------------------------------------------
CGePoint CGePoint::VectorCross (const CGePoint& B) const
{
  CGePoint V;
  V.x = y * B.z - B.y * z;
  V.y = z * B.x - B.z * x;
  return V;
}
*/

//-----------------------------------------------
CGePoint CGePoint::PolarBy (double Angle, double Dist) const
{
  if (_Angle != Angle){
    _Angle  = Angle;
    _cosAng = cos(_Angle);
    _sinAng = sin(_Angle);
  }
  return CGePoint( x+Dist*_cosAng, y+Dist*_sinAng );
}

//-----------------------------------------------
void CGePoint::Polar (double Angle, double Dist)
{
  if (_Angle != Angle){
    _Angle  = Angle;
    _cosAng = cos(_Angle);
    _sinAng = sin(_Angle);
  }
  x += Dist * _cosAng;
  y += Dist * _sinAng;
}



//-----------------------------------------------
CGePoint CGePoint::RotateBy (const CGePoint& Base, double Angle) const
{
  double x1,y1, dx, dy;

  if (_Angle != Angle){
    _Angle  = Angle;
    _cosAng = cos(_Angle);
    _sinAng = sin(_Angle);
  }
  dx = x - Base.x;
  dy = y - Base.y;
  x1 = (dx * _cosAng) - (dy * _sinAng) + Base.x;
  y1 = (dx * _sinAng) + (dy * _cosAng) + Base.y;
  return CGePoint( x1, y1 );
}


//-----------------------------------------------
void CGePoint::Rotate (const CGePoint& Base, double Angle)
{
  double dx, dy;

  if (_Angle != Angle){
    _Angle  = Angle;
    _cosAng = cos(_Angle);
    _sinAng = sin(_Angle);
  }
  dx = x - Base.x;
  dy = y - Base.y;
  x = (dx * _cosAng) - (dy * _sinAng) + Base.x;
  y = (dx * _sinAng) + (dy * _cosAng) + Base.y;
}


//-----------------------------------------------
CGePoint CGePoint::MirrorBy (double a1, double b1) const
{
  double a2, b2, xp, yp;

  a2 = -1.0 / a1;
  b2 = y - x*a2;
  yp = (a1*b2 - a2*b1) / (a1-a2);
  xp = (yp - b1) / a1;
  return CGePoint( xp+xp-x, yp+yp-y );
}

//-----------------------------------------------
void CGePoint::Mirror (double a1, double b1)
{
  double a2, b2, xp, yp;

  a2 = -1.0 / a1;
  b2 = y - x*a2;
  yp = (a1*b2 - a2*b1) / (a1-a2);
  xp = (yp - b1) / a1;
  x = xp + xp - x;
  y = yp + yp - y;
}


//-----------------------------------------------
CGePoint CGePoint::MirrorBy  (const CGeLine& line) const
{
  CGePoint Pt;
  double  a1, b1;

  if (line.IsVertical()){
    Pt.x = line.Start().x + (line.Start().x - x);
    Pt.y = y;
  }else{
    if (line.IsHorizontal()){
      Pt.x = x;
      Pt.y = line.Start().y + (line.Start().y - y);
    }else{
      a1 = (line.End().y - line.Start().y) / (line.End().x - line.Start().x);
      b1 = line.Start().y - line.Start().x * a1;
//      line.GetParamAB( &a1, &b1 );
      Pt = MirrorBy( a1, b1 );
    }
  }
  return Pt;
}


//-----------------------------------------------
void CGePoint::Mirror (const CGeLine& line)
{
  double a1, b1, d;

  if (line.IsVertical()){
    d = line.Start().x - x;
    x = line.Start().x + d;
  }else{
    if (line.IsHorizontal()){
      d = line.Start().y - y;
      y = line.Start().y + d;
    }else{
//      line.GetParamAB( &a1, &b1 );
      a1 = (line.End().y - line.Start().y) / (line.End().x - line.Start().x);
      b1 = line.Start().y - line.Start().x * a1;
      Mirror( a1, b1 );
    }
  }
}

//-----------------------------------------------
CGePoint CGePoint::ReflectBy (const CGePoint& Base) const
{
  double Ang, Len;
  Ang = Angle( Base );
  Len = Distance( Base );
  return PolarBy( Ang, Len + Len );
}

//-----------------------------------------------
void CGePoint::Reflect (const CGePoint& Base)
{
  double Ang, Len;
  Ang = Angle( Base );
  Len = Distance( Base );
  Polar( Ang, Len + Len );
}

//-----------------------------------------------
CGePoint CGePoint::TransformBy (const GE_TRANSPRM* pPrm) const
{
  double tx;
  double ty;
  const GE_TRANSPRM* ptp;
  CGePoint Pt;

  Pt.Set( x, y );
  ptp = pPrm;
  while( ptp ){
    tx = Pt.x - ptp->Orig.x;
    ty = Pt.y - ptp->Orig.y;
    tx *= ptp->Scale.x;
    ty *= ptp->Scale.y;
    Pt.x = (tx * ptp->CosAng) - (ty * ptp->SinAng);
    Pt.y = (tx * ptp->SinAng) + (ty * ptp->CosAng);
    Pt.x += ptp->Dest.x;
    Pt.y += ptp->Dest.y;
    ptp = ptp->pNext;
  }
  return Pt;
}

//-----------------------------------------------
void CGePoint::Transform (const GE_TRANSPRM* pPrm)
{
  double tx;
  double ty;
  const GE_TRANSPRM* ptp;

  ptp = pPrm;
  while( ptp ){
    tx = x - ptp->Orig.x;
    ty = y - ptp->Orig.y;
    tx *= ptp->Scale.x;
    ty *= ptp->Scale.y;
    x = (tx * ptp->CosAng) - (ty * ptp->SinAng);
    y = (tx * ptp->SinAng) + (ty * ptp->CosAng);
    x += ptp->Dest.x;
    y += ptp->Dest.y;
    ptp = ptp->pNext;
  }
}

//-----------------------------------------------
CGePoint CGePoint::TransBackBy (const GE_TRANSPRM* pPrm) const
{
  static double Ang = 0.0;
  static double SinAng = 0.0;
  static double CosAng = 1.0;
  double x2, y2;
  CGePoint Pt;

  if (Ang != pPrm->Angle){
    Ang  = pPrm->Angle;
    SinAng = sin(-Ang);
    CosAng = cos(-Ang);
  }
  Pt.Set( x, y );
  Pt.x -= pPrm->Dest.x;
  Pt.y -= pPrm->Dest.y;
  x2 = (Pt.x * CosAng) - (Pt.y * SinAng);
  y2 = (Pt.x * SinAng) + (Pt.y * CosAng);
  x2 /= pPrm->Scale.x;
  y2 /= pPrm->Scale.y;
  Pt.x = pPrm->Orig.x + x2;
  Pt.y = pPrm->Orig.y + y2;
  return Pt;
}

//-----------------------------------------------
void CGePoint::TransBack (const GE_TRANSPRM* pPrm)
{
  double x2, y2;
  static double Ang = 0.0;
  static double SinAng = 0.0;
  static double CosAng = 1.0;

  if (Ang != pPrm->Angle){
    Ang  = pPrm->Angle;
    SinAng = sin(-Ang);
    CosAng = cos(-Ang);
  }
  x -= pPrm->Dest.x;
  y -= pPrm->Dest.y;
  x2 = (x * CosAng) - (y * SinAng);
  y2 = (x * SinAng) + (y * CosAng);
  x2 /= pPrm->Scale.x;
  y2 /= pPrm->Scale.y;
  x = pPrm->Orig.x + x2;
  y = pPrm->Orig.y + y2;
}


//-----------------------------------------------
// Angle from this point to point (0..2PI)
//-----------------------------------------------
double CGePoint::Angle (const CGePoint& p) const
{
  double a, b;

  a = p.x - x;
  b = p.y - y;
  if (a == 0.0) {
    if (b == 0.0){
      return 0.0;
    }
    return (b>0.0)? GE_PI2 : GE_DEG270;
  }else{
    if (b == 0.0){
      return (a>0.0)? 0.0 : GE_PI;
    }
    if (a > 0.0){
      if (b > 0.0){
        return atan( b / a );
      }else{
        return atan( b / a ) + GE_2PI;
      }
    }else{
      return atan( b / a ) + GE_PI;
    }
  }
}

//-----------------------------------------------
// Angle from point 0,0 to this point (0..2PI)
//-----------------------------------------------
double CGePoint::Angle () const
{
  if (x == 0.0) {
    if (y == 0.0){
      return 0.0;
    }
    return (y>0.0)? GE_PI2 : GE_DEG270;
  }else{
    if (y == 0.0){
      return (x>0.0)? 0.0 : GE_PI;
    }
    if (x > 0.0){
      if (y > 0.0){
        return atan( y / x );
      }else{
        return atan( y / x ) + GE_2PI;
      }
    }else{
      return atan( y / x ) + GE_PI;
    }
  }
}


//-----------------------------------------------
// Left side angle this-Pt1-Pt2 (0..2PI)
//-----------------------------------------------
double CGePoint::LeftAngle (const CGePoint& Pt1, const CGePoint& Pt2) const
{
  double Ang;

  Ang = Pt1.Angle(*this) - Pt1.Angle(Pt2);
  if (Ang < 0) {
    Ang += GE_2PI;
  }
  return Ang;
}

//-----------------------------------------------
// Right side angle this-Pt1-Pt2 (0..2PI)
//-----------------------------------------------
double CGePoint::RightAngle (const CGePoint& Pt1, const CGePoint& Pt2) const
{
  double Ang;

  Ang = Pt1.Angle(Pt2) - Pt1.Angle(*this);
  if (Ang < 0){
    Ang += GE_2PI;
  }
  return Ang;
}

//-----------------------------------------------
// Left side center angle this-Pt1-Pt2 (0..2PI)
//-----------------------------------------------
double CGePoint::LeftCenAngle (const CGePoint& Pt1, const CGePoint& Pt2) const
{
  double Ang1, Ang2, Ang;

  Ang1 = Pt1.Angle(*this);
  Ang2 = Pt1.Angle(Pt2);
  Ang = Ang1 - Ang2;
  if (Ang < 0){
    Ang += GE_2PI;
  }
  Ang2 += Ang/2.0;
  if (Ang2 > GE_2PI){
    Ang2 -= GE_2PI;
  }
  return Ang2;
}

//-----------------------------------------------
// Right side center angle this-Pt1-Pt2 (0..2PI)
//-----------------------------------------------
double CGePoint::RightCenAngle (const CGePoint& Pt1, const CGePoint& Pt2) const
{
  double Ang1, Ang2, Ang;

  Ang1 = Pt1.Angle(Pt2);
  Ang2 = Pt1.Angle(*this);
  Ang = Ang1 - Ang2;
  if (Ang < 0){
    Ang += GE_2PI;
  }
  Ang1 -= Ang/2.0;
  if (Ang1 < 0.0){
    Ang1 += GE_2PI;
  }
  return Ang1;
}

//-----------------------------------------------
// Center angle of bigger lef/rig angle this-Pt1-Pt2 (0..2PI)
//-----------------------------------------------
double CGePoint::CenterAngle (const CGePoint& Pt1, const CGePoint& Pt2, bool* pLeft) const
{
  double Ang1, Ang2, Ang, AngLef, AngRig;

  Ang1 = Pt1.Angle(*this);
  Ang2 = Pt1.Angle(Pt2);
  AngLef = Ang1 - Ang2;
  AngRig = Ang2 - Ang1;
  if (AngLef < 0){
    AngLef += GE_2PI;
  }
  if (AngRig < 0){
    AngRig += GE_2PI;
  }
  if (AngLef > AngRig){
    *pLeft = true;
    Ang = Ang2 + AngLef/2.0;
    if (Ang > GE_2PI){
      Ang -= GE_2PI;
    }
  }else{
    *pLeft = false;
    Ang = Ang2 - AngRig/2.0;
    if (Ang < 0.0){
      Ang += GE_2PI;
    }
  }
  return Ang;
}


//-----------------------------------------------
// Deviation angle from this-Pt1 direction (0..PI) + to left side, - to right side
//-----------------------------------------------
double CGePoint::TurnAngle (const CGePoint& Pt1, const CGePoint& Pt2) const
{
  double Ang;

  Ang = LeftAngle( Pt1, Pt2 );
  if (Ang <= GE_DEG180){
    Ang = GE_DEG180 - Ang;
  }else{
    Ang = -(Ang - GE_DEG180);
  }
  return Ang;
}


//-----------------------------------------------
// Get center of circle that goes through 3 points
//-----------------------------------------------
bool CGePoint::GetCenter (const CGePoint& Pt1, const CGePoint& Pt2, CGePoint* pCenter, double* pRadius) const
{
  CGeArc Arc;

  if (Arc.Set3P( *this, Pt1, Pt2 )){
    *pCenter = Arc.Center();
    *pRadius = Arc.Radius();
    return true;
  }
  return false;
}


//-----------------------------------------------
// Distance from this point to other point along the angle direction
//-----------------------------------------------
double CGePoint::Distance (const CGePoint& pnt, double Angle) const
{
  double x2, dx, dy, Dist;

  dx = pnt.x - x;
  dy = pnt.y - y;
  Dist = _hypot( dx, dy );
  _cosAng = cos(-Angle);
  _sinAng = sin(-Angle);
  x2 = (dx * _cosAng) - (dy * _sinAng);
  if (x2 < 0.0){
    Dist = -Dist;
  }
  return Dist;
}


// Shortest distance from this point to a line
//-----------------------------------------------
double CGePoint::Distance (const CGeLine& line, bool bApparent) const
{
  CGePoint InterPt;
  EnPntPos res;

  res = GetPerpendicular( line, &InterPt );
  if (bApparent){
    return Distance( InterPt );
  }
  if (res == EnPntPos::GE_BEFORE){
    return Distance( line.Start() );
  }
  if (res == EnPntPos::GE_AFTER){
    return Distance( line.End() );
  }
  return Distance( InterPt );
}


/********************************************************************
* Makes perpendicular line from this point to the specified line (Line)
*  pInterPt - (out) intersection point, 
*         if NULL - the point will not be calculated
* Returns:
*  position of the intersection point (pInterPt) on the line (Line)
*  GE_BETWEEN     - between the line's start and end points
*  GE_BEFORE      - before the line's start point
*  GE_AFTER       - after the line's end point
*  GE_ORIGIN      - exactly on the line's start point
*  GE_DESTINATION - exactly on the line's end point
*********************************************************************/
EnPntPos CGePoint::GetPerpendicular (const CGeLine& Line, CGePoint* pInterPt) const
{
  const double Zero = 0.0;  //-0.00000001;
  const double One  = 1.0;  // 1.00000001;
  double   t;
  CGeLine  Line2;
  CGePoint Pt2;

  Pt2 = PolarBy( Line.Angle()+GE_DEG90, 1000000000.0 );
  Line2.SetStart( x, y );
  Line2.SetEnd( Pt2 );
  if (Line.ClassifyLine( Line2, &t ) == EnIntersec::GE_SKEW){
    if (pInterPt){
      // intersection point
      pInterPt->x = Line.Start().x + t * (Line.End().x - Line.Start().x);
      pInterPt->y = Line.Start().y + t * (Line.End().y - Line.Start().y);
    }
    if (Zero < t && t < One){
      return EnPntPos::GE_BETWEEN;
    }
    if (t < Zero){
      return EnPntPos::GE_BEFORE;
    }
    if (One < t){
      return EnPntPos::GE_AFTER;
    }
    if (t == Zero){
      return EnPntPos::GE_ORIGIN;
    }
    if (t == One){
      return EnPntPos::GE_DESTINATION;
    }
  }
  if (pInterPt){
    pInterPt->x = pInterPt->y = 0;
  }
  return EnPntPos::GE_BETWEEN;
}

//-----------------------------------------------
EnPntPos CGePoint::Classify (const CGePoint& pt0, const CGePoint& pt1) const
{
  CGePoint pt2, a, b;
  double sa;

  pt2 = *this;
  a = pt1 - pt0;
  b = pt2 - pt0;
  sa = (a.x * b.y) - (b.x * a.y);
  if (sa > 0.0){
    return EnPntPos::GE_LEFT;
  }
  if (sa < 0.0){
    return EnPntPos::GE_RIGHT;
  }
  if ((a.x*b.x<0.0) || (a.y*b.y<0.0)){
    return EnPntPos::GE_AFTER;
  }
  if (a.Length() < b.Length()){
    return EnPntPos::GE_BEFORE;
  }
  if (pt0 == pt1){
    return EnPntPos::GE_ORIGIN;
  }
  if (pt1 == pt2){
    return EnPntPos::GE_DESTINATION;
  }
  return EnPntPos::GE_BETWEEN;
}






