/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* This code was taken from 'BaseLibs' project
* Class CGePolyline implementation
********************************************************************/
#include "Geom.h"

bool CGePolyline::m_bNoDuplicate = true;

//-----------------------------------------------
CGeVertex::CGeVertex () : COkObject (0,0,0)
{
  m_iParam = 0;
  m_fParam = m_fParam2 = 0.0;
  m_Dist = 0.0;
}

//-----------------------------------------------
CGeVertex::CGeVertex (const CGeVertex* pVerSrc) : COkObject (0,0,0)
{
  m_Pos     = pVerSrc->m_Pos;
  m_iParam  = pVerSrc->m_iParam;
  m_fParam  = pVerSrc->m_fParam;
  m_fParam2 = pVerSrc->m_fParam2;
  m_Dist = 0.0;
}

//-----------------------------------------------
CGeVertex::CGeVertex (const CGePoint& Pnt, int iPrm, double fPrm, double fPrm2) : COkObject (0,0,0)
{
  m_Pos = Pnt;
  m_iParam = iPrm;
  m_fParam = fPrm;
  m_fParam2 = fPrm2;
  m_Dist = 0.0;
}

//-----------------------------------------------
CGeVertex::CGeVertex (double X, double Y, double Z) : COkObject (0,0,0)
{
  m_Pos.Set( X, Y );
  m_iParam = 0;
  m_fParam = Z;
  m_fParam2 = 0.0;
  m_Dist = 0.0;
}



//-----------------------------------------------
CGePolyline::CGePolyline ()
{
  m_bClosed = false;
  m_NumVers = 0;
  m_Zval = 0.0;
  m_bBold = false;
}

//-----------------------------------------------
CGePolyline::CGePolyline (const CGePoint* Pnt, UINT_PTR nPnts, bool bClosed)
{
  UINT_PTR i;
  const CGePoint* pPnt;

  pPnt = Pnt;
  for (i=0; i<nPnts; ++i){
    AddVer( *pPnt );
    ++pPnt;
  }
  m_bClosed = bClosed;
  m_NumVers = 0;
  m_Zval = 0.0;
  m_bBold = false;
}

//-----------------------------------------------
CGePolyline::~CGePolyline ()
{
  Clear();
}


//-----------------------------------------------
void CGePolyline::Clear ()
{
  m_Vers.Clear();
  m_NumVers = 0;
  m_bClosed = false;
}

//-----------------------------------------------
CGeVertex* CGePolyline::AddVer (const CGeVertex* pVerSrc)
{
  CGeVertex* pVer;
  CGeVertex* pVerLast;

  if (m_bNoDuplicate){
    pVerLast = GetLastVer();
    if (pVerLast){
      if (pVerLast->m_Pos.IsEqual( pVerSrc->m_Pos, g_GeTolerance )){
        // duplicate point
        return NULL;
      }
    }
  }
  pVer = (CGeVertex*)m_Vers.Add( new CGeVertex(pVerSrc) );
  if (pVer){
    pVer->m_Dist = pVerSrc->m_Dist;
    ++m_NumVers;
  }
  return pVer;
}

//-----------------------------------------------
CGeVertex* CGePolyline::AddVer (const CGePoint& Pnt, int iPrm, double fPrm, double fPrm2)
{
  CGeVertex* pVer;
  CGeVertex* pVerLast;

  if (m_bNoDuplicate){
    pVerLast = GetLastVer();
    if (pVerLast){
      if (pVerLast->m_Pos.IsEqual( Pnt, g_GeTolerance )){
        // duplicate point
        return NULL;
      }
    }
  }
  pVer = (CGeVertex*)m_Vers.Add( new CGeVertex(Pnt,iPrm,fPrm,fPrm2) );
  if (pVer){
    ++m_NumVers;
  }
  return pVer;
}

//-----------------------------------------------
CGeVertex* CGePolyline::AddVer (double X, double Y, double Z)
{
  CGeVertex* pVer;
  CGeVertex* pVerLast;
  CGePoint Pnt;

  if (m_bNoDuplicate){
    pVerLast = GetLastVer();
    if (pVerLast){
      Pnt.Set(X,Y);
      if (pVerLast->m_Pos.IsEqual( Pnt, g_GeTolerance )){
        // duplicate point
        return NULL;
      }
    }
  }
  pVer = (CGeVertex*)m_Vers.Add( new CGeVertex(X,Y,Z) );
  if (pVer){
    ++m_NumVers;
  }
  return pVer;
}

//-----------------------------------------------
CGeVertex* CGePolyline::AddVer0 ()
{
  CGeVertex* pVer0;
  CGeVertex* pVer;
  CGeVertex* pVerLast;

  pVerLast = GetLastVer();
  if (pVerLast){
    pVer0 = (CGeVertex*)(m_Vers.GetFirst());
    if (pVerLast->m_Pos.IsEqual( pVer0->m_Pos, g_GeTolerance )){
      return NULL;
    }
    pVer = (CGeVertex*)m_Vers.Add( new CGeVertex(pVer0->m_Pos) );
    if (pVer){
      SetClosed( false );
      ++m_NumVers;
    }
    return pVer;
  }
  return NULL;
}


// added vertex before pBeforeVer
// if pBeforeVer==NULL, then new vertex will be added at beginning of the list
//-----------------------------------------------
CGeVertex* CGePolyline::InsertVer (CGeVertex* pBeforeVer, const CGePoint& Pnt, int iPrm, double fPrm, double fPrm2)
{
  CGeVertex* pVer;

  if (pBeforeVer){
    pVer = (CGeVertex*)m_Vers.InsertBefore( pBeforeVer, new CGeVertex(Pnt,iPrm,fPrm,fPrm2) );
  }else{
    pVer = (CGeVertex*)m_Vers.AddFirst( new CGeVertex(Pnt,iPrm,fPrm,fPrm2) );
  }
  if (pVer){
    ++m_NumVers;
  }
  return pVer;
}

//-----------------------------------------------
CGeVertex* CGePolyline::InsertVer (CGeVertex* pBeforeVer, double X, double Y, double Z)
{
  CGeVertex* pVer;

  if (pBeforeVer){
    pVer = (CGeVertex*)m_Vers.InsertBefore( pBeforeVer, new CGeVertex(X,Y,Z) );
  }else{
    pVer = (CGeVertex*)m_Vers.AddFirst( new CGeVertex(X,Y,Z) );
  }
  if (pVer){
    ++m_NumVers;
  }
  return pVer;
}

//-----------------------------------------------
CGeVertex* CGePolyline::InsertVer (double Dist)
{
  CGeVertex* pVerNew = NULL;
  CGeVertex* pVer;
  CGeVertex* pVerNext;
  double X, Y, Len, Len2, DistCoef;

  pVer = GetFirstVer();
  while( pVer ){
    pVerNext = GetNextVer( pVer );
    if (pVerNext){
      if (pVer->m_Dist == Dist){
        return pVer;
      }
      if (pVerNext->m_Dist == Dist){
        return pVerNext;
      }
      if ((pVer->m_Dist < Dist) && (Dist < pVerNext->m_Dist)){
        Len = pVerNext->m_Dist - pVer->m_Dist;
        Len2 = Dist - pVer->m_Dist;
        DistCoef = Len2 / Len;
        X = pVer->m_Pos.x + (DistCoef * (pVerNext->m_Pos.x - pVer->m_Pos.x));
        Y = pVer->m_Pos.y + (DistCoef * (pVerNext->m_Pos.y - pVer->m_Pos.y));
        pVerNew = (CGeVertex*)m_Vers.InsertAfter( pVer, new CGeVertex(X,Y) );
        pVerNew->m_Dist = Dist;
        return pVerNew;
      }
    }
    pVer = pVerNext;
  }
  if (m_bClosed){
    pVer = GetLastVer();
    pVerNext = GetFirstVer();
    Len = pVer->m_Pos.Distance( pVerNext->m_Pos );
    pVerNext->m_Dist = pVer->m_Dist + Len;
    if (pVer->m_Dist == Dist){
      return pVer;
    }
    if (pVerNext->m_Dist == Dist){
      return pVerNext;
    }
    if ((pVer->m_Dist < Dist) && (Dist < pVerNext->m_Dist)){
      Len = pVerNext->m_Dist - pVer->m_Dist;
      Len2 = Dist - pVer->m_Dist;
      DistCoef = Len2 / Len;
      X = pVer->m_Pos.x + (DistCoef * (pVerNext->m_Pos.x - pVer->m_Pos.x));
      Y = pVer->m_Pos.y + (DistCoef * (pVerNext->m_Pos.y - pVer->m_Pos.y));
      pVerNew = (CGeVertex*)m_Vers.InsertAfter( pVer, new CGeVertex(X,Y) );
      pVerNew->m_Dist = Dist;
      return pVerNew;
    }
  }
  return NULL;
}

//-----------------------------------------------
bool CGePolyline::SetStartVer (CGeVertex* pVer0)
{
  CGePolyline Pline2;
  CGeVertex*  pVer;
  COkID       idDel(438474711);
  UINT_PTR    nDels = 0;

  pVer = GetFirstVer();
  while( pVer ){
    if (pVer == pVer0){
      break;
    }else{
      Pline2.AddVer( pVer );
      pVer->SetID( idDel );
      ++nDels;
    }
    pVer = GetNextVer( pVer );
  }
  if (nDels){
    m_Vers.RemoveID( idDel );
    pVer = Pline2.GetFirstVer();
    while( pVer ){
      AddVer( pVer );
      pVer = Pline2.GetNextVer( pVer );
    }
    return true;
  }
  return false;
}

//-----------------------------------------------
void CGePolyline::SetVers (const CGePoint* Pnt, UINT_PTR nPnts, bool bClosed, bool bBackOrder)
{
  UINT_PTR i, j;
  const CGePoint* pPnt;

  pPnt = Pnt;
  Clear();
  if (nPnts >= 2){
    if (bBackOrder){
      if (bClosed){
        AddVer( *Pnt );
        j = nPnts - 1;
        pPnt = Pnt + j;
        for (i=0; i<nPnts; ++i){
          AddVer( *pPnt );
          --pPnt;
        }
      }else{
        j = nPnts - 1;
        pPnt = Pnt + j;
        for (i=0; i<nPnts; ++i){
          AddVer( *pPnt );
          --pPnt;
        }
      }
    }else{
      pPnt = Pnt;
      for (i=0; i<nPnts; ++i){
        AddVer( *pPnt );
        ++pPnt;
      }
    }
    if (nPnts > 2){
      m_bClosed = bClosed;
    }
  }
}

//-----------------------------------------------
void CGePolyline::SetRect (const CGePoint& Pt1, const CGePoint& Pt2, double Width)
{
  double Ang;
  double W2;
  CGePoint Pnt;

  if (Width > 0.0){
    Clear();
    Ang = Pt1.Angle( Pt2 ) + GE_DEG90;
    W2 = Width / 2.0;
    Pnt = Pt1.PolarBy( Ang, -W2 );
    AddVer( Pnt );
    Pnt = Pt1.PolarBy( Ang, W2 );
    AddVer( Pnt );
    Pnt = Pt2.PolarBy( Ang, W2 );
    AddVer( Pnt );
    Pnt = Pt2.PolarBy( Ang, -W2 );
    AddVer( Pnt );
    SetClosed( true );
  }
}

//-----------------------------------------------
void CGePolyline::CopyFrom (const CGePolyline& Pline2)
{
  CGeVertex* pVer2;

  Clear();
  pVer2 = Pline2.GetFirstVer();
  while( pVer2 ){
    AddVer( pVer2 );
    pVer2 = Pline2.GetNextVer( pVer2 );
  }
  m_bBold = Pline2.m_bBold;
  m_Zval = Pline2.m_Zval;
  SetClosed( Pline2.IsClosed() );
}

//-----------------------------------------------
UINT_PTR CGePolyline::AddFrom (const CGePolyline& Pline2)
{
  UINT_PTR nVers;
  CGeVertex* pVer2;

  nVers = 0;
  pVer2 = Pline2.GetFirstVer();
  while( pVer2 ){
    if (AddVer( pVer2->m_Pos )){
      ++nVers;
    }
    pVer2 = Pline2.GetNextVer( pVer2 );
  }
  return nVers;
}

// returns a number of added vertices
//-----------------------------------------------
UINT_PTR CGePolyline::AddFrom (const CGePolyline& Pline2, int Param)
{
  UINT_PTR nVers;
  CGeVertex* pVer;
  CGeVertex* pVer2;

  nVers = 0;
  pVer2 = Pline2.GetFirstVer();
  while( pVer2 ){
    pVer = AddVer( pVer2->m_Pos );
    if (pVer){
      pVer->m_iParam = Param;
      ++nVers;
    }
    pVer2 = Pline2.GetNextVer( pVer2 );
  }
  return nVers;
}



// For "closed" polyline - reset m_bClosed and add end vertex
//-----------------------------------------------
void CGePolyline::Unclose ()
{
  CGeVertex* pVer;

  if (m_bClosed){
    m_bClosed = false;
    pVer = GetFirstVer();
    AddVer( pVer->m_Pos, pVer->m_iParam, pVer->m_fParam, pVer->m_fParam2 );
  }
}

// For open polyline - make it "closed" if first and last vertex are equal
// Returns true if polyline becomes closed
//-----------------------------------------------
bool CGePolyline::TryClose ()
{
  CGeVertex* pVer0;
  CGeVertex* pVer2;

  if (!m_bClosed){
    pVer0 = GetFirstVer();
    pVer2 = GetLastVer();
    if (pVer0 != pVer2){
      if (pVer0->m_Pos.IsEqual( pVer2->m_Pos )){
        DelLastVer();
        m_bClosed = true;
      }
    }
  }
  return m_bClosed;
}

// returns vertex which becomes first after the deletion
//-----------------------------------------------
CGeVertex* CGePolyline::DelFirstVer ()
{
  CGeVertex* pVerNext;
  CGeVertex* pVer;

  pVer = GetFirstVer();
  if (pVer){
    pVerNext = GetNextVer( pVer );
    if (pVerNext){
      if (m_Vers.RemoveFirst()){
        --m_NumVers;
        if (m_NumVers < 2){
          m_bClosed = false;
        }
        return pVerNext;
      }
    }
  }
  return NULL;
}

// returns vertex which becomes last after the deletion
//-----------------------------------------------
CGeVertex* CGePolyline::DelLastVer ()
{
  CGeVertex* pVerPrev;
  CGeVertex* pVer;

  pVer = GetLastVer();
  if (pVer){
    pVerPrev = GetPrevVer( pVer );
    if (pVerPrev){
      if (m_Vers.RemoveLast()){
        --m_NumVers;
        if (m_NumVers < 2){
          m_bClosed = false;
        }
        return pVerPrev;
      }
    }
  }
  return NULL;
}

//-----------------------------------------------
UINT_PTR CGePolyline::DelVersID (COkID Id)
{
  UINT_PTR nDelVers;
  nDelVers = m_Vers.RemoveID( Id );
  if (nDelVers){
    m_NumVers = m_Vers.NumRecs();
  }
  return nDelVers;
}

//-----------------------------------------------
UINT_PTR CGePolyline::GetPoints (CGePoint* Pnt, UINT_PTR MaxPnts) const
{
  UINT_PTR i;
  CGePoint* pPnt;
  CGeVertex* pVer;

  i = 0;
  pPnt = Pnt;
  pVer = GetFirstVer();
  while( pVer ){
    if (i < MaxPnts){
      *pPnt = pVer->m_Pos;
      ++pPnt;
      ++i;
    }else{
      break;
    }
    pVer = GetNextVer( pVer );
  }
  return i;  // number of copied points
}

//-----------------------------------------------
bool CGePolyline::GetFirstPoint (CGePoint* pPnt, double* pDirAngle) const
{
  CGeVertex* pVer;

  pVer = GetFirstVer();
  if (pVer){
    *pPnt = pVer->m_Pos;
    if (pDirAngle){
      pVer = GetNextVer( pVer );
      if (pVer){
        *pDirAngle = pPnt->Angle( pVer->m_Pos );
        return true;
      }
    }else{
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
bool CGePolyline::GetLastPoint (CGePoint* pPnt, double* pDirAngle, bool bMeanClosed) const
{
  CGeVertex* pVer0;
  CGeVertex* pVer;

  pVer = GetLastVer();
  if (pVer){
    if (IsClosed() && bMeanClosed){
      pVer0 = GetFirstVer();
      *pPnt = pVer0->m_Pos;
      if (pVer0 == pVer){
        return false;
      }
      if (pDirAngle){
        *pDirAngle = pVer->m_Pos.Angle( *pPnt );
      }
      return true;
    }else{
      *pPnt = pVer->m_Pos;
      if (pDirAngle){
        pVer = GetPrevVer( pVer );
        if (pVer){
          *pDirAngle = pVer->m_Pos.Angle( *pPnt );
          return true;
        }
      }else{
        return true;
      }
    }
  }
  return false;
}

//-----------------------------------------------
bool CGePolyline::GetMidPoint (CGePoint* pPnt, double* pDirAngle) const
{
  CGeVertex* pVer;
  CGeVertex* pVer2;
  UINT_PTR nVers, i, k;

  nVers = NumVers();
  if (nVers >= 2){
    k = (nVers / 2) - 1;
    i = 0;
    pVer = GetFirstVer();
    while( pVer ){
      if (i == k){
        pVer2 = GetNextVer( pVer );
        *pPnt = pVer->m_Pos.Center( pVer2->m_Pos );
        *pDirAngle = pVer->m_Pos.Angle( pVer2->m_Pos );
        return true;
      }
      ++i;
      pVer = GetNextVer( pVer );
    }
  }
  return false;
}

//-----------------------------------------------
CGeRect CGePolyline::GeExtents () const
{
  CGeRect Rect;
  GetGeExtents( &Rect );
  return Rect; 
}


//-----------------------------------------------
void CGePolyline::GetGeExtents (CGeRect* pRect) const
{
  double Xmin, Xmax, Ymin, Ymax, x, y;
  CGeVertex* pVer;

  pVer = GetFirstVer();
  if (pVer){
    Xmin = Xmax = pVer->m_Pos.x;
    Ymin = Ymax = pVer->m_Pos.y;
    pVer = GetNextVer( pVer );
    while( pVer ){
      x = pVer->m_Pos.x;
      y = pVer->m_Pos.y;
      if (Xmin > x) Xmin = x;
      if (Xmax < x) Xmax = x;
      if (Ymin > y) Ymin = y;
      if (Ymax < y) Ymax = y;
      pVer = GetNextVer( pVer );
    }
    pRect->Set( Xmin, Ymin, Xmax, Ymax ); 
  }else{
    pRect->SetNull();
  }
}

//-----------------------------------------------
void CGePolyline::GetXmin (double* pX) const
{
  double x, Xmin;
  CGeVertex* pVer;

  pVer = GetFirstVer();
  if (pVer){
    Xmin = pVer->m_Pos.x;
    pVer = GetNextVer( pVer );
    while( pVer ){
      x = pVer->m_Pos.x;
      if (Xmin > x) Xmin = x;
      pVer = GetNextVer( pVer );
    }
    *pX = Xmin;
  }else{
    *pX = 0.0;
  }
}

//-----------------------------------------------
void CGePolyline::GetXmax (double* pX) const
{
  double x, Xmax;
  CGeVertex* pVer;
  pVer = GetFirstVer();
  if (pVer){
    Xmax = pVer->m_Pos.x;
    pVer = GetNextVer( pVer );
    while( pVer ){
      x = pVer->m_Pos.x;
      if (Xmax < x) Xmax = x;
      pVer = GetNextVer( pVer );
    }
    *pX = Xmax;
  }else{
    *pX = 0.0;
  }
}

//-----------------------------------------------
void CGePolyline::GetCenter (CGePoint* pPnt) const
{
  UINT_PTR   nLines;
  double     ex, ey, el;
  double     X, Y, Perim; 
  bool       bCont;
  CGeVertex* pVerNext;
  CGeVertex* pVer;

  nLines = 0;
  X = Y = Perim = 0.0;
  bCont = true;
  pVer = GetFirstVer();
  if (pVer){
    *pPnt = pVer->m_Pos;
    while( bCont ){
      pVerNext = GetNextVer( pVer );
      if (pVerNext == NULL){
        pVerNext = GetFirstVer();
        bCont = false;
      }
      ++nLines;
      ex = (pVer->m_Pos.x + pVerNext->m_Pos.x) / 2.0;
      ey = (pVer->m_Pos.y + pVerNext->m_Pos.y) / 2.0;
      el = pVer->m_Pos.Distance( pVerNext->m_Pos );
      X += (ex * el);
      Y += (ey * el);
      Perim += el;
      pVer = pVerNext;
    }
    if (nLines >= 3){
      pPnt->x = X / Perim;
      pPnt->y = Y / Perim;
    }
  }
}



//-----------------------------------------------
double CGePolyline::Length (bool bForceClosed) const
{
  double   Len, d;
  CGePoint Pt0, PtPrev;
  CGeVertex* pVer;

  Len = 0.0;
  pVer = GetFirstVer();
  if (pVer){
    Pt0 = PtPrev = pVer->m_Pos;
    pVer = GetNextVer( pVer );
    while( pVer ){
      d = PtPrev.Distance( pVer->m_Pos );
      Len += d;
      PtPrev = pVer->m_Pos;
      pVer = GetNextVer( pVer );
    }
    if (m_bClosed || bForceClosed){
      d = PtPrev.Distance( Pt0 );
      Len += d;
    }
  }
  return Len;
}

//-----------------------------------------------
bool CGePolyline::IsEmpty () const
{
  if (GetFirstVer()){
    return false;
  }
  return true;
}


//-----------------------------------------------
double CGePolyline::Area (bool bAbs) const
{
  double     S, Ar;
  CGeVertex* pVer;
  CGeVertex* pVerPrev;
  CGeVertex* pVerNext;

  Ar = 0.0;
  pVer = GetFirstVer();
  if (m_NumVers < 3){
    return 0.0;
  }
  while( pVer ){
    pVerPrev = GetPrevVer( pVer );
    if (pVerPrev == NULL){
      pVerPrev = GetLastVer();
    }
    pVerNext = GetNextVer( pVer );
    if (pVerNext == NULL){
      pVerNext = GetFirstVer();
    }
    S = pVer->X() * (pVerPrev->Y() - pVerNext->Y());
    Ar += S;
    pVer = GetNextVer( pVer );
  }
  if (bAbs){
    return fabs( Ar / 2.0 );
  }else{
    return Ar / 2.0;
  }
}

//-----------------------------------------------
bool CGePolyline::IsCW () const
{
  double Ar;
  Ar = Area( false );
  if (Ar > 0.0){
    // clockwise
    return true;
  }
  return false;
}

//-----------------------------------------------
bool CGePolyline::IsEqual (const CGePolyline& Pline2, double Delta) const
{
  CGeVertex* pVer;
  CGeVertex* pVer2;
  if ((Pline2.m_NumVers == m_NumVers) && (m_NumVers > 0)){
    // find a start point
    pVer = GetFirstVer();
    pVer2 = Pline2.GetFirstVer();
    while( pVer2 ){
      if (pVer->m_Pos.IsEqual( pVer2->m_Pos, Delta )){
        // start point is found
        for ( ; ; ){
          pVer = GetNextVer( pVer );
          if (pVer == NULL){
            return true;
          }
          pVer2 = Pline2.GetNextVer( pVer2 );
          if (pVer2 == NULL){
            pVer2 = Pline2.GetFirstVer();
          }
          if (pVer->m_Pos.IsEqual( pVer2->m_Pos, Delta ) == false){
            return false;
          }
        }
      }
      pVer2 = Pline2.GetNextVer( pVer2 );
    }
  }
  return false;
}

/********************************************************************
* Check if point is inside of polygon
*   Pnt  - input point
*   bCross - if TRUE then on-boundary point is treated as "inside"
* Returns
*   0 - outside
*   1 - inside
*   2 - on a polyline
*********************************************************************/
INT_PTR CGePolyline::Contains (const CGePoint& Pnt) const
{
  bool    bInside;
  double  kf,x, x0,y0, x1,y1;
  double  px, py;
  CGeVertex* pVer;

  bInside = false;
  px = Pnt.x;
  py = Pnt.y;
  if (m_NumVers < 3){
    return 0;
  }
  pVer = GetLastVer();
  x0 = pVer->X(); 
  y0 = pVer->Y(); 
  pVer = GetFirstVer();
  while( pVer ){
    x1 = pVer->X(); 
    y1 = pVer->Y(); 
    if (px <= x0 || px <= x1){
      if ((y0 <= py) && (py < y1) || (y1 < py) && (py <= y0)){
        if (y1 == y0){
          if ((py-g_GeTolerance < y1) && (y1 < py+g_GeTolerance)){
            return 2;   // on the polyline 
          }
        }else{
          if (x1 == x0){
            x = x0;
          }else{
            kf = (py - y0) / (y1 - y0);
            x = x0 + (kf * (x1 - x0));
          }
          if ((px-g_GeTolerance < x) && (x < px+g_GeTolerance)){
            return 2;   // on the polyline 
          }
          if (px < x){
            bInside = !bInside;
          }
        }
      }
    }
    x0 = x1;
    y0 = y1;
    pVer = GetNextVer( pVer );
  }
  return (bInside)? 1 : 0;   // 1-inside  0-outside 
}

//-----------------------------------------------
INT_PTR geContainsPoint (const CGePoint* Ver, UINT_PTR nVers, const CGePoint& Pnt)
{
  bool      bInside;
  double    kf,x, x0,y0, x1,y1;
  double    px, py;
  const CGePoint* pVer;
  UINT_PTR  i;

  bInside = false;
  px = Pnt.x;
  py = Pnt.y;
  if (nVers < 3){
    return 0;
  }
  pVer = Ver + nVers - 1;
  x0 = pVer->x; 
  y0 = pVer->y; 
  pVer = Ver;
  for (i=0; i<nVers; ++i){
    x1 = pVer->x; 
    y1 = pVer->y; 
    if (px <= x0 || px <= x1){
      if ((y0 <= py) && (py < y1) || (y1 < py) && (py <= y0)){
        if (y1 == y0){
          if ((py-g_GeTolerance < y1) && (y1 < py+g_GeTolerance)){
            return 2;   // on the polyline 
          }
        }else{
          if (x1 == x0){
            x = x0;
          }else{
            kf = (py - y0) / (y1 - y0);
            x = x0 + (kf * (x1 - x0));
          }
          if ((px-g_GeTolerance < x) && (x < px+g_GeTolerance)){
            return 2;   // on the polyline 
          }
          if (px < x){
            bInside = !bInside;
          }
        }
      }
    }
    x0 = x1;
    y0 = y1;
    ++pVer;
  }
  return (bInside)? 1 : 0;   // 1-inside  0-outside 
}


//  bool     Contains (const CGePolyline& Pline, bool bClosed) const;
/********************************************************************
* Check if other polygon is inside of this polygon
*   Pline - input polyline
* Returns
*   false - outside
*   true - inside
*********************************************************************/
bool CGePolyline::Contains (const CGePolyline& Pline2) const
{
  UINT_PTR   np2;
  CGeLine    Line;
  CGeVertex* pVerPrev;
  CGeVertex* pVer;

  np2 = Pline2.NumVers();
  if ((m_NumVers < 3) || (np2 < 2)){
    return false;
  }
  // check if contains vertices
  pVer = Pline2.GetFirstVer();
  while( pVer ){
    if (Contains( pVer->m_Pos ) == 0){
      // the point is outside
      return false;
    }
    pVer = Pline2.GetNextVer( pVer );
  }
  // check if crosses lines
  pVerPrev = Pline2.GetFirstVer();
  pVer = Pline2.GetNextVer( pVerPrev );
  while( pVer ){
    Line.Set( pVerPrev->m_Pos, pVer->m_Pos );
    if (Crosses( Line )){
      return false;
    }
    pVerPrev = pVer;
    pVer = Pline2.GetNextVer( pVer );
  }
  if (Pline2.IsClosed()){
    pVerPrev = Pline2.GetLastVer();
    pVer = Pline2.GetFirstVer();
    Line.Set( pVerPrev->m_Pos, pVer->m_Pos );
    if (Crosses( Line )){
      return false;
    }
  }
  return true;    // contains
}

//-----------------------------------------------
bool CGePolyline::Contains (const CGeRect& Rect) const
{
  CGePolyline Pline;
  Rect.GetPolyline( &Pline );
  return Contains( Pline );
}


//-----------------------------------------------
bool CGePolyline::Crosses (const CGeRect& Rect) const
{
  return Rect.Crosses( *this );
}


//-----------------------------------------------
bool CGePolyline::Crosses (const CGeLine& Line) const
{
  CGeLine    Line2;
  EnIntersec Inter;
  CGeVertex* pVer;
  CGeVertex* pVerPrev = GetFirstVer();

  if (m_NumVers < 2){
    return false;
  }
  pVer = GetNextVer( pVerPrev );
  while( pVer ){
    Line2.Set( pVerPrev->m_Pos, pVer->m_Pos );
    Inter = Line2.GetIntersection( Line, NULL );
    if (Inter == EnIntersec::GE_SKEW_CROSS){
      return true;
    }
    pVerPrev = pVer;
    pVer = GetNextVer( pVer );
  }
  if (m_bClosed){
    pVerPrev = GetLastVer();
    pVer = GetFirstVer();
    Line2.Set( pVerPrev->m_Pos, pVer->m_Pos );
    Inter = Line2.GetIntersection( Line, NULL );
    if (Inter == EnIntersec::GE_SKEW_CROSS){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
void CGePolyline::Move (double dx, double dy)
{
  CGeVertex* pVer;
  pVer = GetFirstVer();
  while( pVer ){
    pVer->m_Pos.Move( dx, dy );
    pVer = GetNextVer( pVer );
  }
}

//-----------------------------------------------
void CGePolyline::Polar (double Ang, double Dist)
{
  double dx, dy;
  CGeVertex* pVer;

  dx = Dist * cos(Ang);
  dy = Dist * sin(Ang);
  pVer = GetFirstVer();
  while( pVer ){
    pVer->m_Pos.x += dx;
    pVer->m_Pos.y += dy;
    pVer = GetNextVer( pVer );
  }
}


//-----------------------------------------------
void CGePolyline::Rotate (const CGePoint& BasePt, double Ang)
{
  double dx, dy, x0, y0;
  double cosAng, sinAng;
  CGeVertex* pVer;

  x0 = BasePt.x;
  y0 = BasePt.y;
  cosAng = cos( Ang );
  sinAng = sin( Ang );
  pVer = GetFirstVer();
  while( pVer ){
    dx = pVer->m_Pos.x - x0;
    dy = pVer->m_Pos.y - y0;
    pVer->m_Pos.x = (dx * cosAng) - (dy * sinAng) + x0;
    pVer->m_Pos.y = (dx * sinAng) + (dy * cosAng) + y0;
    pVer = GetNextVer( pVer );
  }
}

//-----------------------------------------------
void CGePolyline::RotateMove (const CGePoint& Pt0, double Ang, const CGePoint& Off)
{
  double dx, dy, x0, y0, dx2, dy2;
  double cosAng, sinAng;
  CGeVertex* pVer;

  x0 = Pt0.x;
  y0 = Pt0.y;
  dx2 = Pt0.x + Off.x;
  dy2 = Pt0.y + Off.y;
  cosAng = cos( Ang );
  sinAng = sin( Ang );
  pVer = GetFirstVer();
  while( pVer ){
    dx = pVer->m_Pos.x - x0;
    dy = pVer->m_Pos.y - y0;
    pVer->m_Pos.x = (dx * cosAng) - (dy * sinAng) + dx2;
    pVer->m_Pos.y = (dx * sinAng) + (dy * cosAng) + dy2;
    pVer = GetNextVer( pVer );
  }
}



//-----------------------------------------------
void CGePolyline::Scale (const CGePoint& BasePt, double ScaleX, double ScaleY)
{
  CGeVertex* pVer;
  pVer = GetFirstVer();
  while( pVer ){
    pVer->m_Pos.Scale( BasePt, ScaleX, ScaleY );
    pVer = GetNextVer( pVer );
  }
}


//-----------------------------------------------
void CGePolyline::Mirror (double a1, double b1)
{
  CGeVertex* pVer;
  pVer = GetFirstVer();
  while( pVer ){
    pVer->m_Pos.Mirror( a1, b1 );
    pVer = GetNextVer( pVer );
  }
}


//-----------------------------------------------
void CGePolyline::Mirror (const CGeLine& line)
{
  CGeVertex* pVer;
  pVer = GetFirstVer();
  while( pVer ){
    pVer->m_Pos.Mirror( line );
    pVer = GetNextVer( pVer );
  }
}

//-----------------------------------------------
void CGePolyline::Transform (const GE_TRANSPRM* pPrm, bool bExcludeNegPrm)
{
  CGeVertex* pVer;

  pVer = GetFirstVer();
  if (bExcludeNegPrm){
    while( pVer ){
      if (pVer->m_fParam >= 0.0){
        pVer->m_Pos.Transform( pPrm );
      }
      pVer = GetNextVer( pVer );
    }
  }else{
    while( pVer ){
      pVer->m_Pos.Transform( pPrm );
      pVer = GetNextVer( pVer );
    }
  }
}

//-----------------------------------------------
void CGePolyline::Transform (const CGe3dPoint& Pnt0, const CGe3dPoint& Zaxis)
{
  CGeVertex* pVer;
  CGe3dPoint Pt3d, Xaxis, Yaxis;

  pVer = GetFirstVer();
  Zaxis.GetArbitraryAxes( &Xaxis, &Yaxis );
  while( pVer ){
    Pt3d.x = pVer->m_Pos.x;
    Pt3d.y = pVer->m_Pos.y;
    Pt3d.z = 0.0;
    Pt3d.Transform( Xaxis, Yaxis, Zaxis );
    pVer->m_Pos.x = Pt3d.x + Pnt0.x;
    pVer->m_Pos.y = Pt3d.y + Pnt0.y;
    pVer->SetZ( Pt3d.z + Pnt0.z );
    pVer = GetNextVer( pVer );
  }
}

/*
// get polyline of 3D circle
//-----------------------------------------------
void CGeCircle::GetPolyline (CGe3dPoint* pNormVec, CGePolyline* pPline, UINT_PTR NumPoints) const
{
  CGePoint Pt0 = m_Center;
  CGe3dPoint Pt3d, Xaxis, Yaxis, Zaxis;

  pNormVec->GetArbitraryAxes( &Xaxis, &Yaxis );

  Pt3d.Transform( Xaxis, Yaxis, Zaxis );

  m_Center
  GetPolyline( pPline, NumPoints );

}
*/

//-----------------------------------------------
bool CGePolyline::TrimStart (double Dist, const CGePoint& Pnt)
{
  CGeVertex* pVerNext;
  CGeVertex* pVer;

  pVer = GetFirstVer();
  if (pVer){
//    UpdateDist();
m1:
    while( pVer ){
      if (pVer->m_Dist < Dist){
        pVerNext = GetNextVer( pVer );
        if (pVerNext){
          if (Dist < pVerNext->m_Dist){
            pVer->m_Pos = Pnt;
//            UpdateDist();
            return true;
          }
        }else{
          return false;
        }
        m_Vers.Remove( pVer );
        --m_NumVers;
        pVer = GetFirstVer();
        goto m1;
      }
      pVer = GetNextVer( pVer );
    }
  }
  return false;
}

//-----------------------------------------------
bool CGePolyline::TrimEnd (double Dist, const CGePoint& Pnt)
{
  CGeVertex* pVerPrev;
  CGeVertex* pVer;

  pVer = GetLastVer();
  if (pVer){
m1:
    while( pVer ){
      if (Dist < pVer->m_Dist){
        pVerPrev = GetPrevVer( pVer );
        if (pVerPrev){
          if (pVerPrev->m_Dist < Dist){
            pVer->m_Pos = Pnt;
            return true;
          }
        }else{
          return false;
        }
        m_Vers.Remove( pVer );
        --m_NumVers;
        pVer = GetLastVer();
        goto m1;
      }
      pVer = GetPrevVer( pVer );
    }
  }
  return false;
}


//-----------------------------------------------
bool CGePolyline::TrimMiddle (double Dist1, const CGePoint& Pnt1, double Dist2, const CGePoint& Pnt2, CGePolyline* pPline2)
{
  pPline2->CopyFrom( *this );
  if (TrimEnd( Dist1,  Pnt1 )){
    if (pPline2->TrimStart( Dist2, Pnt2 )){
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
bool CGePolyline::TrimPolyline (double Dist1, double Dist2, CGePolyline* pOutPline)
{
  CGeVertex* pVer;
  CGeVertex* pVerNext;
  double     Len, Len2, DistCoef, X, Y;
  UINT_PTR   Step = 0;

  pOutPline->Clear();
  pVer = GetFirstVer();
  while( pVer ){
    pVerNext = GetNextVer( pVer );
    if (pVerNext){
      if (Step == 0){
        if (pVer->m_Dist == Dist1){
          pOutPline->AddVer( pVer->m_Pos );
          Step = 1;
        }else
        if ((pVer->m_Dist < Dist1) && (Dist1 < pVerNext->m_Dist)){
          Len = pVerNext->m_Dist - pVer->m_Dist;
          Len2 = Dist1 - pVer->m_Dist;
          DistCoef = Len2 / Len;
          X = pVer->m_Pos.x + (DistCoef * (pVerNext->m_Pos.x - pVer->m_Pos.x));
          Y = pVer->m_Pos.y + (DistCoef * (pVerNext->m_Pos.y - pVer->m_Pos.y));
          pOutPline->AddVer( X, Y );
          Step = 1;
        }
      }
      if (Step == 1){
        if (pVerNext->m_Dist == Dist2){
          pOutPline->AddVer( pVerNext->m_Pos );
          return true;
        }else{
          if ((pVer->m_Dist < Dist2) && (Dist2 < pVerNext->m_Dist)){
            Len = pVerNext->m_Dist - pVer->m_Dist;
            Len2 = Dist2 - pVer->m_Dist;
            DistCoef = Len2 / Len;
            X = pVer->m_Pos.x + (DistCoef * (pVerNext->m_Pos.x - pVer->m_Pos.x));
            Y = pVer->m_Pos.y + (DistCoef * (pVerNext->m_Pos.y - pVer->m_Pos.y));
            pOutPline->AddVer( X, Y );
            return true;
          }else{
            pOutPline->AddVer( pVerNext->m_Pos );
          }
        }
      }
    }
    pVer = pVerNext;
  }
  
/*
  if (Step == 1){
    pVer = Pline.GetFirstVer();
    while( pVer ){
      pVerNext = Pline.GetNextVer( pVer );
      if (pVerNext){
        LinR.Set( pVer->m_Pos, pVerNext->m_Pos );
        if (LinR.GetInterPoint( Lin1, &InterPt )){
          pOutPline->AddVertex( InterPt );
          return true;
        }else{
          pOutPline->AddVertex( pVerNext->m_Pos );
        }
      }
      pVer = pVerNext;
    }
  }
*/
  return false;
}



//-----------------------------------------------
// Returns a distance from specified point to this polyline
// Dist < 0. means Pnt is placed at left side of the polyline
// Dist > 0. means Pnt is placed at right side of the polyline
//-----------------------------------------------
double CGePolyline::DistToPoint (const CGePoint& Pnt) const
{
  double    Ang, Len, Dist, AbsDist;
  double    MinDist;
  double    AbsMinDist;
  CGePoint  A, B, Pnt2;
  CGeVertex* pVerNext;
  CGeVertex* pVer;

  MinDist = 0.0;
  AbsMinDist = 1000000000000000000000.0;
  pVer = GetFirstVer();
  while( pVer ){
    pVerNext = GetNextVer( pVer );
    if (pVerNext == NULL){
      if (IsClosed()){
        pVerNext = GetFirstVer();
      }else{
        break;
      }
    }
    A = pVer->m_Pos;
    B = pVerNext->m_Pos;
    Ang = A.Angle( B );
    Len = A.Distance( B );
    Pnt2 = Pnt.RotateBy( A, -Ang );
    if ((A.x <= Pnt2.x) && (Pnt2.x <= (A.x+Len))){
      Dist = A.y - Pnt2.y;
      AbsDist = fabs( Dist );
      if (AbsMinDist > AbsDist){
        AbsMinDist = AbsDist;
        MinDist = Dist;
      }
    }
    pVer = GetNextVer( pVer );
  }
  return MinDist;
}


//-----------------------------------------------
void CGePolyline::Reverse ()
{
  if (m_bClosed){
    m_Vers.ShiftForward();
  }
  m_Vers.Reverse();
}


//-----------------------------------------------
CGeVertex* CGePolyline::GetVerByIndex (UINT_PTR iVer) const
{
  UINT_PTR i;
  CGeVertex* pVer;

  i = 0;
  pVer = GetFirstVer();
  if (iVer < m_NumVers){
    while( pVer ){
      if (i == iVer){
        return pVer;
      }
      ++i;
      pVer = GetNextVer( pVer );
    }
  }
  return NULL;
}

//-----------------------------------------------
CGeVertex* CGePolyline::GetVerByPos (const CGePoint& PickPt, double Delta) const
{
  CGeVertex* pVer;
  pVer = GetFirstVer();
  while( pVer ){
    if (pVer->m_Pos.IsEqual( PickPt, Delta )){
      return pVer;
    }
    pVer = GetNextVer( pVer );
  }
  return NULL;
}


//-----------------------------------------------
bool CGePolyline::GetPointAt (UINT_PTR iPnt, CGePoint* pPnt) const
{
  CGeVertex* pVer;

  pVer = GetVerByIndex( iPnt );
  if (pVer){
    *pPnt = pVer->m_Pos;
    return true;
  }
  return false;
}
