/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* This code was taken from 'BaseLibs' project
* Class CGeRect implementation
********************************************************************/
#include <math.h>
#include <stdlib.h>
#include <memory.h>
#include "Geom.h"

//-----------------------------------------------
void CGeRect::Set (const CGePoint& Center, double Width, double Height)
{
  double dx = Width / 2.0;
  double dy = Height / 2.0;
  m_Left   = Center.x - dx;
  m_Bottom = Center.y - dy;
  m_Right  = Center.x + dx;
  m_Top    = Center.y + dy;
}

//-----------------------------------------------
void CGeRect::Set (const CGePoint& Center, double Delta)
{
  m_Left   = Center.x - Delta;
  m_Bottom = Center.y - Delta;
  m_Right  = Center.x + Delta;
  m_Top    = Center.y + Delta;
}

//-----------------------------------------------
bool CGeRect::SetWidth (double w, bool bCenter)
{
  double xc, w2;
  if (w > 0.0){
    if (bCenter){
      xc = (m_Right + m_Left) / 2.0;
      w2 = w / 2.0;
      m_Left = xc - w2;
      m_Right = xc + w2;
    }else{
      m_Right = m_Left + w;
    }
    return true;
  }
  return false;
}

//-----------------------------------------------
bool CGeRect::SetHeight (double h, bool bCenter)
{
  double yc, h2;
  if (h > 0.){
    if (bCenter){
      yc = (m_Top + m_Bottom) / 2.0;
      h2 = h / 2.0;
      m_Bottom = yc - h2;
      m_Top = yc + h2;
    }else{
      m_Top = m_Bottom + h;
    }
    return true;
  }
  return false;
}

//-----------------------------------------------
void CGeRect::SetCenter (const CGePoint& CenPt)
{
  CGePoint Pt0 = Center();
  double   dx = CenPt.x - Pt0.x;
  double   dy = CenPt.y - Pt0.y;
  m_Left   += dx;
  m_Bottom += dy;
  m_Right  += dx;
  m_Top    += dy;
}

//-----------------------------------------------
void CGeRect::SetSize (double W, double H)
{
  CGePoint Pt0 = Center();
  double   dx = W / 2.0;
  double   dy = H / 2.0;
  m_Left   = Pt0.x - dx;
  m_Bottom = Pt0.y - dy;
  m_Right  = Pt0.x + dx;
  m_Top    = Pt0.y + dy;
}

//-----------------------------------------------
bool CGeRect::SetLeft (double X, bool bConstWidth)
{
  double W;
  if (bConstWidth){
    W = m_Right - m_Left;
    m_Left = X;
    m_Right = m_Left + W;
    return true;
  }else{
    if (X < m_Right){
      m_Left = X;
      return true;
    }
    return false;
  }
}

//-----------------------------------------------
bool CGeRect::SetRight (double X, bool bConstWidth)
{
  double W;
  if (bConstWidth){
    W = m_Right - m_Left;
    m_Right = X;
    m_Left = m_Right - W;
    return true;
  }else{
    if (m_Left < X){
      m_Right = X;
      return true;
    }
    return false;
  }
}

//-----------------------------------------------
bool CGeRect::SetBottom (double Y, bool bConstHeight)
{
  double H;
  if (bConstHeight){
    H = m_Top - m_Bottom;
    m_Bottom = Y;
    m_Top = m_Bottom + H;
    return true;
  }else{
    if (Y < m_Top){
      m_Bottom = Y;
      return true;
    }
    return false;
  }
}

//-----------------------------------------------
bool CGeRect::SetTop (double Y, bool bConstHeight)
{
  double H;
  if (bConstHeight){
    H = m_Top - m_Bottom;
    m_Top = Y;
    m_Bottom = m_Top - H;
    return true;
  }else{
    if (m_Bottom < Y){
      m_Top = Y;
      return true;
    }
    return false;
  }
}

//-----------------------------------------------
void CGeRect::SetExtents (const CGePoint& Pt0, const CGePoint& Pt1, const CGePoint& Pt2)
{
  m_Left = m_Right = Pt0.x;
  m_Bottom = m_Top = Pt0.y;
  // X
  if (m_Left > Pt1.x) m_Left = Pt1.x;
  if (m_Left > Pt2.x) m_Left = Pt2.x;
  if (m_Right < Pt1.x) m_Right = Pt1.x;
  if (m_Right < Pt2.x) m_Right = Pt2.x;
  // Y
  if (m_Bottom > Pt1.y) m_Bottom = Pt1.y;
  if (m_Bottom > Pt2.y) m_Bottom = Pt2.y;
  if (m_Top < Pt1.y) m_Top = Pt1.y;
  if (m_Top < Pt2.y) m_Top = Pt2.y;
}

//-----------------------------------------------
void CGeRect::SetExtents (const CGe3dPoint& Pt0, const CGe3dPoint& Pt1, const CGe3dPoint& Pt2)
{
  m_Left = m_Right = Pt0.x;
  m_Bottom = m_Top = Pt0.y;
  // X
  if (m_Left > Pt1.x) m_Left = Pt1.x;
  if (m_Left > Pt2.x) m_Left = Pt2.x;
  if (m_Right < Pt1.x) m_Right = Pt1.x;
  if (m_Right < Pt2.x) m_Right = Pt2.x;
  // Y
  if (m_Bottom > Pt1.y) m_Bottom = Pt1.y;
  if (m_Bottom > Pt2.y) m_Bottom = Pt2.y;
  if (m_Top < Pt1.y) m_Top = Pt1.y;
  if (m_Top < Pt2.y) m_Top = Pt2.y;
}


//-----------------------------------------------
void CGeRect::Normalize ()
{
  double val;
  if (m_Left > m_Right){
    val = m_Left;
    m_Left = m_Right;
    m_Right = val;
  }
  if (m_Bottom > m_Top){
    val = m_Bottom;
    m_Bottom = m_Top;
    m_Top = val;
  }
}

//-----------------------------------------------
double CGeRect::Ratio () const 
{
  double H = Height();
  if (H > 0.0){
    return Width() / H;
  }else{
    return 1000000000.;
  }
}

//-----------------------------------------------
double CGeRect::Diagonal () const
{
  double W = Width();
  double H = Height();
  double Len = _hypot( W, H );
  return Len;
}

//-----------------------------------------------
void CGeRect::GetPoints (CGePoint* pPnt) const
{
  pPnt[0] = LefBot();
  pPnt[1] = LefTop();
  pPnt[2] = RigTop();
  pPnt[3] = RigBot();
}

//-----------------------------------------------
void CGeRect::Get8Points (CGePoint* pPnt) const
{
  CGePoint Pt0 = Center();
  pPnt[0].Set( m_Left , m_Bottom );
  pPnt[1].Set( m_Left , Pt0.y );
  pPnt[2].Set( m_Left , m_Top);
  pPnt[3].Set( Pt0.x  , m_Top);
  pPnt[4].Set( m_Right, m_Top);
  pPnt[5].Set( m_Right, Pt0.y );
  pPnt[6].Set( m_Right, m_Bottom );
  pPnt[7].Set( Pt0.x  , m_Bottom );
}

//-----------------------------------------------
void CGeRect::GetPolyline (CGePolyline* pPline) const
{
  pPline->Clear();
  pPline->AddVer( LefBot() );
  pPline->AddVer( LefTop() );
  pPline->AddVer( RigTop() );
  pPline->AddVer( RigBot() );
  pPline->SetClosed();
}

//-----------------------------------------------
bool CGeRect::GetAndRect (const CGeRect& Rect2, CGeRect* pOutRect) const
{
  double L, B, R, T;
  if (Crosses( Rect2 )){
    L = (Rect2.m_Left > m_Left)? Rect2.m_Left : m_Left;
    R = (Rect2.m_Right < m_Right)? Rect2.m_Right : m_Right;
    B = (Rect2.m_Bottom > m_Bottom)? Rect2.m_Bottom : m_Bottom;
    T = (Rect2.m_Top < m_Top)? Rect2.m_Top : m_Top;
    pOutRect->Set( L, B, R, T );
    return true;
  }
  return false;
}

//-----------------------------------------------
void CGeRect::GetExtRect (const GE_TRANSPRM* pPrm, CGeRect* pOutRect) const
{
  CGePoint Pnt;
  Pnt = LefBot();
  Pnt.Transform( pPrm );
  pOutRect->Set( Pnt, 0.0 ); 
  Pnt = LefTop();
  Pnt.Transform( pPrm );
  pOutRect->UpdateExtentsBy( Pnt );
  Pnt = RigTop();
  Pnt.Transform( pPrm );
  pOutRect->UpdateExtentsBy( Pnt );
  Pnt = RigBot();
  Pnt.Transform( pPrm );
  pOutRect->UpdateExtentsBy( Pnt );
}

//-----------------------------------------------
bool CGeRect::Contains (const CGeLine& Line) const
{
  if (Contains(Line.Start()) && Contains(Line.End())){
    return true;
  }
  return false;
}

//-----------------------------------------------
bool CGeRect::Contains (const CGePoint& Pnt, double Delta) const
{
  static double L, B, R, T;
  L = m_Left - Delta; 
  B = m_Bottom - Delta; 
  R = m_Right + Delta; 
  T = m_Top + Delta;
  if ((L < Pnt.x) && (Pnt.x < R) && (B < Pnt.y) && (Pnt.y < T)){
    return true;
  }
  return false;
}

//-----------------------------------------------
bool CGeRect::NotContains (const CGePoint& Pnt, double Delta) const
{
  static double L, B, R, T;
  L = m_Left - Delta; 
  B = m_Bottom - Delta; 
  R = m_Right + Delta; 
  T = m_Top + Delta;
  if ((L < Pnt.x) && (Pnt.x < R) && (B < Pnt.y) && (Pnt.y < T)){
    return false;
  }
  return true;
}


//-----------------------------------------------
bool CGeRect::Crosses (const CGeLine& Line) const
{
  double   coef, x, y;
  CGePoint Pt0 = Line.Start();
  CGePoint Pt1 = Line.End();
  double   dx = Pt1.x - Pt0.x;
  double   dy = Pt1.y - Pt0.y;

  if (dy == 0.0){
    if (m_Bottom <= Pt0.y  &&  Pt0.y <= m_Top){
      if (dx > 0.0){
        if (m_Left <= Pt1.x  &&  Pt0.x <= m_Right){
          return true;
        }
      }else{
        if (m_Left <= Pt0.x  &&  Pt1.x <= m_Right){
          return true;
        }
      }
    }
    return false;
  }
  if (dx == 0){
    if (m_Left <= Pt0.x  &&  Pt0.x <= m_Right){
      if (dy > 0){
        if (m_Bottom <= Pt1.y  &&  Pt0.y <= m_Top){
          return true;
        }
      }else{
        if (m_Bottom <= Pt0.y  &&  Pt1.y <= m_Top){
          return true;
        }
      }
    }
    return false;
  }

  // test for intersection with Top and Bottom edges of selecting rectangle
  if (dy > 0){
    y = m_Bottom;
    if (Pt0.y <= y  &&  y <= Pt1.y){
      coef = (y - Pt0.y) / dy;
      x = Pt0.x + (Pt1.x - Pt0.x) * coef;
      if (m_Left <= x  &&  x <= m_Right){
        return true;
      }
    }
    y = m_Top;
    if (Pt0.y <= y  &&  y <= Pt1.y){
      coef = (y - Pt0.y) / dy;
      x = Pt0.x + (Pt1.x - Pt0.x) * coef;
      if (m_Left <= x  &&  x <= m_Right){
        return true;
      }
    }
  }else{
    y = m_Bottom;
    if (Pt1.y <= y  &&  y <= Pt0.y){
      coef = (y - Pt1.y) / -dy;
      x = Pt1.x + (Pt0.x - Pt1.x) * coef;
      if (m_Left <= x  &&  x <= m_Right){
        return true;
      }
    }
    y = m_Top;
    if (Pt1.y <= y  &&  y <= Pt0.y){
      coef = (y - Pt1.y) / -dy;
      x = Pt1.x + (Pt0.x - Pt1.x) * coef;
      if (m_Left <= x  &&  x <= m_Right){
        return true;
      }
    }
  }

  // test for intersection with Left and right edges of selecting rectangle
  if (dx > 0){
    x = m_Left;
    if (Pt0.x <= x  &&  x <= Pt1.x){
      coef = (x - Pt0.x) / dx;
      y = Pt0.y + (Pt1.y - Pt0.y) * coef;
      if (m_Bottom <= y  &&  y <= m_Top){
        return true;
      }
    }
    x = m_Right;
    if (Pt0.x <= x  &&  x <= Pt1.x){
      coef = (x - Pt0.x) / dx;
      y = Pt0.y + (Pt1.y - Pt0.y) * coef;
      if (m_Bottom <= y  &&  y <= m_Top){
        return true;
      }
    }
  }else{
    x = m_Left;
    if (Pt1.x <= x  &&  x <= Pt0.x){
      coef = (x - Pt1.x) / -dx;
      y = Pt1.y + (Pt0.y - Pt1.y) * coef;
      if (m_Bottom <= y  &&  y <= m_Top){
        return true;
      }
    }
    x = m_Right;
    if (Pt1.x <= x  &&  x <= Pt0.x){
      coef = (x - Pt1.x) / -dx;
      y = Pt1.y + (Pt0.y - Pt1.y) * coef;
      if (m_Bottom <= y  &&  y <= m_Top){
        return true;
      }
    }
  }
  return false;
}


//-----------------------------------------------
bool CGeRect::Crosses (const CGeCircle& Circ, bool bPoint) const
{
  return Circ.Crosses( *this, bPoint );
}


//-----------------------------------------------
bool CGeRect::Crosses (const CGeArc& Arc, bool bPoint) const
{
  return Arc.Crosses( *this, bPoint );
}

//-----------------------------------------------
bool CGeRect::Crosses (const CGePolyline& Pline) const
{
  CGeLine    Line;
  CGeVertex* pVer;
  CGeVertex* pVerPrev = Pline.GetFirstVer();;

  if (pVerPrev){
    pVer = Pline.GetNextVer( pVerPrev );
    if (pVer){
      while( pVer ){
        Line.Set( pVerPrev->m_Pos, pVer->m_Pos );
        if (Crosses( Line )){
          return true;
        }
        pVerPrev = pVer;
        pVer = Pline.GetNextVer( pVer );
      }
      if (Pline.IsClosed()){
        pVerPrev = Pline.GetLastVer();
        pVer = Pline.GetFirstVer();
        Line.Set( pVerPrev->m_Pos, pVer->m_Pos );
        if (Crosses( Line )){
          return true;
        }
      }
    }
  }
  return false;
}

/*
//-----------------------------------------------
bool CGeRect::Crosses (const CGePolylinA& Pline) const
{
  CGeLine  line;
  CGePoint pnt;
  UINT_PTR np = Pline.NumPoints();
  UINT_PTR i;

  Pline.GetPointAt( 0, &pnt );
  line.SetStart( pnt );
  for (i=1; i<np; i++){
    Pline.GetPointAt( i, &pnt );
    line.SetEnd( pnt );
    if (Crosses( line )){
      return true;
    }
    line.SetStart( pnt );
  }
  return false;
}
*/


//-----------------------------------------------
void CGeRect::Scale (double Coef, bool bByCenter)
{
  CGePoint cen;
  double   dx, dy;

  if (Coef > 0.0){
    if (bByCenter){
      cen = Center();
      dx = Width() / 2.0 * Coef;
      dy = Height() / 2.0 * Coef;
      Set( cen.x - dx, cen.y + dy, cen.x + dx, cen.y - dy );
    }else{
      m_Left *= Coef;
      m_Bottom *= Coef;
      m_Right *= Coef;
      m_Top *= Coef;
    }
  }
}

// Increase rect to include rotated rect
//-----------------------------------------------
void CGeRect::InflateRot (double Angle)
{
  CGePolyline Pline;
  CGeRect Rect2;
  CGePoint Pt0 = Center();

  GetPolyline( &Pline );
  Pline.Rotate( Pt0, Angle );
  Pline.GetGeExtents( &Rect2 );
  m_Left   = Rect2.m_Left;
  m_Bottom = Rect2.m_Bottom;
  m_Right  = Rect2.m_Right;
  m_Top    = Rect2.m_Top;
}


//-----------------------------------------------
CGeRect CGeRect::MoveBy  (double dx, double dy) const
{
  CGeRect Rect;
  Rect.m_Left   = m_Left + dx; 
  Rect.m_Bottom = m_Bottom + dy; 
  Rect.m_Right  = m_Right + dx; 
  Rect.m_Top    = m_Top + dy;
  return Rect;
}

//-----------------------------------------------
CGeRect CGeRect::ScaleBy (double Coef, bool bByCenter) const
{
  CGeRect Rect = *this;
  Rect.Scale( Coef, bByCenter );
  return Rect;
}

//-----------------------------------------------
void CGeRect::MirrorX (double Xe)
{
  double  dX, x1, x2;

  dX = Xe - m_Left;
  x1 = Xe + dX;
  dX = Xe - m_Right;
  x2 = Xe + dX;
  m_Left = x2;
  m_Right = x1;
}




/********************************************************************
* Clip line
*   Start, End - input line
*   OutPt1, OutPt2 - output points of clipped line
* Returns:
*   0 - Input line is completely outside, OutPt1 & OutPt2 not used
*   1 - Input line is completely inside,  OutPt1=Start, OutPt2=End
*   2 - Input line was clipped by Start side,  OutPt1!=Start, OutPt2=End
*   3 - Input line was clipped by End side,  OutPt1=Start, OutPt2!=End
*   4 - Input line was clipped by both sides, OutPt1!=Start, OutPt2!=End
*********************************************************************/
INT_PTR CGeRect::ClipLine (const CGePoint& Start, const CGePoint& End, CGePoint* pOutPt1, CGePoint* pOutPt2) const
{
  double  tX1 = Start.x;
  double  tY1 = Start.y;
  double  tX2 = End.x;
  double  tY2 = End.y;
  double  tDX=0;
  double  tDY=0;
  INT_PTR tPos1=0;
  INT_PTR tPos2=0;

  // test - if limits of input line is completely outside 
  // we test it first because it will be most often case
  if (tX1 < tX2){
    if (tX1 > m_Right || tX2 < m_Left){
      return 0;
    }
  }else{
    if (tX2 > m_Right || tX1 < m_Left){
      return 0;
    }
  }
  if (tY1 < tY2){
    if (tY1 > m_Top || tY2 < m_Bottom){
      return 0;
    }
  }else{
    if (tY2 > m_Top || tY1 < m_Bottom){
      return 0;
    }
  }

  tPos1 = (m_Left <= tX1 && tX1 <= m_Right && m_Bottom <= tY1 && tY1 <= m_Top)? true : false;
  tPos2 = (m_Left <= tX2 && tX2 <= m_Right && m_Bottom <= tY2 && tY2 <= m_Top)? true : false;

  pOutPt1->Set( tX1, tY1 );
  pOutPt2->Set( tX2, tY2 );

  if (tPos1 && tPos2){
    return 1;  // completely inside
  }

  if (tPos2){
    // Inp2 is inside, Inp1 is outside 
    if (pOutPt1->y > m_Top){
      // clip above
      pOutPt1->Set( tX2 + (tX1 - tX2) * (m_Top - tY2) / (tY1 - tY2),  m_Top );
    }else{
      if (pOutPt1->y < m_Bottom){
        // clip below
        pOutPt1->Set( tX2 + (tX1 - tX2) * (tY2 - m_Bottom) / (tY2 - tY1),  m_Bottom );
      }
    }
    if (pOutPt1->x > m_Right){
      // clip right
      pOutPt1->Set( m_Right, tY2 + (tY1 - tY2) * (m_Right - tX2) / (tX1 - tX2) );
    }else{
      if (pOutPt1->x < m_Left){
        // clip Left
        pOutPt1->Set( m_Left,  tY2 + (tY1 - tY2) * (tX2 - m_Left) / (tX2 - tX1) );
      }
    }
    return 2; // Input line was clipped by Start side
  }

  if (tPos1){
    // Inp1 is inside, Inp2 is outside
    if (pOutPt2->y > m_Top){
      // clip above
      pOutPt2->Set( tX1 + (tX2 - tX1) * (m_Top - tY1) / (tY2 - tY1),  m_Top );
    }else{
      if (pOutPt2->y < m_Bottom){
        // clip below
        pOutPt2->Set( tX1 + (tX2 - tX1) * (tY1 - m_Bottom) / (tY1 - tY2),  m_Bottom );
      }
    }
    if (pOutPt2->x > m_Right){
      // clip right
      pOutPt2->Set( m_Right,  tY1 + (tY2 - tY1) * (m_Right - tX1) / (tX2 - tX1) );
    }else{
      if (pOutPt2->x < m_Left){
        // clip Left
        pOutPt2->Set( m_Left,  tY1 + (tY2 - tY1) * (tX1 - m_Left) / (tX1 - tX2) );
      }
    }
    return 3; // Input line was clipped by End side
  }

  // Now we need to test for case 4 - Input line is clipped 
  // by both sides, OutPt1!=Start, OutPt2!=End 

  // clip by Left & right
  if (tX1 < tX2){
    tDX = tX2 - tX1;
    tDY = tY2 - tY1;
    if (tX1 < m_Left){
      tY1 = tY1 + tDY * (m_Left - tX1) / tDX;
      if (tY1 > m_Top && tY2 > m_Top || tY1 < m_Bottom && tY2 < m_Bottom){
        return 0;   // input line is completely outside
      }
      tX1 = m_Left;
      tDX = tX2 - tX1;
      tDY = tY2 - tY1;
    }
    if (m_Right < tX2){
      tY2 = tY1 + tDY * (m_Right - tX1) / tDX;
      if (tY1 > m_Top && tY2 > m_Top || tY1 < m_Bottom && tY2 < m_Bottom){
        return 0;   // input line is completely outside
      }
      tX2 = m_Right;
    }
  }else{
    tDX = tX1 - tX2;
    tDY = tY1 - tY2;
    if (tX2 < m_Left){
      tY2 = tY2 + tDY * (m_Left - tX2) / tDX;
      if (tY1 > m_Top && tY2 > m_Top || tY1 < m_Bottom && tY2 < m_Bottom){
        return 0;   // input line is completely outside  
      }
      tX2 = m_Left;
      tDX = tX1 - tX2;
      tDY = tY1 - tY2;
    }
    if (m_Right < tX1){
      tY1 = tY2 + tDY * (m_Right - tX2) / tDX;
      if (tY1 > m_Top && tY2 > m_Top || tY1 < m_Bottom && tY2 < m_Bottom){
        return 0;   // input line is completely outside
      }
      tX1 = m_Right;
    }
  }

  // clip by Top & Bottom
  if (tY1 < tY2){
    tDX = tX2 - tX1;
    tDY = tY2 - tY1;
    if (tY1 < m_Bottom){
      tX1 = tX1 + tDX * (m_Bottom - tY1) / tDY;
      if (tX1 > m_Right && tX2 > m_Right || tX1 < m_Left && tX2 < m_Left){
        return 0;   // input line is completely outside
      }
      tY1 = m_Bottom;
      tDX = tX2 - tX1;
      tDY = tY2 - tY1;
    }
    if (m_Top < tY2){
      tX2 = tX1 + tDX * (m_Top - tY1) / tDY;
      if (tX1 > m_Right && tX2 > m_Right || tX1 < m_Left && tX2 < m_Left){
        return 0;   // input line is completely outside
      }
      tY2 = m_Top;
    }
  }else{
    tDX = tX1 - tX2;
    tDY = tY1 - tY2;
    if (tY2 < m_Bottom){
      tX2 = tX2 + tDX * (m_Bottom - tY2) / tDY;
      if (tX1 > m_Right && tX2 > m_Right || tX1 < m_Left && tX2 < m_Left){
        return 0;   // input line is completely outside
      }
      tY2 = m_Bottom;
      tDX = tX1 - tX2;
      tDY = tY1 - tY2;
    }
    if (m_Top < tY1){
      tX1 = tX2 + tDX * (m_Top - tY2) / tDY;
      if (tX1 > m_Right && tX2 > m_Right || tX1 < m_Left && tX2 < m_Left){
        return 0;   // input line is completely outside
      }
      tY1 = m_Top;
    }
  }
  pOutPt1->Set( tX1, tY1 );
  pOutPt2->Set( tX2, tY2 );
  return 4; // Input line was clipped at both ends
}

//-----------------------------------------------
INT_PTR CGeRect::ClipLine (const CGeLine& Line, CGeLine* pOutLine) const
{
  CGePoint OutPt1, OutPt2;
  INT_PTR ret = ClipLine( Line.Start(), Line.End(), &OutPt1, &OutPt2 );
  pOutLine->Set( OutPt1, OutPt2 );
  return ret;
}




#define MAX_CLIPVER 16000

// point state for _ClipPntState
#define OUTSIDE  0
#define BORDER   1
#define INSIDE   2
#define DISABLED 100

// temporary variables used to clip Polyline
static CGePoint _ClipPntBuf[MAX_CLIPVER+8];   // buffer for output polyline
static BYTE     _ClipPntState[MAX_CLIPVER+8];   // state of vertex, OUTSIDE, BORDER, INSIDE
static UINT_PTR _iver = 0;              // counter of vertexes 

/********************************************************************
* Clip polyline
*   Pline - input polyline 
*   bClosed - if false - use Pline.IsClosed(), if true - forced closed
* Returns:
*  -1 - error
*   0 - polyline is completely outside of the clip rect
*   1 - polyline was clipped
*   2 - polyline is completely inside of the clip rect
*********************************************************************/
INT_PTR CGeRect::ClipPolyline (const CGePolyline& Pline, bool bClosed) const
{
  INT_PTR    Ret, i, n;
  INT_PTR    Outside = 0;
  INT_PTR    Inside = 0;
  INT_PTR    Clipped = 0;
  BYTE       State;
  bool       bEnd = false;
  bool       bOpenStart = false;
  CGePoint   Pt0, Pt1;
  CGeVertex* pVerNext;
  CGeVertex* pVer1;
  CGeVertex* pVer0 = Pline.GetFirstVer();

  _iver = 0;
  if (pVer0 == NULL){
    return -1;  // error
  }
  pVer1 = Pline.GetNextVer( pVer0 );
  if (pVer1 == NULL){
    // polyline has only 1 vertex
    if (Contains( pVer0->m_Pos )){
      _ClipPntBuf[0] = pVer0->m_Pos;
      _iver = 1;
      return 2;  // vertex is inside of clip rect
    }else{
      return 0;  // vertex is outside
    }
  }
  if (bClosed == false){
    bClosed = Pline.IsClosed();
  }
  while( pVer1 ){
    if (bEnd){
      pVerNext = NULL;
    }else{
      pVerNext = Pline.GetNextVer( pVer1 );
      if (pVerNext == NULL){
        if (bClosed){
          pVerNext = Pline.GetFirstVer();
          bEnd = true;
        }
      }
    }
    Ret = ClipLine( pVer0->m_Pos, pVer1->m_Pos, &Pt0, &Pt1 );
    switch( Ret ){
      case 0:  // Input line is completely outside
        ++Outside;
        break;
      case 1:  // Input line is completely inside
        if ((_iver == 0) && bClosed){
          bOpenStart = true;
        }
        _ClipPntBuf[_iver] = pVer0->m_Pos;
        _ClipPntState[_iver] = INSIDE;
        ++_iver;
        if (pVerNext == NULL){
          if (bOpenStart && Clipped){
            goto m1;
          }
          _ClipPntBuf[_iver] = pVer1->m_Pos;
          _ClipPntState[_iver] = INSIDE;
          ++_iver;
        }
        ++Inside;
        break;
      case 2:  // Input line was clipped by Start side
        _ClipPntBuf[_iver] = Pt0;
        _ClipPntState[_iver] = BORDER;
        ++_iver;
        if (pVerNext == NULL){
          if (bOpenStart && Clipped){
            goto m1;
          }
          _ClipPntBuf[_iver] = pVer1->m_Pos;
          _ClipPntState[_iver] = INSIDE;
          ++_iver;
        }
        ++Clipped;
        break;
      case 3:  // Input line was clipped by End side
        if ((_iver == 0) && bClosed){
          bOpenStart = true;
        }
        _ClipPntBuf[_iver] = pVer0->m_Pos;
        _ClipPntState[_iver] = INSIDE;
        ++_iver;
        _ClipPntBuf[_iver] = Pt1;
        _ClipPntState[_iver] = BORDER;
        ++_iver;
        _ClipPntState[_iver] = OUTSIDE;
        ++_iver;
        Clipped++;
        break;
      case 4:  // Input line was clipped by both sides
        _ClipPntBuf[_iver] = Pt0;
        _ClipPntState[_iver] = BORDER;
        ++_iver;
        _ClipPntBuf[_iver] = Pt1;
        _ClipPntState[_iver] = BORDER;
        ++_iver;
        _ClipPntState[_iver] = OUTSIDE;
        ++_iver;
        ++Clipped;
        break;
    }
    pVer0 = pVer1;
    pVer1 = pVerNext;
    if (_iver > MAX_CLIPVER){
      break;  // exit the cycle
    }
  }
  if (Clipped){
    return 1;
  }
  if (Outside == 0){
    return 2;  // completely inside
  }
  return 0;  // completely outside
m1:
  n = _iver;
  for (i=0; i<n; ++i){
    _ClipPntBuf[_iver] = _ClipPntBuf[i];
    _ClipPntState[_iver] = State = _ClipPntState[i];
    ++_iver;
    _ClipPntState[i] = DISABLED;
    if (State == BORDER){
      // offset _ClipPnt buffer from
//      n = i
      break;
    }
  }
  return 1;
}






/********************************************************************
* Clip polygon
*   Polygon - input polygon 
* Returns:
*  -1 - error
*   0 - polygon is completely outside of the clip rect
*   1 - polygon was clipped
*   2 - polygon is completely inside of the clip rect
*********************************************************************/
INT_PTR CGeRect::ClipPolygon (const CGePolyline& Polygon) const
{
  CGePoint  StartPt;
  CGePoint* pPnt;
  UINT_PTR  i, n;
  INT_PTR   ret;
  bool      bClockwise = Polygon.IsCW();
  CKusok*   Kusok = NULL; 
  CKusok*   pKusok = NULL; 
  CKusok*   pKusok0 = NULL; 
  UINT_PTR  nKus;

  ret = ClipPolyline( Polygon, true );
  if (ret == 0){
    // polygon is completely outside of the clip rect
    _ClipPntBuf[0] = LefBot();
    _ClipPntState[0] = BORDER;
    if (bClockwise){
      _ClipPntBuf[1] = LefTop();
      _ClipPntState[1] = BORDER;
      _ClipPntBuf[2] = RigTop();
      _ClipPntState[2] = BORDER;
      _ClipPntBuf[3] = RigBot();
      _ClipPntState[3] = BORDER;
    }else{
      _ClipPntBuf[1] = RigBot();
      _ClipPntState[1] = BORDER;
      _ClipPntBuf[2] = RigTop();
      _ClipPntState[2] = BORDER;
      _ClipPntBuf[3] = LefTop();
      _ClipPntState[3] = BORDER;
    }
    _iver = 4;
    return 0;
  }
  if (ret == 2){
    // polygon is completely inside of the clip rect
    return 2;
  }
  if (ret == 1){
    // count number of polylines
    nKus = 0;
    pPnt = GetClippedPolyline( true, &n );
    while( pPnt ){
      ++nKus;
      pPnt = GetClippedPolyline( false, &n );
    }
    if (nKus > 0){
      Kusok = new CKusok[nKus];
      i = 0;
      // get vertices
      pPnt = GetClippedPolyline( true, &n );
      while( pPnt ){
        Kusok[i].Set( pPnt, n, *this, bClockwise );
        ++i;
        pPnt = GetClippedPolyline( false, &n );
      }

      _iver = 0;
      for ( ; ; ){ //while( 1 ){
        pKusok0 = Clip_GetFirstKusok( Kusok, nKus );
        if (pKusok0){
          pKusok = pKusok0;
          while( pKusok ){
            pKusok->m_bSelected = true;
            n = pKusok->m_nPoints;
            for (i=0; i<n; ++i){
              _ClipPntBuf[_iver] = pKusok->m_Point[i];
              _ClipPntState[_iver] = 1;
              ++_iver;
            }
            pKusok = Clip_GetNextKusok( Kusok, nKus, pKusok );
            if (pKusok == pKusok0){
              break;
            }
          }
          Clip_DeleteSelKusok( Kusok, nKus );
          _ClipPntBuf[_iver].Set(0,0);
          _ClipPntState[_iver] = 0;
          ++_iver;
        }else{
          break;
        }
      }
      delete[] Kusok;
    }
    return 1;
  }
  return -1;  // error
}

//-----------------------------------------------
CKusok* CGeRect::Clip_GetFirstKusok (CKusok* Kusok, UINT_PTR nKus) const
{
  UINT_PTR i;
  CKusok*  ptr = Kusok;

  for (i=0; i<nKus; ++i,++ptr){
    if (ptr->m_bDeleted == false){
      return ptr;
    }
  }
  return NULL;
}

//-----------------------------------------------
CKusok* CGeRect::Clip_GetNextKusok (CKusok* Kusok, UINT_PTR nKus, CKusok* pKusok0) const
{
  CGePoint Pt1, Pnt[4], CornPt[4];
  UINT_PTR i, n, nCorns=0;
  CKusok*  ptr = Kusok;
  CKusok*  pFound = NULL;
  double   d, dmin=100000000000.0;

  for (i=0; i<nKus; ++i,++ptr){
    if (ptr->m_bDeleted){
      continue;
    }
    Pt1 = ptr->m_StartPt;
    // get distance from Pt0 to Pt1 along m_CornerPt[] array
    d = pKusok0->GetDist( Pt1, Pnt, &n );
    if (d < dmin){
      dmin = d;
      pFound = ptr;
      memcpy( CornPt, Pnt, sizeof(CornPt) );
      nCorns = n;
    }
  }
  for (i=0; i<nCorns; ++i){
    _ClipPntBuf[_iver] = CornPt[i];
    _ClipPntState[_iver] = 1;
    ++_iver;
  }
  return pFound;
}


//-----------------------------------------------
void CGeRect::Clip_DeleteSelKusok (CKusok* Kusok, UINT_PTR nKus) const
{
  UINT_PTR i=0;
  for (; i<nKus; ++i){
    if (Kusok[i].m_bSelected){
      Kusok[i].m_bDeleted = true;
    }
  }
}

//-----------------------------------------------
CGePoint* CGeRect::GetClippedPolygon (bool bFirst, UINT_PTR* pNumPoints) const
{
  static UINT_PTR iStart=0;
  UINT_PTR   i, n;
  CGePoint*  pPoints = NULL;

  // start point
  if (bFirst){
    iStart = 0;
  }
  for (i=iStart; i<_iver; ++i){
    if (_ClipPntState[i]==BORDER || _ClipPntState[i]==INSIDE){
      iStart = i;
      break;
    }
  }
  if (i >= _iver){
    *pNumPoints = 0;
    return NULL;
  }
  // count a number of polyline points
  for (i=iStart+1; i<_iver; ++i){
    // state of vertex, 2-outside, 1-border (clipped), 0-inside
    if (_ClipPntState[i] == 0){
      n = i - iStart;
      pPoints = _ClipPntBuf + iStart;
      iStart  = i + 1;
      *pNumPoints = n;
      return pPoints;
    }
  }
  n = _iver - iStart;
  pPoints = _ClipPntBuf + iStart;
  iStart  = _iver;
  *pNumPoints = n;
  return pPoints;
}

//-----------------------------------------------
CGePoint* CGeRect::GetClippedPolyline (bool bFirst, UINT_PTR* pNumPoints) const
{
  static UINT_PTR iStart=0;
  UINT_PTR   i, n;
  CGePoint*  pPoints = NULL;

  // start point
  if (bFirst){
    iStart = 0;
  }
  for (i=iStart; i<_iver; ++i){
    if (_ClipPntState[i] == DISABLED){
      continue;
    }
    if (_ClipPntState[i] != OUTSIDE){
      iStart = i;
      break;
    }
  }
  if (i >= _iver){
    *pNumPoints = 0;
    return NULL;
  }
  // count a number of polyline points
  for (i=iStart+1; i<_iver; ++i){
    // state of vertex, 2-outside, 1-border (clipped), 0-inside
    if (_ClipPntState[i] == OUTSIDE){
      n = i - iStart;
      pPoints = _ClipPntBuf + iStart;
      iStart  = i + 1;
      *pNumPoints = n;
      return pPoints;
    }
  }
  n = _iver - iStart;
  pPoints = _ClipPntBuf + iStart;
  iStart  = _iver;
  *pNumPoints = n;
  return pPoints;
}



//*******************************************************************
//
// Class CKusok
//
//*******************************************************************
CKusok::CKusok ()
{
  m_Point = NULL;
  m_nPoints = 0;
  m_bDeleted = false;
  m_bSelected = false;
}

//-----------------------------------------------
CKusok::~CKusok ()
{
  if (m_Point){
    delete[] m_Point;
  }
}

//-----------------------------------------------
void CKusok::Set (const CGePoint* Pnt, UINT_PTR nPnts, const CGeRect& Rect, bool bCW)
{
  double Xmin = Rect.Left();
  double Ymin = Rect.Bottom();
  double Xmax = Rect.Right();
  double Ymax = Rect.Top();
  if (m_Point){
    delete[] m_Point;
  }
  m_nPoints = nPnts;
  m_Point = new CGePoint[nPnts];
  memcpy( m_Point, Pnt, sizeof(CGePoint) * nPnts ); 
  m_StartPt = Pnt[0];
  m_EndPt = Pnt[nPnts-1];
  if (m_EndPt.x == Xmin){
    // on left border
    if (bCW){
      if (m_EndPt.y < Ymax){
        m_CornerPt[0] = Rect.LefTop();
        m_CornerPt[1] = Rect.RigTop();
        m_CornerPt[2] = Rect.RigBot();
        m_CornerPt[3] = Rect.LefBot();
        return;
      }
    }else{
      if (m_EndPt.y > Ymin){
        m_CornerPt[0] = Rect.LefBot();
        m_CornerPt[1] = Rect.RigBot();
        m_CornerPt[2] = Rect.RigTop();
        m_CornerPt[3] = Rect.LefTop();
        return;
      }
    }
  }
  if (m_EndPt.y == Ymax){
    // on top border
    if (bCW){
      if (m_EndPt.x < Xmax){
        m_CornerPt[0] = Rect.RigTop();
        m_CornerPt[1] = Rect.RigBot();
        m_CornerPt[2] = Rect.LefBot();
        m_CornerPt[3] = Rect.LefTop();
        return;
      }
    }else{
      if (m_EndPt.x > Xmin){
        m_CornerPt[0] = Rect.LefTop();
        m_CornerPt[1] = Rect.LefBot();
        m_CornerPt[2] = Rect.RigBot();
        m_CornerPt[3] = Rect.RigTop();
        return;
      }
    }
  }
  if (m_EndPt.x == Xmax){
    // on right border
    if (bCW){
      if (m_EndPt.y > Ymin){
        m_CornerPt[0] = Rect.RigBot();
        m_CornerPt[1] = Rect.LefBot();
        m_CornerPt[2] = Rect.LefTop();
        m_CornerPt[3] = Rect.RigTop();
        return;
      }
    }else{
      if (m_EndPt.y < Ymax){
        m_CornerPt[0] = Rect.RigTop();
        m_CornerPt[1] = Rect.LefTop();
        m_CornerPt[2] = Rect.LefBot();
        m_CornerPt[3] = Rect.RigBot();
        return;
      }
    }
  }
  if (m_EndPt.y == Ymin){
    // on bottom border
    if (bCW){
      if (m_EndPt.x > Xmin){
        m_CornerPt[0] = Rect.LefBot();
        m_CornerPt[1] = Rect.LefTop();
        m_CornerPt[2] = Rect.RigTop();
        m_CornerPt[3] = Rect.RigBot();
        return;
      }
    }else{
      if (m_EndPt.x < Xmax){
        m_CornerPt[0] = Rect.RigBot();
        m_CornerPt[1] = Rect.RigTop();
        m_CornerPt[2] = Rect.LefTop();
        m_CornerPt[3] = Rect.LefBot();
        return;
      }
    }
  }
}

// get distance from m_EndPt to Pnt along m_CornerPt[] array
//-----------------------------------------------
double CKusok::GetDist (const CGePoint& Pnt, CGePoint* CornPt, UINT_PTR* pnCorns) const
{
  double   Dist = 0;;
  double   d, ang;
  UINT_PTR i, nc;
  CGePoint Pt0 = m_EndPt;
  CGePoint Pt1, Pt2, Pnt2;

  nc = 0;
  for (i=0; i<4; ++i){
    Pt1 = m_CornerPt[i];
    ang = Pt0.Angle( Pt1 );
    if ((Pt0.x == Pnt.x) && (Pnt.x == Pt1.x)){
      Pt2 = Pt1.RotateBy( Pt0, -ang );
      Pnt2 = Pnt.RotateBy( Pt0, -ang );
      if (Pt0.x < Pnt2.x){
        d = Pnt2.x - Pt0.x;
        Dist += d;
        break;
      }
    }
    if ((Pt0.y == Pnt.y) && (Pnt.y == Pt1.y)){
      Pt2 = Pt1.RotateBy( Pt0, -ang );
      Pnt2 = Pnt.RotateBy( Pt0, -ang );
      if (Pt0.x < Pnt2.x){
        d = Pnt2.x - Pt0.x;
        Dist += d;
        break;
      }
    }
    CornPt[nc] = Pt1;
    ++nc;
    d = Pt0.Distance( Pt1 );
    Dist += d;
    Pt0 = Pt1;
  }
  *pnCorns = nc;
  return Dist;
}

