/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* This code was taken from 'BaseLibs' project
* Some useful utility functions
********************************************************************/
#include <math.h>
#include <stdlib.h>
#include "Geom.h"


/********************************************************************
* Generates quadratic B-spline curve between control points A,B,C
*   nLines - number of lines per curve
*   GenVer - pointer to array of result points (point C is not added)
*   pnGenVer - pointer to number of points in GenVer
*   MaxGenVer - max number of points in GenVer
*********************************************************************/
bool geGenSegmentBSpline2 (const CGePoint& A, const CGePoint& B, const CGePoint& C, UINT_PTR nLines,
                            CGePoint* GenVer, UINT_PTR* pnGenVer, UINT_PTR MaxGenVer, double* pLen) //, bool bIncludeA)
{
  double    t, t2, cf1, cf2, cf3;
  UINT_PTR  i, iVer;
  CGePoint* pGenVer;
  CGePoint  PrevPos;

  if (pLen){
    *pLen = 0.0;
  }
  iVer = *pnGenVer;
  if ((iVer + nLines) < MaxGenVer){
    pGenVer = GenVer + iVer;
//    if (bIncludeA){
      *pGenVer = A;
      PrevPos = A;
      ++pGenVer;
      ++iVer;
//    }
    for (i=1; i<nLines; ++i){
      t = (double)i / (double)nLines;
      t2 = 1.0 - t;
      cf1 = t2 * t2;
      cf2 = 2.0 * t * t2;
      cf3 = t * t;
      pGenVer->x = (cf1 * A.x) + (cf2 * B.x) + (cf3 * C.x);
      pGenVer->y = (cf1 * A.y) + (cf2 * B.y) + (cf3 * C.y);
      if (pLen){
        *pLen += PrevPos.Distance( *pGenVer );
        PrevPos = *pGenVer;
      }
      ++iVer;
      ++pGenVer;
    }
    *pnGenVer = iVer;
    return true;
  }else{
    // no memory for all result points
    return false;
  }
}

/********************************************************************
* Generates quadratic B-spline curve between control points A,B,C
*   nLines - number of lines per curve 
*   pOutPline - output polyline (point C not added)
*********************************************************************/
void geGenSegmentBSpline2 (const CGePoint& A, const CGePoint& B, const CGePoint& C, 
                           UINT_PTR nLines, CGePolyline* pOutPline)
{
  double   t, t2, cf1, cf2, cf3;
  UINT_PTR i;
  CGePoint Pnt;
  CGeVertex* pVerLast = pOutPline->GetLastVer();

  if (pVerLast){
    if (pVerLast->m_Pos != A){
      pOutPline->AddVer( A );
    }
  }else{
    pOutPline->AddVer( A );
  }
  for (i=1; i<nLines; ++i){
    t = (double)i / (double)nLines;
    t2 = 1.0 - t;
    cf1 = t2 * t2;
    cf2 = 2.0 * t * t2;
    cf3 = t * t;
    Pnt.x = (cf1 * A.x) + (cf2 * B.x) + (cf3 * C.x);
    Pnt.y = (cf1 * A.y) + (cf2 * B.y) + (cf3 * C.y);
    pOutPline->AddVer( Pnt );
  }
}


/********************************************************************
* Generates quadratic B-spline
*   Ver     - array of control vertexes
*   nVer    - number of points in Ver array
*   bClosed - input polyline is closed if true
*   nLines  - number of lines per each spline segment
*   GenVer    - pointer to array of result points
*   pnGenVer  - pointer to number of points in GenVer
*   MaxGenVer - max number of points in GenVer
*********************************************************************/
void geGenBSpline2 (const CGePoint* Ver, UINT_PTR nVer, bool bClosed, UINT_PTR nLines,
                    CGePoint* GenVer, UINT_PTR* pnGenVer, UINT_PTR MaxGenVer) //, CGePoint* pMidPt, double* pMidAng)

{
  CGePoint  A, B, C, Pt2;
  double    t, omt, cf1, cf2, cf3;
  UINT_PTR  i, j, k;
  UINT_PTR  np = 0;    // counter of generated points
  UINT_PTR  i0 = *pnGenVer;  // start index of output array
  UINT_PTR  nGenVer = i0;  // start index of output array
  CGePoint* pGVer = GenVer+i0;
  const CGePoint* pVer = Ver;
  const CGePoint* pVer0 = Ver;
  const CGePoint* pVer1 = Ver+1;
  const CGePoint* pVer2;

/*
  if (pMidPt){
    pMidPt->x = -1000000000000000.1;
  }
*/
  if (nVer < 2){
    return;
  }
  if (nVer == 2){
    *pGVer = Ver[0];
    ++pGVer;
    ++nGenVer;
    *pGVer = Ver[1];
    ++pGVer;
    ++nGenVer;
/*
    if (pMidPt){
      pMidPt->x = (Ver[0].x + Ver[1].x) / 2.0;
      pMidPt->y = (Ver[0].y + Ver[1].y) / 2.0;
      *pMidAng = Ver[0].Angle( Ver[1] );
    }
*/
    return;
  }
  if (nLines <= 1){
    for (i=0; i<nVer; ++i) {
      *pGVer = *pVer;
      ++pGVer;
      ++pVer;
      ++nGenVer;
      if (++np >= MaxGenVer){
        *pnGenVer = nGenVer;
        return;
      }
    }
    *pnGenVer = nGenVer;
    return;
  }
  for (i=0; i<nVer; ++i){
    if (i == 0){
      if (bClosed){
        pVer2 = Ver + nVer - 1;
        A.x = (pVer0->x + pVer2->x) / 2.0;
        A.y = (pVer0->y + pVer2->y) / 2.0;
        B = *pVer0;
        C.x = (pVer0->x + pVer1->x) / 2.0;
        C.y = (pVer0->y + pVer1->y) / 2.0;
      }else{
        A = *pVer0;
        B = *pVer1;
        pVer2 = Ver + 2;
        if (nVer == 3){
          C = *pVer2;
        }else{
          C.x = (pVer1->x + pVer2->x) / 2.0;
          C.y = (pVer1->y + pVer2->y) / 2.0;
        }
        ++i;
        ++pVer;
      }
    }else{
      if (i == nVer-1){
        if (bClosed){
          pVer2 = Ver + i - 1;
          A.x = (pVer->x + pVer2->x) / 2.0;
          A.y = (pVer->y + pVer2->y) / 2.0;
          B = Ver[i];
          C.x = (pVer->x + pVer->x) / 2.0;
          C.y = (pVer->y + pVer->y) / 2.0;
        }else{
          continue;
        }
      }else{
        if ((i == nVer-2) && (bClosed == false)){
          if (nVer == 3){
            // endpoint of non-closed 3 point spline 
            j = nVer - 1;
            *pGVer = *pVer;
            ++pGVer;
            ++nGenVer;
            break;
          }else{
            pVer2 = Ver + i - 1;
            A.x = (pVer->x + pVer2->x) / 2.0;
            A.y = (pVer->y + pVer2->y) / 2.0;
            B = *pVer;
            j = nVer - 1;
            C = Ver[j];
          }
        }else{
          pVer2 = Ver + i - 1;
          A.x = (pVer->x + pVer2->x) / 2.0;
          A.y = (pVer->y + pVer2->y) / 2.0;
          B = Ver[i];
          pVer2 = Ver + i + 1;
          C.x = (pVer->x + pVer2->x) / 2.0;
          C.y = (pVer->y + pVer2->y) / 2.0;
        }
      }
    }
    // generate segment between A-B-C points
    *pGVer = A;
    ++pGVer;
    ++nGenVer;
    for (k=1; k<nLines; ++k){
      t = (double)k / (double)nLines;
      omt = 1.0 - t;
      cf1 = omt * omt;
      cf2 = 2.0 * t * omt;
      cf3 = t * t;
      pGVer->x = (cf1 * A.x) + (cf2 * B.x) + (cf3 * C.x);
      pGVer->y = (cf1 * A.y) + (cf2 * B.y) + (cf3 * C.y);
      ++pGVer;
      ++nGenVer;
      ++np;
      if (np >= MaxGenVer){
        return;
      }
    }
    ++pVer;
/*
    if (pMidPt){
      t = 0.5;
      omt = 0.5 - t;
      cf1 = omt * omt;
      cf2 = 2.0 * t * omt;
      cf3 = t * t;
      pMidPt->x = (cf1 * A.x) + (cf2 * B.x) + (cf3 * C.x);
      pMidPt->y = (cf1 * A.y) + (cf2 * B.y) + (cf3 * C.y);
      t = 0.501;
      omt = 0.5 - t;
      cf1 = omt * omt;
      cf2 = 2.0 * t * omt;
      cf3 = t * t;
      Pt2.x = (cf1 * A.x) + (cf2 * B.x) + (cf3 * C.x);
      Pt2.y = (cf1 * A.y) + (cf2 * B.y) + (cf3 * C.y);
      *pMidAng = pMidPt->Angle( Pt2 );
    }
*/
  }
  // endpoint
  if (bClosed){
    *pGVer = GenVer[i0]; 
  }else{
    *pGVer = Ver[nVer-1];
  }
  ++pGVer;
  ++nGenVer;
  *pnGenVer = nGenVer;
}



/********************************************************************
* Generates cubic B-spline curve for segment B,C
* between  control points A,B,C,D
*   nLines - number of lines per curve
*   GenVer - pointer to array of result points (without last point)
*   pnGenVer - pointer to number of points in GenVer
*   MaxGenVer - max number of points in GenVer
*********************************************************************/
bool geGenSegmentBSpline3 (const CGePoint& A, const CGePoint& B, const CGePoint& C, const CGePoint& D, 
                           UINT_PTR nLines, CGePoint* GenVer, UINT_PTR* pnGenVer, UINT_PTR MaxGenVer)
{
  double    t, a0,a1,a2,a3, b0,b1,b2,b3;
  UINT_PTR  i, iVer;
  CGePoint* pGenVer;

  iVer = *pnGenVer;
  if ((iVer + nLines) < MaxGenVer){
    pGenVer = GenVer + iVer;
    for (i=0; i<nLines; ++i){
      t = (double)i / (double)nLines;
      a0 = (A.x + 4.0 * B.x + C.x) / 6.0;
      b0 = (A.y + 4.0 * B.y + C.y) / 6.0;
      a1 = (C.x - A.x) / 2.0;
      b1 = (C.y - A.y) / 2.0;
      a2 = (A.x - 2.0 * B.x + C.x) / 2.0;
      b2 = (A.y - 2.0 * B.y + C.y) / 2.0;
      a3 = (-A.x + 3.0 * (B.x - C.x) + D.x) / 6.0;
      b3 = (-A.y + 3.0 * (B.y - C.y) + D.y) / 6.0;
      pGenVer->x = ((a3 * t + a2) * t + a1) * t + a0;
      pGenVer->y = ((b3 * t + b2) * t + b1) * t + b0;
      ++pGenVer;
      ++iVer;
    }
    *pnGenVer = iVer;
    return true;
  }else{
    // no memory for all result points
    return false;
  }
}

/********************************************************************
* Generates cubic B-spline curve for segment B,C
* between  control points A,B,C,D
*   nLines - number of result points 
*   pOutPline - output polyline (without last point C)
*********************************************************************/
void geGenSegmentBSpline3 (const CGePoint& A, const CGePoint& B, const CGePoint& C, const CGePoint& D, 
                           UINT_PTR nLines, CGePolyline* pOutPline)
{
  double   t, a0,a1,a2,a3, b0,b1,b2,b3;
  UINT_PTR i;
  CGePoint Pnt;

  for (i=0; i<nLines; ++i){
    t = (double)i / (double)nLines;
    a0 = (A.x + 4.0 * B.x + C.x) / 6.0;
    b0 = (A.y + 4.0 * B.y + C.y) / 6.0;
    a1 = (C.x - A.x) / 2.0;
    b1 = (C.y - A.y) / 2.0;
    a2 = (A.x - 2.0 * B.x + C.x) / 2.0;
    b2 = (A.y - 2.0 * B.y + C.y) / 2.0;
    a3 = (-A.x + 3.0 * (B.x - C.x) + D.x) / 6.0;
    b3 = (-A.y + 3.0 * (B.y - C.y) + D.y) / 6.0;
    Pnt.x = ((a3 * t + a2) * t + a1) * t + a0;
    Pnt.y = ((b3 * t + b2) * t + b1) * t + b0;
    pOutPline->AddVer( Pnt );
  }
}


/********************************************************************
* Generates quadratic Bezier curve between control points A,B,C
*   nLines - number of lines per curve
*   pOutPline - output polyline (without first & last points)
*********************************************************************/
void geGenSegmentBezier2 (const CGePoint& A, const CGePoint& B, const CGePoint& C, 
                          UINT_PTR nLines, CGePolyline* pOutPline)
{
  double   t, ot, cf0, cf1, cf2;
  UINT_PTR i;
  CGePoint Pnt;

  for (i=1; i<nLines; ++i){
    t = (double)i / (double)nLines;
    ot = 1.0 - t;
    cf0 = ot * ot;
    cf1 = 2.0 * t * ot;
    cf2 = t * t;
    Pnt.x = (cf0*A.x) + (cf1*B.x) + (cf2*C.x);
    Pnt.y = (cf0*A.y) + (cf1*B.y) + (cf2*C.y);
    pOutPline->AddVer( Pnt );
  }
}


//-----------------------------------------------
bool geGenSegmentBezier2 (const CGePoint& A, const CGePoint& B, const CGePoint& C, 
                          UINT_PTR nLines, CGePoint* GenVer, UINT_PTR* pnGenVer, UINT_PTR MaxGenVer)
{
  double    t, ot, cf0, cf1, cf2;
  UINT_PTR  i, iVer;
  CGePoint  Pnt;
  CGePoint* pGenVer;

  iVer = *pnGenVer;
  if ((iVer + nLines) < MaxGenVer){
    pGenVer = GenVer + iVer;
    for (i=1; i<nLines; ++i){
      t = (double)i / (double)nLines;
      ot = 1.0 - t;
      cf0 = ot * ot;
      cf1 = 2.0 * t * ot;
      cf2 = t * t;
      pGenVer->x = (cf0*A.x) + (cf1*B.x) + (cf2*C.x);
      pGenVer->y = (cf0*A.y) + (cf1*B.y) + (cf2*C.y);
      ++pGenVer;
      ++iVer;
    }
    *pnGenVer = iVer;
    return true;
  }else{
    // no memory for all result points
    return false;
  }
}

//  points A,B,C have weights
//-----------------------------------------------
void geGenSegmentBezier2 (const CGePoint& A, double Wa, const CGePoint& B, double Wb, const CGePoint& C, double Wc,
                          UINT_PTR nLines, CGePolyline* pOutPline)
{
  double   t, ot, cf0, cf1, cf2;
  UINT_PTR i;
  CGePoint Pnt;

  if (Wa == 0.0){
    Wa = 1.0;
  }
  if (Wb == 0.0){
    Wb = 1.0;
  }
  if (Wc == 0.0){
    Wc = 1.0;
  }
  pOutPline->AddVer( A );
  for (i=1; i<nLines; ++i){
    t = (double)i / (double)nLines;
    ot = 1.0 - t;
    cf0 = ot * ot * Wa;
    cf1 = 2.0 * t * ot * Wb;
    cf2 = t * t * Wc;
    Pnt.x = (cf0*A.x + cf1*B.x + cf2*C.x) / (cf0 + cf1 + cf2);
    Pnt.y = (cf0*A.y + cf1*B.y + cf2*C.y) / (cf0 + cf1 + cf2);
    pOutPline->AddVer( Pnt );
  }
}



/********************************************************************
* Generates cubic Bezier curve between control points A,B,C,D
*   nLines - number of lines per curve
*   pOutPline - output polyline (without first & last points)
*********************************************************************/
void geGenSegmentBezier3 (const CGePoint& A, const CGePoint& B, const CGePoint& C, const CGePoint& D, 
                          UINT_PTR nLines, CGePolyline* pOutPline)
{
  double   t, ot, t2, ot2, cf0, cf1, cf2, cf3;
  UINT_PTR i;
  CGePoint Pnt;

  for (i=1; i<nLines; ++i){
    t = (double)i / (double)nLines;
    ot = 1.0 - t;
    t2 = t * t;
    ot2 = ot * ot;
    cf0 = ot * ot2;        // ot * ot * ot
    cf1 = 3.0 * t * ot2;   // t * ot * ot * 3.0
    cf2 = 3.0 * t2 * ot;   // t * t * ot * 3.0
    cf3 = t * t2;          // t * t * t
    Pnt.x = (cf0*A.x) + (cf1*B.x) + (cf2*C.x) + (cf3*D.x);
    Pnt.y = (cf0*A.y) + (cf1*B.y) + (cf2*C.y) + (cf3*D.y);
    pOutPline->AddVer( Pnt );
  }
}

//-----------------------------------------------
bool geGenSegmentBezier3 (const CGePoint& A, const CGePoint& B, const CGePoint& C, const CGePoint& D, 
                          UINT_PTR nLines, CGePoint* GenVer, UINT_PTR* pnGenVer, UINT_PTR MaxGenVer)
{
  double    t, ot, t2, ot2, cf0, cf1, cf2, cf3;
  UINT_PTR  i, iVer;
  CGePoint  Pnt;
  CGePoint* pGenVer;

  iVer = *pnGenVer;
  if ((iVer + nLines) < MaxGenVer){
    pGenVer = GenVer + iVer;
    for (i=1; i<nLines; ++i){
      t = (double)i / (double)nLines;
      ot = 1.0 - t;
      t2 = t * t;
      ot2 = ot * ot;
      cf0 = ot * ot2;        // ot * ot * ot
      cf1 = 3.0 * t * ot2;   // t * ot * ot * 3.0
      cf2 = 3.0 * t2 * ot;   // t * t * ot * 3.0
      cf3 = t * t2;          // t * t * t
      pGenVer->x = (cf0*A.x) + (cf1*B.x) + (cf2*C.x) + (cf3*D.x);
      pGenVer->y = (cf0*A.y) + (cf1*B.y) + (cf2*C.y) + (cf3*D.y);
      ++pGenVer;
      ++iVer;
    }
    *pnGenVer = iVer;
    return true;
  }else{
    // no memory for all result points
    return false;
  }
}

//  points A,B,C,D have weights
//-----------------------------------------------
void geGenSegmentBezier3 (const CGePoint& A, double Wa, const CGePoint& B, double Wb, const CGePoint& C, double Wc, const CGePoint& D, double Wd,
                          UINT_PTR nLines, CGePolyline* pOutPline)
{
  double   t, ot, cf0, cf1, cf2, cf3;
  UINT_PTR i;
  CGePoint Pnt;

  if (Wa == 0.0){
    Wa = 1.0;
  }
  if (Wb == 0.0){
    Wb = 1.0;
  }
  if (Wc == 0.0){
    Wc = 1.0;
  }
  if (Wd == 0.0){
    Wd = 1.0;
  }
  pOutPline->AddVer( A );
  for (i=1; i<nLines; ++i){
    t = (double)i / (double)nLines;
    ot = 1.0 - t;
    cf0 = ot * ot * ot * Wa;
    cf1 = t * ot * ot * 3.0 * Wb;
    cf2 = t * t * ot * 3.0 * Wc;
    cf3 = t * t * t * Wd;
    Pnt.x = (cf0*A.x + cf1*B.x + cf2*C.x + cf3*D.x) / (cf0 + cf1 + cf2 + cf3);
    Pnt.y = (cf0*A.y + cf1*B.y + cf2*C.y + cf3*D.y) / (cf0 + cf1 + cf2 + cf3);
    pOutPline->AddVer( Pnt );
  }
}


/********************************************************************
* Generates Fit Spline
*   pInPline - input polyline
*   pPline   - result polyline
*   nLines  - number of generated linear segments for each spline segment
*********************************************************************/
void geGenFitSpline (const CGePolyline* pInPline, CGePolyline* pOutPline, UINT_PTR nLines)
{
  CGePoint   Pt;
  double     px0=0.0;
  double     qx0=0.0;
  double     py0=0.0;
  double     qy0=0.0;
  double     xA,xB,xC, yA,yB,yC;
  double     h, h_i, h_i1, k, m;
  double     t,d,delta;
  INT_PTR    i, n, nVers;
  double*    P =NULL;
  double*    Q =NULL;
  double*    a1=NULL;
  double*    b1=NULL;
  double*    c1=NULL;
  double*    d1=NULL;
  double*    a2=NULL;
  double*    b2=NULL;
  double*    c2=NULL;
  double*    d2=NULL;
  double*    dist=NULL;
  CGeVertex* pVer;
  CGeVertex* pVerPrev;
  CGeVertex* pVerNext;
  UINT_PTR   nPnt, iPnt, j;

  nVers = pInPline->NumVers();
  nPnt = (nVers - 1) * (nLines+1) + 2;
  pOutPline->Clear();

  n = nVers + 1;
  // auxiliary arrays
  P  = (double*)malloc(sizeof(double)*n);
  Q  = (double*)malloc(sizeof(double)*n);
  // auxiliary arrays for coefficients
  a1 = (double*)malloc(sizeof(double)*n);
  b1 = (double*)malloc(sizeof(double)*n);
  c1 = (double*)malloc(sizeof(double)*n);
  d1 = (double*)malloc(sizeof(double)*n);
  a2 = (double*)malloc(sizeof(double)*n);
  b2 = (double*)malloc(sizeof(double)*n);
  c2 = (double*)malloc(sizeof(double)*n);
  d2 = (double*)malloc(sizeof(double)*n);
  dist=(double*)malloc(sizeof(double)*n);
  --n;

  // distances between vertices
  i = 0;
  pVerPrev = pInPline->GetFirstVer();
  pVer = pInPline->GetNextVer( pVerPrev );
  while( pVer ){
    d = pVerPrev->m_Pos.Distance( pVer->m_Pos ); 
    if (d == 0.){
      d = 0.0000001;
    }
    dist[i] = d;
    ++i;
    pVerPrev = pVer;
    pVer = pInPline->GetNextVer( pVer );
  }

  // calculate spline coefficients
  P[0] = px0;
  Q[0] = qx0;
  i = 1;
  pVerPrev = pInPline->GetFirstVer();
  pVer = pInPline->GetNextVer( pVerPrev );
  pVerNext = pInPline->GetNextVer( pVer );
  while( pVerNext ){
    xA = pVerPrev->X();
    xB = pVer->X();
    xC = pVerNext->X();
    h_i  = dist[i-1];
    h_i1 = dist[i];
    k = 6.0*((xB-xA)/h_i - (xC-xB)/h_i1);
    m = -2.0*(h_i + h_i1) - h_i*P[i-1];
    P[i] = h_i1 / m;
    Q[i] = (k + h_i*Q[i-1]) / m;
    pVerPrev = pVer;
    pVer = pVerNext;
    pVerNext = pInPline->GetNextVer( pVerNext );
    ++i;
  }
  n = nVers;
  c1[n-1] = 0;
  for (i=n-1; i>=0; --i) {
    if (i <= n-2){
      c1[i] = (P[i] * c1[i+1]) + Q[i];
    }
  }
  i = 1;
  pVerPrev = pInPline->GetFirstVer();
  pVer = pInPline->GetNextVer( pVerPrev );
  while( pVer ){
    xA = pVerPrev->X();
    xB = pVer->X();
    h = dist[i-1];
    d1[i] = (c1[i] - c1[i-1]) / h;
    a1[i] = xB;
    b1[i] = h*c1[i]/2.0 - h*h*d1[i]/6.0 + (xB-xA)/h;
    pVerPrev = pVer;
    pVer = pInPline->GetNextVer( pVer );
    ++i;
  }

  // ---- the same by Y ----------------
  P[0] = py0;
  Q[0] = qy0;
  i = 1;
  pVerPrev = pInPline->GetFirstVer();
  pVer = pInPline->GetNextVer( pVerPrev );
  pVerNext = pInPline->GetNextVer( pVer );
  while( pVerNext ){
    yA = pVerPrev->Y();
    yB = pVer->Y();
    yC = pVerNext->Y();
    h_i  = dist[i-1];
    h_i1 = dist[i];
    k = 6.0*((yB-yA)/h_i - (yC-yB)/h_i1);
    m = -2.0*(h_i + h_i1) - h_i*P[i-1];
    P[i] = h_i1 / m;
    Q[i] = (k + h_i*Q[i-1]) / m;
    pVerPrev = pVer;
    pVer = pVerNext;
    pVerNext = pInPline->GetNextVer( pVerNext );
    ++i;
  }
  n = nVers;
  c2[n-1] = 0;
  for (i=n-1; i>=0; --i) {
    if (i <= n-2){
      c2[i] = (P[i] * c2[i+1]) + Q[i];
    }
  }
  i = 1;
  pVerPrev = pInPline->GetFirstVer();
  pVer = pInPline->GetNextVer( pVerPrev );
  while( pVer ){
    yA = pVerPrev->Y();
    yB = pVer->Y();
    h = dist[i-1];
    d2[i] = (c2[i] - c2[i-1]) / h;
    a2[i] = yB;
    b2[i] = h*c2[i]/2.0 - h*h*d2[i]/6.0 + (yB-yA)/h;
    pVerPrev = pVer;
    pVer = pInPline->GetNextVer( pVer );
    ++i;
  }

  // ---- Generate spline ----------------
  iPnt = 0;
  pVer = pInPline->GetFirstVer();
  pOutPline->AddVer( pVer->m_Pos );
  ++iPnt;
  for (i=1; i<nVers; ++i){
    delta = dist[i-1] / (double)nLines;
    t = -dist[i-1];
    for (j=1; j<nLines; ++j){
      t += delta;
      Pt.x = ((d1[i]*t/6.0 + c1[i]/2.0)*t + b1[i])*t + a1[i];
      Pt.y = ((d2[i]*t/6.0 + c2[i]/2.0)*t + b2[i])*t + a2[i];
      pOutPline->AddVer( Pt );
      ++iPnt;
    }
    Pt.x = a1[i];
    Pt.y = a2[i];
    pOutPline->AddVer( Pt );
    ++iPnt;
  }

  free( P );
  free( Q );
  free( a1 );
  free( b1 );
  free( c1 );
  free( d1 );
  free( a2 );
  free( b2 );
  free( c2 );
  free( d2 );
  free( dist );
}

//-----------------------------------------------
bool geGenFitSpline (const CGePolyline* pInPline, UINT_PTR nLines, 
                     CGePoint* GenVer, UINT_PTR* pnGenVer, UINT_PTR MaxGenVer)
{
  CGePoint   Pt;
  double     px0=0.0;
  double     qx0=0.0;
  double     py0=0.0;
  double     qy0=0.0;
  double     xA,xB,xC, yA,yB,yC;
  double     h, h_i, h_i1, k, m;
  double     t,d,delta;
  INT_PTR    i, n, nVers, n2;
  double*    P =NULL;
  double*    Q =NULL;
  double*    a1=NULL;
  double*    b1=NULL;
  double*    c1=NULL;
  double*    d1=NULL;
  double*    a2=NULL;
  double*    b2=NULL;
  double*    c2=NULL;
  double*    d2=NULL;
  double*    dist=NULL;
  CGeVertex* pVer;
  CGeVertex* pVerPrev;
  CGeVertex* pVerNext;
  UINT_PTR   nPnt, iPnt, j, iVer;
  CGePoint*  pGenVer;

  iVer = *pnGenVer;
  if ((iVer + nLines) >= MaxGenVer){
    return false;
  }
  pGenVer = GenVer + iVer;

  nVers = pInPline->NumVers();
  nPnt = ((nVers - 1) * (nLines+1)) + 2;

  n2 = nVers + 2;
  // auxiliary arrays
  P  = (double*)malloc(sizeof(double)*n2);
  Q  = (double*)malloc(sizeof(double)*n2);
  // auxiliary arrays for coefficients
  a1 = (double*)malloc(sizeof(double)*n2);
  b1 = (double*)malloc(sizeof(double)*n2);
  c1 = (double*)malloc(sizeof(double)*n2);
  d1 = (double*)malloc(sizeof(double)*n2);
  a2 = (double*)malloc(sizeof(double)*n2);
  b2 = (double*)malloc(sizeof(double)*n2);
  c2 = (double*)malloc(sizeof(double)*n2);
  d2 = (double*)malloc(sizeof(double)*n2);
  dist=(double*)malloc(sizeof(double)*n2);
  n = nVers;

  // distances between vertices
  i = 0;
  pVerPrev = pInPline->GetFirstVer();
  pVer = pInPline->GetNextVer( pVerPrev );
  while( pVer ){
    d = pVerPrev->m_Pos.Distance( pVer->m_Pos ); 
    if (d == 0.){
      d = 0.0000001;
    }
    dist[i] = d;
    ++i;
    pVerPrev = pVer;
    pVer = pInPline->GetNextVer( pVer );
  }

  // calculate spline coefficients
  P[0] = px0;
  Q[0] = qx0;
  i = 1;
  pVerPrev = pInPline->GetFirstVer();
  pVer = pInPline->GetNextVer( pVerPrev );
  pVerNext = pInPline->GetNextVer( pVer );
  while( pVerNext ){
    xA = pVerPrev->X();
    xB = pVer->X();
    xC = pVerNext->X();
    h_i  = dist[i-1];
    h_i1 = dist[i];
    k = 6.0*((xB-xA)/h_i - (xC-xB)/h_i1);
    m = -2.0*(h_i + h_i1) - h_i*P[i-1];
    P[i] = h_i1 / m;
    Q[i] = (k + h_i*Q[i-1]) / m;
    pVerPrev = pVer;
    pVer = pVerNext;
    pVerNext = pInPline->GetNextVer( pVerNext );
    ++i;
  }
  n = nVers;
  c1[n-1] = 0;
  for (i=n-1; i>=0; --i) {
    if (i <= n-2){
      c1[i] = (P[i] * c1[i+1]) + Q[i];
    }
  }
  i = 1;
  pVerPrev = pInPline->GetFirstVer();
  pVer = pInPline->GetNextVer( pVerPrev );
  while( pVer ){
    xA = pVerPrev->X();
    xB = pVer->X();
    h = dist[i-1];
    d1[i] = (c1[i] - c1[i-1]) / h;
    a1[i] = xB;
    b1[i] = h*c1[i]/2.0 - h*h*d1[i]/6.0 + (xB-xA)/h;
    pVerPrev = pVer;
    pVer = pInPline->GetNextVer( pVer );
    ++i;
  }

  // ---- the same by Y ----------------
  P[0] = py0;
  Q[0] = qy0;
  i = 1;
  pVerPrev = pInPline->GetFirstVer();
  pVer = pInPline->GetNextVer( pVerPrev );
  pVerNext = pInPline->GetNextVer( pVer );
  while( pVerNext ){
    yA = pVerPrev->Y();
    yB = pVer->Y();
    yC = pVerNext->Y();
    h_i  = dist[i-1];
    h_i1 = dist[i];
    k = 6.0*((yB-yA)/h_i - (yC-yB)/h_i1);
    m = -2.0*(h_i + h_i1) - h_i*P[i-1];
    P[i] = h_i1 / m;
    Q[i] = (k + h_i*Q[i-1]) / m;
    pVerPrev = pVer;
    pVer = pVerNext;
    pVerNext = pInPline->GetNextVer( pVerNext );
    ++i;
  }
  n = nVers;
  c2[n-1] = 0;
  for (i=n-1; i>=0; --i) {
    if (i <= n-2){
      c2[i] = P[i] * c2[i+1] + Q[i];
    }
  }
  i = 1;
  pVerPrev = pInPline->GetFirstVer();
  pVer = pInPline->GetNextVer( pVerPrev );
  while( pVer ){
    yA = pVerPrev->Y();
    yB = pVer->Y();
    h = dist[i-1];
    d2[i] = (c2[i] - c2[i-1]) / h;
    a2[i] = yB;
    b2[i] = h*c2[i]/2.0 - h*h*d2[i]/6.0 + (yB-yA)/h;
    pVerPrev = pVer;
    pVer = pInPline->GetNextVer( pVer );
    ++i;
  }

  // ---- Generate spline ----------------
  iPnt = 0;
  pVer = pInPline->GetFirstVer();
  *pGenVer = pVer->m_Pos;
  ++pGenVer;
  ++iVer;
  ++iPnt;
  for (i=1; i<nVers; ++i){
    delta = dist[i-1] / (double)nLines;
    t = -dist[i-1];
    for (j=1; j<nLines; ++j){
      t += delta;
      Pt.x = ((d1[i]*t/6.0 + c1[i]/2.0)*t + b1[i])*t + a1[i];
      Pt.y = ((d2[i]*t/6.0 + c2[i]/2.0)*t + b2[i])*t + a2[i];
      *pGenVer = Pt;
      ++pGenVer;
      ++iVer;
      ++iPnt;
    }
    Pt.x = a1[i];
    Pt.y = a2[i];
    *pGenVer = Pt;
    ++pGenVer;
    ++iVer;
    ++iPnt;
  }

  free( P );
  free( Q );
  free( a1 );
  free( b1 );
  free( c1 );
  free( d1 );
  free( a2 );
  free( b2 );
  free( c2 );
  free( d2 );
  free( dist );

  *pnGenVer = iVer;
  return true;
}


//-----------------------------------------------
bool geGetFilletPrm (const CGePoint& A, const CGePoint& B, const CGePoint& C, double Radius,
                        double* pTang, double* pBiss, double* pAngle)
{
  double ang1, ang2, arot, alfa, T, biss;

  if (Radius <= 0.0){
    return false;
  }
  // get arc angle (rotation angle, A,B,C)
  ang1 = A.Angle( B );
  ang2 = B.Angle( C );
  // arc angle, + to left(counter-clockwise), - to right (clockwise), radian 
 	arot = ang2 - ang1;
	if (arot < -GE_DEG180) arot+=GE_DEG360;
	if (arot >  GE_DEG180) arot-=GE_DEG360;
 	alfa = fabs( arot ) / 2.0;    // absolute angle of the arc
  T = Radius * tan( alfa );
  biss = _hypot(T,Radius) - Radius;
  if (pTang){
    *pTang = T;
  }
  if (pBiss){
    *pBiss = biss;
  }
  if (pAngle){
    *pAngle = alfa;
  }
  return true;
}


//-----------------------------------------------
bool geGetFilletBulge (const CGePoint& A, const CGePoint& B, const CGePoint& C, double Radius,
                          CGePoint* pOutPt1, CGePoint* pOutPt2, double* pBulge, CGePoint* pOutPtMid)
{
  double   ang1, ang2, d1, d2, arot, alfa, h, dist, bulge, T, Tmax, ang;
  CGePoint ptm, pt2;
  static double Delta = 0.0001;

  if (Radius <= 0.0){
    return false;
  }
  // get arc angle (rotation angle, A,B,C)
  ang1 = A.Angle( B );
  ang2 = B.Angle( C );
  d1 = A.Distance( B );
  d2 = B.Distance( C );
  Tmax = (d1 < d2)? d1 : d2;
  if (Tmax < Delta){
    return false;
  }
  // arc angle, + to left(counter-clockwise), - to right (clockwise), radian 
 	arot = ang2 - ang1;
  if (arot == 0.0){
    return false;
  }
	if (arot < -GE_DEG180) arot+=GE_DEG360;
	if (arot >  GE_DEG180) arot-=GE_DEG360;
 	alfa = fabs( arot ) / 2.0;    // absolute angle of the arc
m1:
  T = Radius * tan( alfa );
  if (T > Tmax){
    T -= Delta;
    if (T < Tmax){
      Radius -= Delta;
      goto m1;
    }
    return false;
  }
  // start point of the curve
  ang = ang1 - GE_DEG180;
  pOutPt1->x = B.x + (T * cos(ang));
  pOutPt1->y = B.y + (T * sin(ang));
  // end point of the curve
  pOutPt2->x = B.x + (T * cos(ang2));
  pOutPt2->y = B.y + (T * sin(ang2));

  h = Radius * (1.0 - cos(alfa));
  dist = _hypot( pOutPt1->x - pOutPt2->x, pOutPt1->y - pOutPt2->y ) / 2.0;
  if (pBulge){
    if (arot < 0.0){
      bulge = -h / dist;
    }else{
      bulge = h / dist;
    }
    *pBulge = bulge;
  }
  if (pOutPtMid){
    pt2 = *pOutPt2;
    ptm = pOutPt1->Center( pt2 );
    ang1 = pOutPt1->Angle( pt2 ) - GE_PI2;
//    dist = pOutPt1->Distance( pt2 ) / 2.0 * h;
    if (arot < 0.0){
      ptm.Polar( ang1, -h );
    }else{
      ptm.Polar( ang1, h );
    }
    *pOutPtMid = ptm;
  }
  return true;
}


//-----------------------------------------------
double geGetRadiusByTang (const CGePoint& A, const CGePoint& B, const CGePoint& C, double Tang)
{
  double ang1, ang2, arot, alfa, Radius;

  if (Tang <= 0.0){
    return 0.0;
  }
  // get arc angle (rotation angle, A,B,C)
  ang1 = A.Angle( B );
  ang2 = B.Angle( C );
  // arc angle, + to left(counter-clockwise), - to right (clockwise), radian 
 	arot = ang2 - ang1;
	if (arot < -GE_DEG180) arot+=GE_DEG360;
	if (arot >  GE_DEG180) arot-=GE_DEG360;
 	alfa = fabs( arot ) / 2.0;    // absolute angle of the arc
  Radius = Tang / tan( alfa );
  return Radius;
}


//-----------------------------------------------
double geGetRadiusByBisector (const CGePoint& A, const CGePoint& B, const CGePoint& C, double Bis)
{
  double ang1, ang2, arot, alfa, Radius, ta;

  if (Bis <= 0.0){
    return 0.0;
  }
  // get arc angle (rotation angle, A,B,C)
  ang1 = A.Angle( B );
  ang2 = B.Angle( C );
  // arc angle, + to left(counter-clockwise), - to right (clockwise), radian 
 	arot = ang2 - ang1;
	if (arot < -GE_DEG180) arot+=GE_DEG360;
	if (arot >  GE_DEG180) arot-=GE_DEG360;
 	alfa = fabs( arot ) / 2.0;    // absolute angle of the arc
  ta = tan( alfa );
  Radius = Bis / (sqrt(ta*ta+1) - 1);
  return Radius;
}

/********************************************************************
* Get Fillet Arc by 3 points and radius
*********************************************************************/
bool geGetFilletArc (const CGePoint& A, const CGePoint& B, const CGePoint& C, double R, CGeArc* pOutArc)
{
  double   ang1, ang2, d1, d2, arot, alfa, T, Tmax, ang;
  CGePoint ptm, pt2, StartPt, CenPt; //EndPt;
  static double Delta = 0.0001;

  // get arc angle (rotation angle, A,B,C)
  ang1 = A.Angle( B );
  ang2 = B.Angle( C );
  d1 = A.Distance( B );
  d2 = B.Distance( C );
  Tmax = (d1 < d2)? d1 : d2;
  if (Tmax < Delta){
    return false;
  }
  Tmax *= 0.95;
  // arc angle, + to left(counter-clockwise), - to right (clockwise), radian 
 	arot = ang2 - ang1;
  if (arot == 0.0){
    return false;
  }
	if (arot < -GE_DEG180) arot+=GE_DEG360;
	if (arot >  GE_DEG180) arot-=GE_DEG360;
 	alfa = fabs( arot ) / 2.0;    // absolute angle of the arc
  
  if (R <= 0.0){
    R = Tmax / tan( alfa );
    T = Tmax;
  }else{
m1:
    T = R * tan( alfa );
    if (T > Tmax){
      T -= Delta;
      if (T < Tmax){
        R -= Delta;
        goto m1;
      }
      return false;
    }
  }
  // start point of the curve
  ang = ang1 - GE_DEG180;
  StartPt.x = B.x + (T * cos(ang));
  StartPt.y = B.y + (T * sin(ang));
  // end point of the curve
//  EndPt.x = B.x + (T * cos(ang2));
//  EndPt.y = B.y + (T * sin(ang2));
  // make fillet arc
  if (arot > 0.0){
    CenPt = StartPt.PolarBy( ang1+GE_DEG90, R );
  }else{
    CenPt = StartPt.PolarBy( ang1-GE_DEG90, R );
  }
  ang = CenPt.Angle( StartPt );
  pOutArc->Set( CenPt, R, ang, arot );
  return true;
}



/********************************************************************
* Get array of fillet arcs between 2 lines
*********************************************************************/
UINT_PTR geGetFilletLines (const CGeLine& Lin1, const CGeLine& Lin2, double R, CGeArc* pOutArc)
{
  UINT_PTR  nArcs = 0;
  UINT_PTR  nPnts, i, nip;
  CGePoint  InterPt[5], Pt1, Pta0, Pta1;
  CGePoint* pPnt;  
  CGeArc*   pArc = pOutArc;
  CGeLine   Lin1a, Lin1b, Lin2a, Lin2b, Line;
  double    dR = R + R + R;
//  double    L1 = Lin1.Length();
  double    A1 = Lin1.Angle() + GE_DEG90;
//  double    L2 = Lin2.Length();
  double    A2 = Lin2.Angle() + GE_DEG90;
  double    Ang;

//  Lin1 = _Lin1;
//  Lin1.Inflate( L1 * 10.0 );
  Lin1a = Lin1b = Lin1;
  Lin1a.Offset( -R );
  Lin1b.Offset( R );

//  Lin2 = _Lin2;
//  Lin2.Inflate( L2 * 10.0 );
  Lin2a = Lin2b = Lin2;
  Lin2a.Offset( -R );
  Lin2b.Offset( R );

  nip = 0;
  pPnt = InterPt;
  for (i=0; i<4; ++i){
    nPnts = 0;
    switch( i ){
      case 0: nPnts = geInter_Line_Line( Lin1a, Lin2a, 0, pPnt ); break;
      case 1: nPnts = geInter_Line_Line( Lin1a, Lin2b, 0, pPnt ); break;
      case 2: nPnts = geInter_Line_Line( Lin1b, Lin2a, 0, pPnt ); break;
      case 3: nPnts = geInter_Line_Line( Lin1b, Lin2b, 0, pPnt ); break;
    }
    if (nPnts){
      ++pPnt;
      ++nip;
      if (nip == 5){
        return 0;
      }
    }
  }
  pPnt = InterPt;
//  A1 += GE_DEG90;
//  A2 += GE_DEG90;
  for (i=0; i<nip; ++i){
    Pt1 = *pPnt; 
    ++pPnt;
    Line.Set( Pt1, A1, dR, dR );
    nPnts = geInter_Line_Line( Line, Lin1, 0, &Pta0 );
    if (nPnts == 1){
      Line.Set( Pt1, A2, dR, dR );
      nPnts = geInter_Line_Line( Line, Lin2, 0, &Pta1 );
      if (nPnts == 1){
        pArc->SetCSE( Pt1, Pta0, Pta1, false );
        Ang = fabs( pArc->AngleArc() );
        if (Ang > GE_PI){
          Ang = GE_2PI - Ang;
          if (pArc->IsCCW()){
            Ang = -Ang;
          }
          pArc->SetArcAngle( Ang );
        }
        ++nArcs;
        ++pArc;
      }
    }
  }
  return nArcs;
}


/********************************************************************
* Get array of fillet arcs between 2 circles
*********************************************************************/
UINT_PTR geGetFilletCircs (const CGeCircle& Cir1, const CGeCircle& Cir2, double R, CGeArc* pOutArc)
{
  UINT_PTR  nArcs = 0;
  UINT_PTR  nPnts, i, j, nip;
  double    R1 = Cir1.Radius();
  double    R2 = Cir2.Radius();
  double    dR = R + R + R;
  double    Ang;
  CGeCircle Cir1o, Cir1i, Cir2o, Cir2i;
  CGeLine   Lin;
  CGePoint  InterPt[12], Pt1, Pta0, Pta1;
  CGePoint* pPnt;  
  CGeArc*   pArc = pOutArc;

  Cir1i.Set( Cir1.Center(), R1-R );
  Cir1o.Set( Cir1.Center(), R1+R );
  Cir2i.Set( Cir2.Center(), R2-R );
  Cir2o.Set( Cir2.Center(), R2+R );
  nip = 0;
  pPnt = InterPt;
  for (i=0; i<4; ++i){
    nPnts = 0;
    switch( i ){
      case 0: nPnts = geInter_Circle_Circle( Cir1i, Cir2i ); break;
      case 1: nPnts = geInter_Circle_Circle( Cir1i, Cir2o ); break;
      case 2: nPnts = geInter_Circle_Circle( Cir1o, Cir2i ); break;
      case 3: nPnts = geInter_Circle_Circle( Cir1o, Cir2o ); break;
    }
    for (j=0; j<nPnts; ++j){
      geInter_GetPoint( j, pPnt );
      ++pPnt;
      ++nip;
      if (nip == 12){
        return 0;
      }
    }
  }
  pPnt = InterPt;
  for (i=0; i<nip; ++i){
    Pt1 = *pPnt; 
    ++pPnt;
    Lin.Set( Cir1.Center(), Pt1 );
    Lin.Inflate( 0.0, dR );
    nPnts = geInter_Line_Circle( Lin, Cir1, 0, &Pta0 );
    if (nPnts == 1){
      Lin.Set( Cir2.Center(), Pt1 );
      Lin.Inflate( 0.0, dR );
      nPnts = geInter_Line_Circle( Lin, Cir2, 0, &Pta1 );
      if (nPnts == 1){
        pArc->SetCSE( Pt1, Pta0, Pta1, false );
        Ang = fabs( pArc->AngleArc() );
        if (Ang > GE_PI){
          Ang = GE_2PI - Ang;
          if (pArc->IsCCW()){
            Ang = -Ang;
          }
          pArc->SetArcAngle( Ang );
        }
        ++nArcs;
        ++pArc;
      }
    }
  }
  return nArcs;
}


/********************************************************************
* Get array of fillet arcs between a circle and a line
*********************************************************************/
UINT_PTR geGetFilletLineCirc (const CGeLine& Lin1, const CGeCircle& Cir1, double R, CGeArc* pOutArc)
{
  UINT_PTR  nArcs = 0;
  UINT_PTR  nPnts, i, j, nip;
  CGePoint  InterPt[12], Pt1, Pta0, Pta1;
  CGePoint* pPnt;  
  CGeArc*   pArc = pOutArc;
  CGeLine   Lin1a, Lin1b, Line;
  double    dR = R + R + R;
  double    A1 = Lin1.Angle() + GE_DEG90;
  double    Ang;
  CGeCircle Cir1o, Cir1i;

  Lin1a = Lin1b = Lin1;
  Lin1a.Offset( -R );
  Lin1b.Offset( R );
  Cir1i.Set( Cir1.Center(), Cir1.Radius()-R );
  Cir1o.Set( Cir1.Center(), Cir1.Radius()+R );
  nip = 0;
  pPnt = InterPt;
  for (i=0; i<4; ++i){
    nPnts = 0;
    switch( i ){
      case 0: nPnts = geInter_Line_Circle( Lin1a, Cir1i  ); break;
      case 1: nPnts = geInter_Line_Circle( Lin1a, Cir1o ); break;
      case 2: nPnts = geInter_Line_Circle( Lin1b, Cir1i ); break;
      case 3: nPnts = geInter_Line_Circle( Lin1b, Cir1o ); break;
    }
    for (j=0; j<nPnts; ++j){
      geInter_GetPoint( j, pPnt );
      ++pPnt;
      ++nip;
      if (nip == 12){
        return 0;
      }
    }
  }
  pPnt = InterPt;
  for (i=0; i<nip; ++i){
    Pt1 = *pPnt; 
    ++pPnt;
    Line.Set( Pt1, A1, dR, dR );
    nPnts = geInter_Line_Line( Line, Lin1, 0, &Pta0 );
    if (nPnts == 1){
      Line.Set( Cir1.Center(), Pt1 );
      Line.Inflate( 0.0, dR );
      nPnts = geInter_Line_Circle( Line, Cir1, 0, &Pta1 );
      if (nPnts == 1){
        pArc->SetCSE( Pt1, Pta0, Pta1, false );
        Ang = fabs( pArc->AngleArc() );
        if (Ang > GE_PI){
          Ang = GE_2PI - Ang;
          if (pArc->IsCCW()){
            Ang = -Ang;
          }
          pArc->SetArcAngle( Ang );
        }
        ++nArcs;
        ++pArc;
      }
    }
  }
  return nArcs;
}


/********************************************************************
* Returns counter-clockwise angle between line A-B
* and 0-X axis.
* result angle is radians 0-2Pi  
*********************************************************************/
double geGetAngle (double Ax, double Ay, double Bx, double By)
{
  double alfa, dx, dy;

  dx = Bx - Ax;
  dy = By - Ay;
  if (dx==0) {
    alfa = (dy>=0)? GE_PI2 : -GE_PI2 ;
  }else{
    alfa = atan (dy/dx);
    if (dx<0) {
      if (dy>=0) alfa+=GE_PI; else alfa-=GE_PI;
    }
  }
  // adjust angle to 0-2Pi range
  if (alfa<0) alfa+=GE_2PI;
  return alfa;
}

/********************************************************************
* Returns turn angle of line 1-2, relative to 0-1 direction
* Result angle is radians 0..+/-Pi
* Positive angle - turn left, negative - turn right
*********************************************************************/
double geGetAngleTurn (double x0, double y0, double x1, double y1, double x2, double y2)
{
  CGePoint Pt0, Pt1, Pt2;
  double   Ang;

  Pt0.Set(x0,y0);
  Pt1.Set(x1,y1);
  Pt2.Set(x2,y2);
  Ang = Pt0.Angle( Pt1 );
  Pt2.Rotate( Pt1, -Ang );
  Ang = Pt1.Angle( Pt2 );
  if (Ang > GE_PI){
    Ang -= GE_2PI;
  }
  return Ang;
}


/********************************************************************
* Normalize Angle to range 0-PI2 radian
*********************************************************************/
double geNormalizeAngle (double Ang)
{
  if (Ang < 0.){
    if (Ang < -1000000.0){
      return 0.0;
    }
    while( Ang < 0. ){
      Ang += GE_2PI;
    }
  }else{
    if (Ang > 1000000.0){
      return 0.0;
    }
    if (GE_2PI < Ang){
      while( GE_2PI < Ang ){
        Ang -= GE_2PI;
      }
    }
  }
  if (Ang < 0.000001 || (GE_2PI-0.000001) < Ang){
    Ang = 0.0;
  }
  return Ang;
}

/********************************************************************
* Returns distance between 2 angles (A,B)
*********************************************************************/
double geGetDeltaAngle (double A, double B)
{
  double r1, r2;
  if (A > B){
    r1 = A - B; 
    r2 = B - A + GE_2PI;
  }else{ 
    r1 = B - A; 
    r2 = A - B + GE_2PI; 
  }
  return (r1>r2)? r2 : r1;
}

/********************************************************************
* Check if 2 angles look to the same direction
*********************************************************************/
bool geIsSameAngle (double Ang1, double Ang2, double Delta)
{
  double dA;
  if (Ang1 == Ang2){
    return true;
  }
  dA = geGetDeltaAngle( Ang1, Ang2 );
  return (dA < Delta)? true : false;
}



/********************************************************************
*  Returns round direction value for points ABC
*  >0 - ABC has counter-clockwise direction
*  <0 - ABC has clockwise direction 
*  =0 - ABC has linear direction
*********************************************************************/
double geRoundValue (const CGePoint& A, const CGePoint& B, const CGePoint& C)
{
//  return (A.x*(B.y-C.y) + B.x*(C.y-A.y) + C.x*(A.y-B.y));
  return (A.x-C.x)*(B.y-C.y) - (B.x-C.x)*(A.y-C.y);
}


/********************************************************************
*  Calculates an area of PolyPolygon
*  Pgon - array of polygons
*  nPgons - number of polygons
*********************************************************************/
double geGetArea (const CGePolyline* Pgon, int nPgons)
{
  int    i, j;
  double Area, TotalArea;

  TotalArea = 0.0;
  for (i=0; i<nPgons; ++i){
    Area = Pgon[i].Area();
    if (Area > 0.){
      for (j=0; j<nPgons; ++j){
        if (i != j){
          if (Pgon[j].Contains( Pgon[i] )){
            Area = -Area;
          }
        }
      }
      TotalArea += Area;
    }
  }
  return TotalArea;
}


//-----------------------------------------------
// Returns 2 offset points of vertex B
// where Dist - distance from B
// pPlef or pPrig can be NULL
//-----------------------------------------------
void geGetWidthPoints (const CGePoint* pA, const CGePoint* pB, const CGePoint* pC, double Dist, CGePoint* pPlef, CGePoint* pPrig)
{
  double ang, ang1, ang2, d, dx, dy, cf;
//  double maxDist = Dist * g_GeMiterLimit;

  if (pA == NULL){
    ang = pB->Angle( *pC );
    ang += GE_DEG90;
    dx = Dist * cos(ang);
    dy = Dist * sin(ang);
    if (pPlef){
      pPlef->x = pB->x + dx;
      pPlef->y = pB->y + dy;
    }
    if (pPrig){
      pPrig->x = pB->x - dx;
      pPrig->y = pB->y - dy;
    }
    return;
  }
  if (pC == NULL){
    ang = pA->Angle( *pB );
    ang += GE_DEG90;
    dx = Dist * cos(ang);
    dy = Dist * sin(ang);
    if (pPlef){
      pPlef->x = pB->x + dx;
      pPlef->y = pB->y + dy;
    }
    if (pPrig){
      pPrig->x = pB->x - dx;
      pPrig->y = pB->y - dy;
    }
    return;
  }
  ang1 = pB->Angle( *pC );
  ang2 = pA->LeftAngle( *pB, *pC );
  ang2 /= 2.;
  // calculate distance along bisector
  ang = fabs( ang2 - GE_DEG90 );
  d = cos( ang );
  cf = (d==0)? 1000.0 : 1.0/d ;
  if (cf > g_GeMiterLimit){
    if (cf > 100){
      // the angle too sharp, cut the end by perpendicular
      ang = pA->Angle( *pB );
      ang += GE_DEG90;
      dx = Dist * cos(ang);
      dy = Dist * sin(ang);
      if (pPlef){
        pPlef->x = pB->x + dx;
        pPlef->y = pB->y + dy;
      }
      if (pPrig){
        pPrig->x = pB->x - dx;
        pPrig->y = pB->y - dy;
      }
      return;
    }else{
      d = Dist * g_GeMiterLimit;
    }
  }else{
    d = Dist * cf;
  }
  // get width points of middle vertex
  ang = ang1 + ang2;
  dx = d * cos(ang);
  dy = d * sin(ang);
  if (pPlef){
    pPlef->x = pB->x + dx;
    pPlef->y = pB->y + dy;
  }
  if (pPrig){
    pPrig->x = pB->x - dx;
    pPrig->y = pB->y - dy;
  }
}

//-----------------------------------------------
// Returns an offset point from vertex B
// Dist < 0. means OffPnt will be placed at left side of A-B-C
// Dist > 0. means OffPnt will be placed at right side of A-B-C
//-----------------------------------------------
void geGetOffsetPoint (const CGePoint* pA, const CGePoint* pB, const CGePoint* pC, double Dist, CGePoint* pOffPnt)
{
  double ang, ang1, ang2, d, dx, dy, cf; 
//  double maxDist = Dist * g_GeMiterLimit;

  if (pA == NULL){
    ang = pB->Angle( *pC );
    ang -= GE_DEG90;
    dx = Dist * cos(ang);
    dy = Dist * sin(ang);
    pOffPnt->x = pB->x + dx;
    pOffPnt->y = pB->y + dy;
    return;
  }
  if (pC == NULL){
    ang = pA->Angle( *pB );
    ang -= GE_DEG90;
    dx = Dist * cos(ang);
    dy = Dist * sin(ang);
    pOffPnt->x = pB->x + dx;
    pOffPnt->y = pB->y + dy;
    return;
  }
  ang1 = pB->Angle( *pC );
  ang2 = pA->LeftAngle( *pB, *pC );
  ang2 /= 2.0;
  // calculate distance along bisector
  ang = fabs( ang2 - GE_DEG90 );
  d = cos( ang );
  cf = (d==0.0)? 1000.0 : 1.0/d ;
  if (cf > g_GeMiterLimit){
    if (cf > 100.0){
      // the angle too sharp, cut the end by perpendicular
      ang = pA->Angle( *pB );
      ang -= GE_DEG90;
      dx = Dist * cos(ang);
      dy = Dist * sin(ang);
      pOffPnt->x = pB->x + dx;
      pOffPnt->y = pB->y + dy;
      return;
    }else{
      d = Dist * g_GeMiterLimit;
    }
  }else{
    d = Dist * cf;
  }
  // get width points of middle vertex
  ang = ang1 + ang2;
  dx = d * cos(ang);
  dy = d * sin(ang);
  pOffPnt->x = pB->x - dx;
  pOffPnt->y = pB->y - dy;
}


// Transform model space length (drw units) into window space length (pixels)
//-----------------------------------------------
UINT_PTR geTransformSize (double Len0, const GE_TRANSPRM* pPrm, double PixelSize)
{
  UINT_PTR sz;
  CGePoint Pt0, Pt1;
  double   Len, Len2;

  if (PixelSize <= 0.0){
    return 0;
  }
  if (pPrm){
    Pt0.Set( 0.0, 0.0 );
    Pt1.Set( Len0, 0.0 );
    Pt0.Transform( pPrm );
    Pt1.Transform( pPrm );
    Len = Pt0.Distance( Pt1 );
    if (pPrm->Scale.x != pPrm->Scale.y){
      Pt0.Set(0,0);
      Pt1.Set(0,Len0);
      Pt0.Transform( pPrm );
      Pt1.Transform( pPrm );
      Len2 = Pt0.Distance( Pt1 );
      if (Len < Len2){
        Len = Len2;
      }
    }
  }else{
    Len = Len0;
  }
  sz = (UINT_PTR)((Len / PixelSize) + 0.5);
  return sz;
}

// Transform angle (drw units)
//-----------------------------------------------
double geTransformAngle (double Angle, const GE_TRANSPRM* pPrm)
{
  double NewAng;
  CGePoint Pt0, Pt1;

  NewAng = Angle;
  if (pPrm){
    Pt0.Set( 0.0, 0.0 );
    Pt1 = Pt0.PolarBy( Angle, 100.0 );
    Pt0.Transform( pPrm );
    Pt1.Transform( pPrm );
    NewAng = Pt0.Angle( Pt1 );
  }
  return NewAng;
}

//-----------------------------------------------
// Return true if transform has uniform scale
//-----------------------------------------------
bool geTransHasUniscale (const GE_TRANSPRM* pPrm)
{
  const GE_TRANSPRM* ptp;
  bool bUniscale = true;

  ptp = pPrm;
  while( ptp ){
    if (ptp->Scale.x != ptp->Scale.y){
      bUniscale = false;
      break;
    }
    ptp = ptp->pNext;
  }
  return bUniscale;
}



//-----------------------------------------------
bool geSnapOrtho (const CGePoint& BasePt, const CGePoint& Pnt2, double SnapDist, CGePoint* pSnapPt)
{
  double   dx, dy;
  CGePoint Pt;

  Pt = Pnt2;
  dx = fabs(Pt.x - BasePt.x);
  dy = fabs(Pt.y - BasePt.y);
  if (dx==0.0 && dy==0.0){
    return false;
  }
  if (dx < dy){
    Pt.x = BasePt.x;
    if (SnapDist > 0.0){
      dy = Pt.y - BasePt.y;
      if (dy > 0.0){
        Pt.y = BasePt.y + floor( dy / SnapDist + 0.5)  * SnapDist;
      }else{
        Pt.y = BasePt.y - floor( -dy / SnapDist + 0.5)  * SnapDist;
      }
    }
  }else{
    Pt.y = BasePt.y;
    if (SnapDist > 0.0){
      dx = Pt.x - BasePt.x;
      if (dx > 0.){
        Pt.x = BasePt.x + floor( dx / SnapDist + 0.5) * SnapDist;
      }else{
        Pt.x = BasePt.x - floor( -dx / SnapDist + 0.5) * SnapDist;
      }
    }
  }
  *pSnapPt = Pt;
  return true;
}


//-----------------------------------------------
// Reverse points order in the array Pnt
//-----------------------------------------------
void geReversePoints (CGePoint* Pnt, UINT_PTR nPnts)
{
  CGePoint  Ptt;
  CGePoint* pPnt1;
  CGePoint* pPnt2;
  UINT_PTR  nPnts2;
  UINT_PTR  i;

  pPnt1 = Pnt;
  pPnt2 = Pnt + nPnts - 1;
  nPnts2 = nPnts >> 1;
  for (i=0; i<nPnts2; ++i){
    Ptt = *pPnt1;
    *pPnt1 = *pPnt2;
    *pPnt2 = Ptt;
    ++pPnt1;
    --pPnt2;
  }
}

//-----------------------------------------------
// Rotate an array of points
//-----------------------------------------------
void gePointsRotate (CGePoint* _pPnt, UINT_PTR nPnts, const CGePoint& Pt0, double Angle)
{
  UINT_PTR i;
  CGePoint* pPnt = _pPnt;
  for (i=0; i<nPnts; ++i){
    pPnt->Rotate( Pt0, Angle );
    ++pPnt;
  }
}

//-----------------------------------------------
// Move an array of points
//-----------------------------------------------
void gePointsMove (CGePoint* _pPnt, UINT_PTR nPnts, double DX, double DY)
{
  UINT_PTR i;
  CGePoint* pPnt = _pPnt;
  for (i=0; i<nPnts; ++i){
    pPnt->Move( DX, DY );
    ++pPnt;
  }
}

//-----------------------------------------------
// Returns true if 4 points are on a line
// Pnt - array of 4 points
//-----------------------------------------------
bool geIsLine4p (CGePoint* Pnt)
{
  CGeLine Line;
  double  Delta, Len, Ang, y0, dy;

  y0 = Pnt[0].y;
  Line.Set( Pnt[0], Pnt[3] );
  Ang = Line.Angle();
  Len = Line.Length();
  Delta = Len / 1000.0;
  Pnt[1].Rotate( Pnt[0], -Ang );
  dy = Pnt[1].y - y0;
  if (fabs(dy) > Delta){
    return false;
  }
  Pnt[2].Rotate( Pnt[0], -Ang );
  dy = Pnt[2].y - y0;
  if (fabs(dy) > Delta){
    return false;
  }
  // 4 poins are on one line
  return true;
}
