/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* This code was taken from 'BaseLibs' project
* Classes for 2D geometric shapes
*********************************************************************/
#ifndef _OK_GEOM_
#define _OK_GEOM_

#ifndef _INC_MATH
  #include <math.h>
#endif
#ifndef _INC_STDLIB
  #include <stdlib.h>
#endif
#ifndef _OK_LISTOBJ_
  #include "ListObj.h"
#endif
#ifndef _OK_BASETYPES_
  #include "okBaseTypes.h"
#endif


#define GE_FLT_MAX  3.4E+38
#define GE_DBL_MAX  1.7E+308

// angles in radians
//            3,14159265358979323846 2643383279 5028841971 6939937510 5820974944 5923078164 0628620899 8628034825 3421170679
#define GE_PI 3.14159265358979323846    // 180
#define GE_PI2 (GE_PI / 2.0)    // 1,57079632679489661923    90
#define GE_PI4 (GE_PI / 4.0)    // 0,785398163397448309615   45
#define GE_2PI (GE_PI+GE_PI)    // 6,28318530717958647692   360

// aka degrees
#define GE_DEG1    GE_PI / 180.0
#define GE_DEG2    GE_PI / 90.0
#define GE_DEG3    GE_PI / 60.0
#define GE_DEG4    GE_PI / 45.0
#define GE_DEG5    GE_PI / 36.0
#define GE_DEG10   GE_PI / 18.0
#define GE_DEG20   GE_PI /  9.0
#define GE_DEG22_5 GE_PI /  8.0
#define GE_DEG30   GE_PI /  6.0
#define GE_DEG36   GE_PI /  5.0
#define GE_DEG45   GE_PI4
#define GE_DEG90   GE_PI2
#define GE_DEG135  (GE_PI2+GE_PI4)
#define GE_DEG180  GE_PI
#define GE_DEG225  (GE_PI+GE_PI4)
#define GE_DEG270  (GE_PI+GE_PI2)
#define GE_DEG315  (GE_2PI-GE_PI4)
#define GE_DEG360  GE_2PI


// Radian -> Degree
#define GE_RAD_TO_DEG   57.2957795130823208768
// Degree -> Radian
#define GE_DEG_TO_RAD    0.01745329251994329577


// Intersection results
enum class EnIntersec{
  GE_COLLINEAR=0, 
  GE_PARALLEL=1, 
  GE_SKEW=2, 
  GE_SKEW_CROSS=3, 
  GE_SKEW_NO_CROSS=4
};

// Type of point's position relative to line
enum class EnPntPos{
  GE_LEFT=0, 
  GE_RIGHT=1, 
  GE_BEFORE=2, 
  GE_AFTER=3, 
  GE_BETWEEN=4, 
  GE_ORIGIN=5, 
  GE_DESTINATION=6
};

class CGeLine;
class CGeRect;
class CGePolyline;
class CGeCircle;
class CGeArc;
class CGeEllipse;
class CGe3dPoint;

struct GE_TRANSPRM;

#ifndef _GE_GLOBALS
  extern double g_GeTolerance;
  extern double g_GeTolClass;   // tolerance for CGeLine::ClassifyLine() and CGeLine::ClassifyPoint()
  extern double g_GeMiterLimit;
#endif

//=============================================================================
//
//  CGePoint_
//
//=============================================================================
class CGePoint {
public:
  double x;
  double y;
public:
    CGePoint ();
    CGePoint (const CGePoint& Pnt);
    CGePoint (double x, double y);

  void  Set (double x, double y);
  void  Set (const CGePoint& Pnt);
  void  SetVector (double Angle);
  void  SetVector (const CGePoint& Pnt);
  void  SetVector (const CGePoint& pt0, const CGePoint& pt1);

  double VectorLength    () const;
  void   VectorScale     (double Scale);
  void   VectorNormalize ();

  // Returns TRUE if distance between this point and 'Pnt2'
  // less then 'delta' value
  bool  IsEqual (const CGePoint& Pnt, double Delta=g_GeTolerance) const;
  bool  IsNotEqual (const CGePoint& Pnt, double Delta=g_GeTolerance) const;

  bool  IsNull    () const;
  bool  IsNotNull () const;

  // Functions/binary-operators that return points or sizes
  CGePoint  MoveBy      (double dx, double dy) const;
  CGePoint  MoveBy      (const CGePoint& Off) const;
  CGePoint  PolarBy     (double Angle, double Dist) const;
  CGePoint  RotateBy    (const CGePoint& Base, double Angle) const;
  CGePoint  ScaleBy     (const CGePoint& Base, double ScaleX, double ScaleY) const;
  CGePoint  ScaleBy     (const CGePoint& Base, double Scal) const;
  CGePoint  MirrorBy    (double a1, double b1) const;
  CGePoint  MirrorBy    (const CGeLine& line) const;
  CGePoint  ReflectBy   (const CGePoint& Base) const;
  CGePoint  TransformBy (const GE_TRANSPRM* pPrm) const;
  CGePoint  TransBackBy (const GE_TRANSPRM* pPrm) const;

  // Functions/assignement-operators that modify this point
  void      Move        (double dx, double dy);
  void      Move        (const CGePoint& Off);
  void      Polar       (double Angle, double Dist);
  void      Rotate      (const CGePoint& Base, double Angle);
  void      Scale       (const CGePoint& Base, double ScaleX, double ScaleY);
  void      Scale       (const CGePoint& Base, double Scal);
  void      Mirror      (double a1, double b1);
  void      Mirror      (const CGeLine& line);
  void      Reflect     (const CGePoint& Base);
  void      Transform   (const GE_TRANSPRM* pPrm);
  void      TransBack   (const GE_TRANSPRM* pPrm);

  bool      operator == (const CGePoint& p) const;
  bool      operator != (const CGePoint& p) const;
  CGePoint  operator +  (const CGePoint& p) const;
  CGePoint  operator -  (const CGePoint& p) const;
  bool      operator <  (const CGePoint& p) const;
  bool      operator >  (const CGePoint& p) const;

  // Vector length
  double Length () const;

  // Distance from this point to other point
  double Distance (const CGePoint& Pnt) const;
  // Distance from this point to other point along the angle direction
  double Distance (const CGePoint& Pnt, double Angle) const;

  // Shortest distance from this point to a line
  double Distance (const CGeLine& line, bool bApparent) const;

  // Angle from this point to 'Pnt' (0..2PI)
  double Angle (const CGePoint& Pnt) const;
  // Angle from point 0,0 to this point (0..2PI)
  double Angle () const;

  // Left side angle this-p1-p2 (0..2PI)
  double LeftAngle  (const CGePoint& Pnt1, const CGePoint& Pnt2) const;
  // Right side angle this-p1-p2 (0..2PI)
  double RightAngle (const CGePoint& Pnt1, const CGePoint& Pnt2) const;

  // Left side center angle this-p1-p2 (0..2PI)
  double LeftCenAngle  (const CGePoint& Pnt1, const CGePoint& Pnt2) const;
  // Right side center angle this-p1-p2 (0..2PI)
  double RightCenAngle (const CGePoint& Pnt1, const CGePoint& Pnt2) const;
  // Center angle of bigger lef/rig angle this-p1-p2 (0..2PI)
  double CenterAngle   (const CGePoint& Pnt1, const CGePoint& Pnt2, bool* pLeft) const;

  // deviation angle from this-p1 direction (0..PI) + to left side, - to right side
  double TurnAngle (const CGePoint& pnt1, const CGePoint& pnt2) const;

  // Get center point between this point to other point
  CGePoint Center (const CGePoint& Pnt) const;

  // Get center of circle that goes through 3 points
  bool GetCenter (const CGePoint& pt1, const CGePoint& pt2, CGePoint* pCenter, double* pRadius) const;

  // Makes perpendicular line from this point to the specified Line
  EnPntPos GetPerpendicular (const CGeLine& Line, CGePoint* pInterPt) const;

  EnPntPos Classify (const CGePoint& pt0, const CGePoint& pt1) const;

  const CGePoint&  Point () const;
};



//-------------------------------------------------------------------
// Inline functions
//-------------------------------------------------------------------
inline      CGePoint::CGePoint () : x(0.), y(0.) {}
inline      CGePoint::CGePoint (const CGePoint& src) : x(src.x), y(src.y) {}
inline      CGePoint::CGePoint (double xx, double yy) : x(xx), y(yy) {}
inline void CGePoint::Set (double xx, double yy) {x=xx; y=yy;}
inline void CGePoint::Set (const CGePoint& Pnt) {x = Pnt.x; y=Pnt.y;}
inline bool CGePoint::IsEqual(const CGePoint& Pt2, double Delta) const {
  if ((x-Delta <= Pt2.x) && (Pt2.x <= x+Delta) && (y-Delta <= Pt2.y) && (Pt2.y <= y+Delta)) {
    return true;
  }
  return false;
}
inline bool CGePoint::IsNotEqual(const CGePoint& p2, double delta) const {
  if ((x-delta <= p2.x) && (p2.x <= x+delta) && (y-delta <= p2.y) && (p2.y <= y+delta)) {
    return false;
  }
  return true;
}
inline bool     CGePoint::IsNull   () const {return (x==0. && y==0.)? true : false;}
inline bool     CGePoint::IsNotNull() const {return (x==0. && y==0.)? false : true;}
inline CGePoint CGePoint::MoveBy (double dx, double dy) const {return CGePoint(x+dx, y+dy);}
inline CGePoint CGePoint::MoveBy (const CGePoint& Off) const {return CGePoint(x+Off.x, y+Off.y);}
inline CGePoint CGePoint::ScaleBy (const CGePoint& Base, double Scx, double Scy) const {
  return CGePoint( Base.x + (x - Base.x) * Scx, Base.y + (y - Base.y) * Scy );
}
inline CGePoint CGePoint::ScaleBy (const CGePoint& Base, double Sc) const {
  return CGePoint( Base.x + (x - Base.x) * Sc, Base.y + (y - Base.y) * Sc );
}
inline void CGePoint::Move (double dx, double dy) {x+=dx; y+=dy;}
inline void CGePoint::Move (const CGePoint& Off)  {x+=Off.x; y+=Off.y;}
inline void CGePoint::Scale (const CGePoint& Base, double Scx, double Scy) {
  x = Base.x + (x - Base.x) * Scx;
  y = Base.y + (y - Base.y) * Scy;
}
inline void CGePoint::Scale (const CGePoint& Base, double Sc) {
  x = Base.x + (x - Base.x) * Sc;
  y = Base.y + (y - Base.y) * Sc;
}
inline bool     CGePoint::operator == (const CGePoint& p) const {return (x==p.x && y==p.y);}
inline bool     CGePoint::operator != (const CGePoint& p) const {return (x!=p.x || y!=p.y);}
inline CGePoint CGePoint::operator + (const CGePoint& p) const {return CGePoint(x+p.x, y+p.y);}
inline CGePoint CGePoint::operator - (const CGePoint& p) const {return CGePoint(x-p.x, y-p.y);}
inline bool     CGePoint::operator < (const CGePoint& p) const {return (x<p.x || (x==p.x && y<p.y));}
inline bool     CGePoint::operator > (const CGePoint& p) const {return (x>p.x || (x==p.x && y>p.y));}
inline CGePoint CGePoint::Center (const CGePoint& p) const {return CGePoint( (x+p.x)/2.0, (y+p.y)/2.0 );}
inline double   CGePoint::Length () const {return _hypot( x, y );}
inline double   CGePoint::Distance  (const CGePoint& Pnt) const {return _hypot( x-Pnt.x, y-Pnt.y );}
inline const CGePoint&  CGePoint::Point () const {return *this;}


// parameters of in-block transformation
//-----------------------------------------------
struct GE_TRANSPRM {
  CGePoint Orig;         // origin point (World Coord System)
  CGePoint Dest;         // destination point (WCS)
  CGePoint Scale;        // scale along 0-X, 0-Y axis before rotation
  double   Angle=0.0;    // rotation angle
  double   CosAng=0.0;   // cosine of rotation angle
  double   SinAng=0.0;   // sine of rotation angle
  INT_PTR  IntVar=0;
  HANDLE   hBlkRef=0;    // handle to inserted block
  double   DblVar=0.0;
  double   DblVar2=0.0;
  double   DblVar3=0.0;
  double   DblVar4=0.0;
  const GE_TRANSPRM* pNext = NULL;    // pointer to next transform parameters or NULL if end
};





//=============================================================================
//
//  CGeLine_
//
//=============================================================================
class CGeLine {
public:
  CGePoint m_Start;
  CGePoint m_End;
public:
    CGeLine ();
    CGeLine (const CGeLine& line);
    CGeLine (const CGePoint& pnt1, const CGePoint& pnt2);
    CGeLine (double x0, double y0, double x1, double y1);

  CGePoint Start        () const;
  CGePoint End          () const;
  double   StartX       () const;
  double   StartY       () const;
  double   EndX         () const;
  double   EndY         () const;
  CGePoint Center       () const;
  double   Length       () const;
  double   Angle        () const;   // Angle from Start to End (0..2PI)
  double   Angle        (const CGeLine& Line2) const;
  double   dx           () const;
  double   dy           () const;
  CGePoint MinExtents   () const;
  CGePoint MaxExtents   () const;
  CGeRect  GeExtents    () const;
  void     GetGeExtents (CGeRect* pRect) const;
  bool     IsVertical   () const;
  bool     IsHorizontal () const;
  bool     IsDot        () const;
  bool     IsNotDot     () const;
  bool     IsDot        (double Delta) const;
  void     Save    (BYTE* pBuf) const;
  void     Restore (const BYTE* pBuf);
  
  const CGeLine&  Line  () const;

  EnPntPos   ClassifyPoint   (const CGePoint& Pnt) const;
  EnPntPos   ClassifyPoint2  (const CGePoint& Pnt, double Delta) const;
  EnIntersec ClassifyLine    (const CGeLine& Line2, double* pT) const;
  EnIntersec Intersect       (const CGeLine& Line2, double* pT) const;
  EnIntersec GetIntersection (const CGeLine& Line2, CGePoint* pInterPt, CGePoint* pNextPt=NULL, double* pRightAngle=NULL, double Zero1=0., double One1=1., double Zero2=0., double One2=1.) const;
//  EnIntersec GetIntersection (const CGeXline& Xline, CGePoint* pInterPt, double Zero=0., double One=1.) const;
  int        GetIntersection (const CGeCircle& Circ, CGePoint* pInterPt) const;
  int        GetIntersection (const CGeArc& Arc, CGePoint* pInterPt) const;
  bool       GetInterPoint   (const CGeLine& Line2, CGePoint* pInterPt, UINT_PTR Apparent=0) const;
  bool       IsIntersect     (const CGeLine& Line2) const;
  bool       IsIntersect     (const CGeLine& Line1, const CGeLine& Line2, const CGeLine& Line3, const CGeLine& Line4) const;
  bool       IsEqual         (const CGeLine& Line2, double Delta=g_GeTolerance) const;
  bool       IsEqualBack     (const CGeLine& Line2, double Delta=g_GeTolerance) const;
  UINT_PTR   GetConnection   (const CGeLine& Line2, double Delta=g_GeTolerance) const;
  bool       Crosses         (const CGeRect& Rect) const;
  void       GetParamAB      (double* A, double* B) const;
  double     DistToPoint     (const CGePoint& Pnt) const;
  double     GetNearPoint    (const CGePoint& Pnt, CGePoint* pPntNear, double* pDistFromStart) const;
  void       GetPerpendPoint (const CGePoint& Pnt, CGePoint* pOutPnt, double* pAlongDist, double* pPerpDist) const;
  double     GetAlongPointA  (double Dist, CGePoint* pPnt) const;
  void       GetAlongPointR  (double DistCoef, CGePoint* pPnt) const;
  void       GetPline        (double Delta, CGePolyline* pOutPline, bool bClosed);
  UINT_PTR   Divide (UINT_PTR nParts, CGePoint* pOutPnt, bool bStartEnd) const;
  bool       SplitByPoint    (const CGePoint& Pnt, double MinLen, CGeLine* pLine1, CGeLine* pLine2) const;

  // modify geometry by properties
  void  Set        (double x1, double y1, double x2, double y2);
  void  Set        (const CGeLine& line);
  void  Set        (const CGePoint& Pt0, const CGePoint& Pt1);
  void  Set        (const CGePoint* Pnts);
  void  Set        (const CGePoint& Pt0, double Ang, double Len1, double Len2);
  void  SetStart   (double x, double y);
  void  SetStart   (const CGePoint& Pnt);
  void  SetStartX  (double x);
  void  SetStartY  (double y);
  void  SetEnd     (double x, double y);
  void  SetEnd     (const CGePoint& Pnt);
  void  SetEnd     (const CGePoint* Pnt, UINT_PTR nPnts, UINT_PTR* piSel);
  void  SetEndPolar(double Angle, double Dist);
  void  SetEndX    (double x);
  void  SetEndY    (double y);
  void  SetLength  (double Len);
  void  SetAngle   (double Ang);

  // modify geometry by operations
  void  Move       (double dx, double dy);
  void  Move       (const CGePoint& Off);
  void  Offset     (double Dist);
  void  Rotate     (const CGePoint& Pt0, double Angle);
  void  Rotate     (UINT_PTR iVer, double Angle);
  void  Scale      (const CGePoint& Pt0, double ScaleX, double ScaleY);
  void  Scale      (const CGePoint& Pt0, double Scal);
  void  Mirror     (const CGeLine& Line);
  void  CutByX     (double x, int Mode);  // 0-start, 1-end
  void  Transform  (const GE_TRANSPRM* pPrm);
  void  Inflate    (double Len);
  void  Inflate    (double Len0, double Len1);
  bool  Compensate (double Len0, double Len1, double MinLen);
  void  InflateR   (double Coef);
  void  InflateR   (double Coef0, double Coef1);
  void  Flip       ();
  void  Reverse    ();
  void  SwapPoints ();
  void  Rot90ccw   ();
  void  Rot90cw    ();
  void  ShiftToStart ();
  void  ShiftToEnt   ();
  void  AdjustFillet (const CGePoint& Pt0, const CGePoint& Pt1);
  void  AdjustPoint  (const CGePoint& Pt);
};

//-----------------------------------------------
inline CGeLine::CGeLine () : m_Start(0.0,0.0), m_End(1.0,0.0) {}
inline CGeLine::CGeLine (const CGeLine& line) {Set(line);}
inline CGeLine::CGeLine (const CGePoint& Pt0, const CGePoint& Pt1) : m_Start(Pt0), m_End(Pt1) {}
inline CGeLine::CGeLine (double x0, double y0, double x1, double y1) : m_Start(x0,y0), m_End(x1,y1) {}
inline void CGeLine::Set (double x0, double y0, double x1, double y1) {
  m_Start.Set( x0, y0 );
  m_End.Set( x1, y1 );
}
inline void CGeLine::Set (const CGeLine& line) {
  m_Start = line.m_Start; 
  m_End = line.m_End;
}
inline void CGeLine::Set (const CGePoint& Pt0, const CGePoint& Pt1) {
  m_Start = Pt0; 
  m_End = Pt1;
}
inline void CGeLine::Set (const CGePoint* Pnts) {
  m_Start = *Pnts; 
  m_End = *(Pnts+1);
}
inline void     CGeLine::SetStart  (double x, double y) {m_Start.Set(x,y);}
inline void     CGeLine::SetStartX (double x) {m_Start.x=x;}
inline void     CGeLine::SetStartY (double y) {m_Start.y=y;}
inline void     CGeLine::SetStart  (const CGePoint& Pt) {m_Start=Pt;}
inline void     CGeLine::SetEnd    (double x, double y) {m_End.Set(x,y);}
inline void     CGeLine::SetEndX   (double x) {m_End.x=x;}
inline void     CGeLine::SetEndY   (double y) {m_End.y=y;}
inline void     CGeLine::SetEnd    (const CGePoint& Pt) {m_End=Pt;}
inline CGePoint CGeLine::Center    () const {return m_Start.Center( m_End );}
inline double   CGeLine::dx        () const {return (m_End.x - m_Start.x);}
inline double   CGeLine::dy        () const {return (m_End.y - m_Start.y);}
inline double   CGeLine::Length    () const {return m_Start.Distance( m_End );}
inline CGePoint CGeLine::Start     () const {return m_Start;}
inline CGePoint CGeLine::End       () const {return m_End;}
inline double   CGeLine::StartX    () const {return m_Start.x;}
inline double   CGeLine::StartY    () const {return m_Start.y;}
inline double   CGeLine::EndX      () const {return m_End.x;}
inline double   CGeLine::EndY      () const {return m_End.y;}
inline CGePoint CGeLine::MinExtents () const {
  return CGePoint( __min( m_Start.x, m_End.x ), __min( m_Start.y, m_End.y ) );
}
inline CGePoint CGeLine::MaxExtents () const {
  return CGePoint( __max( m_Start.x, m_End.x ), __max( m_Start.y, m_End.y ) );
}
// Angle from pt0 to pt1 (+/- PI)
inline double CGeLine::Angle () const {return m_Start.Angle( m_End );}
inline const  CGeLine& CGeLine::Line () const {return *this;}
inline bool   CGeLine::IsVertical () const {return (m_Start.x==m_End.x)?true:false;}
inline bool   CGeLine::IsHorizontal () const {return (m_Start.y==m_End.y)?true:false;}
inline bool   CGeLine::IsDot () const {return (m_Start.x==m_End.x && m_Start.y==m_End.y)?true:false;}
inline bool   CGeLine::IsNotDot () const {return (m_Start.x==m_End.x && m_Start.y==m_End.y)?false:true;}
inline void   CGeLine::Move (const CGePoint& Off) {Move(Off.x,Off.y);}
inline void   CGeLine::SwapPoints () {Flip();}
inline void   CGeLine::Reverse    () {Flip();}
inline void   CGeLine::Inflate (double Len) {Inflate(Len,Len);}
inline void   CGeLine::InflateR (double Coef) {InflateR(Coef,Coef);}
inline void   CGeLine::ShiftToStart () {m_Start=m_End;}
inline void   CGeLine::ShiftToEnt   () {m_End=m_Start;}


//=============================================================================
//
//  CGeRect_
//
//=============================================================================
class CKusok {  // used to clip polygon
public:
  CGePoint* m_Point;
  UINT_PTR  m_nPoints;
  CGePoint  m_StartPt;
  CGePoint  m_EndPt;
  CGePoint  m_CornerPt[4];  // from m_EndPt and around rect
  bool      m_bDeleted;
  bool      m_bSelected;
public:
    CKusok ();
    ~CKusok ();
  void   Set (const CGePoint* Pnt, UINT_PTR nPnts, const CGeRect& Rect, bool bCW);
  double GetDist (const CGePoint& Pt1, CGePoint* CornPt, UINT_PTR* pnCorns) const;
};
//=============================================================================
class CGeRect {
  double m_Left;
  double m_Bottom;
  double m_Right;
  double m_Top;
public:
    CGeRect ();
    CGeRect (double Lef, double Bot, double Rig, double Top);
    CGeRect (const CGeRect& rect);
    CGeRect (const CGePoint& pnt1, const CGePoint& pnt2);
    CGeRect (const CGePoint& Center, double Width, double Height);
    CGeRect (const CGePoint& Center, double Delta);

  void Set (double Lef, double Bot, double Rig, double Top);
  void Set (const CGePoint& pnt1, const CGePoint& pnt2);
  void Set (const CGePoint& Center, double Width, double Height);
  void Set (const CGePoint& Center, double Delta);
  bool SetWidth  (double w, bool bCentered);
  bool SetHeight (double h, bool bCentered);
  void SetCenter (const CGePoint& Center);
  void SetSize   (double Width, double Height);
  bool SetLeft   (double X, bool bConstWidth=true);
  bool SetRight  (double X, bool bConstWidth=true);
  bool SetBottom (double Y, bool bConstHeight=true);
  bool SetTop    (double Y, bool bConstHeight=true);
  void SetNull   ();
  void SetExtents (const CGePoint& Pt0, const CGePoint& Pt1, const CGePoint& Pt2);
  void SetExtents (const CGe3dPoint& Pt0, const CGe3dPoint& Pt1, const CGe3dPoint& Pt2);
  void UpdateExtentsBy (double X, double Y);
  void UpdateExtentsBy (const CGePoint& pt);
  void UpdateExtentsBy (const CGeRect& rect);

  double   Width  () const;
  double   Height () const;
  double   Ratio  () const;  // W/H
  double   Diagonal () const;
  double   Area   () const;
  double   Length () const;
  CGePoint Center () const;
  CGePoint LefBot () const;
  CGePoint LefTop () const;
  CGePoint RigTop () const;
  CGePoint RigBot () const;
  CGePoint LefMid () const;
  CGePoint RigMid () const;
  CGePoint MidBot () const;
  CGePoint MidTop () const;
  double   Left   () const;
  double   Bottom () const;
  double   Right  () const;
  double   Top    () const;
  void     Get    (double* pLef, double* pBot, double* pRig, double* pTop) const;
  void     GetPoints   (CGePoint* pPnt) const;
  void     Get8Points  (CGePoint* pPnt) const;
  void     GetPolyline (CGePolyline* pPline) const;
  bool     GetAndRect (const CGeRect& Rect2, CGeRect* pOutRect) const;
  void     GetExtRect (const GE_TRANSPRM* pPrm, CGeRect* pOutRect) const;

  bool  Contains  (const CGePoint& Pnt) const;
  bool  Contains  (const CGePoint& Pnt, double Delta) const;
  bool  NotContains (const CGePoint& Pnt) const;
  bool  NotContains (const CGePoint& Pnt, double Delta) const;
  bool  Contains  (double X, double Y) const;
  bool  Contains2 (const CGePoint& Pnt) const;
  bool  Contains2 (double X, double Y) const;
  bool  Contains  (const CGeLine& Line) const;
  bool  Contains  (const CGeRect& Rect) const;
  bool  Crosses   (const CGePoint& Pnt) const;
  bool  Crosses   (const CGeLine& Line) const;
  bool  Crosses   (const CGeCircle& Circ, bool bPoint) const;
  bool  Crosses   (const CGeArc& Arc, bool bPoint) const;
  bool  Crosses   (const CGePolyline& Pline) const;
  bool  Crosses   (const CGeRect& Rect) const;
  bool  IsNull    () const;
  bool  IsNotNull () const;
  bool  IsPoint   () const;
  bool  IsEqual   (const CGeRect& Rect) const;

  void    Scale   (double Coef, bool bByCenter);
  void    Inflate (double Delta);
  void    InflateRot (double Angle);
  void    Move    (double dx, double dy);
  void    Move    (const CGePoint& Off);
  CGeRect MoveBy  (double dx, double dy) const;
  CGeRect MoveBy  (const CGePoint& Off) const;
  CGeRect ScaleBy (double Coef, bool bByCenter) const;
  void    MirrorX (double Xe);

  INT_PTR   ClipLine     (const CGePoint& Start, const CGePoint& End, CGePoint* pOutPt1, CGePoint* pOutPt2) const;
  INT_PTR   ClipLine     (const CGeLine& Line, CGeLine* pOutLine) const;
  INT_PTR   ClipPolyline (const CGePolyline& Pline, bool bForceClosed=false) const;
  INT_PTR   ClipPolygon  (const CGePolyline& Polygon) const;
  CGePoint* GetClippedPolygon  (bool bFirst, UINT_PTR* pNumPoints) const;
  CGePoint* GetClippedPolyline (bool bFirst, UINT_PTR* pNumPoints) const;

  const CGeRect& Rect () const;
  bool  operator == (const CGeRect& rect2) const;

private:
  void   Normalize ();
  CKusok* Clip_GetFirstKusok  (CKusok* Kusok, UINT_PTR nKus) const;
  CKusok* Clip_GetNextKusok   (CKusok* Kusok, UINT_PTR nKus, CKusok* pKusok) const;
  void    Clip_DeleteSelKusok (CKusok* Kusok, UINT_PTR nKus) const;
};


//-----------------------------------------------
inline const CGeRect&  CGeRect::Rect () const {return *this;}
inline CGeRect::CGeRect () {m_Left=m_Bottom=m_Right=m_Top=0.0;}
inline CGeRect::CGeRect (double lef, double bot, double rig, double top) {Set(lef,bot,rig,top);}
inline CGeRect::CGeRect (const CGeRect& rect) {m_Left=rect.m_Left; m_Bottom=rect.m_Bottom; m_Right=rect.m_Right; m_Top=rect.m_Top;}
inline CGeRect::CGeRect (const CGePoint& Center, double Wid, double Hei) {Set(Center,Wid,Hei);}
inline CGeRect::CGeRect (const CGePoint& Center, double Delta) {Set(Center,Delta);}
inline CGeRect::CGeRect (const CGePoint& pnt11, const CGePoint& pnt22) {Set(pnt11,pnt22);}
inline void CGeRect::Set (double lef, double bot, double rig, double top){
  if (lef < rig){
    m_Left = lef;
    m_Right = rig;
  }else{
    m_Left = rig;
    m_Right = lef;
  }
  if (bot < top){
    m_Bottom = bot;
    m_Top = top;
  }else{
    m_Bottom = top;
    m_Top = bot;
  }
}
inline void CGeRect::Set (const CGePoint& pt1, const CGePoint& pt2) {
  m_Left = pt1.x;  m_Bottom = pt1.y;
  m_Right = pt2.x; m_Top = pt2.y;
  Normalize();
}
inline void CGeRect::SetNull   () {Set(0.,0.,0.,0.);}
inline void CGeRect::UpdateExtentsBy (double X, double Y){
  if (m_Left  > X) m_Left  =X;
  if (m_Right < X) m_Right =X;
  if (m_Bottom> Y) m_Bottom=Y;
  if (m_Top   < Y) m_Top   =Y;
}
inline void CGeRect::UpdateExtentsBy (const CGePoint& pt) {
  if (m_Left  > pt.x) m_Left  =pt.x;
  if (m_Right < pt.x) m_Right =pt.x;
  if (m_Bottom> pt.y) m_Bottom=pt.y;
  if (m_Top   < pt.y) m_Top   =pt.y;
}
inline void CGeRect::UpdateExtentsBy (const CGeRect& rect) {
  if (m_Left  > rect.m_Left)   m_Left  =rect.m_Left;
  if (m_Right < rect.m_Right)  m_Right =rect.m_Right;
  if (m_Bottom> rect.m_Bottom) m_Bottom=rect.m_Bottom;
  if (m_Top   < rect.m_Top)    m_Top   =rect.m_Top;
}
inline double   CGeRect::Width  () const {return m_Right - m_Left;}
inline double   CGeRect::Height () const {return m_Top - m_Bottom;}
inline double   CGeRect::Area   () const {return Width()*Height();}
inline double   CGeRect::Length () const {return (Width()+Height())*2.0;}
inline CGePoint CGeRect::Center () const {return CGePoint( (m_Right+m_Left)/2.0, (m_Top+m_Bottom)/2.0 );}
inline CGePoint CGeRect::LefBot () const {return CGePoint( m_Left, m_Bottom );}
inline CGePoint CGeRect::LefTop () const {return CGePoint( m_Left, m_Top );}
inline CGePoint CGeRect::RigTop () const {return CGePoint( m_Right, m_Top );}
inline CGePoint CGeRect::RigBot () const {return CGePoint( m_Right, m_Bottom );}
inline CGePoint CGeRect::LefMid () const {return CGePoint( m_Left, (m_Top+m_Bottom)/2.0 );}
inline CGePoint CGeRect::RigMid () const {return CGePoint( m_Right, (m_Top+m_Bottom)/2.0 );}
inline CGePoint CGeRect::MidBot () const {return CGePoint( (m_Right+m_Left)/2.0, m_Bottom );}
inline CGePoint CGeRect::MidTop () const {return CGePoint( (m_Right+m_Left)/2.0, m_Top );}
inline double   CGeRect::Left   () const {return m_Left;}
inline double   CGeRect::Bottom () const {return m_Bottom;}
inline double   CGeRect::Right  () const {return m_Right;}
inline double   CGeRect::Top    () const {return m_Top;}
inline bool CGeRect::Contains (const CGePoint& Pnt) const {   // false - outside or on-border
  return ((m_Left < Pnt.x) && (Pnt.x < m_Right) && (m_Bottom < Pnt.y) && (Pnt.y < m_Top))? true : false;
}
inline bool CGeRect::NotContains (const CGePoint& Pnt) const {   // true - outside
  return ((m_Left < Pnt.x) && (Pnt.x < m_Right) && (m_Bottom < Pnt.y) && (Pnt.y < m_Top))? false : true;
}
inline bool CGeRect::Contains (double X, double Y) const {   // false - outside or on-border
  return ((m_Left < X) && (X < m_Right) && (m_Bottom < Y) && (Y < m_Top))? true : false;
}
inline bool CGeRect::Contains2 (const CGePoint& Pnt) const {   // false - outside
  return ((m_Left <= Pnt.x) && (Pnt.x <= m_Right) && (m_Bottom <= Pnt.y) && (Pnt.y <= m_Top))? true : false;
}
inline bool CGeRect::Contains2 (double X, double Y) const {   // false - outside
  return ((m_Left <= X) && (X <= m_Right) && (m_Bottom <= Y) && (Y <= m_Top))? true : false;
}
inline bool CGeRect::Contains (const CGeRect& r2) const {
  return (m_Left <= r2.m_Left && r2.m_Right <= m_Right && m_Bottom <= r2.m_Bottom && r2.m_Top <= m_Top)? true : false;
}
inline bool CGeRect::Crosses (const CGePoint& Pnt) const {
  return (m_Left == Pnt.x || Pnt.x == m_Right || m_Bottom == Pnt.y || Pnt.y == m_Top)? true : false;
}
inline bool CGeRect::Crosses  (const CGeRect& r2) const {
  return (r2.m_Right >= m_Left && r2.m_Left <= m_Right && r2.m_Bottom <= m_Top && r2.m_Top >= m_Bottom)? true : false;
}
inline bool CGeRect::IsNull () const {
  return (m_Left==0.0 && m_Bottom==0.0 && m_Right==0.0 && m_Top==0.0)? true : false;
}
inline bool CGeRect::IsNotNull () const {
  return (m_Left==0.0 && m_Bottom==0.0 && m_Right==0.0 && m_Top==0.0)? false: true;
}
inline bool CGeRect::IsPoint () const {return (m_Left==m_Right && m_Bottom==m_Top)?true:false;}
inline bool CGeRect::IsEqual (const CGeRect& R2) const {
  return (m_Left==R2.m_Left && m_Right==R2.m_Right && m_Bottom==R2.m_Bottom && m_Top==R2.m_Top)? true : false;
}
inline void CGeRect::Inflate (double Delta) {m_Left-=Delta; m_Bottom-=Delta; m_Right+=Delta; m_Top+=Delta;}
inline void CGeRect::Move (double dx, double dy) {m_Left+=dx; m_Bottom+=dy; m_Right+=dx; m_Top+=dy;}
inline void CGeRect::Move (const CGePoint& Off) {m_Left+=Off.x; m_Bottom+=Off.y; m_Right+=Off.x; m_Top+=Off.y;}
inline CGeRect CGeRect::MoveBy (const CGePoint& Off) const {return MoveBy(Off.x,Off.y);}
inline bool CGeRect::operator == (const CGeRect& r2) const {
  return (r2.m_Right==m_Right && r2.m_Left==m_Left && r2.m_Bottom==m_Bottom && r2.m_Top==m_Top)?true:false;
}
inline void CGeRect::Get (double* pLef, double* pBot, double* pRig, double* pTop) const {
  *pLef=m_Left; *pBot=m_Bottom; *pRig=m_Right; *pTop=m_Top;
}


//=============================================================================
//
//  CGePolyline_
//
//=============================================================================
class CGeVertex : public COkObject {
public:
  CGePoint m_Pos;     // position
  int      m_iParam;
  double   m_fParam;
  double   m_fParam2;
  double   m_Dist;  // distance from beginning of polyline
public:
    CGeVertex ();
    CGeVertex (const CGeVertex* pVerSrc);
    CGeVertex (const CGePoint& Pnt, int iPrm=0, double fPrm=0.0, double fPrm2=0.0);
    CGeVertex (double X, double Y, double Z=0.0);
  double X () const {return m_Pos.x;}
  double Y () const {return m_Pos.y;}
  double Z () const {return m_fParam;}
  void   SetZ (double Z) {m_fParam=Z;}
  const CGeVertex* PrevVer() const {return (CGeVertex*)Prev();}
  const CGeVertex* NextVer() const {return (CGeVertex*)Next();}
};
//-----------------------------------------------
class CGePolyline {
  COkList  m_Vers;       // list of vertices (CGeVertex objects)
  UINT_PTR m_NumVers;    // number of polyline points
  bool     m_bClosed;
public:
  double   m_Zval;       // used for isolines (COkTIN::GenIsolines)
  bool     m_bBold;      // used to mark bold isoline
  static bool m_bNoDuplicate;  // don't add duplicate point
public:
    CGePolyline ();
    CGePolyline (const CGePoint* Pnts, UINT_PTR nPnts, bool bClosed);
    ~CGePolyline ();

  void       Clear       ();
  CGeVertex* AddVer      (const CGeVertex* pVerSrc);
  CGeVertex* AddVer      (const CGePoint& Pnt, int iPrm=0, double fPrm=0.0, double fPrm2=0.0);
  CGeVertex* AddVer      (double X, double Y, double Z=0.0);
  CGeVertex* AddVer0     ();
  CGeVertex* InsertVer   (CGeVertex* pBeforeVer, const CGePoint& Pnt, int iPrm=0, double fPrm=0.0, double fPrm2=0.0);
  CGeVertex* InsertVer   (CGeVertex* pBeforeVer, double X, double Y, double Z=0.0);
  CGeVertex* InsertVer   (double Dist);
  bool       SetStartVer (CGeVertex* pVer);
  void       SetVers     (const CGePoint* Pnt, UINT_PTR nPnts, bool bClosed=false, bool bBackOrder=false);
  void       SetPoints   (const CGePoint* Pnt, UINT_PTR nPnts, bool bClosed=false, bool bBackOrder=false);
  void       SetRect     (const CGePoint& Pt1, const CGePoint& Pt2, double Width);
  void       CopyFrom    (const CGePolyline& Pline);
  UINT_PTR   AddFrom     (const CGePolyline& Pline);
  UINT_PTR   AddFrom     (const CGePolyline& Pline, int Param);
  void       SetClosed   (bool bClosed=true);
  void       Unclose     ();
  bool       TryClose    ();
  CGeVertex* DelFirstVer ();
  CGeVertex* DelLastVer  ();
  UINT_PTR   DelVersID   (COkID Id);

  UINT_PTR   CountVers   ();
  UINT_PTR   CountPoints ();
  UINT_PTR   NumVers     () const;
  UINT_PTR   NumPoints   () const;
  CGeVertex* GetFirstVer () const;
  CGeVertex* GetNextVer  (const CGeVertex* p) const;
  CGeVertex* GetLastVer  () const;
  CGeVertex* GetPrevVer  (const CGeVertex* p) const;
  CGeVertex* GetVerByIndex (UINT_PTR iVer) const;
  CGeVertex* GetVerByPos   (const CGePoint& PickPt, double Delta) const;
  UINT_PTR   GetPoints     (CGePoint* Pnts, UINT_PTR MaxPnts) const;
  bool       GetFirstPoint (CGePoint* pPnt, double* pDirAngle=NULL) const;
  bool       GetLastPoint  (CGePoint* pPnt, double* pDirAngle=NULL, bool bMeanClosed=true) const;
  bool       GetMidPoint   (CGePoint* pPnt, double* pDirAngle=NULL) const;

  bool GetPointAt (UINT_PTR iPnt, CGePoint* pPnt) const;

  CGeRect GeExtents    () const;
  void    GetGeExtents (CGeRect* pRect) const;
  void    GetXmin      (double* pX) const;
  void    GetXmax      (double* pX) const;
  void    GetCenter    (CGePoint* pPnt) const;
  double  Length       (bool bForceClosed=false) const;
  double  Area         (bool bAbs=true) const;
  bool    IsEmpty      () const;
  bool    IsCW         () const;
  bool    IsCCW        () const;
  bool    IsEqual      (const CGePolyline& Pline, double Delta=g_GeTolerance) const;
  bool    IsClosed     () const;
  bool    IsNotClosed  () const;
  INT_PTR Contains     (const CGePoint& Pnt) const;
  bool    Contains     (const CGePolyline& Pline) const;
  bool    Contains     (const CGeRect& Rect) const;
  bool    Crosses      (const CGeRect& Rect) const;
  bool    Crosses      (const CGeLine& Line) const;

  void    Move         (double dx, double dy);
  void    Move         (const CGePoint& Off);
  void    Polar        (double Angle, double Dist);
  void    Rotate       (const CGePoint& BasePt, double Angle);
  void    RotateMove   (const CGePoint& BasePt, double Angle, const CGePoint& Off);
  void    Scale        (const CGePoint& BasePt, double ScaleX, double ScaleY);
  void    Mirror       (double a1, double b1);
  void    Mirror       (const CGeLine& Line);
  void    Transform    (const GE_TRANSPRM* pPrm, bool bExcludeNegPrm=false);
  void    Transform    (const CGe3dPoint& Pnt0, const CGe3dPoint& NormVec);
  void    Reverse      ();
  bool    TrimStart    (double Dist, const CGePoint& Pnt);
  bool    TrimEnd      (double Dist, const CGePoint& Pnt);
  bool    TrimMiddle   (double Dist1, const CGePoint& Pnt1, double Dist2, const CGePoint& Pnt2, CGePolyline* pPline2);
  bool    TrimPolyline (double Dist1, double Dist2, CGePolyline* pOutPline);
  bool    Offset       (double Delta);
  UINT_PTR MakeFreeEnds ();

  bool MakeBSpline2  (CGePolyline* pOutPline, bool bClosed, int Quality) const;
  bool MakeBSpline3  (CGePolyline* pOutPline, bool bClosed, int Quality) const;
  bool MakeFitSpline (CGePolyline* pOutPline, bool bClosed, int Quality) const;
  bool MakeRoundVers (CGePolyline* pOutPline, bool bClosed, int Quality);
  bool MakeBulges    (CGePolyline* pOutPline, bool bClosed, int Quality);
  bool MakeLinQuad   (CGePolyline* pOutPline, bool bClosed, int Quality);
  bool MakeBezier    (CGePolyline* pOutPline, bool bClosed, int Quality);

  double   DistToPoint (const CGePoint& Pnt) const;

  bool     GetDistPoint    (double PtDist, CGePoint* pPnt, double* pDirAngle) const;
  double   GetNearPoint    (const CGePoint& Pnt, CGePoint* pPntNear, double* pDistFromStart) const;
  bool     GetPerpPoint    (const CGePoint& BasePnt, const CGePoint& Pnt, double Delta, CGePoint* pPntPerp) const;
  bool     GetOpposPoint   (double Dist, double MaxLen, CGePoint* pPnt, CGePoint* pBasePt=NULL) const;
  bool     GetOffsetPline  (double OffDist, CGePolyline* pPline2) const;
  bool     GetArcPline     (CGePolyline* pPline, double Delta) const;
  bool     GetY            (double X, double* pY) const;

  UINT_PTR GetCoarsePline  (CGePolyline* pPline, double Delta) const;
  bool     CheckDeviation  (CGeVertex* pVer1, CGeVertex* pVer2, double Delta) const;
  UINT_PTR Divide          (UINT_PTR nParts, CGePoint* pOutPnt, bool bStartEnd) const;

  UINT_PTR GetIntersection (const CGeLine& Line, CGePoint* OutPoint, double* OutDist, UINT_PTR MaxPoints) const;
  UINT_PTR GetIntersection (const CGePolyline& Pline, CGePoint* OutPoint, double* OutDist, UINT_PTR MaxPoints) const;
  UINT_PTR GetIntersection (const CGeCircle& Circ, CGePoint* OutPoint, double* OutDist, UINT_PTR MaxPoints) const;
  UINT_PTR GetIntersection (const CGeArc& Arc, CGePoint* OutPoint, double* OutDist, UINT_PTR MaxPoints) const;

  void UpdateDist ();
};

//-----------------------------------------------
inline UINT_PTR   CGePolyline::CountVers   () {return m_NumVers=m_Vers.NumRecs();}
inline UINT_PTR   CGePolyline::CountPoints () {return m_NumVers=m_Vers.NumRecs();}
inline UINT_PTR   CGePolyline::NumVers     () const {return m_NumVers;}
inline UINT_PTR   CGePolyline::NumPoints   () const {return m_NumVers;}
inline CGeVertex* CGePolyline::GetFirstVer () const {return (CGeVertex*)m_Vers.GetFirst();}
inline CGeVertex* CGePolyline::GetNextVer  (const CGeVertex* p) const {return (CGeVertex*)m_Vers.GetNext(p);}
inline CGeVertex* CGePolyline::GetLastVer  () const {return (CGeVertex*)m_Vers.GetLast();}
inline CGeVertex* CGePolyline::GetPrevVer  (const CGeVertex* p) const {return (CGeVertex*)m_Vers.GetPrev(p);}
inline bool       CGePolyline::IsCCW       () const {return IsCW()?false:true;}
inline void       CGePolyline::SetClosed   (bool b) {m_bClosed=b;}
inline bool       CGePolyline::IsClosed    () const {return m_bClosed;}
inline bool       CGePolyline::IsNotClosed () const {return !m_bClosed;}
inline void       CGePolyline::Move        (const CGePoint& Off) {Move(Off.x,Off.y);}
inline void       CGePolyline::SetPoints   (const CGePoint* Pnt, UINT_PTR nPnts, bool bClosed, bool bBackOrder) {SetVers(Pnt,nPnts,bClosed,bBackOrder);}

//=============================================================================
//
//  CGeCircle_
//
//=============================================================================
class CGeCircle {
  double   m_Radius;
public:
  CGePoint m_Center;
public:
    CGeCircle ();
    CGeCircle (const CGeCircle& Circ);
    CGeCircle (const CGePoint& Center, double Radius);
    CGeCircle (double x, double y, double Radius);

  CGePoint MinExtents   () const;
  CGePoint MaxExtents   () const;
  CGeRect  GeExtents    () const;
  void     GetGeExtents (CGeRect* pRect) const;

  void     Set          (const CGeCircle& Circ);
  void     Set          (const CGePoint& Center, double Radius);
  void     Set          (double x, double y, double Radius);
  bool     Set3P        (const CGePoint& Pt1, const CGePoint& Pt2, const CGePoint& Pt3);
  bool     Set2P        (const CGePoint& Pt1, const CGePoint& Pt2);
  void     SetCenter    (const CGePoint& Pnt);
  void     SetCenter    (double x, double y);
  bool     SetRadius    (double Radius);
  bool     SetDiameter  (double Diameter);
  bool     SetLength    (double Len);
  bool     SetArea      (double Area);

  CGePoint Center       () const;
  double   Radius       () const;
  double   Diameter     () const;
  double   Length       () const;
  double   Area         () const;
  void     Save    (BYTE* pBuf) const;
  void     Restore (const BYTE* pBuf);

  const CGeCircle&  Circle  () const;

  bool     GetTangentPoints (const CGePoint& BasePt, CGePoint* pPnt1, CGePoint* pPnt2) const;
  bool     GetTangentPoints (const CGeCircle& Circ2, CGePoint* Pnt, int* pnPnts) const;
  void     GetPoint     (double Ang, CGePoint* pPnt) const;
  void     GetPoint     (double Ang, double Ang0, bool bClockwise, CGePoint* pPnt, double* pDirAng) const;
  double   GetNearPoint (const CGePoint& Pnt, CGePoint* pPntNear) const;

  void GetPoints   (CGePoint* Pnt, UINT_PTR NumPoints, double Ang0=0.0, bool bClockwise=true) const;
  void GetPolyline (CGePolyline* pPline, UINT_PTR NumPoints, double Ang0=0.0, bool bClockwise=true) const;

  bool Contains (const CGePoint& Pnt) const;
  bool Contains (const CGeRect& Rect) const;
  bool Crosses  (const CGeRect& Rect, bool bPoint) const;
  UINT_PTR Divide (double Ang0, bool bClockwise, UINT_PTR nParts, CGePoint* pOutPnt) const;

  void Move   (const CGePoint& Off);
  void Move   (double dx, double dy);
  void Rotate (const CGePoint& Pt0, double Angle);
  void Scale  (const CGePoint& Pt0, double ScaleX, double ScaleY);
  void Mirror (const CGeLine& Line);
  bool Transform  (const GE_TRANSPRM* pPrm);
};

//-------------------------------------------------------------------
// Inline functions
//-------------------------------------------------------------------
inline CGeCircle::CGeCircle () {m_Radius = 1.0;}
inline CGeCircle::CGeCircle (const CGeCircle& Circ) {Set(Circ);}
inline CGeCircle::CGeCircle (const CGePoint& Cen, double R) {Set(Cen,R);}
inline CGeCircle::CGeCircle (double x, double y, double R) {Set(x,y,R);}
inline const CGeCircle&  CGeCircle::Circle () const {return *this;}
inline CGePoint    CGeCircle::MinExtents  () const {return CGePoint( m_Center.x - m_Radius, m_Center.y - m_Radius );}
inline CGePoint    CGeCircle::MaxExtents  () const {return CGePoint( m_Center.x + m_Radius, m_Center.y + m_Radius );}
inline void        CGeCircle::Set         (const CGeCircle& Circ) {m_Center=Circ.m_Center; m_Radius=Circ.m_Radius;}
inline void        CGeCircle::Set         (const CGePoint& Pnt, double R) {m_Center=Pnt; SetRadius(R);}
inline void        CGeCircle::Set         (double x, double y, double R) {m_Center.Set(x,y); SetRadius(R);}
inline void        CGeCircle::SetCenter   (const CGePoint& Pnt) {m_Center = Pnt;}
inline void        CGeCircle::SetCenter   (double x, double y) {m_Center.Set(x,y);}
inline CGePoint    CGeCircle::Center      () const {return m_Center;}
inline double      CGeCircle::Radius      () const {return m_Radius;}
inline double      CGeCircle::Diameter    () const {return m_Radius+m_Radius;}
inline double      CGeCircle::Length      () const {return GE_2PI * m_Radius;}
inline double      CGeCircle::Area        () const {return GE_PI * m_Radius* m_Radius;}




//=============================================================================
//
//  CGeArc_
//
//=============================================================================
class CGeArc {
  double   m_Radius;
  double   m_AngStart;
  double   m_AngArc;     // positive - CCW arc direction, negative - ClockWise
public:
  CGePoint m_Center;
  bool     m_bSector;
public:
    CGeArc ();
    CGeArc (const CGeArc& arc);
    CGeArc (const CGePoint& Center, double Radius, double Ang0, double AngArc, bool bSector=false);
    CGeArc (double x, double y, double Radius, double Ang0, double AngArc, bool bSector=false);

  const CGeArc& Arc () const;

  CGeRect GeExtents    () const;
  void    GetGeExtents (CGeRect* pRect) const;

  void Set           (const CGeArc& arc);
  bool Set           (const CGePoint& Center, double Radius, double Ang0, double AngArc, bool bSector=false);
  bool Set3P         (const CGePoint& Start, const CGePoint& Mid, const CGePoint& End);
  bool Set3P         (const CGePoint* Pnt);
  bool SetSDE        (const CGePoint& Start, double DirAngle, const CGePoint& End);
  bool SetSDAR       (const CGePoint& Start, double DirAngle, double AngArc, double Radius);
  bool SetSBE        (const CGePoint& Start, double Bulge, const CGePoint& End);
  bool SetSER        (const CGePoint& Start, const CGePoint& End, double Radius, bool bClockwise, bool bLong=false);
  bool SetSEL        (const CGePoint& Start, const CGePoint& End, double ArcLen, bool bClockwise);
  bool SetSEA        (const CGePoint& Start, const CGePoint& End, double AngArc);
  bool SetSEB        (const CGePoint& Start, const CGePoint& End, double Bulge);
  bool SetSEH        (const CGePoint& Start, const CGePoint& End, double dH);
  bool SetSEC        (const CGePoint& Start, const CGePoint& End, const CGePoint& Center, bool bClockwise);
  bool SetCSE        (const CGePoint& Center, const CGePoint& Start, const CGePoint& End, bool bClockwise);
  bool SetCSA        (const CGePoint& Center, const CGePoint& Start, double AngArc);
  bool SetCSL        (const CGePoint& Center, const CGePoint& Start, double ChordLen, bool bClockwise);
  bool SetCRAA       (const CGePoint& Center, double R, double AngStart, double AngEnd, BOOL bClockwise, bool bSec=false);
  void SetCenter     (const CGePoint& Pnt);
  void SetRadius     (double Radius);
  void SetStartAngle (double Angle);
  bool SetArcAngle   (double Angle);
  void SetSector     (bool bSec=true);
  void IncreaseArcLength (double Delta);

  CGePoint Center       () const;
  CGePoint StartPoint   () const;
  CGePoint EndPoint     () const;
  CGePoint MiddlePoint  () const;
  CGePoint MidPoint     () const;
  double   Radius       () const;
  double   StartAngle   () const;
  double   ArcAngle     () const;
  double   EndAngle     () const;
  double   AngleStart   () const;
  double   AngleArc     () const;
  double   AngleEnd     () const;
  double   Bulge        () const;
  double   StartTangent () const;
  double   EndTangent   () const;
  double   Length       () const;
  double   Height       () const;
  double   ChordLen     () const;
  double   Area         () const;
  bool     IsSector     () const;
  bool     IsArcPoint   (const CGePoint& Pnt) const;
  bool     IsArcPoint2  (const CGePoint& Pnt, double dR) const;
  bool     IsCW         () const;
  bool     IsCCW        () const;
  void     Save    (BYTE* pBuf) const;
  void     Restore (const BYTE* pBuf);

  UINT_PTR GetPoints        (CGePoint* Pnt, double AngStep, UINT_PTR MaxNumPoints) const;
  UINT_PTR GetPointsA       (CGePoint* Pnt, UINT_PTR MaxNumPoints, bool bMin5p=false) const;
  UINT_PTR GetPolylineA     (CGePolyline* pPline, UINT_PTR MaxNumPoints, bool bMin5p=false) const;
  void     GetPointsOld     (CGePoint* Pnt, UINT_PTR NumPoints) const;

  void     GetPointByDist   (double Dist, CGePoint* pPnt, double* pAngle, double* pDirAngle=0) const;
  bool     GetPointByAngle  (double Angle, CGePoint* pPnt, double* pDist) const;
  double   GetDistByPoint   (const CGePoint& Pnt, CGePoint* pArcPt=NULL) const;
  double   GetNearPoint     (const CGePoint& Pnt, CGePoint* pPntNear, double* pDistFromStart) const;
  int      GetTangentPoints (const CGePoint& BasePt, CGePoint* Pnt) const;
  double   DistToPoint (const CGePoint& Pnt, bool* pInsideArcAngle) const;
  UINT_PTR Divide      (UINT_PTR nParts, CGePoint* pOutPnt, bool bStartEnd) const;

  bool Contains (const CGePoint& Pnt) const;
  bool Contains (const CGeRect& Rect) const;
  bool Crosses  (const CGeRect& Rect, bool bPoint) const;

  void Move    (const CGePoint& Off);
  void Move    (double dx, double dy);
  void Rotate  (const CGePoint& Pt0, double Angle);
  void Scale   (const CGePoint& Pt0, double ScaleX, double ScaleY);
  void Mirror  (const CGeLine& Line);
  void Reverse ();
  void Inverse ();
  bool Transform (const GE_TRANSPRM* pPrm);
};

//-------------------------------------------------------------------
// Inline functions
//-------------------------------------------------------------------
inline CGeArc::CGeArc () {m_AngStart=0.0; m_Radius=m_AngArc=1.0; m_bSector=false;}
inline CGeArc::CGeArc (const CGeArc& arc) {Set(arc);}
inline CGeArc::CGeArc (const CGePoint& Cen, double R, double Ang0, double AngArc, bool bSec) {Set(Cen,R,Ang0,AngArc,bSec);}
inline CGeArc::CGeArc (double x, double y, double R, double Ang0, double AngArc, bool bSec) {Set( CGePoint(x,y), R, Ang0, AngArc, bSec );}
inline const CGeArc& CGeArc::Arc () const {return *this;}
inline void CGeArc::Set (const CGeArc& arc) {
  m_Center   = arc.m_Center; 
  m_Radius   = arc.m_Radius; 
  m_AngStart = arc.m_AngStart; 
  m_AngArc   = arc.m_AngArc; 
  m_bSector  = arc.m_bSector;
}
inline void     CGeArc::SetCenter  (const CGePoint& Pnt) {m_Center = Pnt;}
inline void     CGeArc::SetRadius  (double R) {m_Radius = R;}
inline void     CGeArc::SetSector  (bool bSec) {m_bSector = bSec;}
inline bool     CGeArc::SetSEB     (const CGePoint& Start, const CGePoint& End, double Bulge) {return SetSBE(Start,Bulge,End);}
inline bool     CGeArc::Set3P      (const CGePoint* Pnt) {return Set3P(Pnt[0],Pnt[1],Pnt[2]);}
inline CGePoint CGeArc::Center     () const {return m_Center;}
inline double   CGeArc::Radius     () const {return m_Radius;}
inline double   CGeArc::StartAngle () const {return m_AngStart;}
inline double   CGeArc::AngleStart () const {return m_AngStart;}
inline double   CGeArc::ArcAngle   () const {return m_AngArc;}
inline double   CGeArc::AngleArc   () const {return m_AngArc;}
inline double   CGeArc::AngleEnd   () const {return EndAngle();}
inline CGePoint CGeArc::StartPoint () const {return m_Center.PolarBy( m_AngStart, m_Radius );}
inline CGePoint CGeArc::EndPoint   () const {return m_Center.PolarBy( m_AngStart + m_AngArc, m_Radius );}
inline CGePoint CGeArc::MiddlePoint() const {return m_Center.PolarBy( m_AngStart + m_AngArc/2.0, m_Radius );}
inline CGePoint CGeArc::MidPoint   () const {return m_Center.PolarBy( m_AngStart + m_AngArc/2.0, m_Radius );}
inline bool     CGeArc::IsSector   () const {return m_bSector;}
inline bool     CGeArc::IsCW       () const {return (m_AngArc<0.0)?true:false;}
inline bool     CGeArc::IsCCW      () const {return (m_AngArc>=0.0)?true:false;}



//=============================================================================
//
//  CGeEllipse_
//
//=============================================================================
class CGeEllipse {
  double   m_R1;           // radius along rotation angle
  double   m_R2;           // radius perpendicular rotation angle (if m_bNormalized then R2<=R1)
  double   m_Rotation;     // rotation angle, radians
  double   m_AngStart;     // start angle, relative to major axis (m_R1)
  double   m_AngArc;       // arc included angle, counter-clockwise,  negative - clockwise
public:
  CGePoint m_Center;
  bool     m_bSector;      // ignored for full ellipse
  bool     m_bNormalized;  // if true, then always R2<=R1
public:
    CGeEllipse ();
    CGeEllipse (const CGeEllipse& Ell);
    CGeEllipse (const CGePoint& Center, double R1, double R2, 
                double RotAngle=0.0, double Ang0=0.0, double AngArc=0.0, bool bSector=false);
    CGeEllipse (double x, double y, double R1, double R2, 
                double RotAngle=0.0, double Ang0=0.0, double AngArc=0.0, bool bSector=false );

  const CGeEllipse& Ellipse () const;

  CGeRect GeExtents    () const;
  void    GetGeExtents (CGeRect* pRect) const;

  void Set  (const CGeEllipse& Ell);
  bool Set  (const CGePoint& Center, double R1, double R2, 
             double RotAngle=0.0, double Ang0=0.0, double AngArc=0.0, bool bSector=false);
  bool Set2 (const CGePoint& Center, double R1, double R2, 
             double RotAngle, double Ang0, double AngArc, bool bSector=false);
  void SetCenter     (const CGePoint& Pnt);
  bool SetRadius12   (double R1, double R2);
  bool SetRadius1    (double Radius);
  bool SetRadius2    (double Radius);
  bool SetRatio      (double Ratio);
  void SetRotation   (double Angle);
  void SetStartAngle (double Angle);
  void SetArcAngle   (double Angle);
  void SetParams     (double StartPrm, double EndPrm);  // set start & arc angles by parameters (from AutoCAD)
  void SetSector     (bool bSec=true);
  void SetFull       ();
  void IncreaseArcLength (double Delta);

  CGePoint Center       () const;
  CGePoint AnglePoint   (double Angle) const;
  void     AnglePoint   (double Angle, CGePoint* pPnt) const;
  CGePoint ParamPoint   (double Param) const;
  CGePoint StartPoint   () const;
  CGePoint EndPoint     () const;
  double   Radius1      () const;
  double   Radius2      () const;
  double   R1           () const;
  double   R2           () const;
  double   Ratio        () const;
  double   Rotation     () const;
  double   StartAngle   () const;
  double   ArcAngle     () const;
  double   EndAngle     () const;
  double   Length       () const;
  double   Area         () const;
  bool     IsFull       () const;
  bool     IsArc        () const;
  bool     IsSector     () const;
  bool     IsClosed     () const;
  bool     IsArcPoint   (const CGePoint& Pnt) const;
  bool     IsCW         () const;
  bool     IsCCW        () const;
  void     Save    (BYTE* pBuf) const;
  void     Restore (const BYTE* pBuf);
  UINT_PTR GetPoints        (CGePoint* Pnt, double AngStep, UINT_PTR MaxNumPoints) const;
  UINT_PTR GetPointsE       (CGePoint* Pnt, UINT_PTR MaxNumPoints) const;
  UINT_PTR GetPolylineE     (CGePolyline* pPline, UINT_PTR MaxNumPoints) const;
  void     GetPointsOld     (CGePoint* Pnt, UINT_PTR NumPoints) const;
  void     GetPolylineOld   (CGePolyline* pPline, UINT_PTR NumPoints) const;

  void     GetPointByAngle (double Angle, CGePoint* pPnt) const;
  double   GetDistByPoint  (const CGePoint& Pnt, CGePoint* pEllPt=NULL) const;
  void     GetAxisPoints   (CGePoint* Pnt) const;

  bool Contains (const CGePoint& Pnt) const;
  bool Contains (const CGeRect& Rect) const;
  bool Crosses  (const CGeRect& Rect) const;

  void Move   (const CGePoint& Off);
  void Rotate (const CGePoint& Pt0, double Angle);
  void Scale  (const CGePoint& Pt0, double ScaleX, double ScaleY);
  void Mirror (const CGeLine& Line);

  static void ParamToAngle (double Param, double* pAngle, double RatioR2R1);
  static void AngleToParam (double Angle, double* pParam, double RatioR2R1);
};

// Inline functions
//-------------------------------------------------------------------
inline const CGeEllipse& CGeEllipse::Ellipse () const {return *this;}
inline void     CGeEllipse::SetCenter  (const CGePoint& Pnt) {m_Center = Pnt;}
inline void     CGeEllipse::SetSector  (bool bSec) {m_bSector = bSec;}
inline CGePoint CGeEllipse::Center     () const {return m_Center;}
inline double   CGeEllipse::Radius1    () const {return m_R1;}
inline double   CGeEllipse::Radius2    () const {return m_R2;}
inline double   CGeEllipse::R1         () const {return m_R1;}
inline double   CGeEllipse::R2         () const {return m_R2;}
inline double   CGeEllipse::Ratio      () const {return m_R2/m_R1;}
inline double   CGeEllipse::Rotation   () const {return m_Rotation;}
inline double   CGeEllipse::StartAngle () const {return m_AngStart;}
inline double   CGeEllipse::ArcAngle   () const {return m_AngArc;}
inline CGePoint CGeEllipse::StartPoint () const {return AnglePoint( m_AngStart );}
inline CGePoint CGeEllipse::EndPoint   () const {return AnglePoint( EndAngle() );}
inline bool     CGeEllipse::IsArc      () const {return (m_AngArc == 0.0)? false : true;}
inline bool     CGeEllipse::IsFull     () const {return (m_AngArc == 0.0)? true : false;}
inline bool     CGeEllipse::IsSector   () const {return m_bSector;}
inline bool     CGeEllipse::IsCW       () const {return (m_AngArc<0.0)?true:false;}
inline bool     CGeEllipse::IsCCW      () const {return (m_AngArc>=0.0)?true:false;}
inline void     CGeEllipse::SetFull    () {SetArcAngle(0);}
inline void     CGeEllipse::GetPointByAngle (double Angle, CGePoint* pPnt) const {AnglePoint(Angle,pPnt);}



enum class EnPlane {
  GE_PLANE_TOP    = 0, 
  GE_PLANE_BOTTOM = 1, 
  GE_PLANE_LEFT   = 2, 
  GE_PLANE_RIGHT  = 3, 
  GE_PLANE_FRONT  = 4, 
  GE_PLANE_BACK   = 5, 
};

//=============================================================================
//
//  CGe3dPoint_  24 bytes
//
//=============================================================================
class CGe3dPoint {
  static double m_X0;
  static double m_Y0;
  static double m_Z0;
  static BYTE   m_Axis;
  static double m_cosAng;
  static double m_sinAng;
public:
  double x;
  double y;
  double z;
public:
    CGe3dPoint ();
    CGe3dPoint (const CGe3dPoint& Pnt);
    CGe3dPoint (const CGePoint& Pnt);
    CGe3dPoint (double x, double y, double z);

  void   Set (double x, double y, double z);
  void   Set (const CGe3dPoint& Pnt);
  void   Set (const CGePoint& Pnt, double Z);
  void   SetVector (const CGe3dPoint& Pnt);
  void   SetVector (const CGe3dPoint& Pt0, const CGe3dPoint& Pt1);

  // Returns TRUE if distance between this point and 'Pnt2'
  // less then 'delta' value
  bool IsEqual (const CGe3dPoint& pt, double delta=g_GeTolerance) const;

//  static void SetRotate (const CGe3dPoint& A, double theta, double phi, double alpha);
  static void SetRotate (const CGe3dPoint& Pt0, UINT_PTR Axis, double Angle);

  // Functions/binary-operators that return points or sizes
  CGe3dPoint MoveBy (double dx, double dy, double dz) const;
  CGe3dPoint MoveBy (const CGe3dPoint& off) const;
  CGe3dPoint ScaleBy  (const CGe3dPoint& Base, double ScaleX, double ScaleY, double ScaleZ) const;
  CGe3dPoint PolarBy  (double Angle, double Dist, EnPlane Plane) const;
  CGe3dPoint PolarBy  (double AngHorz, double AngVert, double Dist) const;
  CGe3dPoint RotateBy (const CGe3dPoint& Base, double Angle, EnPlane Plane) const;
  CGe3dPoint RotateBy () const;
  CGe3dPoint MirrorBy (double a1, double b1, EnPlane Plane) const;
  CGe3dPoint Center   (const CGe3dPoint& p) const;

  // Functions/assignement-operators that modify this point
  void       Move    (double dx, double dy, double dz);
  void       Move    (const CGe3dPoint& off);
  void       Scale   (const CGe3dPoint& Base, double ScaleX, double ScaleY, double ScaleZ);
  void       Polar   (double Angle, double Dist, EnPlane Plane);
  void       Polar   (double AngHorz, double AngVert, double Dist);
  void       Rotate  (const CGe3dPoint& Base, double Angle, EnPlane Plane);
  void       Rotate  ();
  void       Mirror  (double a1, double b1, EnPlane Plane);

  bool       IsPlanar      (double Delta=0.0) const;
  bool       IsPlaneTop    (double Delta=0.0) const;
  bool       IsPlaneBottom (double Delta=0.0) const;
  bool       IsPlaneRight  (double Delta=0.0) const;
  bool       IsPlaneLeft   (double Delta=0.0) const;
  bool       IsPlaneFront  (double Delta=0.0) const;
  bool       IsPlaneBack   (double Delta=0.0) const;

  bool       IsNull           () const;
  bool       IsNotNull        () const;
  double     VectorLength     () const;
  void       VectorScale      (double Scale);
  void       VectorNormalize  ();
  void       VectorReverse    ();
  CGe3dPoint VectorCross      (const CGe3dPoint& B) const;
  void       GetArbitraryAxes (CGe3dPoint* XAxis, CGe3dPoint* YAxis) const;
  void       Transform        (const CGe3dPoint& Xaxis, const CGe3dPoint& Yaxis, const CGe3dPoint& Zaxis);
  void       Transform        (const CGe3dPoint& Xaxis, const CGe3dPoint& Yaxis, const CGe3dPoint& Zaxis, CGe3dPoint* pOutPt) const;
  CGe3dPoint TransformBy      (const CGe3dPoint& Xaxis, const CGe3dPoint& Yaxis, const CGe3dPoint& Zaxis) const;

  bool       operator == (const CGe3dPoint& p) const;
  bool       operator != (const CGe3dPoint& p) const;
  CGe3dPoint operator +  (const CGe3dPoint& p) const;
  CGe3dPoint operator -  (const CGe3dPoint& p) const;

  double Distance (const CGe3dPoint& Pnt) const;
  double Angle    (const CGe3dPoint& Pnt, EnPlane Plane) const;
  double Angle    (EnPlane Plane) const;
  void   GetPoint2d (CGePoint* pPnt) const;
  void   GetAngles (const CGe3dPoint& Pnt, double* pAngHorz, double* pAngVert) const;

  const CGe3dPoint&  Point () const;
};

//-----------------------------------------------
inline const CGe3dPoint& CGe3dPoint::Point () const {return *this;}
inline CGe3dPoint::CGe3dPoint () : x(0.0), y(0.0), z(0.0) {}
inline CGe3dPoint::CGe3dPoint (const CGe3dPoint& src) : x(src.x), y(src.y), z(src.z) {}
inline CGe3dPoint::CGe3dPoint (double xx, double yy, double zz) : x(xx), y(yy), z(zz) {}
inline void CGe3dPoint::Set (double xx, double yy, double zz) {x=xx; y=yy; z=zz;}
inline void CGe3dPoint::Set (const CGe3dPoint& Pnt) {x=Pnt.x; y=Pnt.y; z=Pnt.z;}
inline void CGe3dPoint::Set (const CGePoint& Pnt, double Z) {x=Pnt.x; y=Pnt.y; z=Z;}
inline bool CGe3dPoint::IsEqual(const CGe3dPoint& p2, double delta) const {
  if (x-delta<=p2.x && p2.x<=x+delta && 
      y-delta<=p2.y && p2.y<=y+delta &&
      z-delta<=p2.z && p2.z<=z+delta) 
  {
    return true;
  }
  return false;
}
inline CGe3dPoint CGe3dPoint::MoveBy (double dx, double dy, double dz) const {return CGe3dPoint(x+dx, y+dy, z+dz);}
inline CGe3dPoint CGe3dPoint::MoveBy (const CGe3dPoint& off) const {return CGe3dPoint(x+off.x, y+off.y, z+off.z);}
inline CGe3dPoint CGe3dPoint::ScaleBy (const CGe3dPoint& Base, double scx, double scy, double scz) const {
  return CGe3dPoint( Base.x+(x-Base.x)*scx, Base.y+(y-Base.y)*scy, Base.z+(z-Base.z)*scz );
}
inline void CGe3dPoint::Move (double dx, double dy, double dz) {x+=dx; y+=dy; z+=dz;}
inline void CGe3dPoint::Move (const CGe3dPoint& off) {x+=off.x; y+=off.y; z+=off.z;}
inline void CGe3dPoint::Scale (const CGe3dPoint& Base, double scx, double scy, double scz) {
  x = Base.x + (x - Base.x) * scx;
  y = Base.y + (y - Base.y) * scy;
  z = Base.z + (z - Base.z) * scz;
}
inline bool       CGe3dPoint::operator ==(const CGe3dPoint& p) const {return (x==p.x && y==p.y && z==p.z);}
inline bool       CGe3dPoint::operator !=(const CGe3dPoint& p) const {return (x!=p.x || y!=p.y || z!=p.z);}
inline CGe3dPoint CGe3dPoint::operator + (const CGe3dPoint& p) const {return CGe3dPoint(x+p.x, y+p.y, z+p.z);}
inline CGe3dPoint CGe3dPoint::operator - (const CGe3dPoint& p) const {return CGe3dPoint(x-p.x, y-p.y, z-p.z);}
inline CGe3dPoint CGe3dPoint::Center (const CGe3dPoint& p) const {return CGe3dPoint( (x+p.x)/2., (y+p.y)/2., (z+p.z)/2. );}
inline bool       CGe3dPoint::IsNull   () const {return (x==0.0 && y==0.0 && z==0.0)? true : false;}
inline bool       CGe3dPoint::IsNotNull() const {return (x==0.0 && y==0.0 && z==0.0)? false : true;}
inline bool       CGe3dPoint::IsPlanar (double Delta) const {return IsPlaneTop(Delta);}
inline void       CGe3dPoint::GetPoint2d (CGePoint* pPnt) const {pPnt->Set(x,y);} 
// Return the length of a vector
//-----------------------------------------------
inline double CGe3dPoint::VectorLength () const
{
  return sqrt( x*x + y*y + z*z );
}

// Uniformly Scale a vector by a specified factor
//-----------------------------------------------
inline void CGe3dPoint::VectorScale (double Scale)
{
  x *= Scale;
  y *= Scale;
  z *= Scale;
}

//-----------------------------------------------
inline void CGe3dPoint::VectorReverse ()
{
  x *= -1.0;
  y *= -1.0;
  z *= -1.0;
}

/********************************************************************
* Generates quadratic B-spline curve between 
* control points A,B,C
*   nLines - number of lines per curve 
*   GenVer - pointer to array of result points (without last point C)
*   pnGenVer - pointer to number of points in GenVer
*   MaxGenVer - max number of points in GenVer
*********************************************************************/
bool geGenSegmentBSpline2 (const CGePoint& A, const CGePoint& B, const CGePoint& C, UINT_PTR nLines,
                           CGePoint* GenVer, UINT_PTR* pnGenVer, UINT_PTR MaxGenVer, double* pLen=NULL);

/********************************************************************
* Generates quadratic B-spline curve between control points A,B,C
*   nLines - number of lines per curve 
*   pOutPline - output polyline (point C not added)
*********************************************************************/
void geGenSegmentBSpline2 (const CGePoint& A, const CGePoint& B, const CGePoint& C, 
                           UINT_PTR nLines, CGePolyline* pOutPline); //, int iPrm=0);

/********************************************************************
* Generates quadratic B-spline
*   Ver     - array of control vertexes
*   nVer    - number of points in Ver array
*   bClosed - input polyline is closed if true
*   nLines  - number of lines for each curve segment
*   GenVer    - pointer to array of result points
*   pnGenVer  - pointer to number of points in GenVer
*   MaxGenVer - max number of points in GenVer
*********************************************************************/
void geGenBSpline2 (const CGePoint* Ver, UINT_PTR nVer, bool bClosed, UINT_PTR nLines,
                    CGePoint* GenVer, UINT_PTR* pnGenVer, UINT_PTR MaxGenVer);


/********************************************************************
* Generates cubic B-spline curve for segment B,C
* between  control points A,B,C,D
*   nLines - number of lines per curve 
*   GenVer - pointer to array of result points (without last point C)
*   pnGenVer - pointer to number of points in GenVer
*   MaxGenVer - max number of points in GenVer
*********************************************************************/
bool geGenSegmentBSpline3 (const CGePoint& A, const CGePoint& B, const CGePoint& C, const CGePoint& D, UINT_PTR nLines,
                           CGePoint* GenVer, UINT_PTR* pnGenVer, UINT_PTR MaxGenVer);

/********************************************************************
* Generates cubic B-spline curve for segment B,C
* between  control points A,B,C,D
*   nLines - number of result points 
*   pOutPline - output polyline (without last point C)
*********************************************************************/
void geGenSegmentBSpline3 (const CGePoint& A, const CGePoint& B, const CGePoint& C, const CGePoint& D, 
                           UINT_PTR nLines, CGePolyline* pOutPline);

/********************************************************************
* Generates quadratic Bezier curve between control points A,B,C
*   nLines - number of lines per curve
*   pOutPline - output polyline (without last point C)
*********************************************************************/
void geGenSegmentBezier2 (const CGePoint& A, const CGePoint& B, const CGePoint& C, 
                          UINT_PTR nLines, CGePolyline* pOutPline);
bool geGenSegmentBezier2 (const CGePoint& A, const CGePoint& B, const CGePoint& C, 
                          UINT_PTR nLines, CGePoint* GenVer, UINT_PTR* pnGenVer, UINT_PTR MaxGenVer);
void geGenSegmentBezier2 (const CGePoint& A, double Wa, const CGePoint& B, double Wb, const CGePoint& C, double Wc,
                          UINT_PTR nLines, CGePolyline* pOutPline);

/********************************************************************
* Generates cubic Bezier curve between control points A,B,C,D
*   nLines - number of lines per curve
*   pOutPline - output polyline (without last point D)
*********************************************************************/
void geGenSegmentBezier3 (const CGePoint& A, const CGePoint& B, const CGePoint& C, const CGePoint& D,
                          UINT_PTR nLines, CGePolyline* pOutPline);
bool geGenSegmentBezier3 (const CGePoint& A, const CGePoint& B, const CGePoint& C, const CGePoint& D,
                          UINT_PTR nLines, CGePoint* GenVer, UINT_PTR* pnGenVer, UINT_PTR MaxGenVer);
void geGenSegmentBezier3 (const CGePoint& A, double Wa, const CGePoint& B, double Wb, const CGePoint& C, double Wc, const CGePoint& D, double Wd, 
                          UINT_PTR nLines, CGePolyline* pOutPline);

/********************************************************************
* Generates Fit Spline
*   pInPline - input polyline
*   pPline   - result polyline
*   Quality  - number of generated linear segments for each spline segment
*********************************************************************/
void geGenFitSpline (const CGePolyline* pInPline, CGePolyline* pPline, UINT_PTR nLines);
bool geGenFitSpline (const CGePolyline* pInPline, UINT_PTR nLines, CGePoint* GenVer, UINT_PTR* pnGenVer, UINT_PTR MaxGenVer);

/********************************************************************
*********************************************************************/
bool geGetFilletPrm (const CGePoint& A, const CGePoint& B, const CGePoint& C, double Radius,
                     double* pTang, double* pBiss, double* pAngle);

/********************************************************************
*********************************************************************/
bool geGetFilletBulge (const CGePoint& A, const CGePoint& B, const CGePoint& C, double Radius,
                       CGePoint* pOutPt1, CGePoint* pOutPt2, double* pBulge, CGePoint* pOutPtMid=NULL);

double geGetRadiusByTang (const CGePoint& A, const CGePoint& B, const CGePoint& C, double Tang);
double geGetRadiusByBisector (const CGePoint& A, const CGePoint& B, const CGePoint& C, double Bis);

// Get Fillet Arc between 2 lines
bool geGetFilletArc (const CGePoint& A, const CGePoint& B, const CGePoint& C, double Radius, CGeArc* pOutArc);

// Get Fillet Arc between 2 arcs
//bool geGetFilletArc (const CGeArc& Arc1, const CGeArc& Arc2, double Radius, CGeArc* pOutArc);
//bool geGetFilletCirc (const CGeCircle& Cir1, const CGeCircle& Cir2, double R, CGeArc* pOutArc1, CGeArc* pOutArc2);

UINT_PTR geGetFilletLines    (const CGeLine& Lin1, const CGeLine& Lin2, double R, CGeArc* pOutArc);
UINT_PTR geGetFilletCircs    (const CGeCircle& Cir1, const CGeCircle& Cir2, double R, CGeArc* pOutArc);
UINT_PTR geGetFilletLineCirc (const CGeLine& Lin, const CGeCircle& Cir, double R, CGeArc* pOutArc);

/********************************************************************
* Returns counter-clockwise angle between line A-B
* and 0-X axis.
* result angle is radians 0-2Pi  
*********************************************************************/
double geGetAngle (double x0, double y0, double x1, double y1);

/********************************************************************
* Returns turn angle of line 1-2, relative to 0-1 direction
* Result angle is radians 0..+/-Pi
* Positive angle - turn left, negative - turn right
*********************************************************************/
double geGetAngleTurn (double x0, double y0, double x1, double y1, double x2, double y2);

/********************************************************************
* Normalize Angle to range 0-PI2 radian
*********************************************************************/
double geNormalizeAngle (double Ang);

/********************************************************************
* Returns distance between 2 angles (A,B)
*********************************************************************/
double geGetDeltaAngle (double A, double B);

/********************************************************************
* Check if 2 angles look to the same direction
*********************************************************************/
bool geIsSameAngle (double Ang1, double Ang2, double Delta);


/********************************************************************
*  Returns round direction value for points ABC
*  >0 - ABC has counter-clockwise direction
*  <0 - ABC has clockwise direction 
*  =0 - ABC has linear direction
*********************************************************************/
double geRoundValue (const CGePoint& A, const CGePoint& B, const CGePoint& C);

/********************************************************************
*  Returns an area of PolyPolygon
*  Pgon - array of polygons
*  nPgons - number of polygons
*********************************************************************/
double geGetArea (const CGePolyline* Pgon, int nPgons);

//-----------------------------------------------
// Returns 2 offset points of vertex B
// where Dist - distance from B
// pPlef or pPrig can be NULL
//-----------------------------------------------
void geGetWidthPoints (const CGePoint* pA, const CGePoint* pB, const CGePoint* pC, double Dist, CGePoint* pPlef, CGePoint* pPrig);

//-----------------------------------------------
// Returns an offset point from vertex B
// Dist < 0. means OffPnt will be placed at left side of A-B-C
// Dist > 0. means OffPnt will be placed at right side of A-B-C
//-----------------------------------------------
void geGetOffsetPoint (const CGePoint* pA, const CGePoint* pB, const CGePoint* pC, double Dist, CGePoint* pOffPnt);

//-----------------------------------------------
#define GE_ARCRES_512  0
#define GE_ARCRES_256  1
#define GE_ARCRES_128  2
#define GE_ARCRES_64   3
#define GE_ARCRES_32   4
#define GE_ARCRES_16   5
#define GE_ARCRES_8    6

class CArcPrm {
public:
  CGePoint m_Center;
  double   m_Radius;
  double   m_Ratio;        // R2 = m_Radius * m_Ratio
  double   m_Ang0;
  double   m_AngArc;
  bool     m_bDirCW;       // clockwise direction
  double   m_AngRot;       // total rotation angle
  int      m_Resolution;   // 0-256, 1-128, 2-64, 3-32, 4-16
public:
    CArcPrm ();
};
UINT_PTR geFastArc (const CArcPrm& Prm, CGePoint* pPnt);


//-----------------------------------------------
// Transform model space length (drw units) into window space length (pixels)
//-----------------------------------------------
UINT_PTR geTransformSize (double Len, const GE_TRANSPRM* pPrm, double PixelSize);

//-----------------------------------------------
// Transform angle (drw units)
//-----------------------------------------------
double geTransformAngle (double Angle, const GE_TRANSPRM* pPrm);

//-----------------------------------------------
// Return true if transform has uniform scale
//-----------------------------------------------
bool geTransHasUniscale (const GE_TRANSPRM* pPrm);

//-----------------------------------------------
// Rotate an array of points
//-----------------------------------------------
void gePointsRotate (CGePoint* pPnt, UINT_PTR nPnts, const CGePoint& Pt0, double Angle);

//-----------------------------------------------
// Move an array of points
//-----------------------------------------------
void gePointsMove (CGePoint* pPnt, UINT_PTR nPnts, double DX, double DY);


//-----------------------------------------------
// Intersections between objects
// Returns number of intersection points
// Get the points with the function geInterGetPoint
//-----------------------------------------------
UINT_PTR geInter_Line_Line      (const CGeLine& Line, const CGeLine& Line2, UINT_PTR Apparent=0, CGePoint* IntPt=NULL);
UINT_PTR geInter_Line_Pline     (const CGeLine& Line, const CGePolyline& Pline, UINT_PTR Apparent=0, CGePoint* IntPt=NULL, UINT_PTR MaxPts=1);
UINT_PTR geInter_Line_Circle    (const CGeLine& Line, const CGeCircle& Circ, UINT_PTR Apparent=0, CGePoint* IntPt=NULL);
UINT_PTR geInter_Line_Arc       (const CGeLine& Line, const CGeArc& Arc, UINT_PTR Apparent=0, CGePoint* IntPt=NULL);
UINT_PTR geInter_Line_Ellipse   (const CGeLine& Line, const CGeEllipse& Ell, UINT_PTR Apparent=0, CGePoint* IntPt=NULL);
                          
void     geInter_BeginPolyPline ();
UINT_PTR geInter_EndPolyPline   ();
UINT_PTR geInter_Pline_Line     (const CGePolyline& Pline, const CGeLine& Line, UINT_PTR Apparent=0);
UINT_PTR geInter_Pline_Pline    (const CGePolyline& Pline, const CGePolyline& Pline2, UINT_PTR Apparent=0);
UINT_PTR geInter_Pline_Circle   (const CGePolyline& Pline, const CGeCircle& Circ, UINT_PTR Apparent=0);
UINT_PTR geInter_Pline_Arc      (const CGePolyline& Pline, const CGeArc& Arc, UINT_PTR Apparent=0);
UINT_PTR geInter_Pline_Ellipse  (const CGePolyline& Pline, const CGeEllipse& Ell, UINT_PTR Apparent=0);

UINT_PTR geInter_Circle_Line    (const CGeCircle& Circ, const CGeLine& Line, UINT_PTR Apparent=0);
UINT_PTR geInter_Circle_Pline   (const CGeCircle& Circ, const CGePolyline& Pline, UINT_PTR Apparent=0);
UINT_PTR geInter_Circle_Circle  (const CGeCircle& Circ, const CGeCircle& Circ2);
UINT_PTR geInter_Circle_Arc     (const CGeCircle& Circ, const CGeArc& Arc, UINT_PTR Apparent=0);
UINT_PTR geInter_Circle_Ellipse (const CGeCircle& Circ, const CGeEllipse& Ell, UINT_PTR Apparent=0);

UINT_PTR geInter_Arc_Line       (const CGeArc& Arc, const CGeLine& Line, UINT_PTR Apparent=0);
UINT_PTR geInter_Arc_Pline      (const CGeArc& Arc, const CGePolyline& Pline, UINT_PTR Apparent=0);
UINT_PTR geInter_Arc_Circle     (const CGeArc& Arc, const CGeCircle& Circ, UINT_PTR Apparent=0);
UINT_PTR geInter_Arc_Arc        (const CGeArc& Arc, const CGeArc& Arc2, UINT_PTR Apparent=0);
UINT_PTR geInter_Arc_Ellipse    (const CGeArc& Arc, const CGeEllipse& Ell, UINT_PTR Apparent=0);

UINT_PTR geInter_Ellipse_Line   (const CGeEllipse& Ell, const CGeLine& Line, UINT_PTR Apparent=0);
UINT_PTR geInter_Ellipse_Pline  (const CGeEllipse& Ell, const CGePolyline& Pline, UINT_PTR Apparent=0);
UINT_PTR geInter_Ellipse_Circle (const CGeEllipse& Ell, const CGeCircle& Circ, UINT_PTR Apparent=0);
UINT_PTR geInter_Ellipse_Arc    (const CGeEllipse& Ell, const CGeArc& Arc, UINT_PTR Apparent=0);
UINT_PTR geInter_Ellipse_Ellipse(const CGeEllipse& Ell, const CGeEllipse& Ell2, UINT_PTR Apparent=0);

bool geInter_GetPoint (UINT_PTR i, CGePoint* pPnt);
void geInter_SetTolerance (double Tolerance);

bool geSnapOrtho (const CGePoint& BasePt, const CGePoint& Pt2, double SnapDist, CGePoint* pSnapPt);  

INT_PTR geContainsPoint (const CGePoint* Ver, UINT_PTR nVers, const CGePoint& Pnt);

//-----------------------------------------------
// Reverse points order in the array Pnt
//-----------------------------------------------
void geReversePoints (CGePoint* Pnt, UINT_PTR nPnts); 


#endif  // _OK_GEOM_
