/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Read AutoCAD drawing
*********************************************************************/
#include "pch.h"

extern COptions g_Options;
extern bool     g_bLeader;   // LoadEnts.cpp
extern LC_FUN_PROGRESS g_pFunProgress;
extern HANDLE   g_hDrw;     // used to pass with g_pFunProgress

double      g_LtScale = 1.0;
int         g_NumEnts = 0;
OdGePoint3d g_ExtPtMin;
OdGePoint3d g_ExtPtMax;
bool        g_bExtents = false;

void Read_Header     (OdDbDatabasePtr pOdDb, HANDLE hLcDb);
void Read_Linetypes  (OdDbDatabasePtr pOdDb, HANDLE hLcDb);
void Read_Layers     (OdDbDatabasePtr pOdDb, HANDLE hLcDb);
void Read_TextStyles (OdDbDatabasePtr pOdDb, HANDLE hLcDb);
void Read_DimStyles  (OdDbDatabasePtr pOdDb, HANDLE hLcDb);
void Read_MlineStyles(OdDbDatabasePtr pOdDb, HANDLE hLcDb);
void Read_BlockDefs  (OdDbDatabasePtr pOdDb, HANDLE hLcDb);
void Read_Layouts    (OdDbDatabasePtr pOdDb, HANDLE hLcDb);
void Read_Entities   (OdDbDatabasePtr pOdDb, HANDLE hLcDb);

void Read_Point      (const OdDbPointPtr pOdPoint, HANDLE hLcDb, HANDLE hBlock);
void Read_Ray        (const OdDbRayPtr pOdRay, HANDLE hLcDb, HANDLE hBlock);
void Read_XLine      (const OdDbXlinePtr pOdXline, HANDLE hLcDb, HANDLE hBlock);
HANDLE Read_Line       (const OdDbLinePtr pOdLine, HANDLE hLcDb, HANDLE hBlock);
HANDLE Read_LwPolyline (const OdDbPolylinePtr pOdPline, HANDLE hLcDb, HANDLE hBlock);
HANDLE Read_2dPolyline (const OdDb2dPolylinePtr pOdPline, HANDLE hLcDb, HANDLE hBlock);
HANDLE Read_3dPolyline (const OdDb3dPolylinePtr pOdPline, HANDLE hLcDb, HANDLE hBlock);
void Read_Mline      (const OdDbMlinePtr pOdMline, HANDLE hLcDb, HANDLE hBlock);
void Read_Spline     (const OdDbSplinePtr pOdSpline, HANDLE hLcDb, HANDLE hBlock);

void Read_Circle     (const OdDbCirclePtr pOdCirc, HANDLE hLcDb, HANDLE hBlock);
void Read_Arc        (const OdDbArcPtr pOdArc,  HANDLE hLcDb, HANDLE hBlock);
void Read_Ellipse    (const OdDbEllipsePtr pOdEllipse, HANDLE hLcDb, HANDLE hBlock);

void Read_Text       (const OdDbTextPtr pOdText, HANDLE hLcDb, HANDLE hBlock);
void Read_MText      (const OdDbMTextPtr pOdMText, HANDLE hLcDb, HANDLE hBlock);
void Read_ArcText    (const OdDbArcAlignedTextPtr atext, HANDLE hLcDb, HANDLE hBlock);
void Read_Attdef     (const OdDbAttributeDefinitionPtr pOdAttdef, HANDLE hLcDb, HANDLE hBlock);
void Read_BlockRef   (OdDbBlockReferencePtr pOdBlkRef, HANDLE hLcDb, HANDLE hBlock);
void Read_MInsert    (OdDbMInsertBlockPtr pMInsert, HANDLE hLcDb, HANDLE hBlock);
void Read_DimAlig    (const OdDbAlignedDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock);
void Read_DimAng2p   (const OdDb2LineAngularDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock);
void Read_DimAng3p   (const OdDb3PointAngularDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock);
void Read_DimAssoc   (const OdDbDimAssocPtr pOdDim, HANDLE hLcDb, HANDLE hBlock);
void Read_DimDiam    (const OdDbDiametricDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock);
void Read_DimOrd     (const OdDbOrdinateDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock);
void Read_DimRad     (const OdDbRadialDimensionPtr pOdDim, HANDLE hLcDb, HANDLE hBlock);
void Read_DimRot     (const OdDbRotatedDimensionPtr pOdDimRot, HANDLE hLcDb, HANDLE hBlock);
void Read_Leader     (const OdDbLeaderPtr pOdLeader, HANDLE hLcDb, HANDLE hBlock);
void Read_MLeader    (const OdDbMLeaderPtr pOdMLeader, HANDLE hLcDb, HANDLE hBlock);
void Read_3dSolid    (OdDb3dSolidPtr pSolid, HANDLE hLcDb, HANDLE hBlock);
void Read_Face3D     (const OdDbFacePtr pOdFace, HANDLE hLcDb, HANDLE hBlock);
void Read_FaceMesh   (const OdDbPolyFaceMeshPtr pOdPFmesh, HANDLE hLcDb, HANDLE hBlock);
void Read_PgonMesh   (const OdDbPolygonMeshPtr pOdMesh, HANDLE hLcDb, HANDLE hBlock);
void Read_Shape      (const OdDbShapePtr pOdShape, HANDLE hLcDb, HANDLE hBlock);
void Read_Hatch      (const OdDbHatchPtr pOdHatch, HANDLE hLcDb, HANDLE hBlock);
void Read_Image      (const OdDbRasterImagePtr pOdImage, HANDLE hLcDb, HANDLE hBlock);
void Read_Solid      (const OdDbSolidPtr pOdSolid, HANDLE hLcDb, HANDLE hBlock);
void Read_Trace      (const OdDbTracePtr pOdTrace, HANDLE hLcDb, HANDLE hBlock);
void Read_Wipeout    (const OdDbWipeoutPtr pOdWipe, HANDLE hLcDb, HANDLE hBlock);
void Read_Region     (OdDbRegionPtr pRgn, HANDLE hLcDb, HANDLE hBlock);
void Read_Table      (const OdDbTablePtr pOdTable, HANDLE hLcDb, HANDLE hBlock);
void Read_Viewport   (const OdDbViewportPtr pOdVport, HANDLE hLcDb, HANDLE hBlock);
void Read_SortEnts   (OdDbSortentsTablePtr pOdSortEnts, HANDLE hLcDb, HANDLE hBlock);


// Define module map for statically linked modules:
#if !defined(_TOOLKIT_IN_DLL_)
//ODRX_DECLARE_STATIC_MODULE_ENTRY_POINT(ModelerModule);
ODRX_DECLARE_STATIC_MODULE_ENTRY_POINT(OdRecomputeDimBlockModule);

ODRX_BEGIN_STATIC_MODULE_MAP()
//ODRX_DEFINE_STATIC_APPLICATION(OdModelerGeometryModuleName, ModelerModule)
ODRX_DEFINE_STATIC_APPLICATION(OdRecomputeDimBlockModuleName, OdRecomputeDimBlockModule)
ODRX_END_STATIC_MODULE_MAP()
#endif


//=============================================================================
bool ReadFile (LPCWSTR szFileName, HANDLE hLcDb, WCHAR* szError)
{
  int   RetCode = 0;
  WCHAR szBuf[320];
  bool  bNoDup = CGePolyline::m_bNoDuplicate;
  OdStaticRxObject<MyServices> svcs;

  memset( szBuf, 0, sizeof(szBuf) );
  svcs.disableOutput( false );
  svcs.setRecomputeDimBlocksRequired( true );
  try 
  {
    ODRX_INIT_STATIC_MODULE_MAP();
    odInitialize( &svcs );
    OdDbDatabasePtr pDb = svcs.readFile( szFileName, true, false, Oda::kShareDenyNo); 
    if (!pDb.isNull()){
      // File Version
      swprintf( szBuf, L"File Version: %d", pDb->originalFileVersion() );
      szBuf[0] = 0;
//      g_hDrw = hLcDb;
//      g_hWnd = ::GetActiveWindow();
//      lcPropPutInt( 0, LC_PROP_G_ANGLE, LC_ANGLE_RADIAN );
      CGePolyline::m_bNoDuplicate = false;
      lcDrwNew( hLcDb, NULL, 0 );
      lcPropPutStr( hLcDb, LC_PROP_DRW_FILENAME, szFileName );
      Read_Header     ( pDb, hLcDb );
      Read_Linetypes  ( pDb, hLcDb );
      Read_Layers     ( pDb, hLcDb );
      Read_TextStyles ( pDb, hLcDb );
      Read_DimStyles  ( pDb, hLcDb );
      Read_MlineStyles( pDb, hLcDb );
      Read_BlockDefs  ( pDb, hLcDb );
      Read_Layouts    ( pDb, hLcDb );
      Read_Entities   ( pDb, hLcDb );
      CGePolyline::m_bNoDuplicate = bNoDup;
    }
  }
  catch( OdError& e )
  {
    wcscpy( szBuf, (WCHAR*)e.description().c_str() );
    RetCode = 101;
  }
  catch (...)
  {
    wcscpy( szBuf, L"Unknown Error." );
    RetCode = 102;
  }
  if (RetCode != 0){
    if (szError && szBuf[0]){
      wcscat( szError, szBuf );
    }
  }
  try
  {
    odUninitialize();
//    swprintf( szBuf, L"ODA lib uninitialized\n" );
  }
  catch( OdError& e )
  {
//    swprintf( szBuf, L"odUninitialize() failed\n" );
    wcscpy( szBuf, (WCHAR*)e.description().c_str() );
    RetCode = 103;
  }
  catch(...)
  {
    wcscpy( szBuf, L"odUninitialize() failed" );
    RetCode = 104;
  }
//mEnd:
  if (RetCode == 0){
    return true;
  }else{
    if (szError && szBuf[0]){
      if (szError[0] == 0){
        wcscat( szError, szBuf );
      }
    }
    return false;
  }
}


//-----------------------------------------------
//-----------------------------------------------
void SetObjHandle (const OdDbObjectId& objectId, HANDLE hObj, int IdProp)
{
  static OdChar szHex[32];
  szHex[0] = 0;
  objectId.getHandle().getIntoAsciiBuffer( szHex );
  lcPropPutStr( hObj, IdProp, (LPCWSTR)szHex );
}

//-----------------------------------------------
HANDLE DrwGetObjectByHandle (const OdDbObjectId& objectId, HANDLE hLcDb, int ObjType)
{
  static OdChar szHex[32];
  szHex[0] = 0;
  objectId.getHandle().getIntoAsciiBuffer( szHex );
  return lcDrwGetObjectByIDH( hLcDb, ObjType, (LPCWSTR)szHex );
}



//-----------------------------------------------
void Read_Header (OdDbDatabasePtr pOdDb, HANDLE hLcDb)
{
  int ival;
  double  dval;
  OdGePoint2d OdPt;
  HANDLE hPntStyle;

  hPntStyle = lcPropGetHandle( hLcDb, LC_PROP_DRW_PNTSTYLE );
  if (hPntStyle){
    ival = pOdDb->getPDMODE();
    lcPropPutInt( hPntStyle, LC_PROP_PSTYLE_PTMODE, ival );
    dval = pOdDb->getPDSIZE();
    lcPropPutFloat( hPntStyle, LC_PROP_PSTYLE_PTSIZE, dval );
  }
  g_LtScale = pOdDb->getLTSCALE();
  if (g_LtScale <= 0.0){
    g_LtScale = 1.0;
  }
  OdDbDatabaseSummaryInfoPtr pSumInfo = oddbGetSummaryInfo( pOdDb );
  lcPropPutStr( hLcDb, LC_PROP_DRW_DESCR, (LPCWSTR)(pSumInfo->getComments().c_str()) );
}


//-----------------------------------------------
void Read_Linetypes (OdDbDatabasePtr pOdDb, HANDLE hLcDb)
{
  WCHAR         szName[256], szBuf[512], szData[1024];
  UINT          i, n, len;
  bool          bAbsAngle;
  double        RotAngle, Scale, Xoff, Yoff;
  HANDLE        hLtype;
  OdUInt16      shapeNumber;
  OdDbObjectId  OdId;
  OdDbTextStyleTableRecordPtr pTextStyle;
  OdDbLinetypeTableRecordPtr  pOdLtype;
  OdDbLinetypeTablePtr        pTable = pOdDb->getLinetypeTableId().safeOpenObject();
  OdDbSymbolTableIteratorPtr  pIter = pTable->newIterator();

  for (pIter->start(); !pIter->done(); pIter->step()){
    pOdLtype = pIter->getRecordId().safeOpenObject();
    wcscpy( szName, (LPCWSTR)pOdLtype->getName().c_str() );
    if (wcsicmp( szName, L"Continuous" ) == 0 ||
        wcsicmp( szName, L"ByBlock" ) == 0 ||
        wcsicmp( szName, L"ByLayer" ) == 0)
    {
      hLtype = lcDrwGetObjectByName( hLcDb, LC_OBJ_LINETYPE, szName );
    }else{
      hLtype = lcDrwAddLinetype( hLcDb, szName, L"" );
    }
    if (hLtype){
      SetObjHandle( pOdLtype->objectId(), hLtype, LC_PROP_TABLE_ID );
      wcscpy( szBuf, (LPCWSTR)pOdLtype->comments().c_str() );
      lcPropPutStr( hLtype, LC_PROP_LINETYPE_DESC, szBuf );
      lcPropPutFloat( hLtype, LC_PROP_LINETYPE_SCALE, g_LtScale );
      n = pOdLtype->numDashes();
      memset( szData, 0, sizeof(szData) );
      if (n > 0){
        for (i=0; i<n; i++){
          okDblToStr( pOdLtype->dashLengthAt(i), szBuf );
          wcscat( szData, szBuf );
          wcscat( szData, L"," );
          shapeNumber = pOdLtype->shapeNumberAt(i);
          wcscpy( szBuf, (LPCWSTR)pOdLtype->textAt(i).c_str() );
          if (shapeNumber || szBuf[0]){
            // complex linetype (with symbols)
            bAbsAngle = pOdLtype->shapeIsUcsOrientedAt(i);
            RotAngle = pOdLtype->shapeRotationAt(i);
            Scale = pOdLtype->shapeScaleAt(i);
            Xoff = pOdLtype->shapeOffsetAt(i).x;
            Yoff = pOdLtype->shapeOffsetAt(i).y;
            OdId = pOdLtype->shapeStyleAt(i);
            if (OdId.isValid()){
              pTextStyle = OdId.safeOpenObject();
              if (shapeNumber){
                swprintf( szBuf, L"[%d,%s", shapeNumber, (WCHAR*)pTextStyle->fileName().c_str() );
              }else{
                swprintf( szBuf, L"[%c%s%c,%s", L'"', (WCHAR*)pOdLtype->textAt(i).c_str(), L'"', (WCHAR*)pTextStyle->getName().c_str() );
              }
              wcscat( szData, szBuf );
              if (Scale > 0){
                okDblToStr( Scale, szBuf );
                wcscat( szData, L",S=" );
                wcscat( szData, szBuf );
              }
              if (RotAngle){
                okDblToStr( RotAngle, szBuf );
                if (bAbsAngle){
                  wcscat( szData, L",A=" );
                  wcscat( szData, szBuf );
                }else{
                  wcscat( szData, L",R=" );
                  wcscat( szData, szBuf );
                }
              }
              if (Xoff){
                okDblToStr( Xoff, szBuf );
                wcscat( szData, L",X=" );
                wcscat( szData, szBuf );
              }
              if (Yoff){
                okDblToStr( Yoff, szBuf );
                wcscat( szData, L",Y=" );
                wcscat( szData, szBuf );
              }
              wcscat( szData, L"]," );
            }
          }
        }
        len = (UINT)wcslen( szData );
        if (len > 1){
          if (szData[len-1] == L',' ){
            szData[len-1] = 0;
          }
          lcPropPutStr( hLtype, LC_PROP_LINETYPE_DATA, szData );
        }
      }
    }
  }
}


//-----------------------------------------------
void Read_Layers (OdDbDatabasePtr pOdDb, HANDLE hLcDb)
{
  WCHAR  szName[256], szColor[32];
  HANDLE hLayer;
  HANDLE hLtype, hLtype0;
  int    LwLcd;
  OdDb::LineWeight           OdLW;
  OdCmColor                  OdColor;
  OdDbLayerTableRecordPtr    pOdLayer;
  OdDbLayerTablePtr          pTable = pOdDb->getLayerTableId().safeOpenObject();
  OdDbSymbolTableIteratorPtr pIter = pTable->newIterator();

  hLtype0 = lcDrwGetFirstObject( hLcDb, LC_OBJ_LINETYPE );   // continuous
  for (pIter->start(); !pIter->done(); pIter->step()){
    pOdLayer = pIter->getRecordId().safeOpenObject();
    wcscpy( szName, (LPCWSTR)pOdLayer->getName().c_str() );
    if (wcsicmp( szName, L"0" ) == 0){
      hLayer = lcPropGetHandle( hLcDb, LC_PROP_DRW_LAYER_0 );
    }else{
      hLayer = lcDrwAddLayer( hLcDb, szName, L"Foreground", NULL, LC_LWEIGHT_DEFAULT );
    }
    if (hLayer){
      SetObjHandle( pOdLayer->objectId(), hLayer, LC_PROP_TABLE_ID ); //&idLayer );
      hLtype = DrwGetObjectByHandle( pOdLayer->linetypeObjectId(), hLcDb, LC_OBJ_LINETYPE );
      if (hLtype){
        lcPropPutHandle( hLayer, LC_PROP_LAYER_LINETYPE, hLtype );
      }else{
        lcPropPutHandle( hLayer, LC_PROP_LAYER_LINETYPE, hLtype0 );
      }
      if (pOdLayer->isFrozen() || pOdLayer->isOff()){
        lcPropPutBool( hLayer, LC_PROP_LAYER_VISIBLE, false );
      }
      if (pOdLayer->isPlottable() == false){
        lcPropPutBool( hLayer, LC_PROP_LAYER_NOPRINT, true );
      }
      if (pOdLayer->isLocked()){
        lcPropPutBool( hLayer, LC_PROP_LAYER_LOCKED, true );
      }
      OdLW = pOdLayer->lineWeight();
      LwLcd = LWeight_DwgToLcd( OdLW );
      lcPropPutInt( hLayer, LC_PROP_LAYER_LWEIGHT, LwLcd );
      OdColor = pOdLayer->color();
      if (OdColor.colorMethod() == OdCmEntityColor::kByACI){
        swprintf( szColor, L"%d", OdColor.colorIndex() );
        lcPropPutStr( hLayer, LC_PROP_LAYER_COLOR, szColor );
//        lcPropPutStr( hLayer, LC_PROP_LAYER_BCOLOR, szColor );
//        lcPropPutStr( hLayer, LC_PROP_LAYER_FCOLOR, szColor );
      }else{
        swprintf( szColor, L"%d,%d,%d", OdColor.red(), OdColor.green(), OdColor.blue() );
        lcPropPutStr( hLayer, LC_PROP_LAYER_COLOR, szColor );
//        lcPropPutStr( hLayer, LC_PROP_LAYER_BCOLOR, szColor );
//        lcPropPutStr( hLayer, LC_PROP_LAYER_FCOLOR, szColor );
      }
    }
  }
}


//-----------------------------------------------
void Read_TextStyles (OdDbDatabasePtr pOdDb, HANDLE hLcDb)
{
  WCHAR    szName[512], szFile[64], szBigfont[256], szExt[16], szTypeFace[128], szBuf[256];
  bool     bBig;
  HANDLE   hStyle;
  bool     bold, italic;
  BOOL     bClosed;
  int      charset, pitchAndFamily;
  OdString                    OdTypeFace;
  OdDbTextStyleTableRecordPtr pOdTStyle;
  OdDbTextStyleTablePtr       pTable = pOdDb->getTextStyleTableId().safeOpenObject();
  OdDbSymbolTableIteratorPtr  pIter = pTable->newIterator();

  memset( szTypeFace, 0, sizeof(szTypeFace) );
  memset( szExt, 0, sizeof(szExt) );
  for (pIter->start(); !pIter->done(); pIter->step()){
    pOdTStyle = pIter->getRecordId().safeOpenObject();
    if (pOdTStyle->isShapeFile()){
      continue;
    }
    wcscpy( szName, (LPCWSTR)pOdTStyle->getName().c_str() );
    wcscpy( szBigfont, (LPCWSTR)pOdTStyle->bigFontFileName().c_str() );
    if (wcslen(szBigfont) > 0){
      bBig = true;
    }else{
      bBig = false;
    }
    wcscpy( szBuf, (LPCWSTR)pOdTStyle->fileName().c_str() );
    if (wcsicmp( szName, L"standard" ) == 0){
      hStyle = lcPropGetHandle( hLcDb, LC_PROP_DRW_TEXTSTYLE_STD );
    }else{
      hStyle = lcDrwAddTextStyle( hLcDb, szName, NULL, false );
    }
    if (hStyle){
      SetObjHandle( pOdTStyle->objectId(), hStyle, LC_PROP_TABLE_ID );
      pOdTStyle->font( OdTypeFace, bold,italic, charset, pitchAndFamily );
      wcscpy( szFile, okGetFileName( szBuf, true ) );  // without path, with extention
      if (szFile[0] == 0){
        wcscpy( szFile, (LPCWSTR)OdTypeFace.c_str() );
        wcscpy( szBuf, szFile );
      }
      lcPropPutStr( hStyle, LC_PROP_TSTYLE_FONT, szFile );
      lcPropPutFloat( hStyle, LC_PROP_TSTYLE_HEIGHT  , pOdTStyle->textSize() );
      lcPropPutFloat( hStyle, LC_PROP_TSTYLE_WSCALE  , pOdTStyle->xScale() );
      lcPropPutFloat( hStyle, LC_PROP_TSTYLE_OBLIQUE , pOdTStyle->obliquingAngle() );
      lcPropPutBool(  hStyle, LC_PROP_TSTYLE_BACKWARD, pOdTStyle->isBackwards() );
      lcPropPutBool(  hStyle, LC_PROP_TSTYLE_UPDOWN  , pOdTStyle->isUpsideDown() );
//      pStyle->SetShapes( pOdTStyle->isShapeFile() );
      memset( szName, 0, sizeof(szName) );
      wcscpy( szName, szBuf );
      if (bBig){
        wcscat( szName, L" + " );
        wcscat( szName, szBigfont );
      }
      lcPropPutStr( hStyle, LC_PROP_TSTYLE_DESC, szName );
      bClosed = lcPropGetBool( hStyle, LC_PROP_TSTYLE_CLOSED );
      if (bClosed){
        lcPropPutBool( hStyle, LC_PROP_TSTYLE_SOLID, true );
      }
    }
  }
}


//-----------------------------------------------
void Read_DimStyles (OdDbDatabasePtr pOdDb, HANDLE hLcDb)
{
  WCHAR    szName[512]; 
  HANDLE   hDimStyle, hTStyle;
  OdDbBlockTableRecordPtr    pOdBlock;
  OdDbObjectId               OdId;
  OdCmColor                  OdColor;
  OdDbDimStyleTableRecordPtr pOdDimStyle;
  OdDbDimStyleTablePtr       pTable = pOdDb->getDimStyleTableId().safeOpenObject();
  OdDbSymbolTableIteratorPtr pIter = pTable->newIterator();
  double scale;

  for (pIter->start(); !pIter->done(); pIter->step()){
    pOdDimStyle = pIter->getRecordId().safeOpenObject();
    wcscpy( szName, (LPCWSTR)pOdDimStyle->getName().c_str() );
    if (wcsicmp( szName, L"Standard" ) == 0){
      hDimStyle = lcPropGetHandle( hLcDb, LC_PROP_DRW_DIMSTYLE_STD );
    }else{
/*
      if (wcsicmp( szName, L"ISO-25" ) == 0){
        hDimStyle = lcPropGetHandle( hLcDb, LC_PROP_DRW_DIMSTYLE_STD );
      }else{
        hDimStyle = lcDrwAddDimStyle( hLcDb, szName );
      }
*/
      hDimStyle = lcDrwAddDimStyle( hLcDb, szName );
    }
    if (hDimStyle){
      SetObjHandle( pOdDimStyle->objectId(), hDimStyle, LC_PROP_TABLE_ID );

      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_ADEC  , pOdDimStyle->dimadec() );
      lcPropPutFloat( hDimStyle, LC_PROP_DIMST_ASZ   , pOdDimStyle->dimasz() );
      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_AUNIT , pOdDimStyle->dimaunit() );
      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_AZIN  , pOdDimStyle->dimazin() );

      if (pOdDimStyle->dimsah() == 0){
        OdId = pOdDimStyle->dimblk();
        if (OdId.isValid()){
          pOdBlock = OdId.safeOpenObject();
          wcscpy( szName, (LPCWSTR)pOdBlock->getName().c_str() );
        }else{
          szName[0] = 0;
        }
        lcPropPutStr( hDimStyle, LC_PROP_DIMST_BLK1, szName );
        lcPropPutStr( hDimStyle, LC_PROP_DIMST_BLK2, szName );

//        hBlock = DrwGetObjectByHandle( OdId, hLcDb, LC_OBJ_BLOCK );
//        if (hBlock){
//          wcscpy( szName, lcPropGetStr( hBlock, LC_PROP_BLOCK_NAME );
//        }
      }else{
        OdId = pOdDimStyle->dimblk1();
        if (OdId.isValid()){
          pOdBlock = OdId.safeOpenObject();
          wcscpy( szName, (LPCWSTR)pOdBlock->getName().c_str() );
        }else{
          szName[0] = 0;
        }
        lcPropPutStr( hDimStyle, LC_PROP_DIMST_BLK1, szName );
        OdId = pOdDimStyle->dimblk2();
        if (OdId.isValid()){
          pOdBlock = OdId.safeOpenObject();
          wcscpy( szName, (LPCWSTR)pOdBlock->getName().c_str() );
        }else{
          szName[0] = 0;
        }
        lcPropPutStr( hDimStyle, LC_PROP_DIMST_BLK2, szName );
      }
      lcPropPutFloat( hDimStyle, LC_PROP_DIMST_CEN   , pOdDimStyle->dimcen() );
      OdColor = pOdDimStyle->dimclrd();
      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_CLRD, OdColor.colorIndex() );
      OdColor = pOdDimStyle->dimclre();
      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_CLRE, OdColor.colorIndex() );
      OdColor = pOdDimStyle->dimclrt();
      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_CLRT, OdColor.colorIndex() );

      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_DEC   , pOdDimStyle->dimdec() );
      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_DSEP  , pOdDimStyle->dimdsep() );
      lcPropPutFloat( hDimStyle, LC_PROP_DIMST_EXE   , pOdDimStyle->dimexe() );
      lcPropPutFloat( hDimStyle, LC_PROP_DIMST_EXO   , pOdDimStyle->dimexo() );
      lcPropPutFloat( hDimStyle, LC_PROP_DIMST_GAP   , pOdDimStyle->dimgap() );
      OdId = pOdDimStyle->dimldrblk();
      if (OdId.isValid()){
        pOdBlock = OdId.safeOpenObject();
        wcscpy( szName, (LPCWSTR)pOdBlock->getName().c_str() );
      }else{
        szName[0] = 0;
      }
      lcPropPutStr(   hDimStyle, LC_PROP_DIMST_LDRBLK, szName );
      lcPropPutFloat( hDimStyle, LC_PROP_DIMST_LFAC  , pOdDimStyle->dimlfac() );
      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_LWD   , pOdDimStyle->dimlwd() );
      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_LWE   , pOdDimStyle->dimlwe() );
//LC_PROP_DIMST_POST          1130  // RW string; Sets a dimension prefix or suffix for the measurement text
      wcscpy( szName, (LPCWSTR)pOdDimStyle->dimpost().c_str() );
      lcPropPutStr(   hDimStyle, LC_PROP_DIMST_POST  , szName );
      lcPropPutFloat( hDimStyle, LC_PROP_DIMST_RND   , pOdDimStyle->dimrnd() );
      scale = pOdDimStyle->dimscale();
      if (scale == 0.0){
        scale = 1.0;
      }
      lcPropPutFloat( hDimStyle, LC_PROP_DIMST_SCALE , scale );
      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_TAD   , pOdDimStyle->dimtad() );
      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_TIH   , pOdDimStyle->dimtih() );
      lcPropPutFloat( hDimStyle, LC_PROP_DIMST_TXT   , pOdDimStyle->dimtxt() );
      OdId = pOdDimStyle->dimtxsty();
      hTStyle = DrwGetObjectByHandle( OdId, hLcDb, LC_OBJ_TEXTSTYLE );
      if (hTStyle == 0){
        hTStyle = lcPropGetHandle( hLcDb, LC_PROP_DRW_TEXTSTYLE_STD );
      }
      lcPropPutHandle( hDimStyle, LC_PROP_DIMST_TXSTY, hTStyle );
      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_LUNIT , pOdDimStyle->dimlunit() );
      lcPropPutInt(   hDimStyle, LC_PROP_DIMST_ZIN   , pOdDimStyle->dimzin() );
    }
  }
}

//-----------------------------------------------
void Read_MlineStyles (OdDbDatabasePtr pOdDb, HANDLE hLcDb)
{
  WCHAR szName[256], szDescr[512], szColor[32];
  int   i, nLines;
  OdDbObjectId OdId;
  OdDbMlineStylePtr pOdMlineStyle;
  OdDbDictionaryPtr pDictionary = pOdDb->getMLStyleDictionaryId().safeOpenObject();
  OdDbDictionaryIteratorPtr pIter = pDictionary->newIterator(); 
  double    Offset;
  OdCmColor OdColor;
  HANDLE    hMlStyle, hLtype;
  
  /**********************************************************************/
  /* Step through the MlineStyle dictionary                             */
  /**********************************************************************/
  for (; !pIter->done(); pIter->next())
  {
    OdId = pIter->objectId();
    pOdMlineStyle = OdDbMlineStyle::cast(OdId.safeOpenObject());
    if (pOdMlineStyle.isNull()){
      continue;
    }
    memset( szName, 0, sizeof(szName) );
    memset( szDescr, 0, sizeof(szDescr) );
    wcsncpy( szName, (LPCWSTR)pOdMlineStyle->name().c_str(), 254 );
    wcsncpy( szDescr, (LPCWSTR)pOdMlineStyle->description().c_str(), 510 );
    if (wcsicmp( szName, L"standard" ) == 0){
      hMlStyle = lcPropGetHandle( hLcDb, LC_PROP_DRW_MLSTYLE_STD );
    }else{
      hMlStyle = lcDrwAddMlineStyle( hLcDb, szName );
    }
    if (hMlStyle){
      SetObjHandle( pOdMlineStyle->objectId(), hMlStyle, LC_PROP_TABLE_ID );
      lcPropPutStr( hMlStyle, LC_PROP_MLSTYLE_DESC, szDescr );
      lcPropPutBool( hMlStyle, LC_PROP_MLSTYLE_JOINTS, pOdMlineStyle->showMiters() );
      lcPropPutBool( hMlStyle, LC_PROP_MLSTYLE_STARTLINE, pOdMlineStyle->startSquareCap() );
      lcPropPutBool( hMlStyle, LC_PROP_MLSTYLE_STARTARC, pOdMlineStyle->startRoundCap() );
      lcPropPutBool( hMlStyle, LC_PROP_MLSTYLE_ENDLINE, pOdMlineStyle->endSquareCap() );
      lcPropPutBool( hMlStyle, LC_PROP_MLSTYLE_ENDARC, pOdMlineStyle->endRoundCap() );
      nLines = pOdMlineStyle->numElements();
      if (nLines > 2){
        for (i=2; i<nLines; ++i){
//          lcMLStyleAddLine( hMlStyle, 0, L"1", 0 );
        }
      }
      for (i=0; i<nLines; ++i){
        pOdMlineStyle->getElementAt( i, Offset, OdColor, OdId );
        lcPropPutInt( hMlStyle, LC_PROP_MLSTYLE_ILINE, i );
        lcPropPutFloat( hMlStyle, LC_PROP_MLSTYLE_OFFSET, Offset );
        OdColorToLcad( OdColor, szColor );
        lcPropPutStr( hMlStyle, LC_PROP_MLSTYLE_COLOR, szColor );
/*
        if (OdColor.colorMethod() == OdCmEntityColor::kByACI){
          swprintf( szColor, L"%d", OdColor.colorIndex() );
          lcPropPutStr( hMlStyle, LC_PROP_MLSTYLE_COLOR, szColor );
        }else{
          swprintf( szColor, L"%d,%d,%d", OdColor.red(), OdColor.green(), OdColor.blue() );
          lcPropPutStr( hMlStyle, LC_PROP_MLSTYLE_COLOR, szColor );
        }
*/
        hLtype = DrwGetObjectByHandle( OdId, hLcDb, LC_OBJ_LINETYPE );
        if (hLtype == NULL){
          hLtype = lcPropGetHandle( hLcDb, LC_PROP_DRW_LINETYPE_CONT );
        }
        lcPropPutHandle( hMlStyle, LC_PROP_MLSTYLE_LTYPE, hLtype );
      }
//      lcMLStyleSortLines( hMlStyle );
    }
  }
}


//-----------------------------------------------
void Read_Entity (OdDbEntityPtr pOdEntity, HANDLE hLcDb, HANDLE hBlock)
{
  bool bNotLeader = true;
  if (pOdEntity->isKindOf( OdDbLine::desc() )){                 // LINE
    OdDbLinePtr line = OdDbLine::cast( pOdEntity );
    Read_Line( line, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbPolyline::desc() )){             // LWPOLYLINE
    OdDbPolylinePtr pline = OdDbPolyline::cast( pOdEntity );
    Read_LwPolyline( pline, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDb2dPolyline::desc() )){           // 2DPOLYLINE
    OdDb2dPolylinePtr pline = OdDb2dPolyline::cast( pOdEntity );
    Read_2dPolyline( pline, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDb3dPolyline::desc() )){           // 3DPOLYLINE
    OdDb3dPolylinePtr pline = OdDb3dPolyline::cast( pOdEntity );
    Read_3dPolyline( pline, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbCircle::desc() )){               // CIRCLE
    OdDbCirclePtr circle = OdDbCircle::cast( pOdEntity );
    Read_Circle( circle, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbArc::desc() )){                  // ARC
    OdDbArcPtr arc = OdDbArc::cast( pOdEntity );
    Read_Arc( arc, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbAttributeDefinition::desc() )){     // ATTRIBUTE DEFINITION
    OdDbAttributeDefinitionPtr attdef = OdDbAttributeDefinition::cast( pOdEntity );
    Read_Attdef( attdef, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbText::desc() )){               // TEXT
    OdDbTextPtr text = OdDbText::cast( pOdEntity );
    Read_Text( text, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbMText::desc() )){              // MTEXT
    OdDbMTextPtr mtext = OdDbMText::cast( pOdEntity );
    Read_MText( mtext, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbSpline::desc() )){             // SPLINE
    OdDbSplinePtr spline = OdDbSpline::cast( pOdEntity );
    Read_Spline( spline, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbBlockReference::desc() )){     // BLOCK REFERENCE
    OdDbBlockReferencePtr block = OdDbBlockReference::cast( pOdEntity );
    if (block->isKindOf( OdDbMInsertBlock::desc() )){         // + MULTIPLY BLOCK REFERENCE
      OdDbMInsertBlockPtr mins = OdDbMInsertBlock::cast( block );
      Read_MInsert( mins, hLcDb, hBlock );
    }else{
      if (block->isKindOf( OdDbTable::desc() )){                // + TABLE
        OdDbTablePtr tbl = OdDbTable::cast( block );
        Read_Table( tbl, hLcDb, hBlock );
      }else{
        Read_BlockRef( block, hLcDb, hBlock );                // + BLOCK REFERENCE
      }
    }
  }else
  if (pOdEntity->isKindOf( OdDbPoint::desc() )){              // POINT
    OdDbPointPtr point = OdDbPoint::cast( pOdEntity );
    Read_Point( point, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbEllipse::desc() )){            // ELLIPSE
    OdDbEllipsePtr ellipse = OdDbEllipse::cast( pOdEntity );
    Read_Ellipse( ellipse, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbArcAlignedText::desc() )){     // ARC TEXT
    OdDbArcAlignedTextPtr atext = OdDbArcAlignedText::cast( pOdEntity );
    Read_ArcText( atext, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbXline::desc() )){              // XLINE
    OdDbXlinePtr xline = OdDbXline::cast( pOdEntity );
    Read_XLine( xline, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbRay::desc() )){                // RAY
    OdDbRayPtr ray = OdDbRay::cast( pOdEntity );
    Read_Ray( ray, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbAlignedDimension::desc() )){        // DIMENSION
    OdDbAlignedDimensionPtr dim = OdDbAlignedDimension::cast( pOdEntity );
    Read_DimAlig( dim, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbRotatedDimension::desc() )){        // DIMENSION
    OdDbRotatedDimensionPtr dim = OdDbRotatedDimension::cast( pOdEntity );
    Read_DimRot( dim, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDb3PointAngularDimension::desc() )){  // DIMENSION
    OdDb3PointAngularDimensionPtr dim = OdDb3PointAngularDimension::cast( pOdEntity );
    Read_DimAng3p( dim, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDb2LineAngularDimension::desc() )){   // DIMENSION
    OdDb2LineAngularDimensionPtr dim = OdDb2LineAngularDimension::cast( pOdEntity );
    Read_DimAng2p( dim, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbDiametricDimension::desc() )){      // DIMENSION
    OdDbDiametricDimensionPtr dim = OdDbDiametricDimension::cast( pOdEntity );
    Read_DimDiam( dim, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbOrdinateDimension::desc() )){       // DIMENSION
    OdDbOrdinateDimensionPtr dim = OdDbOrdinateDimension::cast( pOdEntity );
    Read_DimOrd( dim, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbRadialDimension::desc() )){         // DIMENSION
    OdDbRadialDimensionPtr dim = OdDbRadialDimension::cast( pOdEntity );
    Read_DimRad( dim, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbDimAssoc::desc() )){                // DIMENSION
    OdDbDimAssocPtr dim = OdDbDimAssoc::cast( pOdEntity );
    Read_DimAssoc( dim, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbLeader::desc() )){                  // LEADER
    OdDbLeaderPtr leader = OdDbLeader::cast( pOdEntity );
    Read_Leader( leader, hLcDb, hBlock );
    bNotLeader = false;
  }else
  if (pOdEntity->isKindOf( OdDbHatch::desc() )){                   // HATCH
    OdDbHatchPtr hatch = OdDbHatch::cast( pOdEntity );
    Read_Hatch( hatch, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbRasterImage::desc() )){             // RASTER IMAGE
    OdDbRasterImagePtr image = OdDbRasterImage::cast( pOdEntity );
    if (image->isKindOf( OdDbWipeout::desc() )){                   // WIPEOUT
      OdDbWipeoutPtr vp = OdDbWipeout::cast( pOdEntity );
      Read_Wipeout( vp, hLcDb, hBlock );
    }else{
      Read_Image( image, hLcDb, hBlock );
    }
  }else
  if (pOdEntity->isKindOf( OdDbMLeader::desc() )){                  // MULTILEADER
    OdDbMLeaderPtr mleader = OdDbMLeader::cast( pOdEntity );
    Read_MLeader( mleader, hLcDb, hBlock ); 
  }else
  if (pOdEntity->isKindOf( OdDbShape::desc() )){                   // SHAPE
    OdDbShapePtr shape = OdDbShape::cast( pOdEntity );
    Read_Shape( shape, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbMline::desc() )){                   // MLINE
    OdDbMlinePtr mline = OdDbMline::cast( pOdEntity );
    Read_Mline( mline, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbSolid::desc() )){                   // SOLID
    OdDbSolidPtr solid = OdDbSolid::cast( pOdEntity );
    Read_Solid( solid, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbTrace::desc() )){                   // TRACE
    OdDbTracePtr trace = OdDbTrace::cast( pOdEntity );
    Read_Trace( trace, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbPolyFaceMesh::desc() )){            // POLYFACE MESH
    OdDbPolyFaceMeshPtr fmesh = OdDbPolyFaceMesh::cast( pOdEntity );
    Read_FaceMesh( fmesh, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbPolygonMesh::desc() )){             // POLYGONMESH
    OdDbPolygonMeshPtr pmesh = OdDbPolygonMesh::cast( pOdEntity );
    Read_PgonMesh( pmesh, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbViewport::desc() )){                // VIEWPORT
    OdDbViewportPtr vp = OdDbViewport::cast( pOdEntity );
    Read_Viewport( vp, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbRegion::desc() )){                  // REGION
    OdDbRegionPtr pRgn = OdDbRegion::cast( pOdEntity );
    Read_Region( pRgn, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDbFace::desc() )){                    // 3D FACE
    OdDbFacePtr face = OdDbFace::cast( pOdEntity );
    Read_Face3D( face, hLcDb, hBlock );
  }else
  if (pOdEntity->isKindOf( OdDb3dSolid::desc() )){                 // 3D SOLID
    OdDb3dSolidPtr Solid = OdDb3dSolid::cast( pOdEntity );
    Read_3dSolid( Solid, hLcDb, hBlock );
  }

/*
  else
  if (pOdEntity->isKindOf( OdDbSortentsTable::desc() )){           // SORTENTSTABLE
    OdDbSortentsTablePtr SortEnts = OdDbSortentsTable::cast( pOdEntity );
    Read_SortEnts( SortEnts, hLcDb, hBlock );
  }
*/
  if (bNotLeader){
    g_bLeader = false;
  }
}


//-----------------------------------------------
void Read_BlockDefs (OdDbDatabasePtr pOdDb, HANDLE hLcDb)
{
  WCHAR       szBlockName[256], szPath[300], szBuf[512];
  LPCWSTR     psz = NULL;
  bool        bAttach, bOverlay, bLoaded, bAnon;
  HANDLE      hBlock, hCTabBlock; 
  bool        bModel, bStdBlock;
  OdGePoint3d                Pnt3d;
  OdDbBlockTableRecordPtr    pOdBlock;
  OdDbBlockTablePtr          pTable = pOdDb->getBlockTableId().safeOpenObject();
  OdDbSymbolTableIteratorPtr pBlkIter = pTable->newIterator();
  OdDbObjectId               LayoutId, BlockId, ModelBlockId, CTabId;
  OdDbLayoutPtr              pPaperLayout, pCTabLayout, pModelLayout;
  OdResBufPtr                pRb;
  OdChar                     szBlockHandle[32], szLayoutHandle[32];
  int TabOrder, isActModel;
  int TotalNumEnts = 0;

  szBlockHandle[0] = 0;
  szLayoutHandle[0] = 0;
  szBlockName[0] = 0;
  ModelBlockId = pOdDb->getModelSpaceId();

  pRb = pOdDb->getSysVar( L"CTAB" );
  CTabId = pRb->getObjectId( pOdDb );
  pCTabLayout = CTabId.safeOpenObject( OdDb::kForRead );
  psz = (LPCWSTR)pCTabLayout->getLayoutName().c_str();
  TabOrder = pCTabLayout->getTabOrder();
  isActModel = pCTabLayout->modelType();
  hCTabBlock = 0;

  for (pBlkIter->start(); !pBlkIter->done(); pBlkIter->step()){
    BlockId = pBlkIter->getRecordId();
    BlockId.getHandle().getIntoAsciiBuffer( szBlockHandle );          // Block Handle
    pOdBlock = BlockId.safeOpenObject();
    wcscpy( szBlockName, (LPCWSTR)pOdBlock->getName().c_str() );      // Block Name
    bModel = false;
    bStdBlock = false;
    if (pOdBlock->xrefStatus() == OdDb::kXrfNotAnXref){  // Xref Status
      // the block is not XREF
      if (pOdBlock->isLayout()){
        LayoutId = pOdBlock->getLayoutId();
        if (BlockId == ModelBlockId){
          bModel = true;
          hBlock = lcDrwGetFirstObject( hLcDb, LC_OBJ_BLOCK );    // get Litecad "Model Space" block
          if (isActModel){
            // "Model" layout is active
            hCTabBlock = hBlock;
          }
          pModelLayout = LayoutId.safeOpenObject();
          TabOrder = pModelLayout->getTabOrder();
          pOdBlock->getLayoutId().getHandle().getIntoAsciiBuffer( szLayoutHandle );
        }else{
          hBlock = lcDrwAddBlockPaper( hLcDb, szBlockName, LC_PAPER_A4, LC_PAPER_BOOK, 0, 0 );
          if (hBlock){
            pPaperLayout = LayoutId.safeOpenObject();
            psz = (LPCWSTR)pPaperLayout->getLayoutName().c_str();
            lcPropPutStr( hBlock, LC_PROP_BLOCK_LAYOUTNAME, psz );
            if (!isActModel){
              if (pPaperLayout == pCTabLayout){
                // this Paper layout is active
                hCTabBlock = hBlock;
              }
            }
            TabOrder = pPaperLayout->getTabOrder();
          }
          LayoutId.getHandle().getIntoAsciiBuffer( szLayoutHandle );
        }
        lcPropPutInt( hBlock, LC_PROP_BLOCK_LAYOUTORDER, TabOrder );
        lcPropPutStr( hBlock, LC_PROP_BLOCK_ODA_LAYOUTHANDLE, (LPCWSTR)szLayoutHandle );  // save ODA Layout handle in Litecad layout block
      }else{
        // standard block
        bStdBlock = true;
/*
        if (szName[0]==L'*'){
          // hidden block
          if (szName[1]==L'D' && g_Options.m_Imp_DimAsBlock==false){
            // dimension will be added as object, 
            // therefore don't need to add the dimension block
            continue;
          }else{
            if (szName[1]==L'H' && g_Options.m_Imp_HatchAsBlock==false){
              // hatch will be added as object, 
              // therefore don't need to add the hatch block
              continue;
            }
          }
        }
*/
        hBlock = lcDrwAddBlock( hLcDb, szBlockName, 0,0 );
        bAnon = pOdBlock->isAnonymous();         // Anonymous
        if (bAnon){
          if (szBlockName[1] == L'D'){ // && g_OdaOptions.m_Imp_bDimAsBlock==false){
            wcscpy( szBuf, L"DIMBLOCK__" );
            wcscat( szBuf, szBlockName+1 );
            lcPropPutStr( hBlock, LC_PROP_BLOCK_NAME, szBuf );
          }
//        pBlock->SetHidden( pOdBlock->isAnonymous() );         // Anonymous
        }
      }
      if (hBlock){
        lcPropPutStr( hBlock, LC_PROP_TABLE_ODA_HANDLE, (LPCWSTR)szBlockHandle );  // save ODA block handle in Litecad block
        if (hCTabBlock){
          // set active layout (for Tab)
          lcPropPutHandle( hLcDb, LC_PROP_DRW_LAYOUT, hCTabBlock );
          lcPropPutHandle( hLcDb, LC_PROP_DRW_VISBLOCK, hCTabBlock );
        }
        SetObjHandle( pOdBlock->objectId(), hBlock, LC_PROP_TABLE_ID ); 
        psz = (LPCWSTR)pOdBlock->comments().c_str();
        if (psz[0]){
          lcPropPutStr( hBlock, LC_PROP_BLOCK_DESC, psz );   // Comments
        }
        lcPropPutInt( hBlock, LC_PROP_BLOCK_UNITS, pOdBlock->blockInsertUnits() );     // Block Insert Units
        Pnt3d = pOdBlock->origin();                           // Origin point
        lcPropPutFloat( hBlock, LC_PROP_BLOCK_X, Pnt3d.x );
        lcPropPutFloat( hBlock, LC_PROP_BLOCK_Y, Pnt3d.y );
        if (bStdBlock){
          if (pOdBlock->blockScaling() == OdDbBlockTableRecord::kUniform){
            lcPropPutBool( hBlock, LC_PROP_BLOCK_UFSCALING, true );   // 1 - Block references can be scaled only uniformly.
          }else{
            lcPropPutBool( hBlock, LC_PROP_BLOCK_UFSCALING, false );  // 0 - Block references can be non-uniformly scaled.
          }
        }
      }
    }else{

      // XREF block
      hBlock = lcDrwAddBlock( hLcDb, szBlockName, 0,0 );
      if (hBlock){
        SetObjHandle( pOdBlock->objectId(),  hBlock, LC_PROP_TABLE_ID );    
        lcPropPutStr( hBlock, LC_PROP_TABLE_ODA_HANDLE, (LPCWSTR)szBlockHandle );  // save ODA block handle in Litecad block
//        pBlock->SetHidden( pOdBlock->isAnonymous() );         // Anonymous
        if (pOdBlock->blockScaling() == OdDbBlockTableRecord::kUniform){
          lcPropPutBool( hBlock, LC_PROP_BLOCK_UFSCALING, true );   // 1 - Block references can be scaled only uniformly.
        }else{
          lcPropPutBool( hBlock, LC_PROP_BLOCK_UFSCALING, false );  // 0 - Block references can be non-uniformly scaled.
        }
        lcPropPutInt( hBlock, LC_PROP_BLOCK_UNITS, pOdBlock->blockInsertUnits() );     // Block Insert Units
        Pnt3d = pOdBlock->origin();                           // Origin point
        lcPropPutFloat( hBlock, LC_PROP_BLOCK_X, Pnt3d.x );
        lcPropPutFloat( hBlock, LC_PROP_BLOCK_Y, Pnt3d.y );
//        lcPropPutFloat( hBlock, LC_PROP_BLOCK_Z, Pnt3d.z );
        wcscpy( szPath, (LPCWSTR)pOdBlock->pathName().c_str() );       // Xref Path
        bAttach  = pOdBlock->isFromExternalReference();      // From Xref Attach
        bOverlay = pOdBlock->isFromOverlayReference();       // From Xref Overlay
        bLoaded  = pOdBlock->isUnloaded()? false : true;     // Xref Unloaded
        swprintf( szBuf, L"XREF: %s ", szPath );
        wcscat( szBuf, (LPCWSTR)pOdBlock->comments().c_str() );
        lcPropPutStr( hBlock, LC_PROP_BLOCK_DESC, szBuf );   // Comments
      }
    }
  }
}


//-----------------------------------------------
void Read_Layouts (OdDbDatabasePtr pOdDb, HANDLE hLcDb)
{
  OdDbViewportTablePtr       pVportTable;
  OdDbViewportTableRecordPtr pVportRec;
  OdDbDictionaryPtr          pLayouts;
  OdDbDictionaryIteratorPtr  pIter;
  OdDbBlockTableRecordPtr    pOdBlock, pOdModelBlock; 
  OdDbObjectId    VportId, LayoutId, BlockId, ModelBlockId, OdId;
  OdDbLayoutPtr   pLayout; 
  OdDbViewportPtr pOverallViewport;
  OdGePoint2d     viewCenter;
  OdChar          szOdBuf[32];
  OdInt16         nLayouts;
  PlotRotation    PaperRot;
  LPCWSTR psz = NULL;
  double  W, H, X, Y, MarginL, MarginB, MarginR, MarginT, PaperW, PaperH;
  int     PaperOrient;
  HANDLE  hBlock;

  nLayouts = pOdDb->countLayouts();
  ModelBlockId = pOdDb->getModelSpaceId();

  // Model space layout
  pVportTable = pOdDb->getViewportTableId().safeOpenObject();
  VportId = pVportTable->getActiveViewportId();
  pVportRec = VportId.safeOpenObject();
  viewCenter =  pVportRec->centerPoint();
  H = pVportRec->height(); 
  W = pVportRec->width();
  hBlock = lcPropGetHandle( hLcDb, LC_PROP_DRW_BLOCK_MODEL );
  lcBlockSetViewRect( hBlock, viewCenter.x, viewCenter.y, W, H ); 

  // Paper space layouts
  pLayouts = pOdDb->getLayoutDictionaryId().safeOpenObject();
  pIter = pLayouts->newIterator();
  while( !pIter->done() ){
    LayoutId = pIter->objectId();
    pLayout = LayoutId.safeOpenObject();
    psz = (LPCWSTR)pLayout->getLayoutName().c_str();
    BlockId = pLayout->getBlockTableRecordId();
    pOdBlock = BlockId.safeOpenObject();
    if (pOdBlock.isNull() == false){
      if (BlockId != ModelBlockId){
        BlockId.getHandle().getIntoAsciiBuffer( szOdBuf );
        hBlock = lcDrwGetObjectByIDH( hLcDb, LC_OBJ_BLOCK, (LPCWSTR)szOdBuf );
        if (hBlock){

          // set view area
          OdId = pLayout->overallVportId();
          if (OdId.isValid()){
            pOverallViewport = OdId.safeOpenObject();
            if (pOverallViewport.isNull() == false){
              viewCenter = pOverallViewport->viewCenter();
              H = pOverallViewport->viewHeight();
              W = H / 10.0;
              lcBlockSetViewRect( hBlock, viewCenter.x, viewCenter.y, W, H ); 
              // save handle of the overall viewport
              OdId.getHandle().getIntoAsciiBuffer( szOdBuf );
              lcPropPutStr( hBlock, LC_PROP_BLOCK_ODA_VPORTHANDLE, (LPCWSTR)szOdBuf );  
            }
          }

          // set paper sheet
          pLayout->getPlotOrigin( X, Y );
          pLayout->getPlotPaperMargins( MarginL, MarginB, MarginR, MarginT );
          pLayout->getPlotPaperSize( PaperW, PaperH );
          PaperRot = pLayout->plotRotation();
          if (PaperRot == PlotRotation::k0degrees){
            PaperOrient = LC_PAPER_BOOK;
          }else{
            PaperOrient = LC_PAPER_ALBUM;
            W = PaperW;
            PaperW = PaperH;
            PaperH = W;
          }
          lcBlockSetPaperSize( hBlock, LC_PAPER_CUSTOM, PaperOrient, PaperW, PaperH );
          X = X + MarginL;
          Y = Y + MarginB;
          lcPropPutFloat( hBlock, LC_PROP_PAPER_X0, X );
          lcPropPutFloat( hBlock, LC_PROP_PAPER_Y0, Y );
        }
      }
    }
    pIter->next();
  }
}

//-----------------------------------------------
void Read_Entities (OdDbDatabasePtr pOdDb, HANDLE hLcDb)
{
  OdDbObjectId               BlockId;
  OdDbBlockTableRecordPtr    pOdBlock;
  OdDbBlockTablePtr          pTable = pOdDb->getBlockTableId().safeOpenObject();
  OdDbSymbolTableIteratorPtr pBlkIter = pTable->newIterator();
  OdDbObjectIteratorPtr      pEntIter;
  OdDbEntityPtr              pOdEntity;
  OdDbSortentsTablePtr       pOdSortEnts;
  OdChar                     szBlockHandle[32];
  HANDLE  hBlock;
  WCHAR   szBlockName[256];
  int     TotalNumEnts = 0;
#ifdef _DEBUG
  int iEnt = 0;
#endif

  if (g_pFunProgress){
    for (pBlkIter->start(); ! pBlkIter->done(); pBlkIter->step()){
      pOdBlock = pBlkIter->getRecordId().safeOpenObject();
      pEntIter = pOdBlock->newIterator();
      for (; !pEntIter->done(); pEntIter->step()){
        ++TotalNumEnts;
      }
    }
    (g_pFunProgress)( LC_FP_NITEMS, TotalNumEnts, g_hDrw );
    TotalNumEnts = 0;
  }
  for (pBlkIter->start(); ! pBlkIter->done(); pBlkIter->step()){
    BlockId = pBlkIter->getRecordId();
    pOdBlock = BlockId.safeOpenObject();
    if (pOdBlock->xrefStatus() == OdDb::kXrfNotAnXref){  // Xref Status
      // the block is not XREF
      wcscpy( szBlockName, (LPCWSTR)pOdBlock->getName().c_str() );      // Block Name
      BlockId.getHandle().getIntoAsciiBuffer( szBlockHandle );          // Block Handle
      hBlock = lcDrwGetObjectByIDH( hLcDb, LC_OBJ_BLOCK, (LPCWSTR)szBlockHandle );
      if (hBlock){
        g_NumEnts = 0;
        pEntIter = pOdBlock->newIterator();
#ifdef _DEBUG
        iEnt = 0;
#endif
        for (; !pEntIter->done(); pEntIter->step()) {
          pOdEntity = pEntIter->entity( OdDb::kForRead, false );
          Read_Entity( pOdEntity, hLcDb, hBlock ); 
          if (g_pFunProgress){
            (g_pFunProgress)( LC_FP_ITEM, 0, g_hDrw );
          }
#ifdef _DEBUG
          ++iEnt;
/*
          if (iEnt == 232100){
            iEnt = iEnt;
          }
*/
#endif
        }
        pOdSortEnts = pOdBlock->getSortentsTable( false );
        if (pOdSortEnts.isNull() == false){
          Read_SortEnts( pOdSortEnts, hLcDb, hBlock );
        }
#ifdef _DEBUG
        TotalNumEnts += g_NumEnts;
#endif
      }
    }
  }
}

