/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Read AutoCAD drawing
* Block refs and attributes
*********************************************************************/
#include "pch.h"


//-----------------------------------------------
void Read_Attdef (const OdDbAttributeDefinitionPtr pOdAttdef, HANDLE hLcDb, HANDLE hBlock)
{
  HANDLE     hStyle;
  HANDLE     hAttdef;
  WCHAR      szTag[256], szPrompt[256], szDefVal[4096];
  CGe3dPoint Pt0, Pta;
  double     X, Y, Z, H, WScale, Angle, Oblique;
  int        AlignHor, AlignVer;
  int        Align = LC_TA_LEFBOT;
  int        Mode = 0;
  int        len = pOdAttdef->tag().getLength();

  if (len < 1){
    // string is empty, don't add it
    return;
  }
  hStyle = DrwGetObjectByHandle( pOdAttdef->textStyle(), hLcDb, LC_OBJ_TEXTSTYLE );
  if (hStyle == NULL){
    hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_TEXTSTYLE );  // "Standard" text style
  }
  // Flags
  if (pOdAttdef->isInvisible()){
    Mode |= LC_ATTRIB_HIDDEN;
  }
  if (pOdAttdef->isConstant()){
    Mode |= LC_ATTRIB_CONST;
  }
  if (pOdAttdef->isVerifiable()){
    Mode |= LC_ATTRIB_VERIFY;
  }
  if (pOdAttdef->isPreset()){
    Mode |= LC_ATTRIB_PRESET;
  }
  if (pOdAttdef->lockPositionInBlock()){
    Mode |= LC_ATTRIB_LOCK;
  }
  if (pOdAttdef->isMTextAttributeDefinition()){
    Mode |= LC_ATTRIB_MTEXT;
  }
  wcscpy( szTag, (LPCWSTR)pOdAttdef->tag().c_str() );
  wcscpy( szPrompt, (LPCWSTR)pOdAttdef->prompt().c_str() );
  wcscpy( szDefVal, (LPCWSTR)pOdAttdef->textString().c_str() );
//  len = wcslen( szDefVal );
  Pt0.Set( pOdAttdef->position().x, pOdAttdef->position().y, pOdAttdef->position().z );
  Pta.Set( pOdAttdef->alignmentPoint().x, pOdAttdef->alignmentPoint().y, pOdAttdef->alignmentPoint().z );
  WScale   = pOdAttdef->widthFactor();  // width factor
  H        = pOdAttdef->height();
  Angle    = pOdAttdef->rotation();
  Oblique  = pOdAttdef->oblique();
  AlignHor = pOdAttdef->horizontalMode();
  AlignVer = pOdAttdef->verticalMode();
  switch( AlignHor ){   // horizontal alignment, 0=Left;  1=Center; 2=Right; 3=Align; 4=Middle; 5=Fit
    case 0:
      switch( AlignVer ){ // vertical alignment, 0 = Baseline; 1 = Bottom; 2 = Middle; 3 = Top
        case 0: 
        case 1: Align = LC_TA_LEFBOT; break;
        case 2: Align = LC_TA_LEFCEN; break;
        case 3: Align = LC_TA_LEFTOP; break;
      }
      break;
    case 1:
      switch( AlignVer ){
        case 0: 
        case 1: Align = LC_TA_CENBOT; break;
        case 2: Align = LC_TA_CENTER; break;
        case 3: Align = LC_TA_CENTOP; break;
      }
      break;
    case 2:
      switch( AlignVer ){
        case 0: 
        case 1: Align = LC_TA_RIGBOT; break;
        case 2: Align = LC_TA_RIGCEN; break;
        case 3: Align = LC_TA_RIGTOP; break;
      }
      break;
    case 3:  Align = LC_TA_ALIGNED;  break;
    case 4:  Align = LC_TA_CENTER;  break;
    case 5:  Align = LC_TA_FIT;  break;
  }

  hAttdef = lcBlockAddAttDef( hBlock, Mode, szTag, szPrompt, szDefVal, 0,0, LC_TA_LEFBOT, H, WScale, Angle, Oblique );
  if (hAttdef){
//    lcPropPutFloat( hAttdef, LC_PROP_ENT_Z, Z );
    Read_EntBase( pOdAttdef, hAttdef, hLcDb );
    lcPropPutHandle( hAttdef, LC_PROP_ATT_TSTYLE, hStyle );
    lcPropPutBool(  hAttdef, LC_PROP_ATT_BACKWARD, pOdAttdef->isMirroredInX() );
    lcPropPutBool(  hAttdef, LC_PROP_ATT_UPDOWN, pOdAttdef->isMirroredInY() ); 
    if (Align == LC_TA_LEFBOT){
      lcPropPutFloat( hAttdef, LC_PROP_ATT_X, Pt0.x );
      lcPropPutFloat( hAttdef, LC_PROP_ATT_Y, Pt0.y );
      lcPropPutFloat( hAttdef, LC_PROP_ENT_Z, Pt0.z );
    }else{
      if (Align == LC_TA_ALIGNED || Align == LC_TA_FIT){
        lcPropPutFloat( hAttdef, LC_PROP_ATT_X, Pt0.x );
        lcPropPutFloat( hAttdef, LC_PROP_ATT_Y, Pt0.y );
        lcPropPutFloat( hAttdef, LC_PROP_ENT_Z, Pt0.z );
        lcPropPutFloat( hAttdef, LC_PROP_ATT_XFIT, Pta.x );
        lcPropPutFloat( hAttdef, LC_PROP_ATT_YFIT, Pta.y );
      }else{
        lcPropPutFloat( hAttdef, LC_PROP_ATT_X, Pta.x );
        lcPropPutFloat( hAttdef, LC_PROP_ATT_Y, Pta.y );
        lcPropPutFloat( hAttdef, LC_PROP_ENT_Z, Pta.z );
      }
      lcPropPutInt( hAttdef, LC_PROP_ATT_ALIGN, Align );
    }
    X = pOdAttdef->normal().x;
    Y = pOdAttdef->normal().y;
    Z = pOdAttdef->normal().z;
    if (X!=0.0 || Y!=0.0 || Z!= 1.0){
//      lcEntPutNormal( hAttdef, X, Y, Z );
    }
  }
}

//-----------------------------------------------
void Read_Attrib (const OdDbAttributePtr pOdAttr, HANDLE hLcDb, HANDLE hBlock, HANDLE hBlockRef, const OdDbBlockReferencePtr pOdBlkRef)
{
  HANDLE      hAttrib;
  WCHAR       szTag[256], szValue[4096];
  CGe3dPoint  Pos;
//  UINT        idStyle;
  HANDLE      hStyle;
  CGe3dPoint  Pt0, Pta;
  double      H, WScale, Angle, Oblique, X, Y, Z;
  int         AlignHor, AlignVer;
  int         Align = LC_TA_LEFBOT;
  int         Mode = 0;

  double      coef;
  CGe3dPoint  Off, Off2, OrigPt;
  CGe3dPoint  BlkPt( pOdBlkRef->position().x, pOdBlkRef->position().y, pOdBlkRef->position().z );
  double      BlkAngle = pOdBlkRef->rotation();
  OdGeScale3d BlkScale = pOdBlkRef->scaleFactors();

  wcscpy( szTag, (LPCWSTR)pOdAttr->tag().c_str() );
  wcscpy( szValue, (LPCWSTR)pOdAttr->textString().c_str() );
  Pos.Set( pOdAttr->position().x, pOdAttr->position().y, pOdAttr->position().z );
  hAttrib = lcBlkRefAddAtt( hBlockRef, szTag, szValue );
  if (hAttrib){
    lcPropPutFloat( hAttrib, LC_PROP_ENT_Z,  pOdAttr->position().z );
    Read_EntBase( pOdAttr, hAttrib, hLcDb );
    hStyle = DrwGetObjectByHandle( pOdAttr->textStyle(), hLcDb, LC_OBJ_TEXTSTYLE );
    if (hStyle == NULL){
      hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_TEXTSTYLE );  // "Standard" text style
    }
    // Flags
    if (pOdAttr->isInvisible()){
      Mode |= LC_ATTRIB_HIDDEN;
    }
    if (pOdAttr->isConstant()){
      Mode |= LC_ATTRIB_CONST;
    }
    if (pOdAttr->isVerifiable()){
      Mode |= LC_ATTRIB_VERIFY;
    }
    if (pOdAttr->isPreset()){
      Mode |= LC_ATTRIB_PRESET;
    }
    if (pOdAttr->lockPositionInBlock()){
      Mode |= LC_ATTRIB_LOCK;
    }
    if (pOdAttr->isMTextAttribute()){
      Mode |= LC_ATTRIB_MTEXT;
    }
    Pt0.Set( pOdAttr->position().x, pOdAttr->position().y, pOdAttr->position().z );
    Pta.Set( pOdAttr->alignmentPoint().x, pOdAttr->alignmentPoint().y, pOdAttr->alignmentPoint().z );
    coef = BlkScale.sy / BlkScale.sx;
    WScale = pOdAttr->widthFactor() * coef;  // width factor
    H = pOdAttr->height() / BlkScale.sy;
    if (H < 0.0){
      H = -H;
    }
    Angle    = pOdAttr->rotation() - BlkAngle;
    Oblique  = pOdAttr->oblique();
    AlignHor = pOdAttr->horizontalMode();
    AlignVer = pOdAttr->verticalMode();
    switch( AlignHor ){   // horizontal alignment, 0=Left;  1=Center; 2=Right; 3=Align; 4=Middle; 5=Fit
      case 0:
        switch( AlignVer ){ // vertical alignment, 0 = Baseline; 1 = Bottom; 2 = Middle; 3 = Top
          case 0: 
          case 1: Align = LC_TA_LEFBOT; break;
          case 2: Align = LC_TA_LEFCEN; break;
          case 3: Align = LC_TA_LEFTOP; break;
        }
        break;
      case 1:
        switch( AlignVer ){
          case 0: 
          case 1: Align = LC_TA_CENBOT; break;
          case 2: Align = LC_TA_CENTER; break;
          case 3: Align = LC_TA_CENTOP; break;
        }
        break;
      case 2:
        switch( AlignVer ){
          case 0: 
          case 1: Align = LC_TA_RIGBOT; break;
          case 2: Align = LC_TA_RIGCEN; break;
          case 3: Align = LC_TA_RIGTOP; break;
        }
        break;
      case 3:  Align = LC_TA_ALIGNED;  break;
      case 4:  Align = LC_TA_CENTER;  break;
      case 5:  Align = LC_TA_FIT;  break;
    }

    lcPropPutInt( hAttrib, LC_PROP_ATT_MODE, Mode ); 
    lcPropPutFloat( hAttrib, LC_PROP_ATT_WSCALE, WScale ); 
    lcPropPutFloat( hAttrib, LC_PROP_ATT_H, H );
    lcPropPutFloat( hAttrib, LC_PROP_ATT_ANGLE, Angle );
    lcPropPutFloat( hAttrib, LC_PROP_ATT_OBLIQUE, Oblique );
    lcPropPutHandle( hAttrib, LC_PROP_ATT_TSTYLE, hStyle );
    lcPropPutBool(  hAttrib, LC_PROP_ATT_BACKWARD, pOdAttr->isMirroredInX() );
    lcPropPutBool(  hAttrib, LC_PROP_ATT_UPDOWN, pOdAttr->isMirroredInY() ); 
    if (Align == LC_TA_LEFBOT){
      Off = Pt0 - BlkPt;
    }else{
      if (Align == LC_TA_ALIGNED || Align == LC_TA_FIT){
        Off = Pt0 - BlkPt;
        Off2 = Pta - BlkPt;
        if (-0.0001 < Off2.x && Off2.x < 0.0001){
          Off2.x = 0.0;
        }
        if (-0.0001 < Off2.y && Off2.y < 0.0001){
          Off2.y = 0.0;
        }
        if (-0.0001 < Off2.z && Off2.z < 0.0001){
          Off2.z = 0.0;
        }
        Off2.Rotate( OrigPt, -BlkAngle, EnPlane::GE_PLANE_TOP );
        Off2.Scale( OrigPt, 1.0/BlkScale.sx, 1.0/BlkScale.sy, 1.0/BlkScale.sz );
        lcPropPutFloat( hAttrib, LC_PROP_ATT_XFIT, Off2.x );
        lcPropPutFloat( hAttrib, LC_PROP_ATT_YFIT, Off2.y );
      }else{
        Off = Pta - BlkPt;
      }
    }
    if (-0.0001 < Off.x && Off.x < 0.0001){
      Off.x = 0.0;
    }
    if (-0.0001 < Off.y && Off.y < 0.0001){
      Off.y = 0.0;
    }
    if (-0.0001 < Off.z && Off.z < 0.0001){
      Off.z = 0.0;
    }
    Off.Rotate( OrigPt, -BlkAngle, EnPlane::GE_PLANE_TOP );
    Off.Scale( OrigPt, 1.0/BlkScale.sx, 1.0/BlkScale.sy, 1.0/BlkScale.sz );
    lcPropPutFloat( hAttrib, LC_PROP_ATT_X, Off.x );
    lcPropPutFloat( hAttrib, LC_PROP_ATT_Y, Off.y );
    lcPropPutFloat( hAttrib, LC_PROP_ENT_Z, Off.z );
    if (Align != LC_TA_LEFBOT){
      lcPropPutInt( hAttrib, LC_PROP_ATT_ALIGN, Align );
    }
    X = pOdAttr->normal().x;
    Y = pOdAttr->normal().y;
    Z = pOdAttr->normal().z;
    if (X!=0.0 || Y!=0.0 || Z!= 1.0){
//      lcEntPutNormal( hAttrib, X, Y, Z );
    }
  }
}


//-----------------------------------------------
void Read_BlockRef (OdDbBlockReferencePtr pOdBlkRef, HANDLE hLcDb, HANDLE hBlock)
{
  double  Angle, X, Y, Z, nx, ny, nz;
  bool    bAttribs = false;
  int     iAtt = 0;
//  UINT    idRefBlock;
  OdChar  szIdRefBlock[32];
  HANDLE  hInsert;
  OdGeScale3d           odScale;
  OdDbObjectId          odId;
  OdDbAttributePtr      pOdAttr;
  OdDbObjectIteratorPtr pIter;

  X = pOdBlkRef->position().x;
  Y = pOdBlkRef->position().y;
  Z = pOdBlkRef->position().z;
  odScale = pOdBlkRef->scaleFactors();
  Angle   = pOdBlkRef->rotation();
//  OdHandleToLcID( pOdBlkRef->blockTableRecord(), &idRefBlock );
  pOdBlkRef->blockTableRecord().getHandle().getIntoAsciiBuffer( szIdRefBlock );
  hInsert = lcBlockAddBlockRefIDH( hBlock, (LPCWSTR)szIdRefBlock, X, Y, odScale.sx, Angle );
  if (hInsert){
    lcPropPutFloat( hInsert, LC_PROP_ENT_Z, Z );
    Read_EntBase( pOdBlkRef, hInsert, hLcDb ); //, hBlock );
    nx = pOdBlkRef->normal().x;
    ny = pOdBlkRef->normal().y;
    nz = pOdBlkRef->normal().z;
    if (nx!=0.0 || ny!=0.0 || nz!= 1.0){
      iAtt = iAtt;
//      lcEntPutNormal( hInsert, nx, ny, nz );
    }
    if (odScale.sx != odScale.sy){
      lcPropPutFloat( hInsert, LC_PROP_BLKREF_SCX, odScale.sx );
      lcPropPutFloat( hInsert, LC_PROP_BLKREF_SCY, odScale.sy );
//      lcPropPutFloat( hInsert, LC_PROP_BLKREF_SCZ, odScale.sz );
    }
    // attributes
    pIter = pOdBlkRef->attributeIterator();
    for (pIter->start(); !pIter->done(); pIter->step()) {
      bAttribs = true;
      odId = pIter->objectId();
      pOdAttr = pOdBlkRef->openAttribute( odId, OdDb::kForRead );
      Read_Attrib( pOdAttr, hLcDb, hBlock, hInsert, pOdBlkRef );
/*
      if (iAtt == 23){   // used to debug
        iAtt = iAtt;
      }
*/
      ++iAtt;
    }
  }
}

//-----------------------------------------------
void Read_MInsert (OdDbMInsertBlockPtr pOdMInsert, HANDLE hLcDb, HANDLE hBlock)
{
  double  Angle, X, Y, Z, nx, ny, nz;
  bool    bAttribs = false;
  int     iAtt = 0;
//  UINT    idRefBlock;
  OdChar  szIdRefBlock[32];
  HANDLE  hInsert;
  OdGeScale3d           odScale;
  OdDbObjectId          odId;
  OdDbAttributePtr      pOdAttr;
  OdDbObjectIteratorPtr pIter;

  X = pOdMInsert->position().x;
  Y = pOdMInsert->position().y;
  Z = pOdMInsert->position().z;
  odScale = pOdMInsert->scaleFactors();
  Angle   = pOdMInsert->rotation();
//  OdHandleToLcID( pOdBlkRef->blockTableRecord(), &idRefBlock );
  pOdMInsert->blockTableRecord().getHandle().getIntoAsciiBuffer( szIdRefBlock );
  hInsert = lcBlockAddBlockRefIDH( hBlock, (LPCWSTR)szIdRefBlock, X, Y, odScale.sx, Angle );
  if (hInsert){
    lcPropPutFloat( hInsert, LC_PROP_ENT_Z, Z );
    Read_EntBase( pOdMInsert, hInsert, hLcDb );
    nx = pOdMInsert->normal().x;
    ny = pOdMInsert->normal().y;
    nz = pOdMInsert->normal().z;
    if (nx!=0. || ny!=0. || nz!= 1.){
//      lcEntPutNormal( hInsert, nx, ny, nz );
    }
    if (odScale.sx != odScale.sy){
      lcPropPutFloat( hInsert, LC_PROP_BLKREF_SCX, odScale.sx );
      lcPropPutFloat( hInsert, LC_PROP_BLKREF_SCY, odScale.sy );
//      lcPropPutFloat( hInsert, LC_PROP_BLKREF_SCZ, odScale.sz );
    }
  }
}

//-----------------------------------------------
void Read_Table (const OdDbTablePtr pOdTable, HANDLE hLcDb, HANDLE hBlock)
{
  OdGePoint3d odpt;
  HANDLE hBlkRef;
  OdChar szIdRefBlock[32];

  odpt = pOdTable->position();
  pOdTable->blockTableRecord().getHandle().getIntoAsciiBuffer( szIdRefBlock );
  hBlkRef = lcBlockAddBlockRefIDH( hBlock, (LPCWSTR)szIdRefBlock, odpt.x, odpt.y, 1., 0. );
  if (hBlkRef){
    lcPropPutFloat( hBlkRef, LC_PROP_ENT_Z, odpt.z );
    Read_EntBase( pOdTable, hBlkRef, hLcDb ); //, hBlock );
  }
}
