/********************************************************************
* Project: DWG/DXF plugin for LiteCAD
* Copyright (C) 2020-2021, Oleg Kolbaskin
* All rights reserved.
*
* Read AutoCAD drawing
* Lines and polylines
*********************************************************************/
#include "pch.h"

extern COptions   g_Options;

bool geIsLine4p (CGePoint* Pnt);

//-----------------------------------------------
HANDLE Read_Line (const OdDbLinePtr pOdLine, HANDLE hLcDb, HANDLE hBlock)
{
  double X0 = pOdLine->startPoint().x;
  double Y0 = pOdLine->startPoint().y;
  double Z0 = pOdLine->startPoint().z;
  double X1 = pOdLine->endPoint().x;
  double Y1 = pOdLine->endPoint().y;
  double Z1 = pOdLine->endPoint().z;
  HANDLE hLine = lcBlockAddLine( hBlock, X0, Y0, X1, Y1 );
  if (hLine){
    lcPropPutFloat( hLine, LC_PROP_ENT_Z, Z0 );
    lcPropPutFloat( hLine, LC_PROP_ENT_THICKNESS, pOdLine->thickness() );
    lcPropPutFloat( hLine, LC_PROP_LINE_Z0, Z0 );
    lcPropPutFloat( hLine, LC_PROP_LINE_Z1, Z1 );
    Read_EntBase( pOdLine, hLine, hLcDb ); //, hBlock );
  }
  return hLine;
}

//-----------------------------------------------
void Read_XLine (const OdDbXlinePtr pOdXline, HANDLE hLcDb, HANDLE hBlock)
{
  double X    = pOdXline->basePoint().x;
  double Y    = pOdXline->basePoint().y;
  double Z    = pOdXline->basePoint().z;
  double DirX = pOdXline->unitDir().x;
  double DirY = pOdXline->unitDir().y;
  double DirZ = pOdXline->unitDir().z;
  CGePoint Pt;
  if (DirX!=0.0 || DirY!=0.0){
    Pt.Set( DirX, DirY );
    HANDLE hXline = lcBlockAddXline( hBlock, X, Y, Pt.Angle(), false );
    if (hXline){
      lcPropPutFloat( hXline, LC_PROP_ENT_Z, Z );
      Read_EntBase( pOdXline, hXline, hLcDb );
    }
  }
}

//-----------------------------------------------
void Read_Ray (const OdDbRayPtr pOdRay, HANDLE hLcDb, HANDLE hBlock)
{
  double X    = pOdRay->basePoint().x;
  double Y    = pOdRay->basePoint().y;
  double Z    = pOdRay->basePoint().z;
  double DirX = pOdRay->unitDir().x;
  double DirY = pOdRay->unitDir().y;
  double DirZ = pOdRay->unitDir().z;
  CGePoint Pt;
  if (DirX!=0.0 || DirY!=0.0){
    Pt.Set( DirX, DirY );
    HANDLE hRay = lcBlockAddXline( hBlock, X, Y, Pt.Angle(), true );
    if (hRay){
      lcPropPutFloat( hRay, LC_PROP_ENT_Z, Z );
      Read_EntBase( pOdRay, hRay, hLcDb );
    }
  }
}

//-----------------------------------------------
bool PlineToCircle (const OdDbPolylinePtr pOdPline, CGeCircle* pCirc)
{
  bool     bClosed;
  CGeArc   Arc;
  CGePoint Pt0, Pt1, CenPt;
  double   R, Rad, Len, Delta, Bulge;
  bool     bExit = false;
  UINT     i, nVers;
  UINT_PTR iB = 0;
  UINT_PTR j = 0;
  OdGePoint2d odpt, odpt2;

  Len = Rad = Delta = 0.0;
  nVers = pOdPline->numVerts();
  bClosed = pOdPline->isClosed();
  for (i=0; i<nVers; ++i){
    pOdPline->getPointAt( i, odpt );
    Pt0.Set( odpt.x, odpt.y );
    if ((i+1) < nVers){
      pOdPline->getPointAt( i+1, odpt2 );
    }else{
      if (bClosed){
        pOdPline->getPointAt( 0, odpt2 );
      }else{
        break;
      }
    }
    Pt1.Set( odpt2.x, odpt2.y );
    Bulge = pOdPline->getBulgeAt( i );
    if (Bulge != 0.0){
      if (Arc.SetSBE( Pt0, Bulge, Pt1 )){
        R = Arc.Radius();
        Pt0 = Arc.Center();
        if (iB == 0){
          CenPt = Pt0;
          Rad = R;
          Delta = Rad / 100.0;
          ++j;
        }else{
          if ((Rad-Delta)<R && R<(Rad+Delta)){
            if (CenPt.IsEqual(Pt0,Delta)){
              ++j;
              goto m1;
            }
          }
          return false;
        }
m1:
        ++iB;
      }else{
        return false;
      }
    }else{
      Len += Pt0.Distance( Pt1 );
    }
  }
  if (j >= 2){
    if (Len < Delta){
      if (!bClosed){
        pOdPline->getPointAt( 0, odpt );
        Pt0.Set( odpt.x, odpt.y );
//        Pt1.Set( pVer->m_Pnt.x, pVer->m_Pnt.y );
        if (Pt0.IsEqual( Pt1, Delta ) == false){
          return false;
        }
      }
      pCirc->Set( CenPt, Rad );
      return true;
    }
  }
  return false;
}

//-----------------------------------------------
HANDLE Read_LwPolyline (const OdDbPolylinePtr pOdPline, HANDLE hLcDb, HANDLE hBlock)
{
  HANDLE      hPline = NULL;
  int         i, np;
  bool        bClosed, bHasBulges, bHasWidths;
  double      Bulge, X, Y, Z, StartWidth, EndWidth, ConstWidth, b0, b1, rad;
  CGePoint    Pt0, Pt1, Ptc;
  CGe3dPoint  Pt3d, Xaxis, Yaxis, Zaxis;
  OdGePoint2d odpt, odpt2;
  HANDLE      hVer, hCirc;
  CGeCircle   Circ;

  double nX = pOdPline->normal().x;
  double nY = pOdPline->normal().y;
  double nZ = pOdPline->normal().z;
  if (nX==0.0 && nY==0.0 && nZ==1.0){
  }else{
    i = 0;
  }
//  Zaxis.Set( nX, nY, nZ );

  Zaxis.Set( pOdPline->normal().x, pOdPline->normal().y, pOdPline->normal().z );
  Zaxis.GetArbitraryAxes( &Xaxis, &Yaxis );
//  Pt3d.SetNormal( nX, nY, nZ );
  i = sizeof( Pt3d );
  bClosed     = pOdPline->isClosed();
  bHasBulges  = pOdPline->hasBulges();
  bHasWidths  = pOdPline->hasWidth();
  ConstWidth  = pOdPline->getConstantWidth();
  Z           = pOdPline->elevation();
  np = pOdPline->numVerts();
  if (np >= 2){
    if (g_Options.m_Imp_bPlineToCircle && bHasBulges){
      if (np == 2 && bClosed){
        pOdPline->getPointAt( 0, odpt );
        Pt0.Set( odpt.x, odpt.y );
        b0 = pOdPline->getBulgeAt( 0 );
        pOdPline->getPointAt( 1, odpt );
        Pt1.Set( odpt.x, odpt.y );
        b1 = pOdPline->getBulgeAt( 1 );
        if ((0.99 < b0) && (b0 < 1.01) && (0.99 < b1) && (b1 < 1.01)){
          rad = Pt0.Distance( Pt1 );
          rad -= rad * 0.01;
          if (ConstWidth > rad){
            // create filled circle
            Ptc = Pt0.Center( Pt1 );
            hCirc = lcBlockAddCircle( hBlock, Ptc.x, Ptc.y, rad, true );
            if (hCirc){
              lcPropPutFloat( hCirc, LC_PROP_ENT_Z, Z );
              lcPropPutFloat( hCirc, LC_PROP_ENT_THICKNESS, pOdPline->thickness() );
              Read_EntBase( pOdPline, hCirc, hLcDb );
              X = pOdPline->normal().x;
              Y = pOdPline->normal().y;
              Z = pOdPline->normal().z;
              if (X!=0.0 || Y!=0.0 || Z!= 1.0){
//                lcEntPutNormal( hCirc, X, Y, Z );
              }
              return hCirc;
            }
          }
        }
      }
      if (PlineToCircle( pOdPline, &Circ )){
        hCirc = lcBlockAddCircle( hBlock, Circ.m_Center.x, Circ.m_Center.y, Circ.Radius(), false );
        if (hCirc){
          lcPropPutFloat( hCirc, LC_PROP_ENT_Z, Z );
          lcPropPutFloat( hCirc, LC_PROP_ENT_THICKNESS, pOdPline->thickness() );
          Read_EntBase( pOdPline, hCirc, hLcDb );
          return hCirc;
        }
      }
    }
//    odpt0.set( -1234567.124567, 7654321.7654321 ); //, 98765401.2634 );
    hPline = lcBlockAddPolyline( hBlock, 0, bClosed, false );
    if (hPline){
      Read_EntBase( pOdPline, hPline, hLcDb ); //, hBlock );
      lcPropPutFloat( hPline, LC_PROP_ENT_Z, Z );
      lcPropPutFloat( hPline, LC_PROP_ENT_THICKNESS, pOdPline->thickness() );
      lcPropPutFloat( hPline, LC_PROP_PLINE_WIDTH, ConstWidth );
      for (i=0; i<np; i++){
        pOdPline->getPointAt( i, odpt );
        if ((i+1) < np){
          pOdPline->getPointAt( i+1, odpt2 );
          if (odpt == odpt2){  
            // pass duplicate point
            continue;
          }
        }
//        if (odpt != odpt0){  // forbid duplicate points
          Pt3d.Set( odpt.x, odpt.y, Z );
          Pt3d.Transform( Xaxis, Yaxis, Zaxis );
          if (bHasBulges){
            Bulge = pOdPline->getBulgeAt( i );
          }else{
            Bulge = 0.0;
          }
          if (bHasWidths){
            pOdPline->getWidthsAt( i, StartWidth, EndWidth );
          }else{
            StartWidth = EndWidth = ConstWidth;
          }
          hVer = lcPlineAddVer2( hPline, NULL, Pt3d.x, Pt3d.y, Bulge, StartWidth, EndWidth );
          lcPropPutFloat( hVer, LC_PROP_VER_Z, Pt3d.z );
/*
          hVer = lcPlineAddVer( hPline, NULL, odpt.x, odpt.y, odpt.z );
          if (bHasBulges){
            Bulge = pOdPline->getBulgeAt( i );
            lcPropPutFloat( hVer, LC_PROP_VER_BULGE, Bulge );
          }
          if (bHasWidths){
            pOdPline->getWidthsAt( i, StartWidth, EndWidth );
            lcPropPutFloat( hVer, LC_PROP_VER_W0, StartWidth );
            lcPropPutFloat( hVer, LC_PROP_VER_W1, EndWidth );
          }
*/
//          odpt0 = odpt;
//        }
      }
      X = pOdPline->normal().x;
      Y = pOdPline->normal().y;
      Z = pOdPline->normal().z;
      if (X!=0.0 || Y!=0.0 || Z!= 1.0){
//        lcEntPutNormal( hPline, X, Y, Z );
      }
//      lcEntUpdate( hPline );
    }
  }
  return hPline;
}

//-----------------------------------------------
HANDLE Read_2dPolyline (const OdDb2dPolylinePtr pOdPline, HANDLE hLcDb, HANDLE hBlock)
{
  OdDbObjectIteratorPtr pIter;
  OdDb::Vertex2dType    VerType;
  OdDbEntityPtr         pOdEntity;
  OdDb2dVertexPtr       pOdVertex;
  int            FitType = LC_PLFIT_BULGE;
  int            nVers;
  HANDLE         hPline = NULL;
//  LcDbPolyline*  pPline;
//  LcVertex       Ver;
//  UINT           nVers;
  double       X, Y, Z, StartWidth, EndWidth, Bulge;
  double       ConstWidth = -1000001.;
  bool         bConstWidth = true;
  bool         bClosed = pOdPline->isClosed();
  CGe3dPoint  Pt3d, Xaxis, Yaxis, Zaxis;

  Zaxis.Set( pOdPline->normal().x, pOdPline->normal().y, pOdPline->normal().z );
  Zaxis.GetArbitraryAxes( &Xaxis, &Yaxis );

//  bHasBulges  = pOdPline->hasBulges();
//  bHasWidths  = pOdPline->hasWidth();
//  ConstWidth  = pOdPline->getConstantWidth();
  Z = pOdPline->elevation();
  switch( pOdPline->polyType() ){
    case OdDb::k2dSimplePoly:      FitType = LC_PLFIT_BULGE; break;
    case OdDb::k2dQuadSplinePoly:  FitType = LC_PLFIT_QUAD; break;
    case OdDb::k2dCubicSplinePoly: FitType = LC_PLFIT_CUBIC; break;
    case OdDb::k2dFitCurvePoly:    FitType = LC_PLFIT_SPLINE; break;
  }
  hPline = lcBlockAddPolyline( hBlock, FitType, bClosed, false );
  if (hPline){
    Read_EntBase( pOdPline, hPline, hLcDb ); //, hBlock );
    lcPropPutFloat( hPline, LC_PROP_ENT_Z, Z );
    lcPropPutFloat( hPline, LC_PROP_ENT_THICKNESS, pOdPline->thickness() );
  }else{
    return 0;
  }
  pIter = pOdPline->vertexIterator();
  for (pIter->start(); !pIter->done(); pIter->step()) {
    pOdEntity = pIter->entity( OdDb::kForRead, false );
    if (pOdEntity->isKindOf( OdDb2dVertex::desc() )){
//      memset( &Ver, 0, sizeof(Ver) );
      pOdVertex = OdDb2dVertex::cast( pOdEntity );
      VerType = pOdVertex->vertexType();
      switch( FitType ){
        case LC_PLFIT_BULGE:
//          if (VerType == OdDb::k2dVertex){  // Normal 2D vertex.
            X = pOdVertex->position().x;
            Y = pOdVertex->position().y;
//            z = pOdVertex->position().z;
            Pt3d.Set( X, Y, Z );
            Pt3d.Transform( Xaxis, Yaxis, Zaxis );
            X = Pt3d.x;
            Y = Pt3d.y;
            StartWidth = pOdVertex->startWidth();
            EndWidth   = pOdVertex->endWidth();
            if (StartWidth != EndWidth){
              bConstWidth = false;
            }else{
              if (bConstWidth){
                if (ConstWidth < -1000000.){
                  ConstWidth = StartWidth;
                }else{
                  if (StartWidth != ConstWidth){
                    bConstWidth = false;
                  }
                }
              }
//            }
            Bulge = pOdVertex->bulge();
            lcPlineAddVer2( hPline, NULL, X, Y, Bulge, StartWidth, EndWidth );
          }
          break;

        case LC_PLFIT_QUAD:
        case LC_PLFIT_CUBIC:
          if (VerType == OdDb::k2dSplineCtlVertex){  // Control point for a spline or curve-fit polyline.
            X = pOdVertex->position().x;
            Y = pOdVertex->position().y;
//            z = pOdVertex->position().z;
/*
            StartWidth = pOdVertex->startWidth();
            EndWidth   = pOdVertex->endWidth();
            if (StartWidth != EndWidth){
              bConstWidth = false;
            }else{
              if (Width < -1000000.){
                Width = Ver.m_StartWidth;
                bConstWidth = LC_TRUE;
              }else{
                if (Ver.m_StartWidth != Width){
                  bConstWidth = LC_FALSE;
                }
              }
            }
            pPline->AddVertex( &Ver );
*/
            lcPlineAddVer( hPline, NULL, X, Y );
          }
          break;

        case LC_PLFIT_SPLINE:
          if (VerType == OdDb::k2dVertex || VerType == OdDb::k2dCurveFitVertex){  // Control point for a spline or curve-fit polyline.
            X = pOdVertex->position().x;
            Y = pOdVertex->position().y;
/*
            z = pOdVertex->position().z;
            Ver.m_StartWidth = pOdVertex->startWidth();
            Ver.m_EndWidth   = pOdVertex->endWidth();
            if (Ver.m_StartWidth != Ver.m_EndWidth){
              bConstWidth = LC_FALSE;
            }else{
              if (Width < -1000000.){
                Width = Ver.m_StartWidth;
                bConstWidth = LC_TRUE;
              }else{
                if (Ver.m_StartWidth != Width){
                  bConstWidth = LC_FALSE;
                }
              }
            }
            pPline->AddVertex( &Ver );
*/
            lcPlineAddVer( hPline, NULL, X, Y );
          }
          break;
      }
    }
  }
  if (bConstWidth){
    lcPropPutFloat( hPline, LC_PROP_PLINE_WIDTH, ConstWidth );
  }
  nVers = lcPropGetInt( hPline, LC_PROP_PLINE_NVERS );
#ifdef _DEBUG
  if (nVers == 0){
nVers = 0;
  }
#endif
  return hPline;
}

//-----------------------------------------------
HANDLE Read_3dPolyline (const OdDb3dPolylinePtr pOdPline, HANDLE hLcDb, HANDLE hBlock)
{
  OdDbObjectIteratorPtr pIter;
  OdDb::Vertex3dType    VerType;
  OdDbEntityPtr         pOdEntity;
  OdDb3dPolylineVertexPtr pOdVertex;
  int    FitType = LC_PLFIT_BULGE;
  HANDLE hPline, hVer;
  bool   bClosed = pOdPline->isClosed();
  double X, Y, Z;

  hPline = NULL;
  switch( pOdPline->polyType() ){
    case OdDb::k3dSimplePoly:      FitType = LC_PLFIT_NONE; break;
    case OdDb::k3dQuadSplinePoly:  FitType = LC_PLFIT_QUAD; break;
    case OdDb::k3dCubicSplinePoly: FitType = LC_PLFIT_CUBIC; break;
  }
  hPline = lcBlockAddPolyline( hBlock, FitType, bClosed, false );
  if (hPline){
    Read_EntBase( pOdPline, hPline, hLcDb );
  }else{
    return 0;
  }
  pIter = pOdPline->vertexIterator();
  for (pIter->start(); !pIter->done(); pIter->step()) {
    pOdEntity = pIter->entity( OdDb::kForRead, false );
    if (pOdEntity->isKindOf( OdDb3dPolylineVertex::desc() )){
      pOdVertex = OdDb3dPolylineVertex::cast( pOdEntity );
      VerType = pOdVertex->vertexType();
      if (FitType == LC_PLFIT_NONE){
        if (VerType == OdDb::k3dSimpleVertex || VerType == OdDb::k3dControlVertex){
          X = pOdVertex->position().x;
          Y = pOdVertex->position().y;
          Z = pOdVertex->position().z;
          hVer = lcPlineAddVer( hPline, NULL, X, Y );
          lcPropPutFloat( hVer, LC_PROP_VER_Z, Z );
        }
      }else{
        if (VerType == OdDb::k3dControlVertex){
          X = pOdVertex->position().x;
          Y = pOdVertex->position().y;
          Z = pOdVertex->position().z;
          hVer = lcPlineAddVer( hPline, NULL, X, Y );
          lcPropPutFloat( hVer, LC_PROP_VER_Z, Z );
        }
      }
    }
  }
  return hPline;
}

//-----------------------------------------------
bool IsBezier (const CGePoint* CtrlPnt, UINT nCtrlPnts, CGePolyline* pPline, bool* pbLinear)
{
  UINT n, i, j;
  bool bNoDup = CGePolyline::m_bNoDuplicate;
  bool bLinear = true;
  const CGePoint* pPnt; 
  CGePoint  PtSeg[4];
  double z = 0.0;

  n = nCtrlPnts % 4;
  if (n == 0){
//    if (nKnots == (m_nCtrlPnts+4)){
      j = 0;
      for (i=4; i<nCtrlPnts; i+=4){
        if (CtrlPnt[i].IsEqual( CtrlPnt[i-1] ) == false){
          return false;
        }
        if (bLinear){
          // check if segment is linear
          PtSeg[0].Set( CtrlPnt[i-4].x, CtrlPnt[i-4].y );
          PtSeg[1].Set( CtrlPnt[i-3].x, CtrlPnt[i-3].y );
          PtSeg[2].Set( CtrlPnt[i-2].x, CtrlPnt[i-2].y );
          PtSeg[3].Set( CtrlPnt[i-1].x, CtrlPnt[i-1].y );
          if (geIsLine4p( PtSeg ) == false){
            bLinear = false;
          }
        }
      }
      if (pbLinear){
        *pbLinear = bLinear;
      }
      if (pPline){
        CGePolyline::m_bNoDuplicate = false;
        pPnt = CtrlPnt;
        if (pbLinear && bLinear){
          pPline->AddVer( pPnt->x, pPnt->y, z );
          for (i=3; i<nCtrlPnts; i+=4){
            pPnt = CtrlPnt + i;
            pPline->AddVer( pPnt->x, pPnt->y, z );
          }
        }else{
          j = 0;
          for (i=0; i<nCtrlPnts; ++i){
            if (j == 4){
              j = 1;
            }else{
              pPline->AddVer( pPnt->x, pPnt->y, z );
              ++j;
            }
            ++pPnt;
          }
        }
        CGePolyline::m_bNoDuplicate = bNoDup;
      }
      return true;
//    }
  }
  return false;  
}

//-----------------------------------------------
void Read_Spline (const OdDbSplinePtr pOdSpline, HANDLE hLcDb, HANDLE hBlock)
{
  OdGeVector3d     StartTangent, EndTangent;
  OdGePoint3d      odpt, odpt2;
  OdGePoint3dArray CtrlPoints;
  OdGeDoubleArray  Knots;
  OdGeDoubleArray  Weights;
  OdDbObjectId     idOdObj;
  double   CtrlPtTol, KnotTol, dval;
  int      nCtrPts   = pOdSpline->numControlPoints();
  int      nFitPts   = pOdSpline->numFitPoints();
//  int      nKnots    = pOdSpline->numKnumFitPoints();
  bool     bPlanar   = pOdSpline->isPlanar();
  bool     bClosed   = pOdSpline->isClosed();
  bool     bRational, bPeriodic;
  bool     bHasWeights = false;
  int      Degree    = pOdSpline->degree();
/*
  bool     bRational = pOdSpline->isRational();
  bool     bPeriodic = pOdSpline->isPeriodic();
*/
  int      FitType   = LC_PLFIT_SPLINE;
  int      i, n, n2;
  HANDLE   hPline=0, hVer, hEnt2;
  CGePoint Pnt;
  CGePoint* Pts = NULL;   // for spline control points
  CGePolyline Pline;
  CGeVertex*  pVer;
  bool bLinear;

  OdResult Res;
  OdRxObjectPtrArray EntitySet;
  OdRxObjectPtr pOdEntity;
  OdDb2dPolylinePtr pOd2dPline;
  OdDb3dPolylinePtr pOd3dPline;
  OdDbPolylinePtr   pOdLwPline;
  OdDbLinePtr pOdLine;

  // check for bezier spline
  if (nCtrPts > 0){
    if (bPlanar && Degree==3){
//      if (nKnots == (m_nCtrlPnts+4)){
      // read control points
      Pts = new CGePoint[nCtrPts];
      for (i=0; i<nCtrPts; ++i) {
        pOdSpline->getControlPointAt( i, odpt );
        Pts[i].Set( odpt.x, odpt.y );
      }
      if (IsBezier( Pts, nCtrPts, &Pline, &bLinear )){
        // Bezier curve
        if (bLinear){
          FitType = LC_PLFIT_BULGE;
          hPline = lcBlockAddPolyline( hBlock, FitType, bClosed, false );
          if (hPline){
            Read_EntBase( pOdSpline, hPline, hLcDb );
            pVer = Pline.GetFirstVer();
            while( pVer ){
              hVer = lcPlineAddVer( hPline, NULL, pVer->X(), pVer->Y() );
              lcPropPutFloat( hVer, LC_PROP_VER_Z, pVer->Z() );
              pVer = Pline.GetNextVer( pVer );
            }
            lcPlineEnd( hPline );
          }
        }else{
          hPline = lcBlockAddBezier( hBlock );
          if (hPline){
            Read_EntBase( pOdSpline, hPline, hLcDb );
            pVer = Pline.GetFirstVer();
            while( pVer ){
              hVer = lcBezierAddVer( hPline, NULL, pVer->X(), pVer->Y() );
              pVer = Pline.GetNextVer( pVer );
            }
            lcBezierEnd( hPline );
          }
        }
        delete[] Pts;
        return;
      }
      delete[] Pts;
      Pts = NULL;
    }
  }

  Res = pOdSpline->explodeGeometry( EntitySet );
  if (Res == eOk){
    n = EntitySet.length();
    if (n > 0){
      for (i=0; i<n; ++i){
        hEnt2 = 0;
        pOdEntity = EntitySet.getAt( i );
        if (pOdEntity->isKindOf( OdDb2dPolyline::desc() )){ 
          pOd2dPline = OdDb2dPolyline::cast( pOdEntity );
          hEnt2 = Read_2dPolyline( pOd2dPline, hLcDb, hBlock );
        }else
        if (pOdEntity->isKindOf( OdDb3dPolyline::desc() )){ 
          pOd3dPline = OdDb3dPolyline::cast( pOdEntity );
          hEnt2 = Read_3dPolyline( pOd3dPline, hLcDb, hBlock );
        }else
        if (pOdEntity->isKindOf( OdDbLine::desc() )){                 // LINE
          pOdLine = OdDbLine::cast( pOdEntity );
          hEnt2 = Read_Line( pOdLine, hLcDb, hBlock );
        }else
        if (pOdEntity->isKindOf( OdDbPolyline::desc() )){             // LWPOLYLINE
          pOdLwPline = OdDbPolyline::cast( pOdEntity );
          hEnt2 = Read_LwPolyline( pOdLwPline, hLcDb, hBlock );
        }
        if (hEnt2){
          idOdObj = pOdSpline->objectId();
          SetObjHandle( idOdObj, hEnt2, LC_PROP_ENT_ID );
        }
      }
    }
  }
  return;

  if (nFitPts > 0){
    hPline = lcBlockAddPolyline( hBlock, FitType, bClosed, false );
    if (hPline){
      Read_EntBase( pOdSpline, hPline, hLcDb );
//      pOdSpline->getFitPointAt( 0, odpt0 );
//      n2 = nFitPts - 1;
      for (i=0; i<nFitPts; ++i) {
        pOdSpline->getFitPointAt( i, odpt );
/*
        if (bClosed){
          if (i == n2){
            if (odpt == odpt0){
              break;
            }
          }
        }
*/
        hVer = lcPlineAddVer( hPline, NULL, odpt.x, odpt.y );
        lcPropPutFloat( hVer, LC_PROP_VER_Z, odpt.z );
      }
      pOdSpline->getFitTangents( StartTangent, EndTangent );
      if (StartTangent.x != 0.0 || StartTangent.y != 0.0){
        Pnt.Set( StartTangent.x, StartTangent.y );
        lcPropPutBool( hPline, LC_PROP_PLINE_HASANG0, true );
        lcPropPutFloat( hPline, LC_PROP_PLINE_ANG0, Pnt.Angle() );
      }
      if (EndTangent.x != 0.0 || EndTangent.y != 0.0){
        Pnt.Set( EndTangent.x, EndTangent.y );
        lcPropPutBool( hPline, LC_PROP_PLINE_HASANG2, true );
        lcPropPutFloat( hPline, LC_PROP_PLINE_ANG2, Pnt.Angle() );
      }
    }
  }else{
    if (nCtrPts > 0){
      FitType = LC_PLFIT_CUBIC;
      if (bPlanar && Degree==3){
        n2 = (nCtrPts - 1) / 3;
        n = (n2 * 3) + 1;
        if (n == nCtrPts){
          pOdSpline->getNurbsData( Degree, bRational, bClosed, bPeriodic, CtrlPoints, Knots, Weights, CtrlPtTol, KnotTol );
          n2 = Knots.length();
          if ((nCtrPts + 4) == n2){
            // Bezier curve
            FitType = LC_PLFIT_BEZIER;
            n2 = Weights.length();
            if (n2 == nCtrPts){
              bHasWeights = true;
            }
          }
/*
          k = 0;
          for (i=1; i<nCtrPts; i++){
            pOdSpline->getControlPointAt( i-1, odpt );
            pOdSpline->getControlPointAt( i, odpt2 );
            if (odpt.x==odpt2.x && odpt.y==odpt2.y){
              k++;
            }
          }
*/
        }
      }
      dval = 1.0;
      hPline = lcBlockAddPolyline( hBlock, FitType, bClosed, false );
      if (hPline){
        Read_EntBase( pOdSpline, hPline, hLcDb );
        for (i=0; i<nCtrPts; ++i) {
          pOdSpline->getControlPointAt( i, odpt );
          hVer = lcPlineAddVer( hPline, NULL, odpt.x, odpt.y );
          lcPropPutFloat( hVer, LC_PROP_VER_Z, odpt.z );
          if (FitType == LC_PLFIT_BEZIER){
            if (bHasWeights){
              dval = Weights.getAt( i );
            }
            lcPropPutFloat( hVer, LC_PROP_VER_WEIGHT, dval );
          }
        }
      }
    }
  }
}


//-----------------------------------------------
void Read_Mline (const OdDbMlinePtr pOdMline, HANDLE hLcDb, HANDLE hBlock)
{
/*
  OdGePoint3d OdPt;
  double Scale = pOdMline->scale();
  bool   bClosed = pOdMline->closedMline();
  int    Just = pOdMline->justification();
  int    nVers = pOdMline->numVertices();
  int    i;
  HANDLE hMline, hStyle;

  hMline = lcBlockAddMline( hBlock, 0, bClosed );
  if (hMline){
    Read_EntBase( pOdMline, hMline, hLcDb );
    for (i=0; i<nVers; i++){
      OdPt = pOdMline->vertexAt( i );
      lcMlineAddVer( hMline, 0, OdPt.x, OdPt.y );
    }
    lcPropPutFloat( hMline, LC_PROP_MLINE_SCALE, Scale );
    lcPropPutInt( hMline, LC_PROP_MLINE_JUST, Just );
    hStyle = DrwGetObjectByHandle( pOdMline->style(), hLcDb, LC_OBJ_MLSTYLE );
    if (hStyle == NULL){
      hStyle = lcDrwGetFirstObject( hLcDb, LC_OBJ_MLSTYLE );  // "Standard" multiline style
    }
    lcPropPutHandle( hMline, LC_PROP_MLINE_STYLE, hStyle );
  }
*/
}

